/*  TracePopupP.h

    Private header for  TracePopup  widget class.

    Copyright (C) 1996-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This include file contains the private class declarations for the
  TracePopup widget, a simple trace display widget for Xt.


    Written by      Richard Gooch   15-SEP-1996

    Last updated by Richard Gooch   4-SEP-2001

*/
#ifndef TracePopupP__H
#define TracePopupP__H

#include <X11/IntrinsicP.h>
#include <X11/ShellP.h>
#include <X11/StringDefs.h>

#include <karma_canvas.h>

#include <Xkw/TracePopup.h>

typedef struct _TracePopupPart
{
    /*  Public resources  */
    Bool             verbose;
    Visual           *canvasVisual; 
    KWorldCanvas     worldCanvas;
    XtCallbackList   realiseCallback;
    Kcolourmap       karmaColourmap;
    /*  Private resources  */
    KPixCanvas       pixcanvas;
    iarray           array;
    KCallbackFunc    iarr_destroy_callback;
    double           array_minimum;          /*  Raw (unscaled)         */
    double           array_maximum;          /*  Raw (unscaled)         */
    KwcsAstro        ap;
    Pixel            canvas_foreground;
    Widget           track_label0;
    Widget           track_label1;
    Widget           pswinpopup;
    Widget           overlay_menu;
    int              trace_dim;
    float            *trace;                 /*  Scaled                 */
    float            *weights;
    unsigned int     trace_buf_len;
    flag             smooth;
    flag             auto_zoom;
    flag             x_uninitialised;
    unsigned long    previous_background;
    float            previous_intensity;
    unsigned int     coord_buf_len;
    unsigned int     num_coord_values;       /*  Number of values to display */
    double           *coord_buf;
    unsigned int     num_restr_allocated;
    CONST char       **restr_names;
    double           *restr_values;
    unsigned int     coord_indices_buf_len;
    uaddr            *coord_indices;
    unsigned int     coord_counters_buf_len;
    uaddr            *coord_counters;
    int              mode;                   /*  Join, steps, etc.      */
    flag             valid_data;             /*  Trace data are valid?  */
    unsigned int     num_traces;             /*  For the <<save_cbk>>   */
} TracePopupPart, *TracePopupPartPtr;

typedef struct _TracePopupRec
{
    CorePart core;
    CompositePart composite;
    ShellPart shell;
    WMShellPart wm;
    VendorShellPart vendor;
    TopLevelShellPart topLevel;
    TracePopupPart tracePopup;
} TracePopupRec, *TracePopupPtr;

typedef struct _TracePopupClassPart
{
    int empty;
} TracePopupClassPart;

typedef struct _TracePopupClassRec
{
    CoreClassPart core_class;	
    CompositeClassPart composite_class;
    ShellClassPart shell_class;
    WMShellClassPart wm_shell_class;
    VendorShellClassPart vendor_shell_class;
    TopLevelShellClassPart top_level_shell_class;
    TracePopupClassPart tracePopup_class;
} TracePopupClassRec, *TracePopupClassPtr;

extern TracePopupClassRec tracePopupClassRec;

#endif
