/*  View2Datasets.h

    Private header for  View2Datasets  widget class.

    Copyright (C) 1997-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This include file contains the private class declarations for the
  View2Datasets widget.


    Written by      Richard Gooch   22-JAN-1997

    Last updated by Richard Gooch   7-JUN-2001

*/
#ifndef View2DatasetsP__H
#define View2DatasetsP__H

#include <X11/IntrinsicP.h>
#include <X11/ShellP.h>
#include <X11/StringDefs.h>

#include <Xkw/View2Datasets.h>
#include <karma_contour_def.h>

struct direction_type
{
    ViewableImage *vimages;
    ViewableImage *mag_vimages;
    ViewableImage *pan_vimages;
    KContourImage *pc_cimages;
    KContourImage *rgb_cimages;
    unsigned int num_frames;
    unsigned int frame_number;
};

struct dataset_type
{
    multi_array *multi_desc;
    iarray array_pseudo;
    iarray array_red;
    iarray array_green;
    iarray array_blue;
    unsigned int array_index;
    unsigned int cmap_index;
    KCallbackFunc destroy_func;
    /*  Image caches for 2-D dataset  */
    ViewableImage vimage;                /*  On preferred canvas for image   */
    ViewableImage mag_vimage;            /*  Magnified image on same canvas  */
    ViewableImage pan_vimage;            /*  Panner image on same canvas     */
    KContourImage pc_cimage;             /*  Contours on best-PC canvas      */
    KContourImage rgb_cimage;            /*  Contours on RGB canvas          */
    /*  Image caches for 3-D dataset  */
    struct direction_type xy;
    struct direction_type xz;
    struct direction_type zy;
    double min;
    double max;
    char levels[STRING_LENGTH];
    char negative_contour_colour[STRING_LENGTH];
    char positive_contour_colour[STRING_LENGTH];
    Widget filename_label;
    Widget object_label;
    char object_name[STRING_LENGTH];
    double beam_xpos;
    double beam_ypos;
};

struct pointlist
{
    double *x, *y;
    unsigned int buflen;
    unsigned int num;
};

typedef struct _View2DatasetsPart
{
    /*  Public resources  */
    Bool           createAtInit;
    Bool           verbose;
    String         wmtitle;
    XtCallbackList cursorCallback;
    /*  Private resources  */
    Widget image_display;
    Widget animatepopup;
    Widget trace_winpopup;
    Widget wmshell;
    struct dataset_type set0;
    struct dataset_type set1;
    unsigned int slice;
    unsigned int mode;
    unsigned int profile_dir;
    unsigned int profile_mode;
    unsigned int last_event_code;
    flag auto_apply_state;
    Widget blink_tgl;
    Widget mode_menu;
    Widget profile_dir_menu;
    Widget levels_dlg;
    Widget negative_colour_dlg;
    Widget positive_colour_dlg;
    Widget title_txt;
    Widget auto_title_tgl;
    Widget show_beam_tgl;
    Widget show_beam_name_tgl;
    Widget middle_zoom_tgl;
    Widget right_zoom_tgl;
    Widget intensity_menu;
    Widget link_iscale_tgl;
    Widget link_frames_tgl;
    struct edit_context_list *first_edit_context;
    iarray profile_arr;
    double fit_fwhm;
    struct pointlist profile_points;
    double *gausscurve_xarr, *gausscurve_yarr;
    unsigned int first_intensity;
    double contour_neg_width;
    double contour_pos_width;
    KPixCanvasRefreshArea old_trace_area;  /*  For trace circle  */
    char old_title[STRING_LENGTH];
} View2DatasetsPart, *View2DatasetsPartPtr;

typedef struct _View2DatasetsRec
{
    CorePart core;
    CompositePart composite;
    ShellPart shell;
    WMShellPart wm;
    VendorShellPart vendor;
#ifdef USE_TRANSIENT
    TransientShellPart transient;
#else
    TopLevelShellPart topLevel;
#endif
    View2DatasetsPart view2Datasets;
} View2DatasetsRec, *View2DatasetsPtr;

typedef struct _View2DatasetsClassPart
{
    int empty;
} View2DatasetsClassPart;

typedef struct _View2DatasetsClassRec
{
    CoreClassPart core_class;	
    CompositeClassPart composite_class;
    ShellClassPart shell_class;
    WMShellClassPart wm_shell_class;
    VendorShellClassPart vendor_shell_class;
#ifdef USE_TRANSIENT
    TransientShellClassPart transient_shell_class;
#else
    TopLevelShellClassPart top_level_shell_class;
#endif
    View2DatasetsClassPart view2Datasets_class;
} View2DatasetsClassRec, *View2DatasetsClassPtr;

extern View2DatasetsClassRec view2DatasetsClassRec;

#endif
