/*LINTLIBRARY*/
/*  support.c

    This code provides simple colourmap generation routines.

    Copyright (C) 1996-1998  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for writing simple colour
    values.


    Written by      Richard Gooch   19-MAY-1998: Moved from misc7.c

    Last updated by Richard Gooch   19-JUN-1998: Fixed bug: allow intensity
  value pointers to be NULL.


*/
#include <stdio.h>
#include <math.h>
#include <karma_cf.h>
#include <karma_a.h>
#include <karma_n.h>
#include <karma.h>

#define MAX_INTENSITY 65535


/*EXPERIMENTAL_FUNCTION*/
void cf_colour_table (unsigned int num_cells, unsigned short *reds,
		      unsigned short *greens, unsigned short *blues,
		      unsigned int stride, double x, double y,
		      CONST Colour_struct *lut, unsigned int lut_size)
/*  [SUMMARY] Compute a colourmap using a table.
    <num_cells> The number of colour cells to modify.
    <reds> The array of red intensity values. This may be NULL.
    <greens> The array of green intensity values. This may be NULL.
    <blues> The array of blue intensity values. This may be NULL.
    <stride> The stride (in unsigned shorts) between intensity values in the
    array.
    <x> A parameter used to compute the colour values, ranging from 0.0 to 1.0.
    <y> A parameter used to compute the colour values, ranging from 0.0 to 1.0.
    <lut> The lookup table.
    <lut_size> The size of the lookup table.
    [RETURNS] Nothing.
*/
{

    unsigned int pixel_count;
    int m;
    double shift, slope, xx, yy;

    if ( (x < 0.0) || (x > 1.0) || (y < 0.0) || (y > 1.0) )
    {
	shift = 0.5;
	slope = 1.0;
    }
    else
    {
	shift = x;
	slope = tan (PI_ON_2 * y);
    }
    for (pixel_count = 1;  pixel_count <= num_cells ; pixel_count++)
    {
	xx = 1.0 / (float) (num_cells+1) * (pixel_count);
	yy = slope * (xx-shift) + 0.5;
	if (yy < 0.0)
	{
	    yy = 0.0;
	}
	else
	{
	    if (yy > 1.0)
	    {
		yy = 1.0;
	    }
	}
	m = ( (float) (lut_size - 2) * yy + 1.5 );
	if (m >= lut_size) 
	m = lut_size - 1;
	if (reds != NULL) reds[(pixel_count - 1) * stride] =
			      lut[m].red * MAX_INTENSITY;
	if (greens != NULL) greens[(pixel_count - 1) * stride] =
				lut[m].green * MAX_INTENSITY;
	if (blues != NULL) blues[(pixel_count - 1) * stride] =
			       lut[m].blue * MAX_INTENSITY;
    }
}   /*  End Function cf_colour_table  */
