/*LINTLIBRARY*/
/*  main.c

    This code provides high level connection control/ management routines.

    Copyright (C) 1992-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for managing connections.


    Written by      Richard Gooch   14-SEP-1992

    Updated by      Richard Gooch   27-OCT-1992

    Updated by      Richard Gooch   29-DEC-1992: Added connection management
  tool support and complete read processing.

    Updated by      Richard Gooch   1-JAN-1993: Took account of change to
  ch_gets  .

    Updated by      Richard Gooch   6-JAN-1993: Changed from  #include "*.h"
  to  #include <*.h>

    Updated by      Richard Gooch   7-MAR-1993: Added revision number
  transmission.

    Updated by      Richard Gooch   26-MAR-1993: Made use of new routine:
  ch_open_and_fill_memory  .

    Updated by      Richard Gooch   4-APR-1993: Added version number parameter
  to protocol registration routines.

    Updated by      Richard Gooch   24-JUL-1993: Changed to  ch_open_file  .

    Updated by      Richard Gooch   3-AUG-1993: Fixed bug in
  conn_register_client_protocol which did not limit the maximum number of
  connections to that in the services file.

    Updated by      Richard Gooch   25-AUG-1993: Improved documentation for
  conn_get_serv_connection  and  conn_get_client_connection  .

    Updated by      Richard Gooch   27-AUG-1993: Added draining of input data
  in  serv_connection_input_func  and  conn_attempt_connection  .

    Updated by      Richard Gooch   30-AUG-1993: Fixed introduced bug in
  serv_connection_input_func  which froze server until client sent new data.

    Updated by      Richard Gooch   1-SEP-1993: Added test to ensure read
  callbacks are draining channel data.

    Updated by      Richard Gooch   5-SEP-1993: Removed need for protocols file

    Updated by      Richard Gooch   14-SEP-1993: Made diagnostic for input data
  not being read more helpful.

    Updated by      Richard Gooch   26-SEP-1993: Created
  conn_register_cm_quiescent_func  function.

    Updated by      Richard Gooch   27-SEP-1993: Improved error message in
  cm_command_func  .

    Updated by      Richard Gooch   23-NOV-1993: Removed some dead externs.

    Updated by      Richard Gooch   3-MAY-1994: Added encryption support.

    Updated by      Richard Gooch   16-MAY-1994: Dropped encryption once basic
  setup information is transferred for local connections.

    Updated by      Richard Gooch   21-MAY-1994: Added  #include <karma_ch.h>
  and created  conn_extract_protocols  .

    Updated by      Richard Gooch   26-AUG-1994: Moved  typedef  of Connection
  class to header.

    Updated by      Richard Gooch   4-OCT-1994: Changed to  m_cmp  routine in
    order to avoid having to link with buggy UCB compatibility library in
    Slowaris 2.3

    Updated by      Richard Gooch   3-NOV-1994: Switched to OS_ and MACHINE_
  macros for machine determination.

    Updated by      Richard Gooch   25-NOV-1994: Switched to  cen_  package.

    Updated by      Richard Gooch   26-NOV-1994: Moved to  packages/conn/main.c

    Updated by      Richard Gooch   3-DEC-1994: Made use of  pgp_  package.

    Updated by      Richard Gooch   7-DEC-1994: Stripped declaration of  errno
  and added #include <errno.h>

    Updated by      Richard Gooch   1-JAN-1995: Added per-protocol encryption
  support. Raw protocol version number incremented.

    Updated by      Richard Gooch   7-JAN-1995: Changed from "PRIMARY_KEY" to
  "RAW_PROTOCOL" for raw protocol.

    Updated by      Richard Gooch   23-JAN-1995: Added some casts to keep
  IRIX compiler happy.

    Updated by      Richard Gooch   14-MAR-1995: Changed
  <attempt_connection_to_cm> to send fully qualified hostname if possible.

    Updated by      Richard Gooch   5-MAY-1995: Placate SGI compiler.

    Updated by      Richard Gooch   14-JUN-1995: Made use of <ex_uint>.

    Updated by      Richard Gooch   16-JUN-1995: Made use of
  <r_get_fq_hostname>.

    Updated by      Richard Gooch   10-AUG-1995: Removed printing of error
  message in <conn_attempt_connection> if raw connection failed.

    Updated by      Richard Gooch   30-NOV-1995: Changed from call to <_exit>
  to call to <exit> in <cm_close_func> for IRIX sproc() problems.

    Updated by      Richard Gooch   7-APR-1996: Changed to new documentation
  format.

    Updated by      Richard Gooch   3-JUN-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   23-AUG-1996: Support "local-slow" hostname.

    Updated by      Richard Gooch   14-DEC-1996: Minor documentation changes.

    Updated by      Richard Gooch   17-DEC-1996: Created experimental
  <conn_initialise> and obsoleted <conn_register_managers>.

    Updated by      Richard Gooch   23-APR-1997: Send "karmaPortNumber" Glish
  event in <conn_become_server>.

    Updated by      Richard Gooch   23-MAY-1997: Added channel flush in
  "ping_server" protocol.

    Updated by      Richard Gooch   17-FEB-1997: Added UDP support.

    Updated by      Richard Gooch   19-FEB-1998: Created freelist. Added ATMSVC
  support.

    Updated by      Richard Gooch   20-FEB-1998: Created last pointers for
  client and server lists, to avoid searching for the end of list when
  appending.

    Updated by      Richard Gooch   21-FEB-1998: Implemented QOS parsing.

    Updated by      Richard Gooch   6-APR-1998: Changed function
  <conn_attempt_connection> to return new connection rather than flag.

    Updated by      Richard Gooch   8-SEP-1998: Fixed bug in function
  <conn_attempt_connection>: failed open method would cause abort.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <ex> and <r> packages.

    Updated by      Richard Gooch   8-OCT-1998: Took account of change to
  <<karma_qos>> structure.

    Updated by      Richard Gooch   12-DEC-1998: Took account of change to
  <<karma_qos>> structure: separate Tx and Rx traffic parameters.

    Updated by      Richard Gooch   25-DEC-1998: Tidied up "conn_mngr_control"
  protocol.

    Updated by      Richard Gooch   13-AUG-1999: Made egcs happy.

    Updated by      Richard Gooch   26-JAN-2000: Minor cosmetic changes.

    Updated by      Richard Gooch   27-JAN-2000: Removed redundant ERRSTRINGs
  in error messages printed after failures from <pio_*> functions.

    Updated by      Richard Gooch   14-FEB-2000: Fixed bug in allocating last
  entry of freelist.

    Updated by      Richard Gooch   18-MAR-2001: Minor cosmetic changes.

    Last updated by Richard Gooch   14-SEP-2001: Added version information to
  <conn_extract_protocols>.


*/
#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <karma.h>
#include <karma_conn.h>
#include <karma_chm.h>
#include <karma_pio.h>
#include <karma_chs.h>
#include <karma_cen.h>
#include <karma_pgp.h>
#include <karma_gls.h>
#include <karma_ch.h>
#include <karma_ds_val.h>
#include <karma_st.h>
#include <karma_ex.h>
#include <karma_rp.h>
#include <karma_md.h>
#include <karma_dm.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_r.h>
#include <karma_c.h>
#include <karma_s.h>
#include <karma_n.h>
#include <k_cm_codes.h>
#include <os.h>


#if defined(HAS_SOCKETS) || defined(HAS_COMMUNICATIONS_EMULATION)
#define COMMUNICATIONS_AVAILABLE
#endif

#define ALLOC_SIZE 64
#define USED_MAGIC_NUMBER 1794359023
#define FREE_MAGIC_NUMBER 3537843432U

#define VERIFY_USED(conn) {if (conn == NULL) \
{fprintf (stderr, "NULL connection object\n"); \
 a_prog_bug (function_name); } \
if (conn->magic_number != USED_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid connection object\n"); \
 a_prog_bug (function_name); }}

#define VERIFY_FREE(conn) {if (conn == NULL) \
{fprintf (stderr, "NULL connection object\n"); \
 a_prog_bug (function_name); } \
if (conn->magic_number != FREE_MAGIC_NUMBER) \
{fprintf (stderr, "Invalid freed connection object\n"); \
 a_prog_bug (function_name); }}

#define UNMANAGE_CHANNEL(channel,mch) \
{if (unmanage_channel == chm_unmanage) chm_unmanage_by_handle (mch); \
 else (*unmanage_channel) (channel);}

#define AUTH_PASSWORD_LENGTH 128
#define CONN_MAGIC_NUMBER (unsigned long) 245987237
#define REVISION_NUMBER (unsigned long) 5
#define SETUP_MESSAGE_LENGTH (4 + 4)
#define PROTOCOL_NAME_LENGTH 80
#define PROTOCOL_DEFINITION_MESSAGE_LENGTH (PROTOCOL_NAME_LENGTH + 4)

#define RAW_PROTOCOL "RAW_PROTOCOL"
#define CM_PROTOCOL_VERSION (unsigned int) 2

#define SECURITY_TYPE_KEY (unsigned int) 0
#define SECURITY_TYPE_IDEA (unsigned int) 1
#define SECURITY_TYPE_PGPuserID_IDEA (unsigned int) 2
#define SECURITY_TYPE_DROP_ENCRYPTION (unsigned int) 3

#define SECURITY_STRING_KEY "key-only"
#define SECURITY_STRING_IDEA "IDEA"
#define SECURITY_STRING_PGPuserID_IDEA "PGPuserID-IDEA"
#define SECURITY_STRING_DROP_ENCRYPTION "drop-encryption"


/*  Definition of security types:
        key-only        IDEA used to verify authentication. Data unencrypted
	IDEA            IDEA used to verify authentication and data encrypted
	PGPuserID-IDEA  PGP used for session key transfer, then use IDEA mode
	drop-encryption drop data encryption for a specific protocol
*/

/*  Raw protocol
    Stage 0:  Client connects to server.
    Stage 1:  Client then sends:
      Magic number
      Revision number
    Stage 2:
      Client and server negotiate RAW_PROTOCOL security if applicable
    Stage 3:  Client sends:
      Protocol name [PROTOCOL_NAME_LENGTH]
      Protocol version number
    Stage 4:
      Client and server negotiate protocol security if applicable
    Stage 5:  Server replies:
      Verification string (NULL for success)
      Module name (on success only)
    Stage 6:  Client sends:
      Module name
      Connection type
      Optional datagram negotiation
    Stage 7:  Server reads:
      Module name
      Optional datagram negotiation
*/

/*  Security negotiation sequence:
      key-only:
        Key data is used for IDEA session key. Client and server each send an
	IDEA block of random bytes and then write back the blocks. This both
	helps randomise the IDEA IV and also ensures that both ends are
	authenticated (i.e. a third party has not stepped in). Once both ends
	are authenticated, the encryption is dropped.
      IDEA:
        Same as key-only except that encryption is not dropped unless the
	connection is local
      PGPuserID-IDEA:
        PGP is used for IDEA session key transfer. After that IDEA mode is used
*/

/*  Generic connection structure (server and client)  */
struct connection_type
{
    unsigned int magic_number;
    flag client;
    /*  Server specific fields  */
    flag verified_raw;
    flag verified_security;
    flag verified_protocol_security;
    ChConverter pri_encryption_converter;
    ChConverter pro_encryption_converter;
    /*  General fields  */
    unsigned int type;
    Channel channel;                /*  The channel over which data is sent  */
    Channel datagram_channel;       /*  The channel for datagrams            */
    KManagedChannel main_mch;
    KManagedChannel datagram_mch;
    char *protocol_name;            /*  Pointer copy: do not free            */
    unsigned int *connection_count;
    flag (*read_func) (Connection connection, void **info, Channel channel);
    void (*close_func) (Connection connection, void *info, Channel channel);
    void *info;
    char *module_name;
    unsigned long inet_addr;        /*  IP address of remote host            */
    Connection prev;
    Connection next;
    Connection *list_start;         /*  Pointer to the start of the list     */
    Connection *list_end;
    char idea_block[EN_IDEA_BLOCK_SIZE];
};

/*  Server protocol structures  */
struct serv_protocol_list_type
{
    char *protocol_name;
    unsigned int version;
    unsigned int connection_count;
    unsigned int max_connections;
    flag (*open_func) (Connection connection, void **info);
    flag (*read_func) (Connection connection, void **info, Channel channel);
    void (*close_func) (Connection connection, void *info, Channel channel);
    struct serv_protocol_list_type *next;
    struct auth_host_list_type *authorised_hosts;
};

struct auth_host_list_type
{
    struct host_list_type *host_ptr;
    struct auth_host_list_type *prev;
    struct auth_host_list_type *next;
};

struct host_list_type
{
    unsigned long inet_addr;
    char *hostname;
    struct host_list_type *next;
};

/*  Client protocol structure  */
struct client_protocol_list_type
{
    char *protocol_name;
    unsigned int version;
    unsigned int connection_count;
    unsigned int max_connections;
    flag (*validate_func) (void **info);
    flag (*open_func) (Connection connection, void **info);
    flag (*read_func) (Connection connection, void **info, Channel channel);
    void (*close_func) (Connection connection, void *info, Channel channel);
    struct client_protocol_list_type *next;
};

struct auth_password_list_type
{
    char *protocol_name;
    char password[AUTH_PASSWORD_LENGTH];
    unsigned int security_type;
    struct auth_password_list_type *next;
};


/*  Private data  */
/*  Connection information  */
static flag ran_become_server = FALSE;
static Connection first_serv_connection = NULL;
static Connection last_serv_connection = NULL;
static Connection first_client_connection = NULL;
static Connection last_client_connection = NULL;
static Connection first_free = NULL;
static struct serv_protocol_list_type *serv_protocol_list = NULL;
static struct client_protocol_list_type *client_protocol_list = NULL;


/*  Connection to Connection Management tool  */
static Channel cm_channel = NULL;

/*  Pointers to channel management code  */
static KManagedChannel (*manage_channel) () = NULL;
static void (*unmanage_channel) () = NULL;

static void (*exit_schedule_function) () = NULL;

/*  Function to call when CM tool or shell is quiescent  */
static void (*quiescent_function) () = NULL;

/*  Cryptographically strong Pseudo Random Number Generator  */
static RandPool main_randpool = NULL;

/*  Connection types  */
#define CONNECTION_TYPE_UNDEFINED 0
#define CONNECTION_TYPE_STREAM    1
#define CONNECTION_TYPE_UDP       2
#define CONNECTION_TYPE_ATMSVC    3
static char *type_strings[] =
{
    "undefined",
    "stream",
    "udp",
    "atmsvc",
};


/*  Private functions  */
STATIC_FUNCTION (flag _conn_dock_input_func, (Channel dock, void **info) );
STATIC_FUNCTION (flag _conn_serv_connection_input_func,
		 (Channel channel, void **info) );
STATIC_FUNCTION (flag _conn_verify_raw, (Connection connection) );
STATIC_FUNCTION (flag _conn_verify_raw_security, (Connection connection) );
STATIC_FUNCTION (flag _conn_verify_connection, (Connection conn) );
STATIC_FUNCTION (flag _conn_verify_protocol_security,
		 (Connection connection) );
STATIC_FUNCTION (flag _conn_client_connection_input_func,
		 (Channel channel, void **info) );
STATIC_FUNCTION (void _conn_connection_close_func,
		 (Channel channel, void *info) );
STATIC_FUNCTION (void _conn_dealloc_connection,
		 (Connection connection, Channel channel) );
STATIC_FUNCTION (struct serv_protocol_list_type *_conn_get_serv_protocol_info,
		 (CONST char *protocol_name) );
STATIC_FUNCTION (struct client_protocol_list_type
		 *_conn_get_client_protocol_info,
		 (CONST char *protocol_name) );
STATIC_FUNCTION (struct auth_password_list_type *_conn_get_authinfo,
		 (CONST char *protocol) );
STATIC_FUNCTION (struct auth_password_list_type *_conn_get_password_list,
		 (void) );
STATIC_FUNCTION (void _conn_convert_password_to_bin,
		 (unsigned char *binpassword, CONST char *asciipassword) );
#ifdef dummy
static flag check_host_access (void);
#endif
STATIC_FUNCTION (char *_conn_write_protocol,
		 (Channel channel, CONST char *protocol_name,
		  unsigned int version, unsigned int conn_type) );
STATIC_FUNCTION (flag _conn_respond_to_ping_server_from_client,
		 (Connection connection, void **info, Channel channel) );
STATIC_FUNCTION (flag _conn_register_server_exit,
		 (Connection connection, void **info) );
STATIC_FUNCTION (Connection _conn_get_numbered_connection,
		 (CONST char *protocol_name, unsigned int number,
		  Connection list) );
STATIC_FUNCTION (void _conn_attempt_connection_to_cm, (void) );
STATIC_FUNCTION (flag _conn_cm_command_func, (Channel channel, void **info) );
STATIC_FUNCTION (void _conn_cm_close_func, (Channel channel, void *info) );
STATIC_FUNCTION (void _conn_setup_prng, (void) );
STATIC_FUNCTION (void _conn_randpool_destroy_func, (RandPool rp, void *info) );
STATIC_FUNCTION (ChConverter _conn_get_encryption,
		 (Channel channel, struct auth_password_list_type *authinfo,
		  flag server, char *rand_block) );
STATIC_FUNCTION (Connection _conn_alloc_connection, () );
STATIC_FUNCTION (void _conn_free_connection, (Connection connection) );
STATIC_FUNCTION (flag _conn_parse_hostname,
		 (unsigned int *conn_type, struct karma_qos *qos,
		  char *hostname, unsigned int hostname_length) );
STATIC_FUNCTION (flag _conn_parse_qos, (struct karma_qos *qos, char *string) );
STATIC_FUNCTION (int _conn_parse_bw, (char *string) );
STATIC_FUNCTION (void _conn_show_atm_tfparm,
		 (CONST struct karma_traffic_parameters *tp,
		  CONST char *prefix) );


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
void conn_initialise ( void (*exit_schedule_func) () )
/*  [SUMMARY] Initialise the package.
    [NOTE] This routine MUST be called prior to the [<conn_become_server>]
    routine.
    <exit_shedule_func> The routine which should be called when a clean exit is
    to be scheduled. The prototype function is
    [<conn_PROTO_exit_schedule_func>].
    If this is NULL, then exit(3) is called instead.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "conn_initialise";

    if ( (manage_channel != NULL) || (unmanage_channel != NULL) )
    {
	fprintf (stderr, "Channel managers already registered\n");
	a_prog_bug (function_name);
    }
    manage_channel = chm_manage;
    unmanage_channel = chm_unmanage;
    exit_schedule_function = exit_schedule_func;
    _conn_attempt_connection_to_cm ();
    _conn_get_password_list ();
}   /*  End Function conn_initialise  */

/*OBSOLETE_FUNCTION*/
void conn_register_managers ( flag (*manage_func) (), void (*unmanage_func) (),
			      void (*exit_schedule_func) () )
/*  [SUMMARY] Register the channel management functions.
    [NOTE] This routine MUST be called prior to the [<conn_become_server>]
    routine.
    See the [<chm>] package for information on channel management.
    <manage_func> The routine to manage channels. See the [<chm_manage>]
    routine for the interface definition.
    <unmanage_func> The routine to unmanage channels. See the [<chm_unmanage>]
    routine for the interface definition.
    <exit_shedule_func> The routine which should be called when a clean exit is
    to be scheduled. The prototype function is
    [<conn_PROTO_exit_schedule_func>].
    If this is NULL, then exit(3) is called instead.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "conn_register_managers";

    if ( (manage_channel != NULL) || (unmanage_channel != NULL) )
    {
	fprintf (stderr, "Channel managers already registered\n");
	a_prog_bug (function_name);
    }
    fprintf (stderr, "WARNING: the <%s> function will be removed in\nKarma",
	     function_name);
    fprintf (stderr,
	     " version 2.0. Use the <conn_initialise> function instead.\n");
    manage_channel = (KManagedChannel (*) () ) manage_func;
    unmanage_channel = unmanage_func;
    exit_schedule_function = exit_schedule_func;
    _conn_attempt_connection_to_cm ();
    _conn_get_password_list ();
}   /*  End Function conn_register_managers  */

/*PUBLIC_FUNCTION*/
void conn_register_server_protocol ( CONST char *protocol_name,
				     unsigned int version,
				     unsigned int max_connections,
				     flag (*open_func) (),
				     flag (*read_func) (),
				     void (*close_func) () )
/*  [SUMMARY] Register server protocol support.
    [PURPOSE] This routine will register the support of a new Karma protocol
    for the Karma ports which are being managed by the various routines in the
    [<conn>] package. This routine may be called at any time.
    <protocol_name> The name of the new protocol to support. Note that this is
    only for incoming connections.
    <version> The version number for the protocol. When any changes to the
    protocol are made, this should be increased.
    <max_connections> The maximum number of incoming connections to this server
    If this is 0, an unlimited number of connections is permitted.
    <open_func> The routine which is called when a new connection is opened.
    The prototype function is [<conn_PROTO_open_func>].
    <read_func> The function which will be called when data is ready to be read
    data from the connection. The prototype function is
    [<conn_PROTO_read_func>].
    <close_func> The function which will be called prior to closure of a
    connection. The prototype function is [<conn_PROTO_close_func>].
    [RETURNS] Nothing.
*/
{
    struct serv_protocol_list_type *entry;
    struct serv_protocol_list_type *last_entry;
    static char function_name[] = "conn_register_server_protocol";

    if ( (int) strlen (protocol_name) >= PROTOCOL_NAME_LENGTH )
    {
	fprintf (stderr,"Protocol name: \"%s\" too long. Max: %u characters\n",
		 protocol_name, PROTOCOL_NAME_LENGTH);
	a_prog_bug (function_name);
    }
    /*  Test if protocol already supported  */
    if (_conn_get_serv_protocol_info (protocol_name) != NULL)
    {
	/*  Already supported: error  */
	fprintf (stderr, "Protocol: \"%s\" already supported\n",
			protocol_name);
	a_prog_bug (function_name);
    }
    /*  Allocate new entry  */
    if ( ( entry = m_alloc (sizeof *entry) ) == NULL )
    {
	m_abort (function_name, "new protocol list entry");
    }
    if ( ( entry->protocol_name = st_dup (protocol_name) ) == NULL )
    {
	m_abort (function_name, "protocol name");
    }
    entry->version = version;
    entry->max_connections = max_connections;
    entry->connection_count = 0;
    entry->open_func = open_func;
    entry->read_func = read_func;
    entry->close_func = close_func;
    entry->authorised_hosts = NULL;
    entry->next = NULL;
    if (serv_protocol_list == NULL)
    {
	/*  Create protocol list  */
	serv_protocol_list = entry;
    }
    else
    {
	/*  Append entry to protocol list  */
	for (last_entry = serv_protocol_list; last_entry->next != NULL;
	     last_entry = last_entry->next);
	last_entry->next = entry;
    }
    return;
}   /*  End Function conn_register_server_protocol  */

/*PUBLIC_FUNCTION*/
void conn_register_client_protocol ( CONST char *protocol_name,
				     unsigned int version,
				     unsigned int max_connections,
				     flag (*validate_func) (),
				     flag (*open_func) (),
				     flag (*read_func) (),
				     void (*close_func) () )
/*  [SUMMARY] Register client protocol support.
    [PURPOSE] This routine will register the support of a new Karma protocol
    for outgoing (client) connections. This routine may be called at any time.
    <protocol_name> The name of the new protocol to support. Note that this is
    only for outgoing connections.
    <version> The version number for the protocol. When any changes to the
    protocol are made, this should be increased.
    <max_connections> The maximum number of outgoing connections from this
    client. If this is 0, an unlimited number of connections is permitted.
    <validate_func> The function which will validate the opening of a new
    connection (prior to any attempts to connect to the server). The prototype
    function is [<conn_PROTO_client_validate_func>].
    <open_func> The routine which is called when a new connection is opened.
    The prototype function is [<conn_PROTO_open_func>].
    <read_func> The function which will be called when data is ready to be read
    data from the connection. The prototype function is
    [<conn_PROTO_read_func>].
    <close_func> The function which will be called prior to closure of a
    connection. The prototype function is [<conn_PROTO_close_func>].
    [RETURNS] Nothing.
*/
{
    struct client_protocol_list_type *entry;
    struct client_protocol_list_type *last_entry;
    static char function_name[] = "conn_register_client_protocol";

    if ( (int) strlen (protocol_name) >= PROTOCOL_NAME_LENGTH )
    {
	fprintf (stderr,"Protocol name: \"%s\" too long. Max: %u characters\n",
		 protocol_name, PROTOCOL_NAME_LENGTH);
	a_prog_bug (function_name);
    }
    if (strcmp (protocol_name, "conn_mngr_control") == 0)
    {
	fprintf (stderr, "Client protocol: \"%s\" reserved\n", protocol_name);
	a_prog_bug (function_name);
    }
    /*  Test if protocol already supported  */
    if (_conn_get_client_protocol_info (protocol_name) != NULL)
    {
	/*  Already supported: error  */
	fprintf (stderr, "Protocol: \"%s\" already supported\n",protocol_name);
	a_prog_bug (function_name);
    }
    /*  Allocate new entry  */
    if ( ( entry = m_alloc (sizeof *entry) ) == NULL )
    {
	m_abort (function_name, "new protocol list entry");
    }
    if ( ( entry->protocol_name = st_dup (protocol_name) ) == NULL )
    {
	m_abort (function_name, "protocol name");
    }
    entry->version = version;
    entry->max_connections = max_connections;
    entry->connection_count = 0;
    entry->validate_func = validate_func;
    entry->open_func = open_func;
    entry->read_func = read_func;
    entry->close_func = close_func;
    entry->next = NULL;
    if (client_protocol_list == NULL)
    {
	/*  Create protocol list  */
	client_protocol_list = entry;
    }
    else
    {
	/*  Append entry to protocol list  */
	for (last_entry = client_protocol_list; last_entry->next != NULL;
	     last_entry = last_entry->next);
	last_entry->next = entry;
    }
    return;
}   /*  End Function conn_register_client_protocol  */

/*PUBLIC_FUNCTION*/
Channel conn_get_channel (Connection connection)
/*  [SUMMARY] Extract the channel object associated with a connection.
    <connection> The connection object.
    [RETURNS] The channel object.
    [SEE ALSO] [<conn_get_datagram_channel>].
*/
{
    static char function_name[] = "conn_get_channel";

    VERIFY_USED (connection);
    return (connection->channel);
}   /*  End Function conn_get_channel  */

/*EXPERIMENTAL_FUNCTION*/
Channel conn_get_datagram_channel (Connection connection)
/*  [SUMMARY] Extract the datagram channel associated with a connection.
    <connection> The connection object.
    [RETURNS] The datagram channel object. This may be NULL if the connection
    is a normal stream (i.e. no extra unreliable datagram channel).
    [SEE ALSO] [<conn_get_channel>].
*/
{
    static char function_name[] = "conn_get_datagram_channel";

    VERIFY_USED (connection);
    return (connection->datagram_channel);
}   /*  End Function conn_get_datagram_channel  */

/*PUBLIC_FUNCTION*/
Connection conn_attempt_connection (CONST char *hostname,
				    unsigned int port_number,
				    CONST char *protocol_name)
/*  [SUMMARY] Attempt connection.
    [PURPOSE] This routine will attempt to make a connection to a server. The
    routine always makes a connection using the most efficient transport layer
    available.
    <hostname> The hostname of the machine on which the server is running.
    <port_number> The Karma port number to connect to.
    <protocol_name> The protocol to connect with.
    [RETURNS] The connection object on success, else NULL.
    [NOTE] The appropriate callback functions registered with
    [<conn_register_client_protocol>] will be called.
*/
{
    flag local;
    int bytes;
    unsigned int hostname_length, conn_type;
    unsigned long host_addr, real_addr, ul_tmp;
    Connection new_conn;
    Channel channel;
    struct karma_qos qos;
    void *info = NULL;
    struct client_protocol_list_type *protocol_info;
    char host[STRING_LENGTH], atm_addr[STRING_LENGTH], randval[8];
    static char function_name[] = "conn_attempt_connection";

    if ( (int) strlen (protocol_name) >= PROTOCOL_NAME_LENGTH )
    {
	fprintf (stderr,
		 "%s: protocol name: \"%s\" too long. Max: %u characters\n",
		 function_name, protocol_name, PROTOCOL_NAME_LENGTH);
	a_prog_bug (function_name);
    }
    if (manage_channel == NULL)
    {
	fprintf (stderr, "%s: channel managers not registered\n",
		 function_name);
	a_prog_bug (function_name);
    }
    if (strcmp (protocol_name, "conn_mngr_control") == 0)
    {
	fprintf (stderr, "%s: client protocol: \"%s\" reserved\n",
		 function_name, protocol_name);
	a_prog_bug (function_name);
    }
    if (strcmp (protocol_name, "conn_mngr_stdio") == 0)
    {
	fprintf (stderr, "%s: client protocol: \"%s\" reserved\n",
		 function_name, protocol_name);
	a_prog_bug (function_name);
    }
    hostname_length = strlen (hostname);
    if (hostname_length >= STRING_LENGTH)
    {
	fprintf (stderr, "%s: hostname: \"%s\" is too long\n",
		 function_name, hostname);
	return (NULL);
    }
    m_copy (host, hostname, hostname_length + 1);
    if ( !_conn_parse_hostname (&conn_type, &qos, host, hostname_length) )
	return (NULL);
    if ( ( host_addr = r_get_inet_addr_from_host (host, &local) ) == 0 )
    {
	fprintf (stderr, "%s: error getting host address for: \"%s\"\n",
		 function_name, host);
	return (NULL);
    }
    real_addr = host_addr;
    if ( local && (strcmp (host, "local-slow") != 0) ) host_addr = 0;
    if ( ( protocol_info = _conn_get_client_protocol_info (protocol_name) )
	 == NULL )
    {
	fprintf (stderr, "%s: protocol: \"%s\" not supported\n",
		 function_name, protocol_name);
	return (NULL);
    }
    if ( (protocol_info->max_connections > 0) &&
	 (protocol_info->connection_count >= protocol_info->max_connections) )
    {
	fprintf (stderr,
		 "%s: maximum number of client connections reached for protocol: \"%s\"\n",
		 function_name, protocol_name);
	return (NULL);
    }
    /*  Validate  */
    if (protocol_info->validate_func != NULL)
    {
	if ( !(*protocol_info->validate_func) (&info) )
	{
	    return (NULL);
	}
    }
    if ( ( new_conn = _conn_alloc_connection () ) == NULL )
    {
	m_error_notify (function_name, "new connection object");
	return (NULL);
    }
    new_conn->client = TRUE;
    new_conn->type = conn_type;
    new_conn->protocol_name = protocol_info->protocol_name;
    new_conn->connection_count = &protocol_info->connection_count;
    new_conn->read_func = protocol_info->read_func;
    new_conn->close_func = protocol_info->close_func;
    new_conn->info = info;
    new_conn->inet_addr = host_addr;
    /*  Open connection  */
    if ( ( channel = ch_open_connection (host_addr, port_number) ) == NULL )
    {
	_conn_free_connection (new_conn);
	return (NULL);
    }
    new_conn->channel = channel;
    /*  Send protocol info  */
    if ( ( new_conn->module_name =
	  _conn_write_protocol (new_conn->channel, protocol_name,
				protocol_info->version, conn_type) ) == NULL )
    {
	fprintf (stderr, "%s: error writing authentication information\n",
		 function_name);
	ch_close (new_conn->channel);
	_conn_free_connection (new_conn);
	return (NULL);
    }
    switch (conn_type)
    {
      case CONNECTION_TYPE_STREAM:
	break;
      case CONNECTION_TYPE_UDP:
	if ( !pio_read32 (new_conn->channel, &ul_tmp) )
	{
	    fprintf (stderr, "%s: error reading UDP port number\n",
		     function_name);
	    ch_close (new_conn->channel);
	    _conn_free_connection (new_conn);
	    return (NULL);
	}
	if (ch_read (new_conn->channel, randval, sizeof randval)
	    < sizeof randval)
	{
	    fprintf (stderr, "%s: error reading random value\n",
		     function_name);
	    ch_close (new_conn->channel);
	    _conn_free_connection (new_conn);
	    return (NULL);
	}
	if ( ( channel = ch_udp_connect (NULL, real_addr, ul_tmp) ) == NULL )
	{
	    fprintf (stderr, "%s: error connecting UDP socket\n",
		     function_name);
	    ch_close (new_conn->channel);
	    _conn_free_connection (new_conn);
	    return (NULL);
	}
	while (TRUE)
	{
	    if (ch_write (channel, randval, sizeof randval) < sizeof randval)
	    {
		fprintf (stderr, "%s: error writing to UDP connection\n",
			 function_name);
		ch_close (new_conn->channel);
		ch_close (channel);
		_conn_free_connection (new_conn);
		return (NULL);
	    }
	    if ( !ch_flush (channel) )
	    {
		fprintf (stderr, "%s: error flushing UDP connection\n",
			 function_name);
		ch_close (new_conn->channel);
		ch_close (channel);
		_conn_free_connection (new_conn);
		return (NULL);
	    }
	    if ( dm_native_wait_input (ch_get_descriptor (new_conn->channel),
				       1000) )
	    {
		ch_read (new_conn->channel, randval, 1);
		break;
	    }
	}
	new_conn->datagram_channel = channel;
	break;
      case CONNECTION_TYPE_ATMSVC:
	if ( !pio_read32 (new_conn->channel, &ul_tmp) )
	{
	    fprintf (stderr, "%s: error reading ATM address length\n",
		     function_name);
	    ch_close (new_conn->channel);
	    _conn_free_connection (new_conn);
	    return (NULL);
	}
	if (ul_tmp > STRING_LENGTH)
	{
	    fprintf (stderr,
		     "%s: ATM address length: %lu is larger than: %u\n",
		     function_name, ul_tmp, STRING_LENGTH);
	    ch_close (new_conn->channel);
	    _conn_free_connection (new_conn);
	    return (NULL);
	}
	if (ch_read (new_conn->channel, atm_addr, ul_tmp) < ul_tmp)
	{
	    fprintf (stderr, "%s: error reading ATM address\n",
		     function_name);
	    ch_close (new_conn->channel);
	    _conn_free_connection (new_conn);
	    return (NULL);
	}
	if ( ( channel = ch_atm_connect (atm_addr, ul_tmp, &qos) ) == NULL )
	{
	    fprintf (stderr, "%s: error connecting\n", function_name);
	    ch_close (new_conn->channel);
	    _conn_free_connection (new_conn);
	    return (NULL);
	}
	new_conn->datagram_channel = channel;
	break;
      default:
	fprintf (stderr, "Illegal connection type: %u\n", conn_type);
	a_prog_bug (function_name);
	break;
    }
    new_conn->main_mch = (*manage_channel) (new_conn->channel,
					    (void *) new_conn,
					    _conn_client_connection_input_func,
					    _conn_connection_close_func,
					    ( flag (*) () ) NULL,
					    ( flag (*) () ) NULL);
    if (new_conn->main_mch == NULL)
    {
	ch_close (new_conn->channel);
	if (new_conn->datagram_channel != NULL)
	    ch_close (new_conn->datagram_channel);
	_conn_free_connection (new_conn);
	a_func_abort (function_name, "Could not manage channel");
	return (NULL);
    }
    if (new_conn->datagram_channel != NULL)
    {
	new_conn->datagram_mch =
	    (*manage_channel) (new_conn->datagram_channel, (void *) new_conn,
			       _conn_client_connection_input_func,
			       _conn_connection_close_func,
			       ( flag (*) () ) NULL,
			       ( flag (*) () ) NULL);
	if (new_conn->datagram_mch == NULL)
	{
	    UNMANAGE_CHANNEL (new_conn->channel, new_conn->main_mch);
	    ch_close (new_conn->channel);
	    ch_close (new_conn->datagram_channel);
	    _conn_free_connection (new_conn);
	    a_func_abort (function_name, "Could not manage channel");
	    return (NULL);
	}
    }
    ++*new_conn->connection_count;
    if (protocol_info->open_func != NULL)
    {
	if ( !(*protocol_info->open_func) (new_conn, &new_conn->info) )
	{
	    UNMANAGE_CHANNEL (new_conn->channel, new_conn->main_mch);
	    ch_close (new_conn->channel);
	    --*new_conn->connection_count;
	    _conn_free_connection (new_conn);
	    return (NULL);
	}
    }
    new_conn->list_start = &first_client_connection;
    new_conn->list_end = &last_client_connection;
    /*  Add new connection to end of list  */
    new_conn->prev = last_client_connection;
    if (first_client_connection == NULL) first_client_connection = new_conn;
    else last_client_connection->next = new_conn;
    last_client_connection = new_conn;
    /*  Drain any input on new connection  */
    if ( ( bytes = ch_get_bytes_readable (new_conn->channel) ) < 0 )
    {
	exit (RV_SYS_ERROR);
    }
    if (bytes > 0)
    {
	if (new_conn->read_func == NULL)
	{
	    fprintf (stderr, "Input on new connection not being read\n");
	    a_prog_bug (function_name);
	}
	if ( !_conn_client_connection_input_func (new_conn->channel,
						  (void **) &new_conn) )
	{
	    _conn_dealloc_connection (new_conn, NULL);
	    UNMANAGE_CHANNEL (new_conn->channel, new_conn->main_mch);
	    ch_close (new_conn->channel);
	    return (NULL);
	}
    }
    if (cm_channel != NULL)
    {
	/*  Connected to Connection Management tool  */
	if ( !pio_write32 (cm_channel, CM_LIB_NEW_CONNECTION) )
	{
	    fprintf (stderr, "Error writing command value\n");
	    exit (RV_WRITE_ERROR);
	}
	if ( !pio_write_string (cm_channel, protocol_name) )
	{
	    fprintf (stderr, "%s: error writing protocol name\n",
		     function_name);
	    exit (RV_WRITE_ERROR);
	}
	if ( !pio_write32 (cm_channel, host_addr) )
	{
	    fprintf (stderr, "%s: error writing host Internet address\n",
		     function_name);
	    exit (RV_WRITE_ERROR);
	}
	if ( !pio_write32 (cm_channel, (unsigned long) port_number) )
	{
	    fprintf (stderr, "%s: error writing port number\n",
		     function_name);
	    exit (RV_WRITE_ERROR);
	}
	if ( !pio_write32 (cm_channel,
			   (unsigned long) new_conn & 0xffffffff) )
	{
	    fprintf (stderr, "%s: error writing Connection ID\n",
		     function_name);
	    exit (RV_WRITE_ERROR);
	}
	if ( !ch_flush (cm_channel) )
	{
	    fprintf (stderr, "%s: error flushing channel\t%s\n",
		     function_name, ERRSTRING);
	    exit (RV_WRITE_ERROR);
	}
    }
    return (new_conn);
}   /*  End Function conn_attempt_connection  */

/*PUBLIC_FUNCTION*/
flag conn_close (Connection connection)
/*  [SUMMARY] Close connection.
    [PURPOSE] This routine will close a connection. This will cause the closure
    callback routine registered to be executed.
    <connection> The connection object.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag ok = TRUE;
    Channel channel, datagram_channel;
    KManagedChannel main_mch, datagram_mch;
    static char function_name[] = "conn_close";

    VERIFY_USED (connection);
    if (unmanage_channel == NULL)
    {
	fprintf (stderr, "Channel managers not registered\n");
	a_prog_bug (function_name);
    }
    channel = connection->channel;
    datagram_channel = connection->datagram_channel;
    main_mch = connection->main_mch;
    datagram_mch = connection->datagram_mch;
    _conn_dealloc_connection (connection, NULL);
    UNMANAGE_CHANNEL (channel, main_mch);
    if ( !ch_close (channel) ) ok = FALSE;
    if (datagram_channel != NULL)
    {
	UNMANAGE_CHANNEL (datagram_channel, datagram_mch);
	if ( !ch_close (datagram_channel) ) ok = FALSE;
    }
    return (ok);
}   /*  End Function conn_close  */

/*PUBLIC_FUNCTION*/
flag conn_become_server (unsigned int *port_number, unsigned int retries)
/*  [SUMMARY] Become a Karma server.
    [PURPOSE] This routine will allocate a Karma port for the module so that it
    can operate as a server (able to receive network connections).
    <port_number> The port number to allocate. The routine will write the
    actual port number allocated to this address. This must point to an address
    which lies on an <<int>> boundary.
    <retries> The number of succsessive port numbers to attempt to allocate
    before giving up. If this is 0, then the routine will give up immediately
    if the specified port number is in use.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int num_docks;
    unsigned int dock_count;
    flag return_value = FALSE;
    Channel *port;
    static char function_name[] = "conn_become_server";

    if (manage_channel == NULL)
    {
	fprintf (stderr, "Channel managers not registered\n");
	a_prog_bug (function_name);
    }
    if (!ran_become_server)
    {
	/*  Set up internal protocol support  */
	conn_register_server_protocol("ping_server", 0, 0,
				      ( flag (*) () ) NULL,
				      _conn_respond_to_ping_server_from_client,
				      ( void (*) () ) NULL);
	conn_register_server_protocol ("server_exit", 0, 1,
				       _conn_register_server_exit,
				       ( flag (*) () ) NULL,
				       ( void (*) () ) NULL);
	ran_become_server = TRUE;
    }
    if ( ( port = ch_alloc_port (port_number, retries, &num_docks) ) == NULL )
    {
	a_func_abort (function_name, "Error becoming server");
	return (FALSE);
    }
    for (dock_count = 0; dock_count < num_docks; ++dock_count)
    {
	if ( !(*manage_channel) (port[dock_count], (void *) NULL,
				 _conn_dock_input_func,
				 ( void (*) () ) NULL,
				 ( flag (*) () ) NULL,
				 ( flag (*) () ) NULL) )
	{
	    fprintf (stderr, "Error managing dock: %u\n", dock_count);
	    ch_close (port[dock_count]);
	}
	else
	{
	    /*  At least one dock was managed: call this success  */
	    return_value = TRUE;
	}
    }
    m_free (port);
    if ( return_value && (cm_channel != NULL) )
    {
	/*  Have a dock and are connected to Connection Management tool  */
	if ( !pio_write32 (cm_channel, CM_LIB_PORT_NUMBER) )
	{
	    fprintf (stderr, "%s: error writing command value\n",
		     function_name);
	    exit (RV_WRITE_ERROR);
	}
	if ( !pio_write32 (cm_channel, (unsigned long) *port_number) )
	{
	    fprintf (stderr, "%s: error writing port number\n", function_name);
	    exit (RV_WRITE_ERROR);
	}
	if ( !ch_flush (cm_channel) )
	{
	    fprintf (stderr, "%s: error flushing channel\t%s\n",
		     function_name, ERRSTRING);
	    exit (RV_WRITE_ERROR);
	}
    }
    if ( gls_test_available () )
    {
	int tmp = *port_number;

	gls_send_plain_event ("karmaPortNumber", (CONST char *) &tmp,
			      K_INT, 1, 0, (CONST int *) NULL);
    }
    return (return_value);
}   /*  End Function conn_become_server  */

/*PUBLIC_FUNCTION*/
unsigned int conn_get_num_serv_connections (CONST char *protocol_name)
/*  [SUMMARY] Get number of server connections.
    [PURPOSE] This routine will get the number of connections to this server
    with a specified protocol.
    <protocol_name> The protocol name.
    [RETURNS] The number of current connections. A return value of 0 indicates
    that the protocol is not supported or there are no current connections.
*/
{
    struct serv_protocol_list_type *serv_protocol_info;
    /*static char function_name[] = "conn_get_num_serv_connections";*/

    if ( ( serv_protocol_info = _conn_get_serv_protocol_info (protocol_name) )
	 == NULL )
    {
	/*  Protocol is not supported  */
	return (0);
    }
    return (serv_protocol_info->connection_count);
}   /*  End Function conn_get_num_serv_connections  */

/*PUBLIC_FUNCTION*/
unsigned int conn_get_num_client_connections (CONST char *protocol_name)
/*  [SUMMARY] Get number of client connections.
    [PURPOSE] This routine will get the number of connections from this client
    with a specified protocol.
    <protocol_name> The protocol name.
    [RETURNS] The number of current connections. A return value of 0 indicates
    that the protocol is not supported or there are no current connections.
*/
{
    struct client_protocol_list_type *client_protocol_info;
    /*static char function_name[] = "conn_get_num_client_connections";*/

    if ( ( client_protocol_info =
	   _conn_get_client_protocol_info (protocol_name) ) == NULL )
    {
	/*  Protocol is not supported  */
	return (0);
    }
    return (client_protocol_info->connection_count);
}   /*  End Function conn_get_num_client_connections  */

/*PUBLIC_FUNCTION*/
Connection conn_get_serv_connection (CONST char *protocol_name,
				     unsigned int number)
/*  [SUMMARY] Get a numbered server connection.
    [PURPOSE] This routine will get Nth connection to this server with a
    specified protocol. The first connection is numbered 0.
    <protocol_name> The protocol name.
    <number> The number of the connection to get.
    [RETURNS] The connection object on success, else NULL.
*/
{
    return _conn_get_numbered_connection (protocol_name, number,
					  first_serv_connection);
}   /*  End Function conn_get_serv_connection  */

/*PUBLIC_FUNCTION*/
Connection conn_get_client_connection (CONST char *protocol_name,
				       unsigned int number)
/*  [SUMMARY] Get a numbered client connection.
    [PURPOSE] This routine will get Nth connection from this client with a
    specified protocol. The first connection is numbered 0.
    <protocol_name> The protocol name.
    <number> The number of the connection to get.
    [RETURNS] The connection object on success, else NULL.
*/
{
    return _conn_get_numbered_connection (protocol_name, number,
					  first_client_connection);
}   /*  End Function conn__conn_get_client_connection  */

/*PUBLIC_FUNCTION*/
void *conn_get_connection_info (Connection connection)
/*  [SUMMARY] Get the arbitrary information for a connection.
    <connection> The connection object. The routine aborts the process if the
    connection is not valid.
    [RETURNS] A pointer to the arbitrary information. This may be NULL.
*/
{
    static char function_name[] = "conn_get_connection_info";

    VERIFY_USED (connection);
    return (connection->info);
}   /*  End Function conn_get_connection_info  */

/*PUBLIC_FUNCTION*/
flag conn_controlled_by_cm_tool ()
/*  [SUMMARY] Check if controlled by CM tool.
    [PURPOSE] This routine will determine if the module is being controlled by
    the Connection Management tool.
    [RETURNS] TRUE if the module is controlled, else FALSE
*/
{
    return (cm_channel == NULL) ? FALSE : TRUE;
}   /*  End Function conn_controlled_by_cm_tool  */

/*PUBLIC_FUNCTION*/
char *conn_get_connection_module_name (Connection connection)
/*  [SUMMARY] Get the name of the module at the other end of a connection.
    <connection> The connection object. The routine aborts the process if the
    connection is not valid.
    [RETURNS] A pointer to the module name.
*/
{
    static char function_name[] = "conn_get_connection_module_name";

    VERIFY_USED (connection);
    if (connection->module_name == NULL)
    {
	fprintf (stderr, "Invalid connection module_name\n");
	a_prog_bug (function_name);
    }
    return (connection->module_name);
}   /*  End Function conn_get_connection_module_name  */

/*EXPERIMENTAL_FUNCTION*/
unsigned long conn_get_connection_remote_addr (Connection connection)
/*  [SUMMARY] Get the address of the machine at the other end of a connection.
    <connection> The connection object. The routine aborts the process if the
    connection is not valid.
    [RETURNS] The IP address.
*/
{
    static char function_name[] = "conn_get_connection_remote_addr";

    VERIFY_USED (connection);
    return (connection->inet_addr);
}   /*  End Function conn_get_connection_remote_addr  */

/*PUBLIC_FUNCTION*/
void conn_register_cm_quiescent_func ( void (*func) () )
/*  [SUMMARY] Register callback for quiescence.
    [PURPOSE] This routine will register a callback function to be called when
    the Connection Management tool or shell is quiescent (ie. all modules have
    started and all initial connections made). The function is ONLY called if
    the module is running under the Connection Management tool or shell.
    [NOTE] Only one callback may be registered.
    <quiescent_func> The routine which will be called when the Connection
    Management tool or shell is quiescent. The prototype function is
    [<conn_PROTO_quiescent_func>].
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "conn_register_cm_quiescent_func";

    if ( !conn_controlled_by_cm_tool () )
    {
	fprintf (stderr, "Not controlled by CM tool or shell\n");
	a_prog_bug (function_name);
    }
    if (quiescent_function != NULL)
    {
	fprintf (stderr, "Quiescent callback already registered\n");
	a_prog_bug (function_name);
    }
    quiescent_function = func;
}   /*  End Function conn_register_cm_quiescent_func  */

/*PUBLIC_FUNCTION*/
char **conn_extract_protocols ()
/*  [SUMMARY] Get supported protocols.
    [PURPOSE] This routine will extract all the supported protocols and
    produces a dynamically allocated array of strings which may be later
    displayed.
    [RETURNS] A pointer to a dynamically allocated array of dynamically
    allocated strings on success, else NULL. The end of the array is marked
    with a NULL string pointer.
*/
{
    unsigned int num_protocols;
    char **strings;
    char txt[STRING_LENGTH];
    char tmp1[STRING_LENGTH];
    char tmp2[STRING_LENGTH];
    struct client_protocol_list_type *client_entry;
    struct serv_protocol_list_type *serv_entry;
    static char function_name[] = "conn_extract_protocols";

    /*  Find total number of protocols  */
    /*  Loop through all server protocols  */
    for (serv_entry = serv_protocol_list, num_protocols = 0;
	 serv_entry != NULL;
	 serv_entry = serv_entry->next, ++num_protocols);
    /*  Loop through all client protocols  */
    for (client_entry = client_protocol_list; client_entry != NULL;
	 client_entry = client_entry->next)
    {
	if (_conn_get_serv_protocol_info (client_entry->protocol_name)
	    == NULL )
	{
	    /*  Client protocol not in the server list  */
	    ++num_protocols;
	}
    }
    if ( ( strings = m_alloc ( (num_protocols + 3) * sizeof *strings ) )
	 == NULL )
    {
	m_error_notify (function_name, "array of string pointers");
	return (NULL);
    }
    sprintf (txt, "Raw protocol revision: %ld", REVISION_NUMBER);
    if ( ( strings[0] = st_dup (txt) ) == NULL )
    {
	m_error_notify (function_name, "first string");
	m_free (strings);
	return (NULL);
    }
    if ( ( strings[1] = st_dup
	  ("Protocol_name              serv_max  #serv ver     client_max #client ver")
	  ) == NULL )
    {
	m_error_notify (function_name, "first string");
	m_free (strings[0]);
	m_free (strings);
	return (NULL);
    }
    strings[2] = NULL;
    /*  Process the server protocol list again  */
    for (serv_entry = serv_protocol_list, num_protocols = 0;
	 serv_entry != NULL;
	 serv_entry = serv_entry->next, ++num_protocols)
    {
	if (serv_entry->max_connections == 0)
	    strcpy (tmp1, "unlimited");
	else
	    sprintf (tmp1, "%u", serv_entry->max_connections);
	client_entry =
	    _conn_get_client_protocol_info (serv_entry->protocol_name);
	if (client_entry == NULL)
	{
	    sprintf (txt, "%-26s %-9s %-5u %-3u     -          -       -",
		     serv_entry->protocol_name,
		     tmp1,
		     serv_entry->connection_count,
		     serv_entry->version);
	}
	else
	{
	    if (client_entry->max_connections == 0)
		strcpy (tmp2, "unlimited");
	    else
		sprintf (tmp2, "%u", client_entry->max_connections);
	    sprintf (txt, "%-26s %-9s %-5u %-3u     %-10s %-7u %u",
		     serv_entry->protocol_name,
		     tmp1,
		     serv_entry->connection_count,
		     serv_entry->version,
		     tmp2,
		     client_entry->connection_count,
		     client_entry->version);
	}
	if ( ( strings[num_protocols + 2] = st_dup (txt) ) == NULL )
	{
	    m_error_notify (function_name, "protocol string information");
	    for (num_protocols = 0; strings[num_protocols] != NULL;
		 ++num_protocols) m_free (strings[num_protocols]);
	    m_free (strings);
	    return (NULL);
	}
	strings[num_protocols + 3] = NULL;
    }
    /*  Process the client protocol list again  */
    for (client_entry = client_protocol_list; client_entry != NULL;
	 client_entry = client_entry->next)
    {
	if (_conn_get_serv_protocol_info (client_entry->protocol_name)
	    != NULL )
	{
	    /*  Client protocol is in the server list  */
	    continue;
	}
	if (client_entry->max_connections == 0)
	    strcpy (tmp2, "unlimited");
	else
	    sprintf (tmp2, "%u", client_entry->max_connections);
	sprintf (txt, "%-26s -         -     -       %-10s %-7u %u",
		 client_entry->protocol_name,
		 tmp2,
		 client_entry->connection_count,
		 client_entry->version);
	if ( ( strings[num_protocols + 2] = st_dup (txt) ) == NULL )
	{
	    m_error_notify (function_name, "protocol string information");
	    for (num_protocols = 0; strings[num_protocols] != NULL;
		 ++num_protocols) m_free (strings[num_protocols]);
	    m_free (strings);
	    return (NULL);
	}
	strings[++num_protocols + 2] = NULL;
    }
    return (strings);
}   /*  End Function conn_extract_protocols  */

/*EXPERIMENTAL_FUNCTION*/
void conn_show_connections ()
/*  [SUMMARY] Show all connections.
    [RETURNS] Nothing.
*/
{
    Connection curr;
    struct karma_qos qos;
    char txt[STRING_LENGTH];
    /*static char function_name[] = "conn_show_connections";*/

    /*  First show all the server connections  */
    fprintf (stderr, "Server connections:\n");
    for (curr = first_serv_connection; curr != NULL; curr = curr->next)
    {
	if ( !r_gethostbyaddr (txt, STRING_LENGTH, curr->inet_addr) ) return;
	fprintf (stderr, "  Protocol: \"%s\" Origin: \"%s\"  Channel: %p\n",
		 curr->protocol_name, txt, curr->channel);
	fprintf (stderr, "    Type: %s", type_strings[curr->type]);
	if (curr->datagram_channel == NULL)
	{
	    putc ('\n', stderr);
	    continue;
	}
	fprintf (stderr, "  Datagram Channel: %p\n", curr->datagram_channel);
	ch_get_qos (curr->datagram_channel, &qos);
	switch (curr->type)
	{
	  case CONNECTION_TYPE_UDP:
	    fprintf (stderr, "    sdu=%d\n", qos.tx.sdu_max);
	    break;
	  case CONNECTION_TYPE_ATMSVC:
	    _conn_show_atm_tfparm (&qos.tx, "tx");
	    _conn_show_atm_tfparm (&qos.rx, "rx");
	    break;
	}
    }
    /*  Now show all the client connections  */
    fprintf (stderr, "Client connections:\n");
    for (curr = first_client_connection; curr != NULL; curr = curr->next)
    {
	if ( !r_gethostbyaddr (txt, STRING_LENGTH, curr->inet_addr) ) return;
	fprintf (stderr,
		 "  Protocol: \"%s\" Destination: \"%s\"  Channel: 0x%lx\n",
		 curr->protocol_name, txt, (unsigned long) curr->channel);
	fprintf (stderr, "    Type: %s", type_strings[curr->type]);
	if (curr->datagram_channel == NULL)
	{
	    putc ('\n', stderr);
	    continue;
	}
	fprintf (stderr, "  Datagram Channel: %p\n", curr->datagram_channel);
	ch_get_qos (curr->datagram_channel, &qos);
	switch (curr->type)
	{
	  case CONNECTION_TYPE_UDP:
	    fprintf (stderr, "    sdu=%d\n", qos.tx.sdu_max);
	    break;
	  case CONNECTION_TYPE_ATMSVC:
	    _conn_show_atm_tfparm (&qos.tx, "tx");
	    _conn_show_atm_tfparm (&qos.rx, "rx");
	    break;
	}
    }
}   /*  End Function conn_show_connections  */


/*  Private functions follow  */

/*  Functions to process input on server connections  */

static flag _conn_dock_input_func (Channel dock, void **info)
/*   This routine is called when new input occurs on a dock channel (ie. there
     is a connection request on a dock).
     The channel object is given by  dock  .
     An arbitrary pointer may be written to the storage pointed to by  info  .
     The pointer written here will persist until the channel is unmanaged
     (or a subsequent callback routine changes it).
     The routine returns TRUE if the connection is to remain managed,
     else it returns FALSE (indicating that the connection is to be
     unmanaged).
*/
{
    Connection new_conn;
    static char function_name[] = "_conn_dock_input_func";

    /*  Allocate new connection object  */
    if ( ( new_conn = _conn_alloc_connection () ) == NULL )
    {
	m_error_notify (function_name, "connection object");
	/*  Do not close dock  */
	return (TRUE);
    }
    if ( ( new_conn->channel = ch_accept_on_dock (dock, &new_conn->inet_addr) )
	 == NULL )
    {
	a_func_abort (function_name, "could not accept connection on dock");
	return (FALSE);
    }
    new_conn->type = CONNECTION_TYPE_UNDEFINED;
    new_conn->main_mch = (*manage_channel) (new_conn->channel,
					    (void *) new_conn,
					    _conn_serv_connection_input_func,
					    _conn_connection_close_func,
					    ( flag (*) () ) NULL,
					    ( flag (*) () ) NULL);
    if (new_conn->main_mch == NULL)
    {
	ch_close (new_conn->channel);
	_conn_free_connection (new_conn);
	a_func_abort (function_name, "could not manage channel");
	return (FALSE);
    }
    new_conn->list_start = &first_serv_connection;
    new_conn->list_end = &last_serv_connection;
    /*  Add new connection to end of list  */
    new_conn->prev = last_serv_connection;
    if (first_serv_connection == NULL) first_serv_connection = new_conn;
    else last_serv_connection->next = new_conn;
    last_serv_connection = new_conn;
    return (TRUE);
}   /*  End Function _conn_dock_input_func  */

static flag _conn_serv_connection_input_func (Channel channel, void **info)
/*  This routine is called when new input occurs on a channel.
    The channel object is given by  channel  .
    An arbitrary pointer may be written to the storage pointed to by  info
    The pointer written here will persist until the channel is unmanaged
    (or a subsequent callback routine changes it).
    The routine returns TRUE if the channel is to remain managed and
    open, else it returns FALSE (indicating that the channel is to be
    unmanaged and closed).
*/
{
    int bytes;
    unsigned int my_port, remote_port;
    unsigned long old_read_pos, new_read_pos, dummy, conn_type, rem_addr;
    double rv;
    Connection connection = (Connection) *info;
    struct serv_protocol_list_type *protocol_info;
    char randval[8], rrnd[8];
    static Channel my_atm_dock = NULL;
    static char my_atm_addr[STRING_LENGTH];
    static unsigned int my_atm_addr_len = 0;
    static char function_name[] = "_conn_serv_connection_input_func";

    VERIFY_USED (connection);
    if (!connection->verified_raw)
    {
	if ( !_conn_verify_raw (connection) ) return (FALSE);
	return (TRUE);
    }
    /*  At this point we know that the other end wants to talk the Karma
	protocol  */
    if (!connection->verified_security)
    {
	if ( !_conn_verify_raw_security (connection) ) return (FALSE);
	return (TRUE);
    }
    /*  At this point we know any security requirements have been met for the
	raw protocol  */
    if (connection->protocol_name == NULL)
    {
	/*  No protocol yet: connection setup information needed  */
	if ( !_conn_verify_connection (connection) )
	{
	    /*  Connection was not authorised  */
	    ch_flush (channel);
	    return (FALSE);
	}
	return (TRUE);
    }
    if (!connection->verified_protocol_security)
    {
	if ( !_conn_verify_protocol_security (connection) ) return (FALSE);
	return (TRUE);
    }
    if (connection->module_name == NULL)
    {
	/*  Read module name from client  */
	if ( ( connection->module_name = pio_read_string (channel, NULL) )
	     == NULL )
	{
	    fprintf (stderr, "Error reading module name from connection\n");
	    return (FALSE);
	}
	protocol_info =
	    _conn_get_serv_protocol_info (connection->protocol_name);
	if ( !pio_read32 (channel, &conn_type) )
	{
	    fprintf (stderr, "%s: error reading connection type\n",
		     function_name);
	    return (FALSE);
	}
	connection->type = conn_type;
	switch (conn_type)
	{
	  case CONNECTION_TYPE_STREAM:
	    break;
	  case CONNECTION_TYPE_UDP:
	    if ( ( connection->datagram_channel = ch_udp_alloc (&my_port) )
		 == NULL )
	    {
		fprintf (stderr, "%s: error creating UDP socket\n",
			 function_name);
		return (FALSE);
	    }
	    if ( !pio_write32 (channel, my_port) )
	    {
		fprintf (stderr, "%s: error writing UDP port number\n",
			 function_name);
		return (FALSE);
	    }
	    /*  Need to read message to determine client port number, because
		what the client thinks the port number is may be different to
		what comes down the wire (thanks to NAT)  */
	    rv = n_uniform ();  /*  Primitive authentication  */
	    m_copy (randval, &rv,
		    (sizeof rv < sizeof randval) ? sizeof rv : sizeof randval);
	    if (ch_write (channel, randval, sizeof randval) < sizeof randval)
	    {
		fprintf (stderr, "%s: error writing random number\n",
			 function_name);
		return (FALSE);
	    }
	    if ( !ch_flush (channel) )
	    {
		fprintf (stderr, "%s: error flushing\t%s\n",
			 function_name, ERRSTRING);
		return (FALSE);
	    }
	    while (TRUE)
	    {
		bytes = r_recvfrom
		    (ch_get_descriptor (connection->datagram_channel),
		     rrnd, sizeof rrnd, &rem_addr, &remote_port);
		if (bytes < sizeof rrnd)
		{
		    fprintf (stderr,
			     "%s: error reading message, got : %d bytes\t%s\n",
			     function_name, bytes, ERRSTRING);
		    return (FALSE);
		}
		if (rem_addr != connection->inet_addr)
		{
		    fprintf (stderr,
			     "%s: spurious message on UDP port from: %lx (expected: %lx)\n",
			     function_name, rem_addr, connection->inet_addr);
		    continue;
		}
		if ( !m_cmp (randval, rrnd, sizeof randval) )
		{
		    fprintf (stderr, "%s: spurious message data on UDP port\n",
			     function_name);
		    continue;
		}
		break;
	    }
	    if (ch_udp_connect (connection->datagram_channel,
				connection->inet_addr, remote_port) == NULL)
	    {
		fprintf (stderr, "%s: error creating UDP socket\n",
			 function_name);
		return (FALSE);
	    }
	    /*  Send acknowledgement  */
	    ch_write (channel, (CONST char *) &randval, 1);
	    ch_flush (channel);
	    break;
	  case CONNECTION_TYPE_ATMSVC:
	    if (my_atm_dock == NULL)
	    {
		my_atm_addr_len = STRING_LENGTH;
		my_atm_dock = ch_atm_alloc (my_atm_addr, &my_atm_addr_len);
	    }
	    if (my_atm_dock == NULL)
	    {
		fprintf (stderr, "%s: error creating ATM SVC dock\n",
			 function_name);
		return (FALSE);
	    }
	    if ( !pio_write32 (channel, my_atm_addr_len) )
	    {
		fprintf (stderr, "%s: error writing ATM address length\n",
			 function_name);
		return (FALSE);
	    }
	    if (ch_write (channel, my_atm_addr, my_atm_addr_len)
		< my_atm_addr_len)
	    {
		fprintf (stderr, "%s: error writing ATM address\n",
			 function_name);
		return (FALSE);
	    }
	    if ( !ch_flush (channel) )
	    {
		fprintf (stderr, "%s: error flusing ATM address information\n",
			 function_name);
		return (FALSE);
	    }
	    if ( ( connection->datagram_channel =
		   ch_accept_on_dock (my_atm_dock, &dummy) ) == NULL )
	    {
		fprintf (stderr, "%s: error accepting ATM SVC connection\n",
			 function_name);
		s_check_for_int ();
		return (FALSE);
	    }
	    break;
	  default:
	    fprintf (stderr, "%s: illegal connection type: %lu\n",
		     function_name, conn_type);
	    return (FALSE);
	    /*break;*/
	}
	if (connection->datagram_channel != NULL)
	{
	    connection->datagram_mch =
		(*manage_channel) (connection->datagram_channel,
				   (void *) connection,
				   _conn_serv_connection_input_func,
				   _conn_connection_close_func,
				   ( flag (*) () ) NULL,
				   ( flag (*) () ) NULL);
	    if (connection->datagram_mch == NULL)
	    {
		a_func_abort (function_name, "could not manage channel");
		return (FALSE);
	    }
	}
	if (protocol_info->open_func != NULL)
	{
	    /*  Open function registered  */
	    if ( !(*protocol_info->open_func) (connection, &connection->info) )
	    {
		/*  Failure  */
		return (FALSE);
	    }
	}
	connection->read_func = protocol_info->read_func;
	connection->close_func = protocol_info->close_func;
	return (TRUE);
    }
    if ( ( bytes = ch_get_bytes_readable (channel) ) < 0 )
    {
	exit (RV_SYS_ERROR);
    }
    if ( (bytes > 0) && (connection->read_func == NULL) )
    {
	fprintf (stderr,
		 "Input on \"%s\" connection not being read (no callback)\n",
		 connection->protocol_name);
	a_prog_bug (function_name);
    }
    if ( !ch_tell (channel, &old_read_pos, &dummy) )
    {
	exit (RV_SYS_ERROR);
    }
    while (bytes > 0)
    {
	if ( !(*connection->read_func) (connection, &connection->info,
					channel) )
	{
	    return (FALSE);
	}
	if ( !ch_tell (channel, &new_read_pos, &dummy) )
	{
	    exit (RV_SYS_ERROR);
	}
	if (new_read_pos <= old_read_pos)
	{
	    fprintf (stderr,
		     "Connection read callback for protocol: \"%s\" not draining\n",
		     connection->protocol_name);
	    fprintf (stderr,
		     "with %d bytes to read after a total of %lu bytes read\n",
		     bytes, new_read_pos);
	    a_prog_bug (function_name);
	}
	old_read_pos = new_read_pos;
	if ( ( bytes = ch_get_bytes_readable (channel) ) < 0 )
	{
	    exit (RV_SYS_ERROR);
	}
    }
    return (TRUE);
}   /*  End Function _conn_serv_connection_input_func  */

static flag _conn_verify_raw (Connection connection)
/*  [PURPOSE] This routine will verify a raw Karma connection to ensure that
    the magic number and revision number are correct. It will then setup any
    required security.
    <connection> The connection.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int bytes;
    unsigned long magic_number;
    unsigned long revision_number;
    struct auth_password_list_type *authinfo;
    static char function_name[] = "_conn_verify_raw";

    VERIFY_USED (connection);
    /*  Check if enough data on connection  */
    if ( ( bytes = ch_get_bytes_readable (connection->channel) ) < 0 )
    {
	exit (RV_SYS_ERROR);
    }
    if (bytes < SETUP_MESSAGE_LENGTH)
    {
	/*  Not enough info was sent  */
	pio_write_string (connection->channel, "Bad setup message length");
	fprintf (stderr, "Only: %d bytes of connection setup information ",
		 bytes);
	fprintf (stderr, "sent\n%d bytes are required: connection closed\n",
		 SETUP_MESSAGE_LENGTH);
	return (FALSE);
    }
    /*  Read connection setup information  */
    if ( !pio_read32 (connection->channel, &magic_number) )
    {
	a_func_abort (function_name,
		      "Error reading magic number from connection");
	return (FALSE);
    }
    /*  Check magic number  */
    if (magic_number != CONN_MAGIC_NUMBER)
    {
	/*  Bad magic number  */
	pio_write_string (connection->channel, "Bad magic number");
	fprintf (stderr,
		 "WARNING: Connection attempted with bad magic number: %lu\n",
		 magic_number);
	fprintf (stderr, "Someone may be trying to stuff around\n");
	return (FALSE);
    }
    /*  Get revision number  */
    if ( !pio_read32 (connection->channel, &revision_number) )
    {
	a_func_abort (function_name,
		      "Error reading revision number from connection");
	return (FALSE);
    }
    /*  Check revision number  */
    if (revision_number != REVISION_NUMBER)
    {
	/*  Bad magic number  */
	pio_write_string (connection->channel, "Bad revision number");
	return (FALSE);
    }
    connection->verified_raw = TRUE;
    authinfo = _conn_get_authinfo (RAW_PROTOCOL);
    if (authinfo == NULL)
    {
	/*  No security requirements  */
	connection->verified_security = TRUE;
	return (TRUE);
    }
    if (authinfo->security_type == SECURITY_TYPE_DROP_ENCRYPTION)
    {
	a_func_abort (function_name,
		      "security type: \"drop-encryption\" not valid for RAW_PROTOCOL");
	exit (RV_BAD_DATA);
    }
    /*  Setup security requirements  */
    if ( ( connection->pri_encryption_converter =
	   _conn_get_encryption (connection->channel, authinfo, TRUE,
				 connection->idea_block) ) == NULL )
    {
	return (FALSE);
    }
    if ( !ch_flush (connection->channel) ) return (FALSE);
    /*  At some point the client will send some verification data  */
    connection->verified_security = FALSE;
    return (TRUE);
}   /*  End Function _conn_verify_raw  */

static flag _conn_verify_raw_security (Connection connection)
/*  [PURPOSE] This routine will verify a raw Karma connection to ensure that
    the connection is authorised.
    <connection> The connection.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int bytes;
    struct auth_password_list_type *authinfo;
    char idea_blocks[EN_IDEA_BLOCK_SIZE * 2];
    static char function_name[] = "_conn_verify_raw_security";

    VERIFY_USED (connection);
    /*  Check if enough data on connection  */
    if ( ( bytes = ch_get_bytes_readable (connection->channel) ) < 0 )
    {
	exit (RV_SYS_ERROR);
    }
    if (bytes < EN_IDEA_BLOCK_SIZE * 2)
    {
	/*  Not enough info was sent  */
	pio_write_string (connection->channel,"Bad IDEA block message length");
	fprintf (stderr, "Only: %d bytes of connection IDEA block data ",
		 bytes);
	fprintf (stderr, "sent\n%d bytes are required: connection closed\n",
		 EN_IDEA_BLOCK_SIZE * 2);
	return (FALSE);
    }
    if (ch_read (connection->channel, idea_blocks, EN_IDEA_BLOCK_SIZE * 2) <
	EN_IDEA_BLOCK_SIZE * 2) return (FALSE);
    if ( !m_cmp (idea_blocks + EN_IDEA_BLOCK_SIZE, connection->idea_block,
		 EN_IDEA_BLOCK_SIZE) )
    {
	/*  Sorry, I'm not convinced  */
	fprintf (stderr,
		 "%s: authorisation missmatch: rejected raw conection\n",
		 function_name);
	return (FALSE);
    }
    if (ch_write (connection->channel, idea_blocks, EN_IDEA_BLOCK_SIZE) <
	EN_IDEA_BLOCK_SIZE) return (FALSE);
    if ( !ch_flush (connection->channel) ) return (FALSE);
    authinfo = _conn_get_authinfo (RAW_PROTOCOL);
    switch (authinfo->security_type)
    {
      case SECURITY_TYPE_KEY:
	ch_unregister_converter (connection->pri_encryption_converter);
	connection->pri_encryption_converter = NULL;
	break;
      case SECURITY_TYPE_IDEA:
      case SECURITY_TYPE_PGPuserID_IDEA:
	if ( ch_test_for_local_connection (connection->channel) )
	{
	    ch_unregister_converter (connection->pri_encryption_converter);
	    connection->pri_encryption_converter = NULL;
	}
	break;
      case SECURITY_TYPE_DROP_ENCRYPTION:
	a_func_abort (function_name,
		      "security type: \"drop-encryption\" not valid for RAW_PROTOCOL\n");
	exit (RV_BAD_DATA);
/*
	break;
*/
    }
    connection->verified_security = TRUE;
    return (TRUE);
}   /*  End Function _conn_verify_raw_security  */

static flag _conn_verify_connection (Connection connection)
/*  This routine will verify whether a connection is permitted or not.
    The connection object must be given by  connection  .
    The routine returns TRUE on success, else it returns FALSE.
*/
{
    ChConverter converter;
    int bytes;
    unsigned long protocol_version;
    struct serv_protocol_list_type *protocol_info;
    struct auth_password_list_type *authinfo;
    char protocol_buffer[PROTOCOL_NAME_LENGTH + 1];
    extern char module_name[STRING_LENGTH + 1];
    static char function_name[] = "_conn_verify_connection";

    VERIFY_USED (connection);
    if ( ( bytes = ch_get_bytes_readable (connection->channel) ) < 0 )
    {
	exit (RV_SYS_ERROR);
    }
    if (bytes < PROTOCOL_DEFINITION_MESSAGE_LENGTH)
    {
	/*  Not enough info was sent  */
	pio_write_string (connection->channel,
			  "Bad protocol definition message length");
	fprintf (stderr, "Only: %d bytes of connection setup information ",
		 bytes);
	fprintf (stderr, "sent\n%d bytes are required: connection closed\n",
		 PROTOCOL_DEFINITION_MESSAGE_LENGTH);
	return (FALSE);
    }
    /*  Read protocol name  */
    if (ch_read (connection->channel, protocol_buffer, PROTOCOL_NAME_LENGTH)
	< PROTOCOL_NAME_LENGTH)
    {
	fprintf (stderr, "Error reading protocol name from connection\n");
	return (FALSE);
    }
    protocol_buffer[PROTOCOL_NAME_LENGTH] = '\0';
    /*  Get protocol version number  */
    if ( !pio_read32 (connection->channel, &protocol_version) )
    {
	fprintf (stderr,
		 "Error reading protocol version number from connection\n");
	return (FALSE);
    }
    /*  Compare protocol number with list of supported ones  */
    if ( ( protocol_info = _conn_get_serv_protocol_info (protocol_buffer) )
	 == NULL )
    {
	/*  Protocol is not supported  */
	/*  Write failure message back to client  */
	pio_write_string (connection->channel,
			  "Protocol is not supported by server");
	return (FALSE);
    }
    if (protocol_version != protocol_info->version)
    {
	pio_write_string (connection->channel, "Protocol version mismatch");
	return (FALSE);
    }
    /*  Protocol is supported: check if more connections allowed  */
    if ( ( protocol_info->max_connections > 0 ) &&
	( protocol_info->connection_count
	 >= protocol_info->max_connections ) )
    {
	pio_write_string (connection->channel,
			  "Connection limit reached for protocol");
	return (FALSE);
    }
    connection->protocol_name = protocol_info->protocol_name;
    connection->connection_count = &protocol_info->connection_count;
    ++protocol_info->connection_count;
    connection->verified_protocol_security = FALSE;
    authinfo = _conn_get_authinfo (protocol_buffer);
    if ( (authinfo != NULL) &&
	 (authinfo->security_type == SECURITY_TYPE_DROP_ENCRYPTION) )
    {
	if (connection->pri_encryption_converter != NULL)
	{
	    ch_unregister_converter (connection->pri_encryption_converter);
	    connection->pri_encryption_converter = NULL;
	}
	authinfo = NULL;
    }
    if (authinfo == NULL)
    {
	connection->verified_protocol_security = TRUE;
	/*  Write verification message back to client  */
	pio_write_string (connection->channel, (char *) NULL);
	/*  Send module name back to client  */
	pio_write_string (connection->channel, module_name);
	if ( !ch_flush (connection->channel) ) return (FALSE);
	return (TRUE);
    }
    /*  Setup security requirements  */
    if ( ( converter = _conn_get_encryption (connection->channel, authinfo,
					     TRUE, connection->idea_block) )
	 == NULL ) return (FALSE);
    if ( !ch_flush (connection->channel) ) return (FALSE);
    connection->pro_encryption_converter = converter;
    return (TRUE);
}   /*  End Function _conn_verify_connection  */

static flag _conn_verify_protocol_security (Connection connection)
/*  [PURPOSE] This routine will verify a Karma connection to ensure that
    the connection is authorised for the appropriate protocol.
    <connection> The connection.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int bytes;
    struct auth_password_list_type *authinfo;
    char idea_blocks[EN_IDEA_BLOCK_SIZE * 2];
    extern char module_name[STRING_LENGTH + 1];
    static char function_name[] = "_conn_verify_protocol_security";

    VERIFY_USED (connection);
    /*  Check if enough data on connection  */
    if ( ( bytes = ch_get_bytes_readable (connection->channel) ) < 0 )
    {
	exit (RV_SYS_ERROR);
    }
    if (bytes < EN_IDEA_BLOCK_SIZE * 2)
    {
	/*  Not enough info was sent  */
	pio_write_string (connection->channel,
			  "Bad IDEA block message length");
	fprintf (stderr, "Only: %d bytes of connection IDEA block data ",
		 bytes);
	fprintf (stderr, "sent\n%d bytes are required: connection closed\n",
		 EN_IDEA_BLOCK_SIZE * 2);
	return (FALSE);
    }
    if (ch_read (connection->channel, idea_blocks, EN_IDEA_BLOCK_SIZE * 2) <
	EN_IDEA_BLOCK_SIZE * 2) return (FALSE);
    if ( !m_cmp (idea_blocks + EN_IDEA_BLOCK_SIZE, connection->idea_block,
		 EN_IDEA_BLOCK_SIZE) )
    {
	/*  Client is not allowed  */
	fprintf (stderr,
		 "WARNING: connection attempt failed for protocol: %s\n",
		 connection->protocol_name);
	pio_write_string (connection->channel, "Connection not authorised");
	if ( !ch_flush (connection->channel) ) return (FALSE);
	return (FALSE);
    }
    if (ch_write (connection->channel, idea_blocks, EN_IDEA_BLOCK_SIZE) <
	EN_IDEA_BLOCK_SIZE) return (FALSE);
    if ( !ch_flush (connection->channel) ) return (FALSE);
    if (connection->pri_encryption_converter != NULL)
    {
	ch_unregister_converter (connection->pri_encryption_converter);
	connection->pri_encryption_converter = NULL;
    }
    authinfo = _conn_get_authinfo (connection->protocol_name);
    switch (authinfo->security_type)
    {
      case SECURITY_TYPE_KEY:
	ch_unregister_converter (connection->pro_encryption_converter);
	connection->pro_encryption_converter = NULL;
	break;
      case SECURITY_TYPE_IDEA:
      case SECURITY_TYPE_PGPuserID_IDEA:
	if ( ch_test_for_local_connection (connection->channel) )
	{
	    ch_unregister_converter (connection->pro_encryption_converter);
	    connection->pro_encryption_converter = NULL;
	}
	break;
    }
    connection->verified_protocol_security = TRUE;
    /*  Write verification message back to client  */
    pio_write_string (connection->channel, NULL);
    /*  Send module name back to client  */
    pio_write_string (connection->channel, module_name);
    if ( !ch_flush (connection->channel) ) return (FALSE);
    return (TRUE);
}   /*  End Function _conn_verify_protocol_security  */

static struct serv_protocol_list_type *_conn_get_serv_protocol_info
    (CONST char *protocol_name)
/*  This routine will search the list of supported server protocols for the
    protocol given by  protocol_name  .
    The routine will return a pointer to a protocol information structure if
    the protocol is found in the list, else it returns NULL.
*/
{
    struct serv_protocol_list_type *entry;

    entry = serv_protocol_list;
    while (entry != NULL)
    {
	if (strcmp (protocol_name, entry->protocol_name) == 0)
	{
	    /*  Found it!  */
	    return (entry);
	}
	entry = entry->next;
    }
    /*  Not found  */
    return (NULL);
}   /*  End Function _conn_get_serv_protocol_info  */

static struct client_protocol_list_type *_conn_get_client_protocol_info
    (CONST char *protocol_name)
/*  This routine will search the list of supported client protocols for the
    protocol given by  protocol_name  .
    The routine will return a pointer to a protocol information structure if
    the protocol is found in the list, else it returns NULL.
*/
{
    struct client_protocol_list_type *entry;

    entry = client_protocol_list;
    while (entry != NULL)
    {
	if (strcmp (protocol_name, entry->protocol_name) == 0)
	{
	    /*  Found it!  */
	    return (entry);
	}
	entry = entry->next;
    }
    /*  Not found  */
    return (NULL);
}   /*  End Function _conn_get_client_protocol_info  */

static void _conn_connection_close_func (Channel channel, void *info)
/*  This routine is called when a channel closes.
    The channel object is given by  channel  .
    The arbitrary pointer for the channel will be pointed to by  info  .
    Any unread buffered data in the channel will be lost upon closure. The
    call to this function is the last chance to read this buffered data.
    The routine returns nothing.
*/
{
    Connection connection = (Connection) info;
    Channel main_channel, datagram_channel;
    KManagedChannel main_mch, datagram_mch;
    static char function_name[] = "_conn_connection_close_func";

    /*  Verify connection  */
    VERIFY_USED (connection);
    main_channel = connection->channel;
    datagram_channel = connection->datagram_channel;
    main_mch = connection->main_mch;
    datagram_mch = connection->datagram_mch;
    _conn_dealloc_connection (connection, channel);
    if (datagram_channel == NULL) return;
    if (channel == datagram_channel)
    {
	UNMANAGE_CHANNEL (main_channel, main_mch);
	ch_close (main_channel);
    }
    else
    {
	UNMANAGE_CHANNEL (datagram_channel, datagram_mch);
	ch_close (datagram_channel);
    }
}   /*  End Function _conn_connection_close_func  */

static void _conn_dealloc_connection (Connection connection, Channel channel)
/*  [SUMMARY] Deallocate a connection object.
    <connection> The connection object.
    <channel> The channel that was remotely closed, causing the connection to
    be closed. This may be NULL, indicating that the connection was closed
    locally.
    [NOTE] The routine will NOT unmanage or close the channels associated with
    the connection.
    [NOTE] The close function registered for the connection will be called
    prior to deallocation.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "_conn_dealloc_connection";

    VERIFY_USED (connection);
    /*  First decrement connection count and remove from list,
	otherwise the close_func callback could query for a connection and
	get this one!  */
    if (connection->connection_count != NULL)
    {
	/*  Decrement count of connections for this protocol  */
	--*connection->connection_count;
    }
    /*  Remove connection from list  */
    if (connection->prev == NULL) *connection->list_start = connection->next;
    else connection->prev->next = connection->next;
    if (connection->next == NULL) *connection->list_end = connection->prev;
    else connection->next->prev = connection->prev;
    /*  Call registered closure routine if it exists  */
    if (connection->close_func != NULL)
	(*connection->close_func) (connection, connection->info, channel);
    if (connection->module_name != NULL) m_free (connection->module_name);
    if ( connection->client && (cm_channel != NULL) )
    {
	/*  Have client connection and connected to
	    Connection Management tool  */
	if ( !pio_write32 (cm_channel, CM_LIB_CONNECTION_CLOSED) )
	{
	    fprintf (stderr, "%s: error writing command value\n",
		     function_name);
	    exit (RV_WRITE_ERROR);
	}
	if ( !pio_write32 (cm_channel,
			   (unsigned long) connection & 0xffffffff) )
	{
	    fprintf (stderr, "%s: error writing Connection ID\n",
		     function_name);
	    exit (RV_WRITE_ERROR);
	}
	if ( !ch_flush (cm_channel) )
	{
	    fprintf (stderr, "%s: error flushing channel\t%s\n",
		     function_name, ERRSTRING);
	    exit (RV_WRITE_ERROR);
	}
    }
    m_clear (connection->idea_block, EN_IDEA_BLOCK_SIZE);
    _conn_free_connection (connection);
}   /*  End Function _conn_dealloc_connection  */

static struct auth_password_list_type *_conn_get_authinfo
    (CONST char *protocol)
/*  This routine will determine the authorisation required to gain access to a
    Karma server using the protocol given by  protocol  .
    The routine will write the security type to the storage pointed to by
    security_type  .
    The routine returns the authorisation information.
*/
{
#ifdef COMMUNICATIONS_AVAILABLE
    struct auth_password_list_type *entry;
    struct auth_password_list_type *password_list;
    /*static char function_name[] = "_conn_get_authinfo";*/

    password_list = _conn_get_password_list ();
    if (password_list == NULL)
    {
	/*  No list: no passwords required  */
	return (NULL);
    }
    for (entry = password_list; entry != NULL; entry = entry->next)
    {
	if (strcmp (protocol, entry->protocol_name) == 0)
	{
	    return (entry);
	}
    }
    return (NULL);

#else  /*  COMMUNICATIONS_AVAILABLE  */
    fprintf (stderr, "Operating system does not support communications\n");
    exit (RV_UNDEF_ERROR);
#endif  /*  COMMUNICATIONS_AVAILABLE  */
}   /*  End Function _conn_get_authinfo  */

static struct auth_password_list_type *_conn_get_password_list (void)
/*  [PURPOSE] This routine will process the authorisation file.
    [RETURNS] The password list.
*/
{
#ifdef COMMUNICATIONS_AVAILABLE
    Channel channel;
    flag need_bin_password;
    CONST char *home;
    CONST char *p, *security_name;
    char *password = NULL;
    struct auth_password_list_type *entry;
    char auth_file[STRING_LENGTH];
    char txt[STRING_LENGTH + 1];
    static flag read_list = FALSE;
    static struct auth_password_list_type *password_list = NULL;
    static char function_name[] = "_conn_get_password_list";

    if (read_list) return (password_list);
    read_list = TRUE;
    /*  Read in password list  */
    if ( ( home = r_getenv ("HOME") ) == NULL )
    {
	fprintf (stderr, "Error getting environment variable: HOME\n");
	/*  Return saying no passwords needed  */
	password_list = NULL;
	return (NULL);
    }
    sprintf (auth_file, "%s/.KARMAauthority", home);
    if (access (auth_file, R_OK) != 0)
    {
	/*  File not there  */
	password_list = NULL;
	/*  No passwords required  */
	return (NULL);
    }
    /*  Read file  */
    if ( ( channel = ch_open_file (auth_file, "r") ) == NULL )
    {
	fprintf (stderr, "Error opening file: %s\n", auth_file);
	exit (RV_CANNOT_OPEN);
    }
    while ( chs_get_line (channel, txt, STRING_LENGTH) )
    {
	/*  Not end of file  */
	need_bin_password = FALSE;
	if ( ( entry = m_alloc (sizeof *entry) ) == NULL )
	{
	    m_abort (function_name, "protocol password entry");
	}
	entry->next = NULL;
	if ( ( entry->protocol_name = ex_word (txt, &p) ) == NULL )
	{
	    /*  Error converting line  */
	    ch_close (channel);
	    fprintf (stderr, "Error in file: %s\n", auth_file);
	    exit (RV_BAD_DATA);
	}
	if ( ( security_name = ex_word (p, &p) ) == NULL )
	{
	    fprintf (stderr, "Missing security type\n");
	    ch_close (channel);
	    fprintf (stderr, "Error in file: %s\n", auth_file);
	    exit (RV_BAD_DATA);
	}
	if (strcmp (security_name, SECURITY_STRING_KEY) == 0)
	{
	    entry->security_type = SECURITY_TYPE_KEY;
	}
	else if (strcmp (security_name, SECURITY_STRING_IDEA) == 0)
	{
	    entry->security_type = SECURITY_TYPE_IDEA;
	}
	else if (strcmp (security_name, SECURITY_STRING_PGPuserID_IDEA) == 0)
	{
	    entry->security_type = SECURITY_TYPE_PGPuserID_IDEA;
	}
	else if (strcmp (security_name, SECURITY_STRING_DROP_ENCRYPTION) == 0)
	{
	    entry->security_type = SECURITY_TYPE_DROP_ENCRYPTION;
	}
	else
	{
	    fprintf (stderr, "Illegal security type: \"%s\"\n", security_name);
	    ch_close (channel);
	    fprintf (stderr, "Error in file: %s\n", auth_file);
	    exit (RV_BAD_DATA);
	}
	switch (entry->security_type)
	{
	  case SECURITY_TYPE_IDEA:
	  case SECURITY_TYPE_KEY:
	    _conn_setup_prng ();
	    need_bin_password = TRUE;
	    if ( ( password = ex_word (p, &p) ) == NULL )
	    {
		m_abort (function_name, "password");
	    }
	    if ( (int) strlen (password) / 2 <
		 EN_IDEA_KEY_SIZE + EN_IDEA_BLOCK_SIZE )
	    {
		fprintf (stderr, "IDEA password too short\n");
		ch_close (channel);
		fprintf (stderr, "Error in file: %s\n", auth_file);
		exit (RV_BAD_DATA);
	    }
	    break;
	  case SECURITY_TYPE_PGPuserID_IDEA:
	    _conn_setup_prng ();
	    if ( ( password = ex_word (p, &p) ) == NULL )
	    {
		m_abort (function_name, "password");
	    }
	    break;
	  case SECURITY_TYPE_DROP_ENCRYPTION:
	    password = NULL;
	    break;
	  default:
	    fprintf (stderr, "Illegal security type: %u\n",
		     entry->security_type);
	    a_prog_bug (function_name);
	    break;
	}
	if (need_bin_password)
	{
	    /*  Convert password in file from hex format to binary format  */
	    _conn_convert_password_to_bin ( (unsigned char *) entry->password,
					    password );
	    m_clear ( password, strlen (password) );
	    m_free (password);
	}
	else
	{
	    if (password != NULL)
	    {
		strcpy (entry->password, password);
		m_clear ( password, strlen (password) );
		m_free (password);
	    }
	}
	/*  Add entry to list  */
	entry->next = password_list;
	password_list = entry;
    }
    /*  End of file  */
    ch_close (channel);
    m_clear (txt, STRING_LENGTH);
    return (password_list);

#else  /*  COMMUNICATIONS_AVAILABLE  */
    fprintf (stderr, "Operating system does not support communications\n");
    exit (RV_UNDEF_ERROR);
#endif  /*  COMMUNICATIONS_AVAILABLE  */
}   /*  End Function _conn_get_password_list  */

static void _conn_convert_password_to_bin (unsigned char *binpassword,
					   CONST char *asciipassword)
/*  [PURPOSE] This routine will convert an ASCII password to a binary password.
    <binpassword> Pointer to the storage for the binary password.
    <asciipassword> Pointer to the ASCII password.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    int byte[2];

    m_clear (binpassword, AUTH_PASSWORD_LENGTH);
    for (count = 0; count < AUTH_PASSWORD_LENGTH; ++count)
    {
	byte[0] = tolower (asciipassword[count * 2]);
	byte[1] = tolower (asciipassword[count * 2 + 1]);
	if ( (byte[0] == '\0') || (byte[1] == '\0') ) return;
	if ( !isxdigit (byte[0]) || !isxdigit (byte[1]) )
	{
	    fprintf (stderr,
			    "Password: \"%s\" contains non-hex character\n",
			    asciipassword);
	    fprintf (stderr, "Error in authorisation file\n");
	    exit (RV_BAD_DATA);
	}
	if (byte[0] > '9') byte[0] = byte[0] - 'a' + 0x0a;
	else byte[0] -= '0';
	if (byte[1] > '9') byte[1] = byte[1] - 'a' + 0x0a;
	else byte[1] -= '0';
	binpassword[count] = (byte[0] << 4) + byte[1];
    }
    byte[0] = 0;
    byte[1] = 0;
}   /*  End Function _conn_convert_password_to_bin  */

#ifdef dummy
static flag check_host_access (inet_addr, protocol)
/*  This routine will check if the host with Internet address given by
    inet_addr  is authorised to connect to the server with protocol given by
    protocol  .
    The routine returns TRUE if the host is permitted access,
    else it returns FALSE.
*/
unsigned long inet_addr;
char *protocol;
{
    struct serv_protocol_list_type *protocol_info;
    struct auth_host_list_type *auth_entry;
    static char function_name[] = "check_host_access";

    if (serv_protocol_list == NULL)
    {
	/*  Protocol not supprted  */
	fprintf (stderr, "No protocols supported by server\n");
	a_prog_bug (function_name);
    }
    /*  Search for protocol  */
    if ( ( protocol_info = _conn_get_serv_protocol_info (protocol) ) == NULL )
    {
	/*  Protocol not supported  */
	fprintf (stderr, "Protocol: %u not supported by server\n",
			protocol);
	a_prog_bug (function_name);
    }
    /*  Search if host is authorised  */
    for (auth_entry = protocol_info->authorised_hosts; auth_entry != NULL;
	 auth_entry = auth_entry->next)
    {
	if (inet_addr == auth_entry->host_ptr->inet_addr)
	{
	    /*  Found it  */
	    return (TRUE);
	}
    }
    /*  Host is not authorised  */
    return (FALSE);
}   /*  End Function check_host_access  */
#endif

/*  Functions to process client connection input  */

static flag _conn_client_connection_input_func (Channel channel, void **info)
/*  This routine is called when new input occurs on a channel.
    The channel object is given by  channel  .
    An arbitrary pointer may be written to the storage pointed to by  info
    The pointer written here will persist until the channel is unmanaged
    (or a subsequent callback routine changes it).
    The routine returns TRUE if the channel is to remain managed and
    open, else it returns FALSE (indicating that the channel is to be
    unmanaged and closed).
*/
{
    int bytes;
    unsigned long old_read_pos, new_read_pos, dummy;
    Connection connection = (Connection) *info;
    static char function_name[] = "_conn_client_connection_input_func";

    VERIFY_USED (connection);
    if (connection->read_func == NULL)
    {
	fprintf (stderr,
		 "Input on \"%s\" connection not being read (no callback)\n",
		 connection->protocol_name);
	a_prog_bug (function_name);
    }
    bytes = 1;
    if ( !ch_tell (channel, &old_read_pos, &dummy) ) exit (RV_SYS_ERROR);
    while (bytes > 0)
    {
	if ( !(*connection->read_func) (connection, &connection->info,
					channel) )
	    return (FALSE);
	if ( !ch_tell (channel, &new_read_pos, &dummy) ) exit (RV_SYS_ERROR);
	if (new_read_pos <= old_read_pos)
	{
	    fprintf (stderr,
		     "Connection read callback for protocol: \"%s\" not draining\n",
		     connection->protocol_name);
	    a_prog_bug (function_name);
	}
	old_read_pos = new_read_pos;
	if ( ( bytes = ch_get_bytes_readable (channel) ) < 0 )
	    exit (RV_SYS_ERROR);
    }
    return (TRUE);
}   /*  End Function _conn_client_connection_input_func  */

static char *_conn_write_protocol (Channel channel, CONST char *protocol_name,
				   unsigned int version,unsigned int conn_type)
/*  This routine will write an authentication message for a specified protocol
    to a channel.
    The channel object must be given by  channel  .
    The protocol name must be pointed to by  protocol_name  .
    The version number for the protocol must be given by  version  .
    <conn_type> The connection type.
    The routine returns the module name of the server on success, else it
    returns NULL.
*/
{
    ChConverter pri_converter = NULL;
    ChConverter pro_converter = NULL;
    unsigned int length;
    char *message;
    char protocol_buffer[PROTOCOL_NAME_LENGTH];
    char idea_block_client[EN_IDEA_BLOCK_SIZE];
    char idea_block_serv[EN_IDEA_BLOCK_SIZE];
    struct auth_password_list_type *authinfo;
    extern char module_name[STRING_LENGTH + 1];
    static char function_name[] = "_conn_write_protocol";

    /*  Write magic number  */
    if ( !pio_write32 (channel, CONN_MAGIC_NUMBER) )
    {
	a_func_abort (function_name, "Error writing magic number to channel");
	return (NULL);
    }
    /*  Write revision number  */
    if ( !pio_write32 (channel, REVISION_NUMBER) )
    {
	a_func_abort (function_name,
		      "Error writing revision number to channel");
	return (NULL);
    }
    if ( !ch_flush (channel) )
    {
	fprintf (stderr, "%s: error flushing channel\t%s\n",
		 function_name, ERRSTRING);
	return (NULL);
    }
    /*  Get primary password  */
    if ( ( authinfo = _conn_get_authinfo (RAW_PROTOCOL) ) != NULL )
    {
	if (authinfo->security_type == SECURITY_TYPE_DROP_ENCRYPTION)
	{
	    a_func_abort (function_name,
			  "security type: \"drop-encryption\" not valid for RAW_PROTOCOL\n");
	    exit (RV_BAD_DATA);
	}
	if ( ( pri_converter = _conn_get_encryption (channel, authinfo, FALSE,
						     idea_block_client) )
	     == NULL ) return (NULL);
	if (ch_read (channel, idea_block_serv, EN_IDEA_BLOCK_SIZE) <
	    EN_IDEA_BLOCK_SIZE) return (NULL);
	if (ch_write (channel, idea_block_serv, EN_IDEA_BLOCK_SIZE) <
	    EN_IDEA_BLOCK_SIZE) return (NULL);
	if ( !ch_flush (channel) )
	{
	    fprintf (stderr, "%s: error flushing channel\t%s\n",
		     function_name, ERRSTRING);
	    return (NULL);
	}
	if (ch_read (channel, idea_block_serv, EN_IDEA_BLOCK_SIZE) <
	    EN_IDEA_BLOCK_SIZE) return (NULL);
	if ( !m_cmp (idea_block_client, idea_block_serv, EN_IDEA_BLOCK_SIZE) )
	    return (NULL);
	switch (authinfo->security_type)
	{
	  case SECURITY_TYPE_KEY:
	    ch_unregister_converter (pri_converter);
	    pri_converter = NULL;
	    break;
	  case SECURITY_TYPE_IDEA:
	  case SECURITY_TYPE_PGPuserID_IDEA:
	    if ( ch_test_for_local_connection (channel) )
	    {
		ch_unregister_converter (pri_converter);
		pri_converter = NULL;
	    }
	    break;
	}
    }
    /*  Write protocol name  */
    strncpy (protocol_buffer, protocol_name, PROTOCOL_NAME_LENGTH);
    if (ch_write (channel, protocol_buffer, PROTOCOL_NAME_LENGTH)
	< PROTOCOL_NAME_LENGTH)
    {
	fprintf (stderr, "%s: error writing protocol name\t%s\n",
		 function_name, ERRSTRING);
	return (NULL);
    }
    /*  Write protocol version number  */
    if ( !pio_write32 (channel, (unsigned long) version) )
    {
	a_func_abort (function_name,
		      "Error writing protocol version number to channel");
	return (NULL);
    }
    if ( !ch_flush (channel) )
    {
	fprintf (stderr, "%s: error flushing channel\t%s\n",
		 function_name, ERRSTRING);
	return (NULL);
    }
    authinfo = _conn_get_authinfo (protocol_name);
    if ( (authinfo != NULL) &&
	 (authinfo->security_type == SECURITY_TYPE_DROP_ENCRYPTION) )
    {
	if (pri_converter != NULL) ch_unregister_converter (pri_converter);
	authinfo = NULL;
    }
    if (authinfo != NULL)
    {
	if ( ( pro_converter = _conn_get_encryption (channel, authinfo, FALSE,
						     idea_block_client) )
	     == NULL ) return (NULL);
	if (ch_read (channel, idea_block_serv, EN_IDEA_BLOCK_SIZE) <
	    EN_IDEA_BLOCK_SIZE) return (NULL);
	if (ch_write (channel, idea_block_serv, EN_IDEA_BLOCK_SIZE) <
	    EN_IDEA_BLOCK_SIZE) return (NULL);
	if ( !ch_flush (channel) )
	{
	    fprintf (stderr, "%s: error flushing channel\t%s\n",
		     function_name, ERRSTRING);
	    return (NULL);
	}
	if (ch_read (channel, idea_block_serv, EN_IDEA_BLOCK_SIZE) <
	    EN_IDEA_BLOCK_SIZE) return (NULL);
	if ( !m_cmp (idea_block_client, idea_block_serv,
		     EN_IDEA_BLOCK_SIZE) ) return (NULL);
	if (pri_converter != NULL) ch_unregister_converter (pri_converter);
	switch (authinfo->security_type)
	{
	  case SECURITY_TYPE_KEY:
	    ch_unregister_converter (pro_converter);
	    pro_converter = NULL;
	    break;
	  case SECURITY_TYPE_IDEA:
	  case SECURITY_TYPE_PGPuserID_IDEA:
	    if ( ch_test_for_local_connection (channel) )
	    {
		ch_unregister_converter (pro_converter);
		pro_converter = NULL;
	    }
	    break;
	}
    }
    /*  Get returned message  */
    if ( ( message = pio_read_string (channel, &length) ) == NULL )
    {
	fprintf (stderr, "%s: error reading return message\n", function_name);
	return (NULL);
    }
    if (length > 0)
    {
	fprintf (stderr, "\"%s\" connection refused. Reason: %s\n",
		 protocol_name, message);
	m_free (message);
	return (NULL);
    }
    m_free (message);
    /*  Send module name to server  */
    if ( !pio_write_string (channel, module_name) )
    {
	fprintf (stderr, "%s: error writing module name\n", function_name);
	return (NULL);
    }
    if ( !pio_write32 (channel, conn_type) )
    {
	fprintf (stderr, "%s: error writing connection type\n",
		 function_name);
	return (NULL);
    }
    if ( !ch_flush (channel) )
    {
	fprintf (stderr, "%s: error flushing channel\t%s\n",
		 function_name, ERRSTRING);
	return (NULL);
    }
    /*  Get returned module name  */
    if ( ( message = pio_read_string (channel, NULL) ) == NULL )
    {
	fprintf (stderr, "%s: error reading server module name\n",
		 function_name);
	ch_close (cm_channel);
	exit (RV_UNDEF_ERROR);
    }
    return (message);
}   /*  End Function _conn_write_protocol  */

static flag _conn_respond_to_ping_server_from_client (Connection connection,
						      void **info,
						      Channel channel)
/*  [SUMMARY] Connection read event callback.
    [PURPOSE] This routine will process input from a client on a "ping_server"
    connection.
    <connection> The connection object.
    <info> A pointer to the arbitrary information pointer. This may be modified
    <channel> The channel object on which the read event occurred. This may be
    the normal or datagram channel.
    [RETURNS] TRUE on successful reading, else FALSE (indicating the connection
    should be closed).
    [NOTE] The <<close_func>> will be called if this routine returns FALSE.
*/
{
#ifdef COMMUNICATIONS_AVAILABLE
    int bytes_readable;
    static unsigned int buffer_length = 0;
    static char *buffer = NULL;
    static char function_name[] = "_conn_respond_to_ping_server_from_client";

    VERIFY_USED (connection);
    if ( ( bytes_readable = ch_get_bytes_readable (channel) ) < 0 )
    {
	exit (RV_SYS_ERROR);
    }
    if (bytes_readable < 1)
    {
	fprintf (stderr,
		 "Connection has: %d bytes readable: should be at least 1\n",
		 bytes_readable);
	a_prog_bug (function_name);
    }
    /*  Check if buffer is big enough  */
    if (buffer_length < (unsigned int) bytes_readable)
    {
	/*  Allocate bigger buffer  */
	if (buffer != NULL) m_free (buffer);
	buffer_length = 0;
	if ( ( buffer = m_alloc (bytes_readable) ) == NULL )
	{
	    m_error_notify (function_name, "buffer");
	    return (FALSE);
	}
	buffer_length = bytes_readable;
    }
    /*  Read data  */
    if (ch_read (channel, buffer, bytes_readable) < bytes_readable)
    {
	fprintf (stderr, "%s: error reading: %u bytes from descriptor\t%s\n",
		 function_name, (unsigned int) bytes_readable, ERRSTRING);
	return (FALSE);
    }
    /*  Send data back from whence it came  */
    if (ch_write (channel, buffer, bytes_readable) < bytes_readable)
    {
	fprintf (stderr, "%s: error writing: %u bytes to channel\t%s\n",
		 function_name, (unsigned int) bytes_readable, ERRSTRING);
	return (FALSE);
    }
    if ( !ch_flush (channel) )
    {
	fprintf (stderr, "%s: error flushing channel\t%s\n",
		 function_name, ERRSTRING);
	return (FALSE);
    }
    return (TRUE);

#else  /*  COMMUNICATIONS_AVAILABLE  */
    fprintf (stderr, "Operating system does not support communications\n");
    return (FALSE);
#endif
}   /*  End Function _conn_respond_to_ping_server_from_client   */

static flag _conn_register_server_exit (Connection connection, void **info)
/*  This routine will process input from a client on a  "server_exit"
    connection.
    The connection object must be given by  connection  .
    Any appropriate information will be written to the pointer pointed to by
    info  (unused).
    The routine either returns FALSE (indicating that the connection should be
    closed), or the routine will never exit ( because it calls exit(3) ).
*/
{
    static char function_name[] = "_conn_register_server_exit";

    VERIFY_USED (connection);
    if (exit_schedule_function == NULL)
    {
	exit (RV_OK);
    }
    (*exit_schedule_function) ();
    return (FALSE);
}   /*  End Function _conn_register_server_exit   */

static Connection _conn_get_numbered_connection (CONST char *protocol_name,
						 unsigned int number,
						 Connection list)
/*  This routine will get Nth connection from a connection list with a
    specified protocol.
    The protocol name must be pointed to by  protocol_name  .
    The number of the connection to get must be given by  number  .
    The connection list must be given by  list  .
    The routine returns the connection on success, else it returns NULL.
*/
{
    unsigned int conn_count;
    Connection curr_entry;
    static char function_name[] = "_conn_get_numbered_connection";

    if (list != NULL) VERIFY_USED (list);
    for (conn_count = 0, curr_entry = list; curr_entry != NULL;
	 curr_entry = curr_entry->next)
    {
	if (strcmp (protocol_name, curr_entry->protocol_name) == 0)
	{
	    /*  Found a connection of the correct protocol  */
	    if (conn_count == number)
	    {
		/*  Found it  */
		return (curr_entry);
	    }
	    ++conn_count;
	}
    }
    /*  Not found  */
    return (NULL);
}   /*  End Function _conn_get_numbered_connection  */  

static void _conn_attempt_connection_to_cm (void)
/*  This routine will attempt to connect to the Connection Management tool.
    The routine returns nothing.
*/
{
    Channel stdio;
    int x;
    int y;
    int out_fd;
    unsigned int port_number;
    unsigned long host_addr;
    CONST char *env;
    char *char_ptr;
    char *message;
    char my_hostname[STRING_LENGTH];
    char control_command[STRING_LENGTH];
    char txt[STRING_LENGTH];
    extern char module_name[STRING_LENGTH + 1];
    static char *env_name = "KARMA_CM_CONTROL_COMMAND";
    static char function_name[] = "_conn_attempt_connection_to_cm";

    if (cm_channel != NULL)
    {
	fprintf (stderr,
		 "Connection to Connection Management tool already open\n");
	a_prog_bug (function_name);
    }
    if ( ( env = r_getenv (env_name) ) == NULL )
    {
	/*  Ignore  */
	return;
    }
    if (strlen (env) >= STRING_LENGTH)
    {
	fprintf (stderr, "Environment variable: \"%s\" is too long\n",
		 env_name);
	a_prog_bug (function_name);
    }
    strcpy (control_command, env);
#ifdef HAS_SOCKETS
    sprintf ( txt, "/tmp/%s.log.%d.%d",
		     module_name, (int) getuid (), (int) getpid () );
    if ( ( out_fd = open (txt, O_CREAT | O_TRUNC | O_WRONLY,
			  S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH) ) < 0 )
    {
	fprintf (stderr, "%s: error opening: \"%s\" for output\t%s\n",
		 function_name, txt, ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    /*  Attatch log file to standard output  */
    if (dup2 (out_fd, 1) < 0)
    {
	fprintf (stderr, "%s: error redirecting standard output\t%s\n",
		 function_name, ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    /*  Attatch log file to standard error  */
    if (dup2 (out_fd, 2) < 0)
    {
	fprintf (stderr, "%s: error redirecting standard error\t%s\n",
		 function_name, ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    /*  Close temporary descriptor for log file  */
    if (close (out_fd) != 0)
    {
	fprintf (stderr,
		 "%s: error closing temporary descriptor for log file\t%s\n",
		 function_name, ERRSTRING);
	exit (RV_SYS_ERROR);
    }
#endif
    /*  Get the fully qualified hostname  */
    if ( !r_get_fq_hostname (my_hostname, STRING_LENGTH) ) exit (RV_SYS_ERROR);
    /*  Decode command  */
    for (char_ptr = control_command; *char_ptr != '\0'; ++char_ptr)
    {
	if (*char_ptr == ':')
	{
	    /*  Convert ':' to ' '  */
	    *char_ptr = ' ';
	}
    }
    char_ptr = control_command;
    host_addr = ex_uint (char_ptr, (CONST char **) &char_ptr);
    port_number = ex_uint (char_ptr, (CONST char **) &char_ptr);
    x = ex_int (char_ptr, (CONST char **) &char_ptr);
    y = ex_int (char_ptr, (CONST char **) &char_ptr);
    if (host_addr == 0) strcpy (my_hostname, "localhost");
#ifdef OS_VXMVX
    /*  Append ":vx" to hostname  */
    strcat (my_hostname, ":vx");
#endif
    /*  Connect to Connection Management tool with control protocol  */
    /*  Open connection  */
    if ( ( cm_channel = ch_open_connection (host_addr, port_number) ) == NULL )
    {
	fprintf (stderr,
		 "Could not open control connection to Connection Management tool\n");
	fprintf (stderr, "on host addr: %lu  port: %u\n",
		 host_addr, port_number);
	exit (RV_UNDEF_ERROR);
    }
    /*  Send protocol info  */
    if ( ( message = _conn_write_protocol (cm_channel, "conn_mngr_control", 
					   CM_PROTOCOL_VERSION,
					   CONNECTION_TYPE_STREAM) ) == NULL )
    {
	fprintf (stderr, "Error writing authentication information\n");
	ch_close (cm_channel);
	exit (RV_SYS_ERROR);
    }
    m_free (message);
    if ( !(*manage_channel) (cm_channel, (void *) NULL,
			     _conn_cm_command_func,
			     _conn_cm_close_func,
			     ( flag (*) () ) NULL,
			     ( flag (*) () ) NULL ) )
    {
	ch_close (cm_channel);
	fprintf (stderr, "Could not manage channel");
	exit (RV_UNDEF_ERROR);
    }
    /*  Send information back to Connection Management tool  */
    if ( !pio_write_string (cm_channel, my_hostname) )
    {
	fprintf (stderr, "%s: error writing hostname\n", function_name);
	exit (RV_WRITE_ERROR);
    }
    if ( !pio_write32s (cm_channel, x) )
    {
	fprintf (stderr, "%s: error writing x position\n", function_name);
	exit (RV_WRITE_ERROR);
    }
    if ( !pio_write32s (cm_channel, y) )
    {
	fprintf (stderr, "%s: eror writing y position\n", function_name);
	exit (RV_WRITE_ERROR);
    }
    if ( !pio_write32s ( cm_channel, getpid () ) )
    {
	fprintf (stderr, "%s: error writing PID\n", function_name);
	exit (RV_WRITE_ERROR);
    }
    if ( !ch_flush (cm_channel) )
    {
	fprintf (stderr, "%s: error flushing channel\t%s\n",
		 function_name, ERRSTRING);
	exit (RV_WRITE_ERROR);
    }
    /*  Connect to Connection Management tool with standard IO protocol  */
    /*  Open connection  */
    if ( ( stdio = ch_open_connection (host_addr, port_number) ) == NULL )
    {
	fprintf (stderr,
		 "Could not open stdio connection to Connection Management tool");
	exit (RV_UNDEF_ERROR);
    }
    /*  Send protocol info  */
    if ( ( message = _conn_write_protocol (stdio, "conn_mngr_stdio",
					   CM_PROTOCOL_VERSION,
					   CONNECTION_TYPE_STREAM) ) == NULL )
    {
	fprintf (stderr, "Error writing authentication information\n");
	ch_close (stdio);
	exit (RV_SYS_ERROR);
    }
    m_free (message);
    if ( !pio_write_string (stdio, my_hostname) )
    {
	fprintf (stderr, "%s: error writing hostname\n", function_name);
	exit (RV_WRITE_ERROR);
    }
    if ( !pio_write32s ( stdio, getpid () ) )
    {
	fprintf (stderr, "%s: error writing PID\n", function_name);
	exit (RV_WRITE_ERROR);
    }
    if ( !ch_flush (stdio) )
    {
	fprintf (stderr, "%s: error flushing channel\t%s\n",
		 function_name, ERRSTRING);
	exit (RV_WRITE_ERROR);
    }
    ch_stdin = stdio;
    ch_stdout = stdio;
    ch_stderr = stdio;
}   /*  End Function _conn_attempt_connection_to_cm  */

static flag _conn_cm_command_func (Channel channel, void **info)
/*  This routine is called when new input occurs on the connection to the
    Connection Management tool.
    The channel object is given by  channel  .
    An arbitrary pointer may be written to the storage pointed to by  info
    The pointer written here will persist until the channel is unmanaged
    (or a subsequent callback routine changes it).
    The routine returns TRUE if the channel is to remain managed and
    open, else it returns FALSE (indicating that the channel is to be
    unmanaged and closed). This routine MUST NOT unmanage or close the
    channel given by  channel  .
*/
{
    unsigned long command;
    unsigned long conn_id;
    unsigned long port_number;
    char *hostname;
    char *protocol_name;
    static char function_name[] = "_conn_cm_command_func";

    if (channel != cm_channel)
    {
	fprintf (stderr, "Bad channel\n");
	a_prog_bug (function_name);
    }
    if ( !pio_read32 (channel, &command) )
    {
	fprintf (stderr, "%s: error reading command\n", function_name);
	return (FALSE);
    }
    switch (command)
    {
      case CM_TOOL_ATTEMPT_CONNECTION:
	if ( ( hostname = pio_read_string (channel, NULL) ) == NULL )
	{
	    fprintf (stderr, "%s: error reading hostname\n", function_name);
	    return (FALSE);
	}
	if ( !pio_read32 (channel, &port_number) )
	{
	    fprintf (stderr, "%s: error reading port_number\n", function_name);
	    return (FALSE);
	}
	if ( ( protocol_name = pio_read_string (channel, NULL) ) == NULL )
	{
	    fprintf (stderr, "%s: error reading protocol name\n",
		     function_name);
	    return (FALSE);
	}
	if ( !conn_attempt_connection (hostname, port_number, protocol_name) )
	{
	    fprintf (stderr, "Error attempting connection\n");
	}
	break;
      case CM_TOOL_CLOSE_CONNECTION:
	if ( !pio_read32 (channel, &conn_id) )
	{
	    fprintf (stderr, "%s: error reading command\n", function_name);
	    return (FALSE);
	}
	if ( !conn_close ( (Connection) conn_id ) )
	{
	    fprintf (stderr, "Error closing connection\t%s\n", ERRSTRING);
	}
	break;
      case CM_TOOL_EXIT_MODULE:
	if (exit_schedule_function == NULL) exit (RV_OK);
	else (*exit_schedule_function) ();
	break;
      case CM_TOOL_NOW_QUIESCENT:
	if (quiescent_function != NULL) (*quiescent_function) ();
	break;
      default:
	fprintf (stderr, "%s: illegal command value from CM tool: %lu\n",
		 function_name, command);
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function _conn_cm_command_func  */

static void _conn_cm_close_func (Channel channel, void *info)
/*  This routine is called when the connection to the Connection Management
    tool closes.
    The routine returns TRUE if the channel is to remain managed and
    open, else it returns FALSE (indicating that the channel is to be
    unmanaged and closed). This routine MUST NOT unmanage or close the
    channel given by  channel  .
*/
{
    static char function_name[] = "_conn_cm_close_func";

    if (channel != cm_channel)
    {
	fprintf (stderr, "Bad channel\n");
	a_prog_bug (function_name);
    }
    fprintf (stderr, "PANIC: lost connection to Connection Management tool\n");
    /*  Used to call _exit() but because this leaves sproc() processes lying
	around under IRIX, changed to exit(). Hope this doesn't cause a nasty
	loop.  */
    exit (RV_OK);
}   /*  End Function _conn_cm_close_func  */

static void _conn_setup_prng (void)
/*  This routine will setup the cryptographically strong random number
    generator(s). If already set up, the routine does nothing.
    The routine returns nothing.
*/
{
    KCallbackFunc cbk;
    static char function_name[] = "_conn_setup_prng";

    if (main_randpool == NULL)
    {
	if ( ( main_randpool = rp_create (512, 16, 64, md_md5_transform) )
	     == NULL )
	{
	    m_abort (function_name, "random pool of bytes");
	}
	cbk = ch_tap_io_events (rp_add_time_noise, main_randpool);
	rp_register_destroy_func (main_randpool, _conn_randpool_destroy_func,
				  cbk);
    }
}   /*  End Function _conn_setup_prng  */

static void _conn_randpool_destroy_func (RandPool rp, void *info)
/*  This routine is called when a random pool is destroyed.
    The random pool will be given by  rp  .
    The arbitrary information pointer will be given by  info  .
    The routine returns nothing.
*/
{
    c_unregister_callback ( (KCallbackFunc) info );
}   /*  End Function _conn_randpool_destroy_func  */

static ChConverter _conn_get_encryption
    (Channel channel, struct auth_password_list_type *authinfo,
     flag server, char *rand_block)
/*  [PURPOSE] The routine will set the appropriate channel encryption for a
    connection.
    <channel> The channel for the connection.
    <authinfo> A pointer to the authorisation information for the desired
    protocol.
    <server> If TRUE, the connection is for a server, else a client.
    [RETURNS] An encryption converter function, NULL on failure.
*/
{
    ChConverter converter = NULL;
    flag idea_encryption = FALSE;
    flag clear_session = TRUE;
    unsigned int ciphertext_length, plaintext_length;
    unsigned long data;
    char *ciphertext, *plaintext;
    char *session_key = NULL;  /*  Initialised to keep compiler happy  */
    char *session_iv = NULL;   /*  Initialised to keep compiler happy  */
    char *recipient;
    char idea_session[EN_IDEA_KEY_SIZE + EN_IDEA_BLOCK_SIZE];
    static char function_name[] = "_conn_get_encryption";

    if ( (authinfo == NULL) || (rand_block == NULL) )
    {
	fprintf (stderr, "NULL pointer(s) passed\n");
	a_prog_bug (function_name);
    }
    switch ( authinfo->security_type )
    {
      case SECURITY_TYPE_KEY:
      case SECURITY_TYPE_IDEA:
	idea_encryption = TRUE;
	session_key = authinfo->password;
	session_iv = authinfo->password + EN_IDEA_KEY_SIZE;
	clear_session = FALSE;
	break;
      case SECURITY_TYPE_PGPuserID_IDEA:
	idea_encryption = TRUE;
	if (server)
	{
	    rp_get_bytes (main_randpool, (unsigned char *) idea_session,
			  EN_IDEA_KEY_SIZE + EN_IDEA_BLOCK_SIZE);
	    recipient = authinfo->password;
	    fprintf (stderr, "recipient: \"%s\"\n", recipient);
	    if ( ( ciphertext =
		   pgp_encrypt (idea_session,
				EN_IDEA_KEY_SIZE + EN_IDEA_BLOCK_SIZE,
				(CONST char **) &recipient, 1,
				&ciphertext_length, FALSE) ) == NULL )
	    {
		m_abort (function_name, "encrypted IDEA session key");
	    }
	    if ( !pio_write32 (channel, ciphertext_length) )
	    {
		fprintf (stderr, "%s: error writing length\n", function_name);
		m_clear (idea_session, EN_IDEA_KEY_SIZE + EN_IDEA_BLOCK_SIZE);
		m_clear (ciphertext, ciphertext_length);
		m_free (ciphertext);
		return (NULL);
	    }
	    if (ch_write (channel, ciphertext, ciphertext_length) <
		ciphertext_length)
	    {
		fprintf (stderr,
			 "%s: error writing encrypted IDEA session key\t%s\n",
			 function_name, ERRSTRING);
		m_clear (idea_session, EN_IDEA_KEY_SIZE + EN_IDEA_BLOCK_SIZE);
		m_clear (ciphertext, ciphertext_length);
		m_free (ciphertext);
		return (NULL);
	    }
	    m_clear (ciphertext, ciphertext_length);
	    m_free (ciphertext);
	    fprintf (stderr, "Wrote encrypted IDEA session\n");
	}
	else
	{
	    /*  Client  */
	    if ( !pio_read32 (channel, &data) )
	    {
		fprintf (stderr, "%s: error reading length\n", function_name);
		return (NULL);
	    }
	    ciphertext_length = data;
	    if ( ( ciphertext = m_alloc (ciphertext_length) ) == NULL )
	    {
		m_abort (function_name, "ciphertext");
	    }
	    if (ch_read (channel, ciphertext, ciphertext_length) <
		ciphertext_length)
	    {
		fprintf (stderr, "%s: error reading length\t%s\n",
			 function_name, ERRSTRING);
		m_free (ciphertext);
		return (NULL);
	    }
	    if ( ( plaintext = pgp_decrypt (ciphertext, ciphertext_length,
					    &plaintext_length) ) == NULL )
	    {
		fprintf (stderr, "Error decrypting IDEA session key\n");
		m_free (ciphertext);
		return (NULL);
	    }
	    if (plaintext_length != EN_IDEA_KEY_SIZE + EN_IDEA_BLOCK_SIZE)
	    {
		fprintf (stderr,
			 "Decrypted session key length: %u is not: %u\n",
			 plaintext_length,
			 EN_IDEA_KEY_SIZE + EN_IDEA_BLOCK_SIZE);
		m_clear (plaintext, plaintext_length);
		m_free (plaintext);
		return (NULL);
	    }
	    m_copy (idea_session, plaintext, plaintext_length);
	    m_clear (plaintext, plaintext_length);
	    m_free (plaintext);
	    fprintf (stderr, "Read encrypted IDEA session\n");
	}
	session_key = idea_session;
	session_iv = idea_session + EN_IDEA_KEY_SIZE;
	break;
      default:
	fprintf (stderr, "Illegal security type: %u\n",
		 authinfo->security_type);
	a_prog_bug (function_name);
	break;
    }
    if (idea_encryption)
    {
	if ( ( converter = cen_idea (channel, session_key, session_iv,
				     session_key, session_iv, clear_session) )
	     == NULL )
	{
	    m_abort (function_name, "IDEA encryption");
	}
	/*  Send some random data to get around constant IV problem  */
	rp_get_bytes (main_randpool, (unsigned char *) rand_block,
		      EN_IDEA_BLOCK_SIZE);
	if (ch_write (channel, rand_block, EN_IDEA_BLOCK_SIZE) <
	    EN_IDEA_BLOCK_SIZE)
	{
	    fprintf (stderr, "%s: error writing random block\t%s\n",
		     function_name, ERRSTRING);
	    m_clear (rand_block, EN_IDEA_BLOCK_SIZE);
	    return (NULL);
	}
    }
    return (converter);
}   /*  End Function _conn_get_encryption  */

static Connection _conn_alloc_connection ()
/*  [SUMMARY] Allocate a new connection.
    [RETURNS] The new connection on success, else NULL.
*/
{
    Connection new;
    unsigned int count;
    static char function_name[] = "_conn_alloc_connection";

    /*  Get new entry  */
    if (first_free == NULL)
    {
	/*  Have to allocate some more for the free list  */
	if ( ( new = m_alloc (ALLOC_SIZE * sizeof *new) ) == NULL )
	{
	    m_error_notify (function_name, "new connection entries");
	    return (NULL);
	}
	m_clear (new, ALLOC_SIZE * sizeof *new);
	/*  Add them to the free list  */
	first_free = new;
	for (count = 0; count < ALLOC_SIZE - 1; ++count)
	{
	    new[count].magic_number = FREE_MAGIC_NUMBER;
	    new[count].next = new + count + 1;
	}
	new[count].magic_number = FREE_MAGIC_NUMBER;
    }
    /*  Take one from the free list  */
    new = first_free;
    VERIFY_FREE (new);
    first_free = new->next;
    /*  Fill in entry  */
    new->magic_number = USED_MAGIC_NUMBER;
    new->next = NULL;
    return (new);
}   /*  End Function _conn_alloc_connection  */

static void _conn_free_connection (Connection connection)
/*  [SUMMARY] Add a Connection object to the free list.
    <entry> The Connection object.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "_conn_free_connection";

    VERIFY_USED (connection);
    m_clear (connection, sizeof *connection);
    connection->magic_number = FREE_MAGIC_NUMBER;
    connection->next = first_free;
    first_free = connection;
}   /*  End Function _conn_free_connection  */

static flag _conn_parse_hostname (unsigned int *conn_type,
				  struct karma_qos *qos,
				  char *hostname, unsigned int hostname_length)
/*  [SUMMARY] Parse a hostname.
    <conn_type> The connection type is written here.
    <qos> The QOS information is written here.
    <hostname> The hostname to parse. This is modified.
    <hostname_length> The length of the original hostname.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    char *ptr;
    static char function_name[] = "_conn_parse_hostname";

    *conn_type = CONNECTION_TYPE_STREAM;
    /*  Search for last colon  */
    for (ptr = hostname + hostname_length - 1; ptr >= hostname; --ptr)
	if (*ptr == ':') break;
    if (ptr < hostname) return (TRUE);
    *ptr++ = '\0';
    if (strcmp (ptr, "stream") == 0) return (TRUE);
    if (strcmp (ptr, "udp") == 0)
    {
	*conn_type = CONNECTION_TYPE_UDP;
	return (TRUE);
    }
    if (strncmp (ptr, "atmsvc", 6) == 0)
    {
	*conn_type = CONNECTION_TYPE_ATMSVC;
	if (ptr[6] == ',') ++ptr;
	else if (ptr[6] != '\0')
	{
	    fprintf (stderr, "%s: illegal protocol name: \"%s\"\n",
		     function_name, ptr);
	    return (FALSE);
	}
	return _conn_parse_qos (qos, ptr + 6);
    }
    fprintf (stderr, "%s: cannot parse protocol type: \"%s\"\n",
	     function_name, ptr);
    return (FALSE);
}   /*  End Function _conn_parse_hostname  */

static flag _conn_parse_qos (struct karma_qos *qos, char *string)
/*  [SUMMARY] Parse QOS (Quality Of Service) parameters.
    <qos> The QOS information is written here.
    <string> The string to parse. This is modified.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    char *comma_ptr, *value_ptr;
    static char function_name[] = "_conn_parse_qos";

    m_clear (qos, sizeof *qos);
    for (; (string != NULL) && (*string != '\0'); string = comma_ptr)
    {
	comma_ptr = strchr (string, ',');
	if (comma_ptr != NULL) *comma_ptr++ = '\0';
	if ( ( value_ptr = strchr (string, '=') ) == NULL )
	{
	    fprintf (stderr, "%s: illegal attribute: \"%s\"\n",
		     function_name, string);
	    return (FALSE);
	}
	*value_ptr++ = '\0';
	if (strcmp (string, "txclass") == 0)
	{
	    if (st_icmp (value_ptr, "ubr") == 0)
		qos->tx.traffic_class = K_ATM_TRAFFIC_CLASS_UBR;
	    else if (st_icmp (value_ptr, "cbr") == 0)
		qos->tx.traffic_class = K_ATM_TRAFFIC_CLASS_CBR;
	    else
	    {
		fprintf (stderr, "%s: illegal traffic class: \"%s\"\n",
			 function_name, value_ptr);
		return (FALSE);
	    }
	}
	else if (strcmp (string, "rxclass") == 0)
	{
	    if (st_icmp (value_ptr, "ubr") == 0)
		qos->rx.traffic_class = K_ATM_TRAFFIC_CLASS_UBR;
	    else if (st_icmp (value_ptr, "cbr") == 0)
		qos->rx.traffic_class = K_ATM_TRAFFIC_CLASS_CBR;
	    else
	    {
		fprintf (stderr, "%s: illegal traffic class: \"%s\"\n",
			 function_name, value_ptr);
		return (FALSE);
	    }
	}
	else if (strcmp (string, "txmin") == 0)
	    qos->tx.rate_min = _conn_parse_bw (value_ptr);
	else if (strcmp (string, "txtyp") == 0)
	    qos->tx.rate_typ = _conn_parse_bw (value_ptr);
	else if (strcmp (string, "txmax") == 0)
	    qos->tx.rate_max = _conn_parse_bw (value_ptr);
	else if (strcmp (string, "txsdu") == 0)
	    qos->tx.sdu_max = atoi (value_ptr);
	else if (strcmp (string, "txdvmax") == 0)
	    qos->tx.dv_max = atoi (value_ptr);
	else if (strcmp (string, "rxmin") == 0)
	    qos->rx.rate_min = _conn_parse_bw (value_ptr);
	else if (strcmp (string, "rxtyp") == 0)
	    qos->rx.rate_typ = _conn_parse_bw (value_ptr);
	else if (strcmp (string, "rxmax") == 0)
	    qos->rx.rate_max = _conn_parse_bw (value_ptr);
	else if (strcmp (string, "rxsdu") == 0)
	    qos->rx.sdu_max = atoi (value_ptr);
	else if (strcmp (string, "rxdvmax") == 0)
	    qos->rx.dv_max = atoi (value_ptr);
	else
	{
	    fprintf (stderr, "%s: illegal parameter: \"%s\"\n",
		     function_name, string);
	    return (FALSE);
	}
    }
    return (TRUE);
}   /*  End Function _conn_parse_qos  */

static int _conn_parse_bw (char *string)
/*  [SUMMARY] Parse a bandwidth parameter.
    <string> The ASCII input. This is modified.
    [RETURNS] The bandwidth, in bytes per second.
*/
{
    int mul = 1;
    int div = 1;
    int len = strlen (string);

    switch (string[len - 1])
    {
      case 'b':
	div = 8;
	string[--len] = '\0';
	break;
      case 'B':
	string[--len] = '\0';
	break;
      default:
	break;
    }
    switch (string[len - 1])
    {
      case 'k':
	mul = 1000;
	string[--len] = '\0';
	break;
      case 'M':
	mul = 1000000;
	string[--len] = '\0';
	break;
      default:
	break;
    }
    return (atoi (string) * mul / div);
}   /*  End Function _conn_parse_bw  */

static void _conn_show_atm_tfparm (CONST struct karma_traffic_parameters *tp,
				   CONST char *prefix)
/*  [SUMMARY] Show traffic parameters.
    <tp> The traffic parameters.
    <prefix> The prefix string.
    [RETURNS] Nothing.
*/
{
    switch (tp->traffic_class)
    {
      case K_ATM_TRAFFIC_CLASS_UBR:
	fprintf (stderr, "    %sclass=ubr,%ssdu=%d\n",
		 prefix, prefix, tp->sdu_max);
	break;
      case K_ATM_TRAFFIC_CLASS_CBR:
	fprintf (stderr,
		 "    %sclass=cbr,%smin=%d,%styp=%d,%smax=%d,%ssdu=%d,%sdvmax=%d\n",
		 prefix, prefix, tp->rate_min, prefix, tp->rate_typ,
		 prefix, tp->rate_max, prefix, tp->sdu_max,
		 prefix, tp->dv_max);
	break;
    }
}   /*  End Function _conn_show_atm_tfparm  */
