/*LINTLIBRARY*/
/*  event.c

    This code provides routines to manage Karma data structure events.

    Copyright (C) 1996-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for managing callbacks for
    general data structure supported in Karma.


    Written by      Richard Gooch   26-DEC-1996

    Updated by      Richard Gooch   26-DEC-1996

    Updated by      Richard Gooch   24-NOV-1997: Tried a few experimental
  hacks to make function pointer passing more C++ friendly.

    Last updated by Richard Gooch   28-JUN-2000: Made <<list>> static.


*/
#include <stdio.h>
#include <karma.h>
#define KARMA_DS_EVENT_C
#include <karma_ds.h>
#include <karma_c.h>


/*  Private structure types  */
typedef struct
{
    multi_array *desc;
    CONST char *domain;
    CONST char *name;
} info_type;

typedef flag (*EventFuncType) (void *object, multi_array *multi_desc,
			       CONST char *domain, CONST char *name);


/*  Private functions  */
STATIC_FUNCTION (flag glue_func,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );


/*  Private data  */
static KCallbackList list = NULL;


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
KCallbackFunc ds_event_register_func (flag (*func) (), void *object)
/*  [SUMMARY] Register a multi_array event function.
    <func> The function which is called to process an event. The prototype
    function is [<ds_PROTO_event_func>].
    <object> The object.
    [RETURNS] A KCallbackFunc object. On failure, the process aborts.
*/
{
    return c_register_callback (&list, ( flag (*) () ) glue_func,
				object, (void *) func, FALSE, NULL, FALSE,
				TRUE);
}   /*  End Function ds_event_register_func  */

/*EXPERIMENTAL_FUNCTION*/
flag ds_event_dispatch (multi_array *multi_desc,
			CONST char *domain, CONST char *name)
/*  [SUMMARY] Dispatch a multi_array event.
    <multi_desc> The multi_array descriptor.
    <domain> The domain from where the event was generated.
    <name> The name of the event in its domain.
    [RETURNS] TRUE if one of the callbacks quenched the further delivery of
    callbacks, else FALSE.
*/
{
    info_type info;

    info.desc = multi_desc;
    info.domain = domain;
    info.name = name;
    return c_call_callbacks (list, &info);
}   /*  End Function ds_event_dispatch  */


/*  Private functions follow  */

static flag glue_func (void *object, void *client1_data,
		       void *call_data, void *client2_data)
/*  [SUMMARY] Generic callback.
    [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    info_type *info = (info_type *) call_data;
    EventFuncType func;

    func = (EventFuncType) client1_data;
    return (*func) (object, info->desc, info->domain, info->name);
}   /*  End Function glue_func  */
