/*LINTLIBRARY*/
/*  aips2_read.c

    This code provides an AIPS++ read facility.

    Copyright (C) 1997-1999  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for reading data files in
  AIPS++ (AIPS2) Image format.


    Written by      Richard Gooch   17-OCT-1997

    Updated by      Richard Gooch   12-MAR-1998: Added <<test_dir>> parameter
  to <foreign_aips2_test>.

    Last updated by Richard Gooch   28-MAY-1998: Created experimental function
  <foreign_aips2_run>.


*/
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_r.h>


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
flag foreign_aips2_test (CONST char *dirname, flag test_dir)
/*  [SUMMARY] Test if a directory is an AIPS++ Image file.
    <dirname> The directory name of the AIPS++ dataset name.
    <test_dir> If TRUE will first test if the dataset name is a directory. Set
    this to FALSE only if you know the dataset name is a directory.
    [RETURNS] TRUE if the directory is a AIPS++ Image file, else FALSE.
*/
{
    struct stat statbuf;
    char fname[STRING_LENGTH];

    if (test_dir)
    {
	if (stat (dirname, &statbuf) != 0)
	{
	    if (errno == ENOENT) return (FALSE);
	    fprintf (stderr, "Error statting file: \"%s\"\t%s\n",
		     dirname, ERRSTRING);
	    return (FALSE);
	}
	if ( !S_ISDIR (statbuf.st_mode) ) return (FALSE);
    }
    sprintf (fname, "%s/table.dat", dirname);
    if (stat (fname, &statbuf) != 0)
    {
	if (errno == ENOENT) return (FALSE);
	fprintf (stderr, "Error statting file: \"%s\"\t%s\n", fname,ERRSTRING);
	return (FALSE);
    }
    if ( !S_ISREG (statbuf.st_mode) ) return (FALSE);
    sprintf (fname, "%s/table.info", dirname);
    if (stat (fname, &statbuf) != 0)
    {
	if (errno == ENOENT) return (FALSE);
	fprintf (stderr, "Error statting file: \"%s\"\t%s\n", fname,ERRSTRING);
	return (FALSE);
    }
    if ( !S_ISREG (statbuf.st_mode) ) return (FALSE);
    return (TRUE);
}   /*  End Function foreign_aips2_test  */


/*EXPERIMENTAL_FUNCTION*/
pid_t foreign_aips2_run (CONST char *filename, int stdout_fd, int close_fd)
/*  [SUMMARY] Fork and execute a programme.
    <filename> The input filename for the programme.
    <stdout_fd> The output file descriptor.
    <close_fd> The file descriptor to close.
    [RETURNS] The child process ID on success, else -1.
*/
{
    char txt[STRING_LENGTH];
    char *argv[4];

    sprintf (txt, "in=%s", filename);
    argv[0] = "image2fits";
    argv[1] = txt;
    argv[2] = "out=-";
    argv[3] = NULL;
    return r_fork_and_execvp (0, stdout_fd, 2, 1, &close_fd, argv[0], argv);
}   /*  End Function foreign_aips2_run  */
