/*LINTLIBRARY*/
/*  pnm_read.c

    This code provides a PNM read facility.

    Copyright (C) 1998  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various utility routines for reading images
    in PNM format.


    Written by      Richard Gooch   5-MAY-1998

    Last updated by Richard Gooch   5-MAY-1998


*/
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <karma.h>
#include <karma_foreign.h>
#include <karma_chs.h>
#include <karma_ch.h>
#include <karma_ex.h>
#include <karma_a.h>


/*  Declarations of private functions follow  */


/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
unsigned int foreign_pnm_read_type (Channel channel, flag *binary)
/*  [SUMMARY] Read the type of a PNM file.
    <channel> The channel to read from.
    <binary> The value TRUE is written here if the file is in binary format,
    else FALSE is written here.
    [RETURNS] The type of PNM file.
*/
{
    unsigned int type;
    char txt[STRING_LENGTH];
    static char function_name[] = "foreign_pnm_read_type";

    if ( channel == NULL)
    {
	fprintf (stderr, "NULL pointer passed\n");
	a_prog_bug (function_name);
    }
    /*  Read magic number  */
    if (ch_read (channel, txt, 3) < 3)
    {
	fprintf (stderr, "Error reading\t%s\n", ERRSTRING);
	return (FOREIGN_FILE_FORMAT_UNKNOWN);
    }
    if (strncmp (txt, "P3", 2) == 0)
    {
	type = FOREIGN_FILE_FORMAT_PPM;
	*binary = FALSE;
    }
    else if (strncmp (txt, "P6", 2) == 0)
    {
	type = FOREIGN_FILE_FORMAT_PPM;
	*binary = TRUE;
    }
    else if (strncmp (txt, "P2", 2) == 0)
    {
	type = FOREIGN_FILE_FORMAT_PGM;
	*binary = FALSE;
    }
    else if (strncmp (txt, "P5", 2) == 0)
    {
	type = FOREIGN_FILE_FORMAT_PGM;
	*binary = TRUE;
    }
    else
    {
	fprintf (stderr, "Input not of PNM format\n");
	return (FOREIGN_FILE_FORMAT_UNKNOWN);
    }
    if ( !isspace (txt[2]) )
    {
	fprintf (stderr, "Input not of PNM format (whitespace missing)\n");
	return (FOREIGN_FILE_FORMAT_UNKNOWN);
    }
    return (type);
}   /*  End Function foreign_pnm_read_type  */

/*EXPERIMENTAL_FUNCTION*/
multi_array *foreign_pnm_read (Channel channel)
/*  [SUMMARY] Read a PNM file.
    <channel> The channel to read from.
    [RETURNS] A pointer to the multi_array data structure on success, else
    NULL.
*/
{
    flag binary;
    unsigned int filetype;
    multi_array *multi_desc;
    static char function_name[] = "foreign_pnm_read";

    filetype = foreign_pnm_read_type (channel, &binary);
    switch (filetype)
    {
      case FOREIGN_FILE_FORMAT_PGM:
	if ( ( multi_desc = foreign_pgm_read (channel,
					      FA_PGM_READ_BINARY, binary,
					      FA_PGM_READ_END) )
	     == NULL )
	{
	    fprintf (stderr, "%s: error reading PGM file\n", function_name);
	    ch_close (channel);
	    return (NULL);
	}
	break;
      case FOREIGN_FILE_FORMAT_PPM:
	/*  Read PPM file  */
	if ( ( multi_desc = foreign_ppm_read (channel,
					      FA_PPM_READ_BINARY, binary,
					      FA_PPM_READ_END) )
	    == NULL )
	{
	    fprintf (stderr, "%s: error reading PPM file\n", function_name);
	    ch_close (channel);
	    return (NULL);
	}
	break;
      default:
	return (NULL);
	/*break;*/
    }
    return (multi_desc);
}   /*  End Function foreign_pnm_read  */
