/*LINTLIBRARY*/
/*  misc.c

    This code provides random number generation routines.

    Copyright (C) 1992-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various Karma utility routines which generate random
  numbers.


    Written by      Richard Gooch   3-OCT-1992

    Updated by      Richard Gooch   14-NOV-1992

    Updated by      Richard Gooch   6-JAN-1993: Changed from  #include "*.h"
  to  #include <*.h>

    Updated by      Richard Gooch   23-AUG-1994: XOred in microseconds time
  value when computing seed.

    Updated by      Richard Gooch   4-OCT-1994: Changed to  ?rand48  routines
  in order to avoid having to link with buggy UCB compatibility library in
  Slowaris 2.3

    Updated by      Richard Gooch   26-NOV-1994: Moved to  packages/n/misc.c

    Updated by      Richard Gooch   25-JAN-1995: Added #ifdef OS_ConvexOS

    Updated by      Richard Gooch   12-APR-1996: Changed to new documentation
  format.

    Last updated by Richard Gooch   25-SEP-2001: Added #ifdef OS_Darwin.


*/
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <karma.h>
#include <karma_n.h>

#ifdef OS_ConvexOS
#  define MYRAND ( (double) random () / (double) (long) 0x7fffffff )
#  define MYSEED(seed) srandom(seed)
#endif
#ifdef OS_Darwin
#  define MYRAND ( (double) rand () / (double) (int) RAND_MAX )
#  define MYSEED(seed) srand(seed)
#endif
#ifndef MYRAND
extern double drand48 ();
#  define MYRAND drand48()
#  define MYSEED(seed) srand48(seed)
#endif


/*PUBLIC_FUNCTION*/
double n_gaussian ()
/*  [SUMMARY] Compute a random number with Gaussian distribution.
    [PURPOSE] This routine will compute a random number with a Gaussian
    distribution. The mean is 0.0 and the variance is 1.0
    [RETURNS] The number.
*/
{   
    double val = -6.0;
    int i;

    for (i = 0; i < 12; i++) val += n_uniform ();
    return (val);
}   /*  End Function n_gaussian  */

/*PUBLIC_FUNCTION*/
double n_uniform ()
/*  [SUMMARY] Compute a random number with Uniform distribution.
    [PURPOSE] This routine will compute a random number with a Uniform
    distribution. The range is from 0.0 to 1.0
    [RETURNS] The number.
*/
{   
    struct timeval tv;
    struct timezone tz;
    static int first_time = TRUE;

    if (first_time)
    {
	first_time = FALSE; 
	gettimeofday (&tv, &tz);
	tv.tv_sec ^= tv.tv_usec;
	MYSEED (tv.tv_sec);
    }
    return (MYRAND);
}   /*  End Function n_uniform  */
