/*LINTLIBRARY*/
/*  main.c
    This code implements low level communications.

    Copyright (C) 1992-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/
/*

    This file contains the various utility routines for manipulating Karma
    connections.


    Written by      Richard Gooch   14-AUG-1992

    Updated by      Richard Gooch   29-OCT-1992

    Updated by      Richard Gooch   16-DEC-1992: Added  r_getenv  and  r_setenv

    Updated by      Richard Gooch   25-DEC-1992: Added  r_gethostname  .

    Updated by      Richard Gooch   30-DEC-1992: Added  r_getppid  .

    Updated by      Richard Gooch   6-JAN-1993: Changed from  #include "*.h"
  to  #include <*.h>

    Updated by      Richard Gooch   24-JAN-1993: Declared  connect_to_port  at
  top of file (inside #ifdef).

    Updated by      Richard Gooch   12-APR-1993: Improved documentation for
  r_read  and added blocking of  SIGPIPE  to  r_write  .

    Updated by      Richard Gooch   20-JUL-1993: Added support for disc channel
  for Standard Input under VX/ MVX.

    Updated by      Richard Gooch   24-JUL-1993: Added support for opening
  character special devices and FIFOs by creating  r_open_file  .

    Updated by      Richard Gooch   29-JUL-1993: Added Linux support in
  r_setenv  .

    Updated by      Richard Gooch   20-NOV-1993: Made  alloc_port  tolerant of
  Linux kernels without TCP/IP support.

    Updated by      Richard Gooch   23-NOV-1993: Fixed bug introduced on 20-NOV
  which failed to call  htons(3)  prior to binding in  bind_inet  .

    Updated by      Richard Gooch   11-MAY-1994: Added un-named pipe support by
  creating  r_create_pipe  .

    Updated by      Richard Gooch   20-MAY-1994: Added  CONST  declaration
  where appropriate.

    Updated by      Richard Gooch   13-JUN-1994: Made implementation of
  r_setenv  completely internal.

    Updated by      Richard Gooch   14-AUG-1994: Extended functionality of
  r_open_file  .

    Updated by      Richard Gooch   14-AUG-1994: Improved error messages for
  r_open_file  .

    Updated by      Richard Gooch   3-OCT-1994: Added  #include <sys/filio.h>
  for sun4sol2.

    Updated by      Richard Gooch   4-OCT-1994: Created  mclear  routine in
  order to avoid having to link with buggy UCB compatibility library in
  Slowaris 2.3

    Updated by      Richard Gooch   10-OCT-1994: Fixed Unix connections under
  Slowaris 2.3 now that I see Slowaris falsely reports Unix sockets as named
  pipes.

    Updated by      Richard Gooch   31-OCT-1994: Changed from  sun4sol2  to
  sparc_Solaris

    Updated by      Richard Gooch   3-NOV-1994: Switched to OS_ and MACHINE_
  macros for machine determination.

    Updated by      Richard Gooch   14-NOV-1994: Removed unnecessary error
  message in  r_open_file  .

    Updated by      Richard Gooch   26-NOV-1994: Moved to  packages/r/main.c

    Updated by      Richard Gooch   7-DEC-1994: Stripped declaration of  errno

    Updated by      Richard Gooch   7-JAN-1995: Advertised <r_open_file>
  routine to documentation generator.

    Updated by      Richard Gooch   24-FEB-1995: Fixed logical test in
  <r_open_file>.

    Updated by      Richard Gooch   14-MAR-1995: Made <get_inet_addr_from_host>
  less verbose.

    Updated by      Richard Gooch   5-MAY-1995: Placate SGI compiler.

    Updated by      Richard Gooch   9-JUN-1995: Removed assumption that
  Kword32u is defined (problem on crayPVP machines).

    Updated by      Richard Gooch   15-JUN-1995: Made use of IS_ALIGNED macro.

    Updated by      Richard Gooch   16-JUN-1995: Created <r_get_fq_hostname>.

    Updated by      Richard Gooch   16-JUL-1995: Disabled NIS lookup of Karma
  port number base.

    Updated by      Richard Gooch   10-AUG-1995: Cope with Unix socket not
  existing in <connect_unix>.

    Updated by      Richard Gooch   19-FEB-1996: Cope with yet another error
  code from Slowass 2 in <connect_unix>.

    Updated by      Richard Gooch   13-APR-1996: Changed to new documentation
  format.

    Updated by      Richard Gooch   3-JUN-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch   23-AUG-1996: Support "local-slow" hostname.

    Updated by      Richard Gooch   23-MAY-1997: Fixed bug in
  <accept_connection_on_dock>: IP address was not converted from net to host.

    Updated by      Richard Gooch   12-JUL-1997: Changed boolean test from ||
  to && when checking error from stat call in <r_open_file>.

    Updated by      Richard Gooch   28-JUL-1997: Created <r_get_fd_filetype>.

    Updated by      Richard Gooch   2-SEP-1997: Switched to ERRSTRING.

    Updated by      Richard Gooch   9-SEP-1997: No longer print error message
  in <get_bytes_readable>.

    Updated by      Richard Gooch   24-DEC-1997: Created experimental
  <r_gethostbyaddr> function.

    Updated by      Richard Gooch   28-DEC-1997: Created unpublished
  <r_connect_to_udp> function.

    Updated by      Richard Gooch   29-DEC-1997: Updated <r_gethostbyaddr>
  function to accept address of 0.

    Updated by      Richard Gooch   17-FEB-1998: Created unpublished
  <r_alloc_udp> function.

    Updated by      Richard Gooch   14-MAY-1998: Created experimental function
  <r_create_socketpair>.

    Updated by      Richard Gooch   21-JUN-1998: Switched from CAN_ FORK to
  HAS_SIGNALS.

    Updated by      Richard Gooch   17-SEP-1998: Added some CONST declarations.

    Updated by      Richard Gooch   23-SEP-1998: Created experimental function
  <r_set_nodelay>.

    Updated by      Richard Gooch   2-NOV-1998: Worked around glibc 2 arrogance

    Updated by      Richard Gooch   14-DEC-1998: Increased receive buffer size
  for UDP connections.

    Updated by      Richard Gooch   22-DEC-1998: Changed <r_read> and <r_write>
  functions to use generic size values. Changed <r_close_dock> to return flag.

    Updated by      Richard Gooch   26-FEB-1999: Created experimental function
  <r_open_stdin2>.

    Updated by      Richard Gooch   15-AUG-1999: Made egcs happy.

    Updated by      Richard Gooch   26-JAN-2000: Created experimental functions
  <r_create_exclusive_file> and <r_set_file_lock>.

    Updated by      Richard Gooch   27-JAN-2000: Allow NULL pointers in
  <r_open_file>.

    Updated by      Richard Gooch   7-OCT-2000: Created unpublished
  <r_recvfrom> function.

    Updated by      Richard Gooch   27-APR-2001: Changed function
  <accept_connection_on_dock> to return error rather than exit and don't
  generate error message.

    Last updated by Richard Gooch   4-MAY-2001: Moved __STRICT_ANSI__ down to
  work around changes in kernel 2.4 headers.


*/
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <os.h>
#ifdef HAS_SOCKETS
#  define __STRICT_ANSI__
#  include <sys/socket.h>
#  include <sys/un.h>
#  include <sys/param.h>
#  include <fcntl.h>
#  include <sys/ioctl.h>
#  include <netdb.h>
#  include <netinet/in.h>
#  include <arpa/inet.h>
#  undef __STRICT_ANSI__
#endif
#ifdef HAS_SIGNALS
#  include <signal.h>
#endif
#ifdef OS_VXMVX
#  include <vx/vx.h>
#endif
#ifdef OS_Solaris
#  include <sys/filio.h>  /*  There may be better ways of doing this.  */
#endif
#include <karma.h>
#include <karma_r.h>


/*  Check for brain damaged platforms which don't define  S_ISSOCK
*/
#ifndef S_ISSOCK
#  if defined(S_IFMT) && defined(S_IFSOCK)
#    define S_ISSOCK(m)     (((m) & S_IFMT) == S_IFSOCK)
#  else
!!!! ERROR !!! *** Do not know how to check for sockets ****
#  endif
#endif

#if __GLIBC__ < 2     /*  The arrogance of the Glibc 2 bastards!  */
typedef int socklen_t;
#endif

#define NUM_DOCKS 2

#define SOCK_TYPE_STREAM 1
#define SOCK_TYPE_DGRAM  2

#if defined(HAS_SOCKETS) || defined(HAS_COMMUNICATIONS_EMULATION)
#  define COMMUNICATIONS_AVAILABLE
#endif

#ifdef HAS_SOCKETS
#  define INET_PORT_OFFSET 6200
#  define UNIX_SOCKET_DIR "/tmp/.KARMA_connections"
#  define UNIX_SOCKET_FILE "KARMA"
#  define INDEX_UNIX_DOCK 0
#  define INDEX_TCP_DOCK 1
#  define READ read
#  define WRITE write
#  define GETHOSTNAME gethostname
#  define GETPPID getppid
#endif

#ifdef HAS_COMMUNICATIONS_EMULATION
#  define READ read_from_connection
#  define WRITE write_to_connection
#  define GETHOSTNAME get_hostname
#  define GETPPID get_ppid
#endif


#ifdef OS_VXMVX
#  define MAX_DESCRIPTORS 100
#  define DESCRIPTOR_OFFSET 11000
#  define HR_ALLOC_PORT 0
#  define HR_CONNECT_TO_PORT 1
#  define HR_GET_INET_ADDR 2
#  define HR_OPEN_STDIN 3
#  define HR_GET_HOSTNAME 4
#endif

/*  Structure definitions  */
#ifdef OS_VXMVX
struct descriptor_type
{
    flag open;
    flag received_closure;
    int incoming_message_queue;
    int outgoing_message_queue;
    char *incoming_msg;
    int msg_pos;
    int msg_len;
};
#endif


/*  Private functions  */
STATIC_FUNCTION (void prog_bug, (char *function_name) );
STATIC_FUNCTION (void mclear, (void *buffer, unsigned int length) );
STATIC_FUNCTION (void mcopy,
		 (char *dest, CONST char *source, unsigned int length) );
#ifdef COMMUNICATIONS_AVAILABLE
STATIC_FUNCTION (int *alloc_port,
		 (unsigned int *port_number, unsigned int retries,
		  unsigned int *num_docks) );
#  ifdef HAS_SOCKETS
STATIC_FUNCTION (flag bind_unix, (int sock, unsigned int port_number) );
STATIC_FUNCTION (flag bind_inet, (int sock, unsigned int port_number,
				  char type) );
STATIC_FUNCTION (int connect_unix, (CONST char *filename) );
STATIC_FUNCTION (int connect_inet, (int sock, unsigned long addr,
				unsigned int port_number, char type) );
#  endif
STATIC_FUNCTION (void close_dock, (int dock_index) );
STATIC_FUNCTION (int connect_to_port, (unsigned long addr,
				       unsigned int port_number) );
STATIC_FUNCTION (int accept_connection_on_dock,
		 (int dock_index, unsigned long *addr, flag *local) );
STATIC_FUNCTION (flag close_connection, (int connection) );
STATIC_FUNCTION (int get_bytes_readable, (int connection) );
STATIC_FUNCTION (unsigned long get_inet_addr_from_host,
		 (CONST char *host, flag *local) );
static unsigned long conv_hostname_to_addr ();
STATIC_FUNCTION (int open_stdin, (unsigned int *type) );
#endif
#ifdef HAS_COMMUNICATIONS_EMULATION
static int read_from_connection ();
static int write_to_connection ();
static int get_hostname ();
static void init_control_connection ();
static int find_spare_descriptor ();
#endif


/*  Private data  */
static int inet_port_offset = -1;
static int docks[NUM_DOCKS];
static unsigned int num_docks_open = 0;
static unsigned int allocated_port_number;
#ifdef OS_VXMVX
static int parent_pid = -1;
static int incoming_control_message_queue = -1;
static int outgoing_control_message_queue = -1;
static struct descriptor_type descriptors[MAX_DESCRIPTORS];
#endif

/*  Public data  */
#ifndef HAS_ENVIRON
char **environ = NULL;
#endif


/*  Public routines follow  */

/*PUBLIC_FUNCTION*/
int *r_alloc_port (unsigned int *port_number, unsigned int retries,
		   unsigned int *num_docks)
/*  [SUMMARY] Allocate a Karma port whic can receive connections.
    [PURPOSE] This routine will allocate a Karma port for the module so that it
    can operate as a server (able to receive network connections).
    The routine will create a number of docks for one port. Each dock is an
    alternative access point for other modules to connect to this port.
    The close-on-exec flags of the docks are set such that the docks will
    close on a call to execve(2V).
    The docks are placed into non-blocking mode.
    <port_number> A pointer to the port number to allocate. The routine will
    write the actual port number allocated to this address. This must point to
    an address which lies on an <<int>> boundary.
    <retries> The number of succsessive port numbers to attempt to allocate
    before giving up. If this is 0, then the routine will give up immediately
    if the specified port number is in use.
    <num_docs> The number of docks allocated will be written here. This must
    point to an address which lies on an <<int>> boundary.
    [RETURNS] A pointer to a statically allocated array of docks on success,
    else NULL.
*/
{
    static char function_name[] = "r_alloc_port";

    if (num_docks_open > 0)
    {
	fprintf (stderr, "Port has already been allocated\n");
	prog_bug (function_name);
    }
    /*  Check alignment of  port_number  */
    if ( !IS_ALIGNED (port_number, sizeof *port_number) )
    {
	fputs ("Pointer to port number storage does not lie on an", stderr);
	fputs ("  int  boundary\n", stderr);
	prog_bug (function_name);
    }
    /*  Check alignment of  num_docks  */
    if ( !IS_ALIGNED (num_docks, sizeof *num_docks) )
    {
	fputs ("Pointer to number of docks storage does not lie on", stderr);
	fputs (" an  int  boundary\n", stderr);
	prog_bug (function_name);
    }
#ifdef COMMUNICATIONS_AVAILABLE
    return alloc_port (port_number, retries, num_docks);

#else
    fprintf (stderr, "No communications support\n");
    return (NULL);
#endif
}   /*  End Function r_alloc_port  */

/*PUBLIC_FUNCTION*/
flag r_close_dock (int dock)
/*  [SUMMARY] Close a dock.
    [PURPOSE] This routine will close a dock. If the dock was the last open
    dock for the port, then the entire port is closed and a new port may be
    allocated.
    <dock> The dock to close.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int dock_count;
    static char function_name[] = "r_close_dock";

    if (num_docks_open < 1)
    {
	fprintf (stderr, "No docks are open\n");
	prog_bug (function_name);
    }
#ifdef COMMUNICATIONS_AVAILABLE
    for (dock_count = 0; dock_count < NUM_DOCKS; ++dock_count)
    {
	if (dock == docks[dock_count])
	{
	    /*  Found the dock  */
	    close_dock (dock_count);
	    docks[dock_count] = -1;
	    --num_docks_open;
	    return (TRUE);
	}
    }
    fprintf (stderr, "Dock: %d does not exist\n", dock);
    prog_bug (function_name);

#else
    fprintf (stderr, "No communications support\n");
#endif

    return (FALSE);
}  /*  End Function r_close_dock  */

/*PUBLIC_FUNCTION*/
int r_connect_to_port (unsigned long addr, unsigned int port_number,
		       flag *local)
/*  [SUMMARY] Make a connection to a Karma port on some machine.
    <addr> The address of the machine to connect to. If the value of 0 is
    supplied for the address, the connection is made to a Karma server running
    on the local machine.
    <port_number> The port number to connect to.
    <local> If the connection is made to a port on the local host, then the
    value TRUE will be written here, else the value FALSE will be written here.
    [NOTE] The close-on-exec flags of the socket is set such that the socket
    will close on a call to execve(2V).
    [RETURNS] The file descriptor of the opened connection on success, else -1.
*/
{
#ifdef COMMUNICATIONS_AVAILABLE
    flag local_flag;
    int fd;
    static char function_name[] = "r_connect_to_port";

    /*  Check if self-connect  */
    if (addr == 0)
    {
	/*  Unix domain socket/ connection  */
	if ( (port_number == allocated_port_number) &&
	     (num_docks_open > 0) )
	{
	    fprintf (stderr, "Attempt to connect to oneself!\n");
	    prog_bug (function_name);
	}
	local_flag = TRUE;
    }
    else
    {
	/*  Internet/ TCP domain socket/ connection  */
	if (r_get_inet_addr_from_host ( (char *) NULL, (flag *) NULL ) == addr)
	{
	    /*  Connecting locally  */
	    local_flag = TRUE;
	    if ( (port_number == allocated_port_number) &&
		(num_docks_open > 0) )
	    {
		fprintf (stderr, "Attempt to connect to oneself!\n");
		prog_bug (function_name);
	    }
	}
	else
	{
	    /*  Not connecting locally  */
	    local_flag = FALSE;
	}
    }
    if ( ( fd = connect_to_port (addr, port_number) ) > -1 )
    {
	*local = local_flag;
    }
    return (fd);

#else
    fprintf (stderr, "No communications support\n");
    return (-1);
#endif
}   /*  End Function r_connect_to_port  */

/*UNPUBLISHED_FUNCTION*/
int r_connect_to_udp (int sock, unsigned long addr, unsigned int port_number)
/*  [SUMMARY] Make a connection to a UDP port on some machine.
    <sock> A UDP socket. If this is less than 0, a socket is created.
    <addr> The address of the machine to connect to.
    <port_number> The port number to connect to.
    [NOTE] The close-on-exec flags of the socket is set such that the socket
    will close on a call to execve(2V).
    [RETURNS] The file descriptor of the opened connection on success, else -1.
*/
{
    int fd;
    int buflen = 65536 + 32;

    fd = connect_inet (sock, addr, port_number, SOCK_TYPE_DGRAM);
    if (fd < 0) return (-1);
    if (sock >= 0) return (fd);
    /*  Leave enough space in the socket buffer so MAC address, alignment and
	other things won't steal payload space  */
    if (setsockopt (fd, SOL_SOCKET, SO_RCVBUF, &buflen, sizeof buflen) != 0)
    {
	fprintf (stderr, "Error setting socket options on UDP/IP socket\t%s\n",
		 ERRSTRING);
	close (sock);
	return (-1);
    }
    return (fd);
}   /*  End Function r_connect_to_udp  */

/*UNPUBLISHED_FUNCTION*/
int r_alloc_udp (unsigned int *port_number)
/*  [SUMMARY] Create a UDP socket and bind to a random port number.
    <port_number> The port number the socket was bound to.
    [NOTE] The close-on-exec flags of the socket is set such that the socket
    will close on a call to execve(2V).
    [RETURNS] The file descriptor of the socket on success, else -1.
*/
{
    int sock;
    int buflen = 65536 + 32;
    socklen_t len;
    struct sockaddr_in in_addr;

    len = sizeof in_addr;
    if ( ( sock = socket (AF_INET, SOCK_DGRAM, 0) ) < 0 )
    {
	fprintf (stderr, "Error creating UDP/IP socket\t%s\n", ERRSTRING);
	return (-1);
    }
    /*  Leave enough space in the socket buffer for MAC address, alignment and
	other things won't steal payload space  */
    if (setsockopt (sock, SOL_SOCKET, SO_RCVBUF, &buflen, sizeof buflen) != 0)
    {
	fprintf (stderr, "Error setting socket options on UDP/IP socket\t%s\n",
		 ERRSTRING);
	close (sock);
	return (-1);
    }
    /*  Try to bind to a random port  */
    memset (&in_addr, 0, sizeof in_addr);
    in_addr.sin_family = AF_INET;
    in_addr.sin_addr.s_addr = INADDR_ANY;
    /*  Try to bind to port number  */
    in_addr.sin_port = htons (0);
    if (bind (sock, (struct sockaddr *) &in_addr, (int) sizeof in_addr) != 0)
    {
	fprintf (stderr, "Error binding UDP/IP socket\t%s\n", ERRSTRING);
	if (close (sock) != 0)
	{
	    fprintf (stderr, "Error closing UDP/IP socket\t%s\n", ERRSTRING);
	}
	exit (RV_SYS_ERROR);
    }
    if (getsockname (sock, (struct sockaddr *) &in_addr, &len) != 0)
    {
	fprintf (stderr, "Error getting UDP/IP socket name\t%s\n", ERRSTRING);
	if (close (sock) != 0)
	{
	    fprintf (stderr, "Error closing UDP/IP socket\t%s\n", ERRSTRING);
	}
	exit (RV_SYS_ERROR);
    }
    /*  Set close-on-exec flag  */
    if (fcntl (sock, F_SETFD, 1) == -1)
    {
	fprintf (stderr,
		 "Error setting close-on-exec flag for UDP/IP socket\t%s\n",
		 ERRSTRING);
	if (close (sock) != 0)
	{
	    fprintf (stderr, "Error closing UDP/IP socket\t%s\n", ERRSTRING);
	}
	exit (RV_SYS_ERROR);
    }
    *port_number = ntohs (in_addr.sin_port);
    return (sock);
}   /*  End Function r_alloc_udp  */

/*PUBLIC_FUNCTION*/
int r_accept_connection_on_dock (int dock, unsigned long *addr, flag *local)
/*  [SUMMARY] Accept a connection on a dock.
    <dock> The dock.
    <addr> The address of the host connecting to the dock will be written here.
    <local> If the connection is a local connection, then the routine will
    write the value TRUE here, else it will write the value FALSE.
    [RETURNS] A connection on success, else -1.
*/
{
#ifdef COMMUNICATIONS_AVAILABLE
    unsigned int dock_count;
    static char function_name[] = "r_accept_connection_on_dock";

    if (num_docks_open < 1)
    {
	fprintf (stderr, "No docks are open\n");
	prog_bug (function_name);
    }
    dock_count = 0;
    while ( (dock != docks[dock_count]) && (dock_count < NUM_DOCKS) )
    {
	++dock_count;
    }
    if (dock_count >= NUM_DOCKS)
    {
	fprintf (stderr, "Dock: %d does not exist\n", dock);
	prog_bug (function_name);
    }
    return accept_connection_on_dock (dock_count, addr, local);

#else
    fprintf (stderr, "No communications support\n");
    return (-1);
#endif
}  /*  End Function r_accept_connection_on_dock  */

/*PUBLIC_FUNCTION*/
flag r_close_connection (int connection)
/*  [SUMMARY] Close a connection.
    <connection> The connection to close.
    [RETURNS] TRUE on success, else FALSE.
*/
{
#ifdef COMMUNICATIONS_AVAILABLE
    /*static char function_name[] = "r_close_connection";*/

    return close_connection (connection);

#else
    fprintf (stderr, "No communications support\n");
    return (FALSE);
#endif
}  /*  End Function r_close_connection  */

/*PUBLIC_FUNCTION*/
int r_get_bytes_readable (int connection)
/*  [SUMMARY] Get bytes readable on a connection.
    [PURPOSE] This routine will determine the minimum number of bytes readable
    on a connection. There may be more bytes readable than indicated.
    <connection> The connection.
    [RETURNS] The number of bytes readable on success, else -1.
*/
{
#ifdef COMMUNICATIONS_AVAILABLE
    return get_bytes_readable (connection);

#else
    fprintf (stderr, "No communications support\n");
    return (-1);
#endif
}   /*  End Function r_get_bytes_readable  */

/*PUBLIC_FUNCTION*/
unsigned long r_get_inet_addr_from_host (CONST char *host, flag *local)
/*  [SUMMARY] Get the first listed Internet address of a hostname.
    <host> The hostname.
    <local> If the specified host is the local machine, then the routine will
    write the value TRUE here, else it will write the value FALSE here. If
    this is NULL, nothing is written here.
    [RETURNS] The Internet address on success (in host byte order), else 0.
*/
{
#ifdef COMMUNICATIONS_AVAILABLE
    return get_inet_addr_from_host (host, local);

#else
    fprintf (stderr, "No communications support\n");
    return (0);
#endif
}   /*  End Function r_get_inet_addr_from_host  */

/*PUBLIC_FUNCTION*/
iaddr r_read (int fd, char *buf, uaddr nbytes)
/*  [SUMMARY] Read bytes from a file descriptor.
    [PURPOSE] This routine is similar to the system <<read(2)>> call, except
    that the number of bytes requested is always returned (except on error or
    closure). Hence, if the descriptor references a socket, the routine will
    read as much data as was requested, rather than a lesser amount due to
    packetisation or interrupted system calls.
    <fd> The file descriptor to read from. This descriptor must NOT be set to
    non-blocking IO.
    <buf> The buffer in which to write the data.
    <nbytes> The number of bytes to read.
    [RETURNS] The number of bytes requested on success, the number of bytes
    read on end of file (or closure) and -1 on error.
*/
{
#ifdef COMMUNICATIONS_AVAILABLE
    uaddr total_bytes_read = 0;
    iaddr bytes_read;

    while (total_bytes_read < nbytes)
    {
	errno = 0;
	if ( ( bytes_read = READ (fd, buf + total_bytes_read,
				  nbytes - total_bytes_read) )
	    < 0 )
	{
	    /*  Error  */
#  ifdef EINTR
	    if (errno != EINTR)
	    {
		/*  Unrecoverable error  */
		return (-1);
	    }
#  else
	    return (-1);
#  endif
	}
	else
	{
	    /*  No error condition  */
	    if (bytes_read == 0)
	    {
		/*  End-of-file  */
		return (total_bytes_read);
	    }
	    total_bytes_read += bytes_read;
	}
    }
    return (total_bytes_read);

#else
    fprintf (stderr, "No communications support\n");
    return (-1);
#endif
}   /*  End Function r_read  */

/*PUBLIC_FUNCTION*/
iaddr r_write (int fd, CONST char *buf, uaddr nbytes)
/*  [SUMMARY] Write bytes to a file descriptor.
    [PURPOSE] This routine is similar to the system <<write(2)>> call, except
    that the number of bytes requested is always returned (except on error).
    Hence, if the descriptor references a socket, the routine will write as
    much data as was requested, rather than a lesser amount due to
    packetisation or interrupted system calls.
    <fd> The file descriptor to write to. This descriptor must NOT be set to
    non-blocking IO.
    <buf> The buffer in which to write the data.
    <nbytes> The number of bytes to write.
    [NOTE] The routine will force  SIGPIPE  to be ignored.
    [RETURNS] The number of bytes requested on success, else -1 indicating
    error.
*/
{
#ifdef COMMUNICATIONS_AVAILABLE
    uaddr total_bytes_written = 0;
    iaddr bytes_written;
#  ifdef SIGPIPE
    static flag must_ignore_sigpipe = TRUE;
#  endif

#  ifdef SIGPIPE
    if (must_ignore_sigpipe)
    {
	signal (SIGPIPE, SIG_IGN);
	must_ignore_sigpipe = FALSE;
    }
#  endif
    while (total_bytes_written < nbytes)
    {
	errno = 0;
	if ( ( bytes_written = WRITE (fd, buf + total_bytes_written,
				      nbytes - total_bytes_written) )
	    < 0 )
	{
	    /*  Error  */
#  ifdef EINTR
	    if (errno != EINTR)
	    {
		/*  Unrecoverable error  */
		return (-1);
	    }
#  else
	    return (-1);
#  endif
	}
	else
	{
	    /*  No error condition  */
	    total_bytes_written += bytes_written;
	}
    }
    return (total_bytes_written);

#else
    fprintf (stderr, "No communications support\n");
    return (-1);
#endif
}   /*  End Function r_write  */

/*UNPUBLISHED_FUNCTION*/
int r_recvfrom (int sock, void *buf, int length,
		unsigned long *addr, unsigned int *port_number)
/*  [SUMMARY] Receive a message on a datagram socket
    <sock> The socket.
    <buf> The buffer to write the message to.
    <length> The length of the buffer.
    <addr> The address of the originating machine is written here. This may be
    NULL.
    <port_number> The port number of the originating socket is written here.
    This may be NULL.
    [RETURNS] The number of bytes read on success, else -1.
*/
{
    socklen_t fromlen;
    struct sockaddr_in in_addr;
    static char function_name[] = "r_recvfrom";

    fromlen = sizeof in_addr;
    length = recvfrom (sock, buf, length, 0,
		       (struct sockaddr *) &in_addr, &fromlen);
    if (length < 0) return (-1);
    if (fromlen != sizeof in_addr)
    {
	fprintf (stderr, "%s: addrlen: %d is not: %d\n",
		 function_name, fromlen, sizeof in_addr);
	return (-1);
    }
    if (addr != NULL) *addr = ntohl (in_addr.sin_addr.s_addr);
    if (port_number != NULL) *port_number = ntohs (in_addr.sin_port);
    return (length);
}   /*  End Function r_recvfrom  */

/*PUBLIC_FUNCTION*/
flag r_test_input_event (int connection)
/*  [SUMMARY] Test for activity on a connection.
    [PURPOSE] This routine will test if there is input activity on a
    connection. This activity also covers the case of connection closure. The
    connection descriptor must be given by  connection  .
    [NOTE] This routine is only available on platforms which emulate the
    communications facilities of Unix. It is NOT available on standard Unix
    systems.
    [RETURNS] TRUE if there is some input activity, else FALSE.
*/
{
    static char function_name[] = "r_test_input_event";
#ifdef OS_VXMVX
    void *msg;

    connection -= DESCRIPTOR_OFFSET;
    if ( (connection < 0) || (connection >= MAX_DESCRIPTORS) )
    {
	fprintf (stderr, "Bad descriptor: %d\n",
			connection + DESCRIPTOR_OFFSET);
	prog_bug (function_name);
    }
    if (!descriptors[connection].open)
    {
	fprintf (stderr, "Descriptor: %d not open\n",
			connection + DESCRIPTOR_OFFSET);
	prog_bug (function_name);
    }
    if (descriptors[connection].incoming_msg != NULL)
    {
	return (TRUE);
    }
    if ( ( msg = task_recv_msg (descriptors[connection].incoming_message_queue,
				0, 0) ) == NULL )
    {
	return (FALSE);
    }
    descriptors[connection].incoming_msg = msg;
    descriptors[connection].msg_len = task_sizeof_msg (msg);
    descriptors[connection].msg_pos = 0;
    if (descriptors[connection].msg_len == 0)
    {
	descriptors[connection].received_closure = TRUE;
    }
    return (TRUE);

#else  /*  OS_VXMVX  */
    fprintf (stderr, "No communications emulation\n");
    prog_bug (function_name);
    return (FALSE);
#endif  /*  OS_VXMVX  */
}   /*  End Function r_test_input_event  */

/*PUBLIC_FUNCTION*/
int r_open_stdin (flag *disc)
/*  [SUMMARY] Open the standard input.
    <disc> The routine will write the value TRUE here if the standard input is
    a disc, else it will write FALSE.
    [RETURNS] The descriptor on success, else -1.
*/
{
    int fd;
    unsigned int type;

    if ( ( fd = r_open_stdin2 (&type) ) < 0) return (-1);
    switch (type)
    {
      case KFTYPE_DISC:
	*disc = TRUE;
	break;
      default:
	*disc = FALSE;
	break;
    }
    return (fd);
}   /*  End Function r_open_stdin  */

/*EXPERIMENTAL_FUNCTION*/
int r_open_stdin2 (unsigned int *type)
/*  [SUMMARY] Open the standard input.
    <disc> The routine will write the type of the file here.
    [RETURNS] The descriptor on success, else -1.
*/
{
#ifdef COMMUNICATIONS_AVAILABLE
    return open_stdin (type);

#else
    fprintf (stderr, "No communications support\n");
    return (-1);
#endif
}   /*  End Function r_open_stdin2  */

/*PUBLIC_FUNCTION*/
CONST char *r_getenv (CONST char *name)
/*  [SUMMARY] Get the value of an environment variable.
    <name> The name of the environment variable.
    [RETURNS] A pointer to the value string if present, else NULL.
*/
{
#ifdef HAS_ENVIRON
    return getenv (name);

#else
    int length;
    char **env;
    char cmp[STRING_LENGTH];
    extern char **environ;

    init_control_connection ();
    if (environ == NULL)
    {
	return (NULL);
    }
    strcpy (cmp, name);
    length = strlen (cmp);
    if (cmp[length] != '=')
    {
	cmp[length] = '=';
	cmp[length + 1] = '\0';
	++length;
    }
    for (env = environ; *env != NULL; ++env)
    {
	if (strncmp (cmp, *env, length) == 0)
	{
	    return (*env + length);
	}
    }
    return (NULL);
#endif
}   /*  End Function r_getenv  */

/*PUBLIC_FUNCTION*/
int r_setenv (CONST char *env_name, CONST char *env_value)
/*  [SUMMARY] Set an environment variable.
    [PURPOSE] This routine will provide a consistent interface to set
    environment variables. This is necessary because the "standard" C library
    routines: <<putenv>> or <<setenv>> (depending on the particular standard
    C library supplied with the operating system) are in fact not standard.
    <env_name> The environment variable to create or change.
    <env_value> The string value to set the variable to.
    [RETURNS] 0 on success, else -1.
*/
{
    int num_strings, len;
    char **env;
    char *env_string;
    extern char **environ;
    static char **old_environ = NULL;

    len = strlen (env_name) + 1 + strlen (env_value);
    if ( ( env_string = malloc (len + 1) ) == NULL )
    {
	fprintf (stderr, "Error allocating string\n");
	return (-1);
    }
    strcpy (env_string, env_name);
    strcat (env_string, "=");
    strcat (env_string, env_value);
    for (num_strings = 0; environ[num_strings] != NULL; ++num_strings);
    if ( ( env = (char **) malloc ( sizeof *env * (num_strings + 2) ) )
	== NULL )
    {
	free (env_string);
	fprintf (stderr, "Error allocating environment\n");
	return (-1);
    }
    mcopy ( (char *) env, (char *) environ, num_strings * sizeof *env );
    env[num_strings++] = env_string;
    env[num_strings] = NULL;
    if (old_environ == environ) free ( (char *) environ );
    environ = env;
    old_environ = env;
    return (0);
}   /*  End Function r_setenv  */

/*PUBLIC_FUNCTION*/
void r_gethostname (char *name, unsigned int namelen)
/*  [SUMMARY] This routine will determine the local hostname.
    <name> The hostname will be written here. It is guaranteed to be null
    terminated.
    <namelen> The size of the buffer.
    [RETURNS] Nothing.
*/
{
#ifdef COMMUNICATIONS_AVAILABLE

    if (GETHOSTNAME (name, namelen - 1) != 0)
    {
	fprintf (stderr, "Error getting hostname\t%s\n", ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    name[namelen - 1] = '\0';

#else
    fprintf (stderr, "No communications support\n");
    exit (RV_UNDEF_ERROR);
#endif
}   /*  End Function r_gethostname  */

/*PUBLIC_FUNCTION*/
flag r_get_fq_hostname (char *name, unsigned int namelen)
/*  [SUMMARY] This routine will get the fully qualified local hostname.
    <name> The hostname will be written here. It is guaranteed to be null
    terminated.
    <namelen> The size of the buffer.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    char host[STRING_LENGTH];
    struct hostent *host_ptr;

    r_gethostname (host, STRING_LENGTH);
    /*  Get host info  */
    if ( ( host_ptr = gethostbyname (host) ) == NULL )
    {
	fprintf (stderr, "Host: \"%s\" not in database\n", host);
	return (FALSE);
    }
    strncpy (name, host_ptr->h_name, namelen);
    name[namelen - 1] = '\0';
    return (TRUE);
}   /*  End Function r_get_fq_hostname  */

/*PUBLIC_FUNCTION*/
int r_getppid ()
/*  [SUMMARY] Get the parent process ID.
    [RETURNS] The parent process ID.
*/
{
#ifdef COMMUNICATIONS_AVAILABLE
    int ppid;

    if ( ( ppid = GETPPID () ) < 0 )
    {
	fprintf (stderr, "Error getting parent process ID\t%s\n", ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    return (ppid);

#else
    fprintf (stderr, "No communications support\n");
    exit (RV_UNDEF_ERROR);
#endif
}   /*  End Function r_getppid  */

/*PUBLIC_FUNCTION*/
int r_open_file (CONST char *filename, int flags, int mode,
		 unsigned int *filetype, unsigned int *blocksize)
/*  [SUMMARY] Open a named file.
    [PURPOSE] This routine will open a file. The file may be a regular disc
    file, a named FIFO, a character special device, a Unix domain socket or a
    TCP/IP connection (where supported). This routine provides an enhanced
    interface to the <<open(2)>>, <<socket(2)>> and <<connect(2)>> routines.
    <filename> The pathname of the file to open. This parameter has the same
    meaning as the first parameter to <<open(2)>>.
    Filenames of the form: "//tcpIP/{hostname}:{port}" indicate a connection to
    a TCP/IP port on host {hostname} with raw port number {port} is
    requested.
    <flags> The access flags. This parameter has the same meaning as the second
    parameter to <<open(2)>>.
    <mode> If the file is created, the file protection modes will be determined
    by the combination of the process  umask  and this value. This parameter
    has the same meaning as the third parameter to <<open(2)>>.
    <filetype> The type of the file will be written here. This may be NULL.
    <blocksize> The blocksize (in bytes) of the file will be written here (this
    is only valid for a disc file). If the blocksize cannot be determined, the
    value 0 will be written here. This may be NULL.
    [RETURNS] The file descriptor on success, else -1 and sets <<errno>> with
    the error code.
*/
{
    flag stat_done = TRUE;
    int fd;
    struct stat statbuf;
#ifdef HAS_SOCKETS
    flag local;
    long port;
    unsigned int lfiletype, lblocksize;
    unsigned long addr;
    CONST char *hostname_ptr, *colon_ptr;
    char hostname[STRING_LENGTH];
#endif

    if (filetype == NULL) filetype = &lfiletype;
    if (blocksize == NULL) blocksize = &lblocksize;
#ifdef HAS_SOCKETS
    if (strncmp (filename, "//tcpIP/", 8) == 0)
    {
	hostname_ptr = filename + 8;
	if (flags != O_RDWR)
	{
	    fprintf (stderr, "Bad access flags: %d for TCP/IP connection\n",
		     flags);
	    return (-1);
	}
	if ( ( colon_ptr = (CONST char *) strchr (hostname_ptr, ':') )
	     == NULL )
	{
	    fprintf (stderr, "Bad IP syntax filename: \"%s\"\n",
		     filename);
	    return (-1);
	}
	strncpy (hostname, hostname_ptr, colon_ptr - hostname_ptr);
	hostname[colon_ptr - hostname_ptr] = '\0';
	if ( ( port = atol (colon_ptr + 1) ) < 0 )
	{
	    fprintf (stderr,"Bad port number for IP syntax filename: \"%s\"\n",
		     filename);
	    return (-1);
	}
	if ( ( addr = r_get_inet_addr_from_host (hostname, &local) ) == 0 )
	{
	    return (-1);
	}
	if (local) *filetype = KFTYPE_LOCAL_tcpIP_CONNECTION;
	else *filetype = KFTYPE_REMOTE_tcpIP_CONNECTION;
	*blocksize = 0;
	return connect_inet (-1, addr, port, SOCK_TYPE_STREAM);
    }
#endif
    /*  Get stats on file  */
    if (stat (filename, &statbuf ) != 0)
    {
	if ( (errno != ENOENT) && ( (flags & O_CREAT) == 0 ) )
	{
	    fprintf (stderr, "Error getting file stats for file: \"%s\"\t%s\n",
		     filename, ERRSTRING);
	    return (-1);
	}
	stat_done = FALSE;
    }
    else stat_done = TRUE;
    if (stat_done)
    {
	if ( S_ISDIR (statbuf.st_mode) )
	{
	    fprintf (stderr, "File: \"%s\" is a directory: cannot open\n",
		     filename);
	    return (-1);
	}
	else if ( S_ISREG (statbuf.st_mode) )
	{
	    *filetype = KFTYPE_DISC;
#ifdef OS_IRIX5
	    *blocksize = 0;
#else
	    *blocksize = statbuf.st_blksize;
#endif
	}
	else if ( S_ISCHR (statbuf.st_mode) )
	{
	    *filetype = KFTYPE_CHARACTER;
	    *blocksize = 0;
	}
	else if ( S_ISFIFO (statbuf.st_mode) )
	{
	    *filetype = KFTYPE_FIFO;
	    *blocksize = 0;
	}
	else if ( S_ISSOCK (statbuf.st_mode) )
	{
	    if (flags != O_RDWR)
	    {
		fprintf (stderr, "Bad access flags: %d for Unix socket\n",
			 flags);
		return (-1);
	    }
	    *filetype = KFTYPE_UNIX_CONNECTION;
	    *blocksize = 0;
	}
	else
	{
	    fprintf (stderr, "Illegal file mode: %u\n",
		     (unsigned int) statbuf.st_mode);
	    return (-1);
	}
#ifdef HAS_SOCKETS
	if (*filetype == KFTYPE_UNIX_CONNECTION)
	{
	    return connect_unix (filename);
	}
#  ifdef OS_Solaris
	/*  Stupid Slowaris stats Unix sockets as named pipes. Arrrggghh!  */
	if (*filetype == KFTYPE_FIFO)
	{
	    if ( ( fd = connect_unix (filename) ) >= 0 )
	    {
		*filetype = KFTYPE_UNIX_CONNECTION;
		return (fd);
	    }
	}
#  endif
#endif
    }
    /*  Open file descriptor  */
    if ( ( fd = open (filename, flags, mode) ) < 0 )
    {
	return (-1);
    }
    if (!stat_done)
    {
	if (fstat (fd, &statbuf) != 0)
	{
	    fprintf (stderr, "Error stat'ing descriptor\n");
	    close (fd);
	    return (-1);
	}
	if ( S_ISREG (statbuf.st_mode) )
	{
	    *filetype = KFTYPE_DISC;
#ifdef OS_IRIX5
	    *blocksize = 0;
#else
	    *blocksize = statbuf.st_blksize;
#endif
	}
	else
	{
	    fprintf (stderr,"Illegal file mode: %u (should be regular file)\n",
		     (unsigned int) statbuf.st_mode);
	    close (fd);
	    return (-1);
	}
    }
    return (fd);
}   /*  End Function r_open_file  */

/*PUBLIC_FUNCTION*/
int r_create_pipe (int *read_fd, int *write_fd)
/*  [SUMMARY] Create an un-named pipe.
    [PURPOSE] This routine will create an un-named pipe. This routine provides
    an enhanced interface to the <<pipe(2)>> routine.
    <read_fd> The file descriptor corresponding to the read end of the pipe
    will be written here.
    <write_fd> The file descriptor corresponding to the write end of the pipe
    will be written here.
    [RETURNS] 0 on success, else -1 and sets <<errno>> with the error code.
*/
{
#ifdef HAS_SOCKETS
    int retval;
    int fds[2];
#endif
    /*static char function_name[] = "r_create_pipe";*/

#ifdef HAS_SOCKETS
    retval = pipe (fds);
    *read_fd = fds[0];
    *write_fd = fds[1];
    return (retval);
#else
    fprintf (stderr, "No pipe support\n");
    return (-1);
#endif
}   /*  End Function r_create_pipe  */

/*EXPERIMENTAL_FUNCTION*/
int r_create_socketpair (int *fd1, int *fd2)
/*  [SUMMARY] Create an un-named pair of connected sockets.
    [PURPOSE] This routine will create an un-named pair of connected sockets.
    This routine provides an enhanced interface to the <<socketpair(2)>>
    routine.
    <fd1> One end of the socketpair is written here.
    <fd2> The other end of the socketpair is written here.
    [RETURNS] 0 on success, else -1 and sets <<errno>> with the error code.
*/
{
#ifdef HAS_SOCKETS
    int retval;
    int fds[2];
#endif
    /*static char function_name[] = "r_create_socketpair";*/

#ifdef HAS_SOCKETS
    retval = socketpair (AF_UNIX, SOCK_STREAM, 0, fds);
    *fd1 = fds[0];
    *fd2 = fds[1];
    return (retval);
#else
    fprintf (stderr, "No socket support\n");
    return (-1);
#endif
}   /*  End Function r_create_socketpair  */

/*EXPERIMENTAL_FUNCTION*/
flag r_get_fd_filetype (int fd, unsigned int *type)
/*  [SUMMARY] Get filetype for a descriptor.
    <fd> The file descriptor.
    <type> The filetype is written here.
    [RETURNS] TRUE on success, else FALSE and an error message is displayed.
*/
{
    struct stat statbuf;

    if (fstat (fd, &statbuf) != 0)
    {
	fprintf (stderr, "Error getting stats on descriptor: %d\t%s\n",
		 fd, ERRSTRING);
	return (-1);
    }
    if ( S_ISREG (statbuf.st_mode) ) *type = KFTYPE_DISC;
    else if ( S_ISDIR (statbuf.st_mode) ) *type = KFTYPE_DIRECTORY;
    else if ( S_ISCHR (statbuf.st_mode) ) *type = KFTYPE_CHARACTER;
    else if ( S_ISBLK (statbuf.st_mode) ) *type = KFTYPE_BLOCK;
    else if ( S_ISFIFO (statbuf.st_mode) ) *type = KFTYPE_FIFO;
    else if ( S_ISSOCK (statbuf.st_mode) ) *type = KFTYPE_SOCKET;
    else
    {
	/*  Illegal object  */
	fprintf (stderr, "Illegal descriptor type\n");
	return (FALSE);
    }
    return (TRUE);
}   /*  End Function r_get_fd_filetype  */

/*EXPERIMENTAL_FUNCTION*/
flag r_gethostbyaddr (char *name, unsigned int length, unsigned long addr)
/*  [SUMMARY] Get hostname from Internet Address.
    <name> The hostname is written here.
    <length> The length of the hostname buffer.
    <addr> The Internet Address. If this is 0 the local hostname is given.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned long naddr;
    struct hostent *hent;
    static char function_name[] = "r_gethostbyaddr";

    if (addr == 0) addr = r_get_inet_addr_from_host (NULL, NULL);
    naddr = htonl (addr);
    if ( ( hent = gethostbyaddr ( (char *) &naddr, sizeof naddr, AF_INET) )
	 == NULL )
    {
	fprintf (stderr, "%s: error getting hostname for addr: 0x%lx\t%s\n",
		 function_name, addr, ERRSTRING);
	return (FALSE);
    }
    strncpy (name, hent->h_name, length);
    name[length - 1] = '\0';
    return (TRUE);
}   /*  End Function r_gethostbyaddr  */

/*EXPERIMENTAL_FUNCTION*/
flag r_set_nodelay (int fd)
/*  [SUMMARY] Set the no-delay flag for a descriptor if appropriate.
    <fd> The file descriptor.
    [RETURNS] TRUE on success, else FALSE.
*/
{
#ifdef TCP_NODELAY
    int optval = 1;

    if (setsockopt (fd, IPPROTO_TCP, TCP_NODELAY, &optval, sizeof optval) == 0)
	return (TRUE);
    else return (FALSE);
#else
    return (FALSE);
#endif
}   /*  End Function r_set_nodelay  */

/*EXPERIMENTAL_FUNCTION*/
flag r_create_exclusive_file (CONST char *filename, flag do_wait)
/*  [SUMMARY] Exclusively create a file (typically for locking purposes).
    [PURPOSE] This routine will create a file, guaranteeing that no other
    process (on any machine) created the file.
    <filename> The name of the file to create.
    <do_wait> If TRUE, the routine will wait until the file is created, or a
    failure occurs.
    [RETURNS] TRUE if the file was created exclusively by the process, else
    FALSE.
*/
{
    int fd;
    struct stat statbuf;
    char unique_name[STRING_LENGTH];
    static flag first_time = TRUE;
    static char hostname[STRING_LENGTH];
    static char function_name[] = "r_create_exclusive_file";

    if (first_time)
    {
	first_time = FALSE;
	r_gethostname (hostname, STRING_LENGTH);
    }
    sprintf ( unique_name, "%s.%s.%d", filename, hostname, (int) getpid () );
    if ( ( fd = open (unique_name, O_RDWR | O_CREAT | O_EXCL, 0) ) < 0 )
    {
	fprintf (stderr, "%s: error creating file: \"%s\"\t%s\n",
		 function_name, unique_name, ERRSTRING);
	return (FALSE);
    }
    close (fd);
    while (TRUE)
    {
	if (link (unique_name, filename) == 0)
	{
	    if (stat (unique_name, &statbuf) != 0)
	    {
		fprintf (stderr, "%s: error stating file: \"%s\"\t%s\n",
			 function_name, unique_name, ERRSTRING);
		break;
	    }
	    if (statbuf.st_nlink == 2)
	    {
		unlink (unique_name);
		return (TRUE);
	    }
	}
	if (do_wait) sleep (1);
	else break;
    }
    unlink (unique_name);
    return (FALSE);
}   /*  End Function r_create_exclusive_file  */

/*EXPERIMENTAL_FUNCTION*/
flag r_set_file_lock (int fd, flag lock, flag do_wait)
/*  [SUMMARY] Grab or release the lock on a file.
    <fd> The file descriptor to (un)lock.
    <lock> If TRUE, the file is locked, else it is unlocked.
    <do_wait> If TRUE, the routine will wait until the lock is grabbed.
    [RETURNS] TRUE on success, else FALSE.
*/
{
#ifdef F_SETLK
    struct flock fl;
#endif

#undef LOCKING_WORKS
#ifdef F_SETLK
#define LOCKING_WORKS
    fl.l_type = lock ? F_WRLCK : F_UNLCK;
    fl.l_whence = 0;
    fl.l_start = 0L;
    fl.l_len = 0L;
    if (fcntl (fd, do_wait ? F_SETLKW : F_SETLK, &fl) == -1)
    {
	fprintf (stderr, "Error %slocking file: %d with fcntl\t%s\n",
		 lock ? "" : "un", fd, ERRSTRING);
	return (FALSE);
    }
    return (TRUE);
#endif

#ifndef LOCKING_WORKS
    fprintf (stderr, "File locking not implemented\n");
    return (FALSE);
#endif
}   /*  End Function r_set_file_lock  */


/*  Private functions follow  */

#ifdef HAS_SOCKETS
static unsigned long get_inet_addr_from_host (CONST char *host, flag *local)
/*  This routine will get the first listed Internet address from the hostname
    string pointed to by  host  .
    IF the specified host is the local machine, then the routine will write
    the value TRUE to the storage pointed to by  local  ,else it will write
    the value FALSE here. If this is NULL, nothing is written here.
    The routine returns the Internet address on success (in host byte order),
    else it returns 0.
*/
{
    unsigned long addr;
    static unsigned long local_addr = 0;
    static char local_hostname[MAXHOSTNAMELEN + 1];

    if (local_addr < 1)
    {
	/*  Get my hostname  */
	if (gethostname (local_hostname, MAXHOSTNAMELEN + 1) != 0)
	{
	    fprintf (stderr, "Error getting local hostname\t%s", ERRSTRING);
	    return (0);
	}
	/*  Get my host address  */
	if ( ( local_addr = conv_hostname_to_addr (local_hostname) ) == 0 )
	{
	    fprintf (stderr, "Error getting local host address\n");
	    return (0);
	}
    }
    if ( (host == NULL) || (strcmp (host, "unix") == 0) ||
	 (strcmp (host, "localhost") == 0) ||
	 (strcmp (host, "local-slow") == 0) )
    {
	/*  Local machine  */
	if (local != NULL) *local = TRUE;
	return (local_addr);
    }
    if ( ( addr = conv_hostname_to_addr (host) ) == 0 ) return (0);
    if (local != NULL) *local = (addr == local_addr) ? TRUE : FALSE;
    return (addr);
}   /*  End Function get_inet_addr_from_host  */

static unsigned long conv_hostname_to_addr (host)
/*  This routine will get the first listed Internet address from the hostname
    string pointed to by  host  .This MUST be a valid hostname.
    The routine returns the Internet address on success (in host byte order),
    else it returns 0.
*/
char *host;
{
    unsigned long addr = 0;  /*  Initialised to keep compiler happy  */
    struct hostent *host_ptr;
    static char function_name[] = "__r_conv_hostname_to_addr";

    if ( isascii (host[0]) && isdigit (host[0]) )
    {
	/*  Numeric Internet address  */
	return ntohl ( inet_addr (host) );
    }
    /*  Get host info  */
    if ( ( host_ptr = gethostbyname (host) ) == NULL )
    {
	fprintf (stderr, "Host: \"%s\" not in database\n", host);
	return (0);
    }
    switch (host_ptr->h_length)
    {
#ifdef Kword32u
      case 4:
	addr = (unsigned long) *(Kword32u *) host_ptr->h_addr_list[0];
	break;
#endif
#ifdef Kword64u
      case 8:
	addr = (unsigned long) *(Kword64u *) host_ptr->h_addr_list[0];
	break;
#endif
      default:
	fprintf (stderr, "Host: \"%s\" address length: %d not supported\n",
		 host, host_ptr->h_length);
	prog_bug (function_name);
	break;
    }
    return ntohl (addr);
}   /*  End Function conv_hostname_to_addr  */

static int *alloc_port (unsigned int *port_number, unsigned int retries,
			unsigned int *num_docks)
/*  This routine will allocate a Karma port for the module so that it can
    operate as a server (able to receive network connections).
    The port number to allocate must be pointed to by  port_number  .The
    routine will write the actual port number allocated to this address. This
    must point to an address which lies on an  int  boundary.
    The number of succsessive port numbers to attempt to allocate before giving
    up must be given by  retries  .If this is 0, then the routine will give up
    immediately if the specified port number is in use.
    The routine will create a number of docks for one port. Each dock is an
    alternative access point for other modules to connect to this port.
    The number of docks allocated will be written to the storage pointed to by
    num_docks  .This must point to an address which lies on an  int
    boundary.
    The close-on-exec flags of the docks are set such that the docks will
    close on a call to execve(2V).
    The docks are placed into non-blocking mode.
    The routine returns a pointer to a statically allocated array of docks on
    success, else it returns NULL.
*/
{
    flag bound;
    unsigned int retry_number;
#ifdef DISABLED
    struct servent *service_entry;
#endif

    /*  Get Karma port number for tcp/ ip  */
#ifdef DISABLED
    if (inet_port_offset < 0)
    {
	/*  Get tcp port offset  */
	if ( ( service_entry = getservbyname ("KARMA", "tcp") ) == NULL )
	{
	    inet_port_offset = INET_PORT_OFFSET;
	}
	else
	{
	    inet_port_offset = ntohs (service_entry->s_port);
	    if (INET_PORT_OFFSET != inet_port_offset)
	    {
		fprintf (stderr, "WARNING: NIS sevices database lists Karma");
		fprintf (stderr, " as having port number: %d\n",
			 inet_port_offset);
		fprintf (stderr, "whereas the hardcoded default is: %d\n",
			 INET_PORT_OFFSET);
	    }
	}
    }
#else
    inet_port_offset = INET_PORT_OFFSET;
#endif
    /*  Create Unix domain socket  */
    if ( ( docks[INDEX_UNIX_DOCK] = socket (AF_UNIX, SOCK_STREAM, 0) ) < 0 )
    {
	fprintf (stderr, "Error creating Unix socket\t%s\n", ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    /*  Create TCP/IP socket  */
    if ( ( docks[INDEX_TCP_DOCK] = socket (AF_INET, SOCK_STREAM, 0) ) < 0 )
    {
#ifdef OS_Linux
	if (errno == EINVAL)
	{
	    fprintf (stderr, "No kernel support for IP/TCP\t%s\n", ERRSTRING);
	    fprintf (stderr, "TCP/IP dock not created\n");
	    docks[INDEX_TCP_DOCK] = -1;
	}
	else
#endif
	{
	    fprintf (stderr, "Error creating IP/TCP socket\t%s\n",ERRSTRING);
	    if (close (docks[INDEX_UNIX_DOCK]) != 0)
	    {
		fprintf (stderr, "Error closing Unix socket\t%s\n", ERRSTRING);
	    }
	    exit (RV_SYS_ERROR);
	}    
    }
    /*  Try to bind to port  */
    for (retry_number = 0, bound = FALSE, allocated_port_number = *port_number;
	 (retry_number <= retries) && !bound;
	 ++retry_number)
    {
	/*  Try to bind to another port number  */
	if (docks[INDEX_TCP_DOCK] < 0)
	{
	    /*  No Internet dock: try Unix dock  */
	    bound = bind_unix (docks[INDEX_UNIX_DOCK], allocated_port_number);
	}
	else
	{
	    /*  Have TCP/IP dock: try to bind  */
	    bound = bind_inet (docks[INDEX_TCP_DOCK],
			       allocated_port_number + inet_port_offset,
			       SOCK_TYPE_STREAM);
	}
	if (!bound) ++allocated_port_number;
    }
    if (!bound)
    {
	/*  All slots occupied  */
	if (close (docks[INDEX_UNIX_DOCK]) != 0)
	{
	    fprintf (stderr, "Error closing Unix socket\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	if (close (docks[INDEX_TCP_DOCK]) != 0)
	{
	    fprintf (stderr, "Error closing IP/TCP socket\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	return (NULL);
    }
    if (docks[INDEX_TCP_DOCK] > -1)
    {
	/*  Have bound TCP/IP dock: now bind Unix dock  */
	if ( !bind_unix (docks[INDEX_UNIX_DOCK], allocated_port_number) )
	{
	    fprintf (stderr,
		     "IP/TCP dock bound but could not bind Unix dock\n");
	    exit (RV_SYS_ERROR);
	}
    }
    *port_number = allocated_port_number;
    if (docks[INDEX_TCP_DOCK] < 0)
    {
	num_docks_open = NUM_DOCKS - 1;
    }
    else
    {
	num_docks_open = NUM_DOCKS;
    }
    *num_docks = num_docks_open;
    return (docks);
}   /*  End Function alloc_port  */

static flag bind_unix (int sock, unsigned int port_number)
/*  This routine will bind a unix socket to a particular Karma port number.
    The socket must be given by  sock  .
    The port number must be given by  port_number  .
    The routine returns TRUE on success, else it returns FALSE (indicating the
    address is in use).
*/
{
    struct sockaddr_un un_addr;

    un_addr.sun_family = AF_UNIX;
    sprintf (un_addr.sun_path, "%s/%s%d", UNIX_SOCKET_DIR,
		    UNIX_SOCKET_FILE, port_number);
    mkdir (UNIX_SOCKET_DIR, (mode_t) 0777);
    chmod (UNIX_SOCKET_DIR, (mode_t) 0777);
    unlink (un_addr.sun_path);
    /*  Try to bind to Unix port  */
    if (bind (sock, (struct sockaddr *) &un_addr, (int) sizeof un_addr) != 0)
    {
	/*  Could not bind to port number  */
	if (errno != EADDRINUSE)
	{
	    fprintf (stderr, "Error binding Unix socket\t%s\n", ERRSTRING);
	    if (close (sock) != 0)
	    {
		fprintf (stderr, "Error closing Unix socket\t%s\n", ERRSTRING);
	    }
	    exit (RV_SYS_ERROR);
	}
	/*  Port already in use: say so  */
	return (FALSE);
    }
    /*  Bound  */
    /*  Set close-on-exec flag  */
    if (fcntl (sock, F_SETFD, 1) == -1)
    {
	fprintf (stderr,
		 "Error setting close-on-exec flag for Unix socket\t%s\n",
		 ERRSTRING);
	if (close (sock) != 0)
	{
	    fprintf (stderr, "Error closing Unix socket\t%s\n", ERRSTRING);
	}
	exit (RV_SYS_ERROR);
    }
    if (listen (sock, 2) != 0)
    {
	fprintf (stderr, "Error listening to Unix dock\t%s\n", ERRSTRING);
	if (close (sock) != 0)
	{
	    fprintf (stderr, "Error closing Unix socket\t%s\n", ERRSTRING);
	}
	exit (RV_SYS_ERROR);
    }
    return (TRUE);
}   /*  End Function bind_unix  */

static flag bind_inet (int sock, unsigned int port_number, char type)
/*  [SUMMARY] Bind an Internet socket to a particular Karma port number.
    <sock> The socket.
    <port_number> The port number.
    <type> The type of the dock.
    [RETURNS] TRUE on success, else FALSE (indicating the address is in use).
*/
{
    /*int sock_opt = SO_REUSEADDR;*/
    struct sockaddr_in in_addr;

    in_addr.sin_family = AF_INET;
    in_addr.sin_addr.s_addr = INADDR_ANY;
    /*  Try to bind to port number  */
    in_addr.sin_port = htons (port_number);
    if (bind (sock, (struct sockaddr *) &in_addr, (int) sizeof in_addr) != 0)
    {
	/*  Could not bind to port number  */
	if (errno != EADDRINUSE)
	{
	    fprintf (stderr, "Error binding Internet socket\t%s\n", ERRSTRING);
	    if (close (sock) != 0)
	    {
		fprintf (stderr, "Error closing Internet socket\t%s\n",
			 ERRSTRING);
	    }
	    exit (RV_SYS_ERROR);
	}
	/*  Port already in use: say so  */
	return (FALSE);
    }
    /*  Bound  */
    /*  Set close-on-exec flag  */
    if (fcntl (sock, F_SETFD, 1) == -1)
    {
	fprintf (stderr,
		 "Error setting close-on-exec flag for Internet socket\t%s\n",
		 ERRSTRING);
	if (close (sock) != 0)
	{
	    fprintf (stderr, "Error closing Internet socket\t%s\n", ERRSTRING);
	}
	exit (RV_SYS_ERROR);
    }
    if (type != SOCK_TYPE_STREAM) return (TRUE);
    if (listen (sock, 2) != 0)
    {
	fprintf (stderr, "Error listening to Internet dock\t%s\n", ERRSTRING);
	if (close (sock) != 0)
	{
	    fprintf (stderr, "Error closing Internet socket\t%s\n", ERRSTRING);
	}
	exit (RV_SYS_ERROR);
    }
    return (TRUE);
}   /*  End Function bind_inet  */

static int connect_unix (CONST char *filename)
/*  This routine will connect to a Unix domain port.
    The socket filename must be pointed to by  filename  .
    The close-on-exec flag of the socket is set such that the connection will
    close on a call to execve(2V).
    The routine returns the file descriptor of the opened connection on
    success, else it returns -1
*/
{
    int fd;
    struct sockaddr_un un_addr;

    /*  Create socket  */
    if ( ( fd = socket (AF_UNIX, SOCK_STREAM, 0) ) < 0 )
    {
	fprintf (stderr, "Error creating socket\t%s\n",
			ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    /*  Set close-on-exec flag  */
    if (fcntl (fd, F_SETFD, 1) == -1)
    {
	fprintf (stderr,
		 "Error setting close-on-exec flag for Unix socket\t%s\n",
		 ERRSTRING);
	if (close (fd) != 0)
	{
	    fprintf (stderr, "Error closing Unix socket\t%s\n", ERRSTRING);
	}
	exit (RV_SYS_ERROR);
    }
    /*  Set up addressing info  */
    un_addr.sun_family = AF_UNIX;
    strcpy (un_addr.sun_path, filename);
    /*  Connect to other end (server module)  */
    if (connect (fd, (struct sockaddr *) &un_addr, (int) sizeof un_addr) != 0)
    {
	/*  No connection made  */
	if ( (errno != ECONNREFUSED) && (errno != ENOENT) && (errno != ENXIO) )
	{
	    fprintf (stderr, "Error connecting to Unix socket\t%s\n",
		     ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	if (close (fd) != 0)
	{
	    fprintf (stderr, "Error closing unix socket: %d\t%s\n",
		     fd, ERRSTRING);
	}
	return (-1);
    }
    /*  Return the connection descriptor  */
    return (fd);
}   /*  End Function connect_unix  */

static int connect_inet (int sock, unsigned long addr,
			 unsigned int port_number, char type)
/*  [SUMMARY] Connect to an Internet port.
    <sock> The socket. If this is less than 0, a socket is created.
    <addr> The IP address of the machine to connect to. If this is 0 no
    connection is made and the socket is left in the unconnected state.
    <port_number> The port number to connect to.
    <type> The protocol type (SOCK_TYPE_STREAM or SOCK_TYPE_DGRAM).
    [NOTE] The close-on-exec flag of the socket is set such that the connection
    will close on a call to execve(2V).
    [RETURNS] The file descriptor of the opened connection on success, else -1.
*/
{
    int protocol = -1;
    struct sockaddr_in in_addr;

    /*  Create socket  */
    switch (type)
    {
      case SOCK_TYPE_STREAM:
	protocol = SOCK_STREAM;
	break;
      case SOCK_TYPE_DGRAM:
	protocol = SOCK_DGRAM;
	break;
      default:
	break;
    }
    if (sock < 0)
    {
	if ( ( sock = socket (AF_INET, protocol, 0) ) < 0 )
	{
	    fprintf (stderr, "Error creating socket\t%s\n", ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
    }
    /*  Set close-on-exec flag  */
    if (fcntl (sock, F_SETFD, 1) == -1)
    {
	fprintf (stderr,
		 "Error setting close-on-exec flag for Internet socket\t%s\n",
		 ERRSTRING);
	if (close (sock) != 0)
	{
	    fprintf (stderr, "Error closing Internet socket\t%s\n", ERRSTRING);
	}
	exit (RV_SYS_ERROR);
    }
    if (addr == 0) return (sock);
    /*  Set up addressing info  */
    in_addr.sin_family = AF_INET;
    in_addr.sin_port = htons (port_number);
    in_addr.sin_addr.s_addr = htonl (addr);
    /*  Connect to other end (server module)  */
    if (connect (sock, (struct sockaddr *) &in_addr, sizeof in_addr) != 0)
    {
	/*  No connection made  */
	if (errno != ECONNREFUSED)
	{
	    fprintf (stderr, "Error connecting to Internet socket\t%s\n",
		     ERRSTRING);
	    exit (RV_SYS_ERROR);
	}
	if (close (sock) != 0)
	{
	    fprintf (stderr, "Error closing Internet socket\t%s\n", ERRSTRING);
	}
	return (-1);
    }
    /*  Return the connection descriptor  */
    return (sock);
}   /*  End Function connect_inet  */

static void close_dock (int dock_index)
/*  This routine will close a dock.
    The dock index to close must be given by  dock_index  .
    The routine returns nothing.
*/
{
    struct sockaddr_un un_addr;

    if (close (docks[dock_index]) != 0)
    {
	fprintf (stderr, "Error closing dock: %d\t%s\n",
		 docks[dock_index], ERRSTRING);
	exit (RV_SYS_ERROR);
    }
    if (dock_index == INDEX_UNIX_DOCK)
    {
	sprintf (un_addr.sun_path, "%s/%s%d", UNIX_SOCKET_DIR,
		 UNIX_SOCKET_FILE, allocated_port_number);
	unlink (un_addr.sun_path);
    }
}  /*  End Function close_dock  */

static int connect_to_port (unsigned long addr, unsigned int port_number)
/*  This routine will connect to a server module running on the machine
    with Internet address given by  addr  .
    If the value of 0 is supplied for the address, the connection is made to a
    Karma server running on the local machine.
    The port number to connect to must given by  port_number  .
    The close-on-exec flags of the socket is set such that the socket will
    close on a call to execve(2V).
    The routine returns the file descriptor of the opened connection on
    success, else it returns -1
*/
{
#ifdef DISABLED
    struct servent *service_entry;
#endif
    char filename[STRING_LENGTH];

    /*  Get port number for tcp/ ip  */
#ifdef DISABLED
    if (inet_port_offset < 0)
    {
	/*  Get tcp port offset  */
	if ( ( service_entry = getservbyname ("KARMA", "tcp") ) == NULL )
	{
	    inet_port_offset = INET_PORT_OFFSET;
	}
	else
	{
	    inet_port_offset = ntohs (service_entry->s_port);
	    if (INET_PORT_OFFSET != inet_port_offset)
	    {
		fprintf (stderr, "WARNING: NIS sevices database lists Karma");
		fprintf (stderr, " as having port number: %d\n",
			 inet_port_offset);
		fprintf (stderr, "whereas the hardcoded default is: %d\n",
			 INET_PORT_OFFSET);
	    }
	}
    }
#else
    inet_port_offset = INET_PORT_OFFSET;
#endif
    /*  Try to connect to specified host  */
    if (addr == 0)
    {
	/*  Unix domain socket/ connection  */
	sprintf (filename, "%s/%s%d", UNIX_SOCKET_DIR,
		 UNIX_SOCKET_FILE, port_number);
	/*  Connect to other end (server module)  */
	return connect_unix (filename);
    }
    /*  Internet/ TCP domain socket/ connection  */
    return connect_inet (-1, addr, port_number + inet_port_offset,
			 SOCK_TYPE_STREAM);
}   /*  End Function connect_to_port  */

static int accept_connection_on_dock (int dock_index, unsigned long *addr,
				      flag *local)
/*  This routine will accept a connection on a dock.
    The dock index must be given by  dock_index  .
    The address of the host connecting to the dock will be written to the
    storage pointed to by  addr  .
    If the connection is a local connection, then the routine will write the
    value TRUE to the storage pointed to by  local  ,else it will write
    the value FALSE.
    The routine returns a connection on success, else it returns -1
*/
{
    /*  Unix dock info  */
    socklen_t un_addr_len;
    struct sockaddr_un un_addr;
    /* Internet dock info  */
    socklen_t in_addr_len;
    struct sockaddr_in in_addr;
    /*  General info  */
    int fd = -1;  /*  Initialised to keep compiler happy  */
    static char function_name[] = "_r_accept_connection_on_dock";

    switch (dock_index)
    {
      case INDEX_UNIX_DOCK:
	un_addr_len = sizeof un_addr;
	mclear (&un_addr, un_addr_len);
	/*  Listen for connections  */
	if ( ( fd = accept (docks[dock_index], (struct sockaddr *) &un_addr,
			    &un_addr_len) )
	    < 0 )
	    return (-1);
	*addr = r_get_inet_addr_from_host ( (char *) NULL, (flag *) NULL );
	break;
      case INDEX_TCP_DOCK:
	in_addr_len = sizeof in_addr;
	mclear (&in_addr, in_addr_len);
	/*  Listen for connections  */
	if ( ( fd = accept (docks[dock_index], (struct sockaddr *) &in_addr,
			    &in_addr_len) ) < 0 )
	    return (-1);
	*addr = ntohl (in_addr.sin_addr.s_addr);
	break;
      default:
	fprintf (stderr, "Unknown dock type\n");
	prog_bug (function_name);
	break;
    }
    if (r_get_inet_addr_from_host ( (char *) NULL, (flag *) NULL ) == *addr)
	*local = TRUE;
    else
	*local = FALSE;
    /*  Set close-on-exec flag  */
    if (fcntl (fd, F_SETFD, 1) == -1)
    {
	fprintf (stderr,
		 "%s: error setting close-on-exec flag for descriptor: %d\t%s\n",
		 function_name, fd, ERRSTRING);
	close (fd);
	return (-1);
    }
    return (fd);
}  /*  End Function accept_connection_on_dock  */

static flag close_connection (int connection)
/*  This routine will close a connection.
    The connection to close must be given by  connection  .
    The routine returns TRUE on success, else it returns FALSE.
*/
{
    if (close (connection) == 0) return (TRUE);
    fprintf (stderr, "Error closing descriptor: %d\t%s\n",
	     connection, ERRSTRING);
    return (FALSE);
}  /*  End Function close_connection  */

static int get_bytes_readable (int connection)
/*  This routine will determine the minimum number of bytes readable on a
    connection. There may be more bytes readable than indicated.
    The connection should be given by  connection  .
    The routine returns the number of bytes readable on success,
    else it returns -1
*/
{
    int bytes = 0;

    errno = 0;
    if (ioctl (connection, (int) FIONREAD, (char *) &bytes) == -1)
    {
#ifdef OS_Solaris
	/*  The following test/return for EPIPE is needed for Slowaris.
	    I've also noticed in Slowass 2.5 that consecutive calls can return
	    some non-zero value followed by a lesser (typically zero) value!
	    That is clearly a kernel bug!!
	    Seeing all the changes I've had to make for Slowaris, doesn't it
	    give the feeling there's something wrong with it???  */
	if (errno == EPIPE) return (0);
#endif
	return (-1);
    }
    return (bytes);
}  /*  End Function get_bytes_readable  */

static int open_stdin (unsigned int *type)
/*  This routine will open the standard input.
    The routine will write the value TRUE to the storage pointed to by  disc
    if the standard input is a disc, else it will write FALSE.
    The routine returns the descriptor on success, else it returns -1
*/
{
    struct stat statbuf;

    if (fstat (0, &statbuf) != 0)
    {
	fprintf (stderr, "Error getting stats on input descriptor\t%s\n",
		 ERRSTRING);
	return (-1);
    }
    if ( S_ISREG (statbuf.st_mode) ) *type = KFTYPE_DISC;
    else if ( S_ISCHR (statbuf.st_mode) ) *type = KFTYPE_CHARACTER;
    else if ( S_ISBLK (statbuf.st_mode) ) *type = KFTYPE_BLOCK;
    else if ( S_ISFIFO (statbuf.st_mode) ) *type = KFTYPE_FIFO;
    else
    {
	/*  Illegal object  */
	fprintf (stderr, "Illegal input descriptor\n");
	return (-1);
    }
    return (0);
}   /*  End Function open_stdin  */


#endif  /*  HAS_SOCKETS  */

#ifdef HAS_COMMUNICATIONS_EMULATION

#  ifdef OS_VXMVX
static void init_control_connection ()
/*  This routine will initialise the control connection to the host service
    process.
    This routine may be called repeatedly, only on the first call does it do
    anything.
    The routine returns nothing.
*/
{
    int count;
    int num_env;
    char *ptr;
    void *msg;
    extern char **environ;
    static flag initialised = FALSE;

    if (initialised) return;
    initialised = TRUE;
    if ( ( msg = task_recv_msg (0, 0, 1) ) == NULL )
    {
	fprintf (stderr, "Error receiving first message\n");
	exit (RV_UNDEF_ERROR);
    }
    parent_pid = *(int *) msg;
    outgoing_control_message_queue = *( (int *) msg + 1 );
    /*  Get environment  */
    num_env = *( (int *) msg + 2 );
    ptr = (char *) msg + 3 * sizeof outgoing_control_message_queue;
    if ( ( environ = (char **) malloc ( sizeof *environ * (num_env + 1) ) )
	== NULL )
    {
	fprintf (stderr,
			"Error allocating space for: %d environment strings\n",
			num_env);
	exit (RV_MEM_ERROR);
    }
    for (count = 0; count < num_env; ++count)
    {
	if ( ( environ[count] = strdup (ptr) ) == NULL )
	{
	    fprintf (stderr,
			    "Error allocating space for environment string: \"%s\"\n",
			    ptr);
	    exit (RV_MEM_ERROR);
	}
	ptr += strlen (ptr) + 1;
    }
    environ[count] = NULL;
    task_free_msg (msg);
    if ( ( incoming_control_message_queue = task_create_msg_q () ) < 0 )
    {
	fprintf (stderr, "Error creating control message queue\n");
	exit (RV_UNDEF_ERROR);
    }
    if ( ( msg = task_alloc_msg (sizeof incoming_control_message_queue, 1) )
	== NULL )
    {
	fprintf (stderr, "Error allocating message buffer\n");
	exit (RV_MEM_ERROR);
    }
    *(int *) msg = incoming_control_message_queue;
    task_send_msg (outgoing_control_message_queue, msg);
    /*  Install remote memory copy facility  */
    remote_memcpy_install ();
}   /*  End Function init_control_connection  */

static int *alloc_port (port_number, retries, num_docks)
/*  This routine will allocate a Karma port for the module so that it can
    operate as a server (able to receive network connections).
    The port number to allocate must be pointed to by  port_number  .The
    routine will write the actual port number allocated to this address. This
    must point to an address which lies on an  int  boundary.
    The number of succsessive port numbers to attempt to allocate before giving
    up must be given by  retries  .If this is 0, then the routine will give up
    immediately if the specified port number is in use.
    The routine will create a number of docks for one port. Each dock is an
    alternative access point for other modules to connect to this port.
    The number of docks allocated will be written to the storage pointed to by
    num_docks  .This must point to an address which lies on an  int
    boundary.
    The close-on-exec flags of the docks are set such that the docks will
    close on a call to execve(2V).
    The docks are placed into non-blocking mode.
    The routine returns a pointer to a statically allocated array of docks on
    success, else it returns NULL.
*/
unsigned int *port_number;
unsigned int retries;
unsigned int *num_docks;
{
    flag bound;
    int connect_queue;
    int fd;
    unsigned int retry_number;
    unsigned int dock_count;
    void *msg;

    init_control_connection ();
    if ( ( msg = task_alloc_msg (3 * sizeof (unsigned int), 1) ) == NULL )
    {
	fprintf (stderr, "Error allocating message queue\n");
	exit (RV_UNDEF_ERROR);
    }
    *(unsigned int *) msg = HR_ALLOC_PORT;
    *( (unsigned int *) msg + 1) = *port_number;
    *( (unsigned int *) msg + 2) = retries;
    task_send_msg (outgoing_control_message_queue, msg);
    if ( ( msg = task_recv_msg (incoming_control_message_queue, 0, 1) )
	== NULL )
    {
	fprintf (stderr, "Error receiving control message\n");
	exit (RV_UNDEF_ERROR);
    }
    if ( (num_docks_open = *(unsigned int *) msg) < 1 )
    {
	return (NULL);
    }
    allocated_port_number = *( (unsigned int *) msg + 1 );
    for (dock_count = 0; dock_count < num_docks_open; ++dock_count)
    {
	if ( ( fd = find_spare_descriptor () ) < 0 )
	{
	    fprintf (stderr, "Descriptor limit reached\n");
	    exit (RV_UNDEF_ERROR);
	}
	docks[dock_count] = fd + DESCRIPTOR_OFFSET;
	descriptors[fd].open = TRUE;
	descriptors[fd].received_closure = FALSE;
	if ( ( descriptors[fd].incoming_message_queue =
	      task_create_msg_q () ) < 0 )
	{
	    fprintf (stderr, "Error creating message queue\n");
	    exit (RV_UNDEF_ERROR);
	}
	descriptors[fd].outgoing_message_queue = *( (int *) msg + dock_count
						   + 2);
	descriptors[fd].incoming_msg = NULL;
	descriptors[fd].msg_pos = 0;
	descriptors[fd].msg_len = 0;
    }
    task_free_msg (msg);
    if ( ( msg = task_alloc_msg (num_docks_open * sizeof (int), 1) ) == NULL )
    {
	fprintf (stderr, "Error allocating message queue\n");
	exit (RV_UNDEF_ERROR);
    }
    for (dock_count = 0; dock_count < num_docks_open; ++dock_count)
    {
	fd = docks[dock_count] - DESCRIPTOR_OFFSET;
	*( (int *) msg + dock_count) = descriptors[fd].incoming_message_queue;
    }
    task_send_msg (outgoing_control_message_queue, msg);
    *num_docks = num_docks_open;
    *port_number = allocated_port_number;
    return (docks);
}   /*  End Function alloc_port  */

static void close_dock (dock)
/*  This routine will close a dock.
    The dock to close must be given by  dock  .
    The routine returns nothing.
*/
int dock;
{
    void *msg;
    static char function_name[] = "close_dock";

    init_control_connection ();
    dock -= DESCRIPTOR_OFFSET;
    if ( (dock < 0) || (dock >= MAX_DESCRIPTORS) )
    {
	fprintf (stderr, "Bad descriptor: %d\n",
			dock + DESCRIPTOR_OFFSET);
	prog_bug (function_name);
    }
    if (descriptors[dock].open != TRUE)
    {
	fprintf (stderr, "Descriptor: %d not open\n",
			dock + DESCRIPTOR_OFFSET);
	prog_bug (function_name);
    }
    if ( ( msg = task_alloc_msg (0, 1) ) == NULL )
    {
	fprintf (stderr, "Error allocating message queue\n");
	exit (RV_UNDEF_ERROR);
    }
    task_send_msg (descriptors[dock].outgoing_message_queue, msg);
}  /*  End Function close_dock  */

static int connect_to_port (unsigned long addr, unsigned int port_number)
/*  This routine will connect to a server module running on the machine
    with Internet address given by  addr  .
    If the value of 0 is supplied for the address, the connection is made to a
    Karma server running on the local machine.
    The port number to connect to must given by  port_number  .
    The close-on-exec flags of the socket is set such that the socket will
    close on a call to execve(2V).
    The routine returns the file descriptor of the opened connection on
    success, else it returns -1
*/
{
    int fd;
    int errnum;
    void *msg;

    init_control_connection ();
    if ( ( fd = find_spare_descriptor () ) < 0 )
    {
	fprintf (stderr, "Descriptor limit reached\n");
	return (-1);
    }
    if ( ( descriptors[fd].incoming_message_queue = task_create_msg_q () )
	< 0 )
    {
	fprintf (stderr, "Error creating message queue\n");
	exit (RV_UNDEF_ERROR);
    }
    if ( ( msg = task_alloc_msg (4 * sizeof (unsigned int), 1) ) == NULL )
    {
	fprintf (stderr, "Error allocating message\n");
	exit (RV_UNDEF_ERROR);
    }
    *(unsigned int *) msg = HR_CONNECT_TO_PORT;
    *( (unsigned int *) msg + 1 ) = addr;
    *( (unsigned int *) msg + 2 ) = port_number;
    *( (int *) msg + 3 ) = descriptors[fd].incoming_message_queue;
    task_send_msg (outgoing_control_message_queue, msg);
    if ( ( msg = task_recv_msg (incoming_control_message_queue, 0, 1) )
	== NULL )
    {
	fprintf (stderr, "Error receiving control message\n");
	exit (RV_UNDEF_ERROR);
    }
    if ( (errnum = *(int *) msg) != 0 )
    {
	/*  Could not connect  */
	if (errnum != ECONNREFUSED)
	{
	    /*  Error  */
	    fprintf ( stderr, "Error connecting\t%s\n", STRERROR (errnum) );
	}
	task_free_msg (msg);
	return (-1);
    }
    descriptors[fd].open = TRUE;
    descriptors[fd].received_closure = FALSE;
    descriptors[fd].outgoing_message_queue = *( (int *) msg + 1 );
    descriptors[fd].incoming_msg = NULL;
    descriptors[fd].msg_pos = 0;
    descriptors[fd].msg_len = 0;
    task_free_msg (msg);
    /*  Return the connection descriptor  */
    return (fd + DESCRIPTOR_OFFSET);
}   /*  End Function connect_to_port  */

static int accept_connection_on_dock (int dock_index, unsigned long *addr,
				      flag *local)
/*  This routine will accept a connection on a dock.
    The dock index must be given by  dock_index  .
    The address of the host connecting to the dock will be written to the
    storage pointed to by  addr  .
    If the connection is a local connection, then the routine will write the
    value TRUE to the storage pointed to by  local  ,else it will write
    the value FALSE.
    The routine returns a connection on success, else it returns -1
*/
{
    int fd;
    int errnum;
    int dock;
    int accept_message_queue;
    void *msg;
    void *in_msg;
    static char function_name[] = "accept_connection_on_dock";

    init_control_connection ();
    dock = docks[dock_index];
    dock -= DESCRIPTOR_OFFSET;
    if ( (dock < 0) || (dock >= MAX_DESCRIPTORS) )
    {
	fprintf (stderr, "Bad descriptor: %d\n", dock + DESCRIPTOR_OFFSET);
	prog_bug (function_name);
    }
    if (descriptors[dock].open != TRUE)
    {
	fprintf (stderr, "Descriptor: %d not open\n", dock +DESCRIPTOR_OFFSET);
	prog_bug (function_name);
    }
    if ( (in_msg = descriptors[dock].incoming_msg) == NULL )
    {
	fprintf (stderr, "No request on dock\n");
	prog_bug (function_name);
    }
    accept_message_queue = *(int *) in_msg;
    descriptors[dock].incoming_msg = NULL;
    descriptors[dock].msg_pos = 0;
    descriptors[dock].msg_len = 0;
    if ( ( msg = task_alloc_msg (sizeof (int), 1) ) == NULL )
    {
	fprintf (stderr, "Error allocating message\n");
	exit (RV_UNDEF_ERROR);
    }
    if ( ( fd = find_spare_descriptor () ) < 0 )
    {
	fprintf (stderr, "Descriptor limit reached\n");
	*(int *) msg = -1;
	task_send_msg (accept_message_queue, msg);
	task_free_msg (in_msg);
	return (-1);
    }
    if ( ( descriptors[fd].incoming_message_queue = task_create_msg_q () )
	 < 0 )
    {
	fprintf (stderr, "Error creating message queue\n");
	exit (RV_UNDEF_ERROR);
    }
    *(int *) msg = descriptors[fd].incoming_message_queue;
    task_send_msg (accept_message_queue, msg);
    descriptors[fd].open = TRUE;
    descriptors[fd].received_closure = FALSE;
    descriptors[fd].outgoing_message_queue = *( (int *) in_msg + 1 );
    descriptors[fd].incoming_msg = NULL;
    descriptors[fd].msg_pos = 0;
    descriptors[fd].msg_len = 0;
    *local = *( (flag *) in_msg + 2 );
    *addr = *( (unsigned long *) in_msg + 3 );
    task_free_msg (in_msg);
    /*  Return the connection descriptor  */
    return (fd + DESCRIPTOR_OFFSET);
}   /*  End Function accept_connection_on_dock  */

static flag close_connection (int connection)
/*  This routine will close a connection.
    The connection to close must be given by  connection  .
    The routine returns TRUE on success, else it returns FALSE.
*/
{
    void *msg;
    static char function_name[] = "close_connection";

    connection -= DESCRIPTOR_OFFSET;
    if ( (connection < 0) || (connection >= MAX_DESCRIPTORS) )
    {
	fprintf (stderr, "Bad descriptor: %d\n",
		 connection + DESCRIPTOR_OFFSET);
	prog_bug (function_name);
    }
    if (descriptors[connection].open != TRUE)
    {
	fprintf (stderr, "Descriptor: %d not open\n",
		 connection + DESCRIPTOR_OFFSET);
	prog_bug (function_name);
    }
    descriptors[connection].open = FALSE;
    if ( ( msg = descriptors[connection].incoming_msg ) != NULL )
    {
	/*  There was some previous message  */
	task_free_msg (msg);
	descriptors[connection].incoming_message_queue = NULL;
	descriptors[connection].msg_pos = 0;
    }
    if (descriptors[connection].received_closure)
    {
	/*  Closure message was received: go no further  */
	fprintf (stderr, "%s: Close remote\n", function_name);
	return (TRUE);
    }
    fprintf (stderr, "%s: Close local: send 0 bytes\n", function_name);
    if (descriptors[connection].outgoing_message_queue > -1)
    {
	if ( ( msg = task_alloc_msg (0, 1) ) == NULL )
	{
	    fprintf (stderr, "Error allocating message\n");
	    exit (RV_UNDEF_ERROR);
	}
	task_send_msg (descriptors[connection].outgoing_message_queue, msg);
    }
    return (TRUE);
}   /*  End Function close_connection  */

static int find_spare_descriptor ()
/*  This routine will find a spare descriptor number.
    The routine returns the descriptor on success, else it returns -1
*/
{
    unsigned int count;

    for (count = 0; count < MAX_DESCRIPTORS; ++count)
    {
	if (!descriptors[count].open) return (count);
    }
    return (-1);
}   /*  End Function find_spare_descriptor  */

static int read_from_connection (fd, buf, nbytes)
/*  This routine is similar to the system  read(2)  call, except that the
    number of bytes requested is always returned (except on error). Hence, if
    the descriptor references a socket, the routine will read as much data as
    was requested, rather than a lesser amount due to packetisation or
    interrupted system calls.
    The file descriptor to read from must be given by  fd  .This descriptor
    must NOT be set to non-blocking IO.
    The buffer in which to write the data must be pointed to by  buf  .
    The number of bytes to read must be given by  nbytes  .
    The routine returns the number of bytes requested on success,
    the number of bytes read on end of file and -1 on error.
*/
int fd;
char *buf;
int nbytes;
{
    int byte_count;
    int bytes_left;
    char *tmp = buf;
    void *msg;
    char *ptr;
    static char function_name[] = "read_from_connection";

    fd -= DESCRIPTOR_OFFSET;
    if ( (fd < 0) || (fd >= MAX_DESCRIPTORS) )
    {
	fprintf (stderr, "Bad descriptor: %d\n",
			fd + DESCRIPTOR_OFFSET);
	prog_bug (function_name);
    }
    if (descriptors[fd].open != TRUE)
    {
	fprintf (stderr, "Descriptor: %d not open\n",
			fd + DESCRIPTOR_OFFSET);
	prog_bug (function_name);
    }
    if (descriptors[fd].incoming_msg == NULL)
    {
	if ( ( msg =
	      task_recv_msg (descriptors[fd].incoming_message_queue, 0,
			     1) ) == NULL )
	{
	    fprintf (stderr, "Error reading message\n");
	    exit (RV_UNDEF_ERROR);
	}
	descriptors[fd].incoming_msg = msg;
	descriptors[fd].msg_len = task_sizeof_msg (msg);
	if (descriptors[fd].msg_len == 0)
	{
	    descriptors[fd].received_closure = TRUE;
	}
	descriptors[fd].msg_pos = 0;
    }
    msg = descriptors[fd].incoming_msg;
    bytes_left = (descriptors[fd].msg_len -
		 descriptors[fd].msg_pos);
    if (nbytes >= bytes_left)
    {
	nbytes = bytes_left;
	for (byte_count = 0, ptr = (char *) msg + descriptors[fd].msg_pos;
	     byte_count < nbytes;
	     ++byte_count, ++ptr, ++buf)
	{
	    *buf = *ptr;
	}
	descriptors[fd].incoming_msg = NULL;
	descriptors[fd].msg_pos = 0;
	descriptors[fd].msg_len = 0;
	task_free_msg (msg);
    }
    else
    {
	for (byte_count = 0, ptr = (char *) msg + descriptors[fd].msg_pos;
	     byte_count < nbytes;
	     ++byte_count, ++ptr, ++buf)
	{
	    *buf = *ptr;
	}
	descriptors[fd].msg_pos += nbytes;
    }
    return (nbytes);
}   /*  End Function read_from_connection  */

static int write_to_connection (fd, buf, nbytes)
/*  This routine is similar to the system  write(2)  call, except that the
    number of bytes requested is always returned (except on error). Hence, if
    the descriptor references a socket, the routine will write as much data as
    was requested, rather than a lesser amount due to packetisation or
    interrupted system calls.
    The file descriptor to write from must be given by  fd  .This descriptor
    must NOT be set to non-blocking IO.
    The buffer in which to write the data must be pointed to by  buf  .
    The number of bytes to write must be given by  nbytes  .
    The routine returns the number of bytes requested on success,
    else it returns -1 indicating error.
*/
int fd;
char *buf;
int nbytes;
{
    int byte_count;
    void *msg;
    char *ptr;
    static char function_name[] = "write_to_connection";

    fd -= DESCRIPTOR_OFFSET;
    if ( (fd < 0) || (fd >= MAX_DESCRIPTORS) )
    {
	fprintf (stderr, "Bad descriptor: %d\n",
			fd + DESCRIPTOR_OFFSET);
	prog_bug (function_name);
    }
    if (descriptors[fd].open != TRUE)
    {
	fprintf (stderr, "Descriptor: %d not open\n",
			fd + DESCRIPTOR_OFFSET);
	prog_bug (function_name);
    }
    if ( ( msg = task_alloc_msg (nbytes, 1) ) == NULL )
    {
	fprintf (stderr, "Error allocating message\n");
	exit (RV_UNDEF_ERROR);
    }
    ptr = (char *) msg;
    byte_count = nbytes;
    while (byte_count-- > 0)
    {
	*ptr++ = *buf++;
    }
    task_send_msg (descriptors[fd].outgoing_message_queue, msg);
    return (nbytes);
}   /*  End Function write_to_connection  */

static int get_bytes_readable (int connection)
/*  This routine will determine the minimum number of bytes readable on a
    connection. There may be more bytes readable than indicated.
    The connection should be given by  connection  .
    The routine returns the number of bytes readable on success,
    else it returns -1
*/
{
    int bytes_available;
    void *msg;
    static char function_name[] = "get_bytes_readable";

    connection -= DESCRIPTOR_OFFSET;
    if ( (connection < 0) || (connection >= MAX_DESCRIPTORS) )
    {
	fprintf (stderr, "Bad descriptor: %d\n",
		 connection + DESCRIPTOR_OFFSET);
	prog_bug (function_name);
    }
    if (descriptors[connection].open != TRUE)
    {
	fprintf (stderr, "Descriptor: %d not open\n",
		 connection + DESCRIPTOR_OFFSET);
	prog_bug (function_name);
    }
    msg = descriptors[connection].incoming_msg;
    if (msg == NULL)
    {
	if ( ( msg =
	      task_recv_msg (descriptors[connection].incoming_message_queue, 0,
			     0) ) == NULL )
	{
	    descriptors[connection].incoming_msg = NULL;
	    descriptors[connection].msg_len = 0;
	    descriptors[connection].msg_pos = 0;
	}
	else
	{
	    descriptors[connection].incoming_msg = msg;
	    descriptors[connection].msg_len = task_sizeof_msg (msg);
	    descriptors[connection].msg_pos = 0;
	    if (descriptors[connection].msg_len == 0)
	    {
		descriptors[connection].received_closure = TRUE;
	    }
	}
    }
    if (msg == NULL)
    {
	return (0);
    }
    return (descriptors[connection].msg_len - descriptors[connection].msg_pos);
}  /*  End Function get_bytes_readable  */

static unsigned long get_inet_addr_from_host (CONST char *host, flag *local)
/*  This routine will get the first listed Internet address from the hostname
    string pointed to by  host  .
    IF the specified host is the local machine, then the routine will write
    the value TRUE to the storage pointed to by  local  ,else it will write
    the value FALSE here. If this is NULL, nothing is written here.
    The routine returns the Internet address on success (in host byte order),
    else it returns 0.
*/
{
    unsigned long addr;
    static unsigned long local_addr = 0;

    if (local_addr < 1)
    {
	/*  Get my host address  */
	if ( ( local_addr = conv_hostname_to_addr ("localhost") ) == 0 )
	{
	    fprintf (stderr, "Error getting local host address\n");
	    return (0);
	}
    }
    if ( (host == NULL) || (strcmp (host, "unix") == 0) ||
	(strcmp (host, "localhost") == 0) )
    {
	/*  Local machine  */
	if (local != NULL) *local = TRUE;
	return (local_addr);
    }
    if ( ( addr = conv_hostname_to_addr (host) ) == 0 ) return (0);
    if (local != NULL) *local = (addr == local_addr) ? TRUE : FALSE;
    return (addr);
}   /*  End Function get_inet_addr_from_host  */

static unsigned long conv_hostname_to_addr (host)
/*  This routine will get the first listed Internet address from the hostname
    string pointed to by  host  .
    The routine returns the Internet address on success (in host byte order),
    else it returns 0.
*/
char *host;
{
    int length;
    unsigned long addr;
    void *msg;

    init_control_connection ();
    length = sizeof (unsigned int) + strlen (host) + 1;
    if ( ( msg = task_alloc_msg (length, 1) ) == NULL )
    {
	fprintf (stderr, "Error allocating message\n");
	exit (RV_UNDEF_ERROR);
    }
    *(unsigned int *) msg = HR_GET_INET_ADDR;
    strcpy ( (char *) msg + sizeof (unsigned int), host );
    task_send_msg (outgoing_control_message_queue, msg);
    if ( ( msg = task_recv_msg (incoming_control_message_queue, 0, 1) )
	== NULL )
    {
	fprintf (stderr, "Error receiving control message\n");
	exit (RV_UNDEF_ERROR);
    }
    addr = *(unsigned long *) msg;
    task_free_msg (msg);
    return (addr);
}   /*  End Function conv_hostname_to_addr  */

static int open_stdin (disc)
/*  This routine will open the standard input.
    The routine will write the value TRUE to the storage pointed to by  disc
    if the standard input is a disc, else it will write FALSE.
    The routine returns the descriptor on success, else it returns -1
*/
flag *disc;
{
    int fd;
    int errnum;
    int conn_message_queue;
    void *msg;

    init_control_connection ();
    if ( ( fd = find_spare_descriptor () ) < 0 )
    {
	fprintf (stderr, "Descriptor limit reached\n");
	return (-1);
    }
    if ( ( conn_message_queue = task_create_msg_q () ) < 0 )
    {
	fprintf (stderr, "Error creating message queue\n");
	exit (RV_UNDEF_ERROR);
    }
    if ( ( msg = task_alloc_msg (2 * sizeof (unsigned int), 1) ) == NULL )
    {
	fprintf (stderr, "Error allocating message\n");
	exit (RV_UNDEF_ERROR);
    }
    *(unsigned int *) msg = HR_OPEN_STDIN;
    *( (int *) msg + 1 ) = conn_message_queue;
    task_send_msg (outgoing_control_message_queue, msg);
    if ( ( msg = task_recv_msg (incoming_control_message_queue, 0, 1) )
	== NULL )
    {
	fprintf (stderr, "Error receiving control message\n");
	exit (RV_UNDEF_ERROR);
    }
    if ( (errnum = *(int *) msg) == -1 )
    {
	/*  Standard input is a disc  */
	*disc = TRUE;
	task_free_msg (msg);
	return (0);
    }
    if (errnum != 0)
    {
	fprintf ( stderr, "Error connecting\t%s\n", STRERROR (errnum) );
	task_free_msg (msg);
	return (-1);
    }
    descriptors[fd].open = TRUE;
    descriptors[fd].received_closure = FALSE;
    descriptors[fd].incoming_message_queue = conn_message_queue;
    descriptors[fd].outgoing_message_queue = -1;
    descriptors[fd].incoming_msg = NULL;
    descriptors[fd].msg_pos = 0;
    descriptors[fd].msg_len = 0;
    task_free_msg (msg);
    /*  Return the connection descriptor  */
    *disc = FALSE;
    return (fd + DESCRIPTOR_OFFSET);
}   /*  End Function open_stdin  */

int get_hostname (name, namelen)
/*  This routine will determine the local hostname.
    The buffer to write the hostname to must be pointed to by  name  .
    The length of the buffer must be given by  namelen  .
    The routine returns 0. On failure, the routine exits the process.
*/
char *name;
unsigned int namelen;
{
    void *msg;

    init_control_connection ();
    if ( ( msg = task_alloc_msg (sizeof (unsigned int), 1) ) == NULL )
    {
	fprintf (stderr, "Error allocating message\n");
	exit (RV_UNDEF_ERROR);
    }
    *(unsigned int *) msg = HR_GET_HOSTNAME;
    task_send_msg (outgoing_control_message_queue, msg);
    if ( ( msg = task_recv_msg (incoming_control_message_queue, 0, 1) )
	== NULL )
    {
	fprintf (stderr, "Error receiving control message\n");
	exit (RV_UNDEF_ERROR);
    }
    strncpy (name, msg, namelen);
    task_free_msg (msg);
    return (0);
}   /*  End Function get_hostname  */

static int get_ppid ()
/*  This routine will determine the parent process ID.
    The routine returns the parent process ID.
*/
{
    init_control_connection ();
    return (parent_pid);
}   /*  End Function get_ppid  */

#  endif  /*  OS_VXMVX  */
#endif  /*  HAS_COMMUNICATIONS_EMULATION  */

static void prog_bug (char *function_name)
{
    fprintf (stderr, "Program bug noted in function: %s\n",
		    function_name);
    fprintf (stderr, "Aborting.%c\n", BEL);
    exit (RV_PROGRAM_BUG);
}   /*  End Function prog_bug   */

static void mclear (void *buffer, unsigned int length)
{
    char *ptr = buffer;

    while (length-- > 0) *ptr++ = 0;
}   /*  End Function mclear  */

static void mcopy (char *dest, CONST char *source, unsigned int length)
{
    while (length-- > 0) *dest++ = *source++;
}   /*  End Function mcopy  */
