/*LINTLIBRARY*/
/*  pass_fd.c
    This code implements file descriptor passing.

    Copyright (C) 1998-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/
/*

    This file contains the various utility routines for passing file
    descriptors over connections.


    Written by      Richard Gooch   13-MAY-1998

    Last updated by Richard Gooch   15-FEB-2000: Support building outside Karma


*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#ifdef __KARMA__
#  include <karma.h>
#  include <karma_r.h>
#  include <os.h>
#else
   typedef int flag;
#  ifdef __linux__
#    define HAS_NEWBSD_FD_PASSING
#  endif  /*  Nothing else matters  */
#endif
#ifdef HAS_SYSV_FD_PASSING
#  include <stropts.h>
#endif
#if defined(HAS_OLDBSD_FD_PASSING) || defined(HAS_NEWBSD_FD_PASSING)
#  define __STRICT_ANSI__
#  include <sys/socket.h>
#  undef __STRICT_ANSI__
#  include <sys/uio.h>
#endif


#ifdef TRUE
#  undef TRUE
#endif
#define TRUE 1
#ifdef FALSE
#  undef FALSE
#endif
#define FALSE 0

#define CONTROLLEN ( sizeof (struct cmsghdr) + sizeof (int) )

/*  Public functions follow  */

/*EXPERIMENTAL_FUNCTION*/
flag r_send_fd (int conn_fd, int send_fd)
/*  [SUMMARY] Send a file descriptor over a connection.
    <conn_fd> The connection file descriptor.
    <send_fd> The file descriptor to send.
    [RETURNS] TRUE on success, else FALSE.
*/
{
#if defined(HAS_OLDBSD_FD_PASSING) || defined(HAS_NEWBSD_FD_PASSING)
    char dummy;
    struct msghdr msg;
    struct iovec iov;
#endif
#if defined(HAS_NEWBSD_FD_PASSING)
    struct cmsghdr cmhdr[2];  /*  Evil but easy allocation for FD space  */
#endif

#if defined(HAS_SYSV_FD_PASSING)
    return (ioctl (conn_fd, I_SENDFD, send_fd) == 0) ? TRUE : FALSE;
#endif
#if defined(HAS_OLDBSD_FD_PASSING) || defined(HAS_NEWBSD_FD_PASSING)
    memset (&iov, 0, sizeof iov);
    iov.iov_base = &dummy;
    iov.iov_len = 1;
    memset (&msg, 0, sizeof msg);
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;
#endif
#if defined(HAS_OLDBSD_FD_PASSING)
    msg.msg_accrights = (caddr_t) &send_fd;
    msg.msg_accrightslen = sizeof send_fd;
#endif
#if defined(HAS_NEWBSD_FD_PASSING)
    memset (cmhdr, 0, sizeof cmhdr);
    msg.msg_control = (caddr_t) cmhdr;
    msg.msg_controllen = CONTROLLEN;
    cmhdr[0].cmsg_level = SOL_SOCKET;
    cmhdr[0].cmsg_type = SCM_RIGHTS;
    cmhdr[0].cmsg_len = CONTROLLEN;
    *(int *) (cmhdr + 1) = send_fd;
#endif
#if defined(HAS_OLDBSD_FD_PASSING) || defined(HAS_NEWBSD_FD_PASSING)
    return (sendmsg (conn_fd, &msg, 0) == -1) ? FALSE : TRUE;
#endif
#if !defined(CAN_PASS_FD)
    return (FALSE);
#endif
}   /*  End Function r_send_fd  */

/*EXPERIMENTAL_FUNCTION*/
int r_recv_fd (int conn_fd)
/*  [SUMMARY] Receive a file descriptor over a connection.
    <conn_fd> The connection file descriptor.
    [RETURNS] The received file descriptor on success, else -1.
*/
{
#if defined(HAS_SYSV_FD_PASSING)
    struct strrecvfd recvfd;
#endif
#if defined(HAS_OLDBSD_FD_PASSING) || defined(HAS_NEWBSD_FD_PASSING)
    char dummy;
    struct msghdr msg;
    struct iovec iov;
#endif
#if defined(HAS_OLDBSD_FD_PASSING)
    int recv_fd;
#endif
#if defined(HAS_NEWBSD_FD_PASSING)
    struct cmsghdr cmhdr[2];  /*  Evil but easy allocation for FD space  */
#endif

#if defined(HAS_SYSV_FD_PASSING)
    return (ioctl (conn_fd, I_RECVFD, &recvfd) == 0) ? recvfd.fd : -1;
#endif
#if defined(HAS_OLDBSD_FD_PASSING) || defined(HAS_NEWBSD_FD_PASSING)
    memset (&iov, 0, sizeof iov);
    iov.iov_base = &dummy;
    iov.iov_len = 1;
    memset (&msg, 0, sizeof msg);
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;
#endif
#if defined(HAS_OLDBSD_FD_PASSING)
    msg.msg_accrights = (caddr_t) &recv_fd;
    msg.msg_accrightslen = sizeof recv_fd;
    return ( (recvmsg (conn_fd, &msg, 0) != -1) &&
	     (msg.msg_accrightslen == sizeof recv_fd) ) ? recv_fd : -1;
#endif
#if defined(HAS_NEWBSD_FD_PASSING)
    memset (cmhdr, 0, sizeof cmhdr);
    msg.msg_control = (caddr_t) cmhdr;
    msg.msg_controllen = CONTROLLEN;
    if (recvmsg (conn_fd, &msg, 0) == -1) return (-1);
    if (cmhdr[0].cmsg_level != SOL_SOCKET) return (-1);
    if (cmhdr[0].cmsg_type != SCM_RIGHTS) return (-1);
    if (cmhdr[0].cmsg_len != CONTROLLEN) return (-1);
    return *(int *) (cmhdr + 1);
#endif
#if !defined(CAN_PASS_FD)
    return (FALSE);
#endif
}   /*  End Function r_recv_fd  */
