/*LINTLIBRARY*/
/*  misc.c

    This code provides some sample (recommended) signal handlers.

    Copyright (C) 1992-1999  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains the various Karma utility routines which relate to
  handling signals.


    Written by      Richard Gooch   19-SEP-1992

    Updated by      Richard Gooch   14-NOV-1992

    Updated by      Richard Gooch   6-JAN-1993: Changed from  #include "*.h"
  to  #include <*.h>

    Updated by      Richard Gooch   26-NOV-1994: Moved to  packages/s/misc.c

    Updated by      Richard Gooch   13-APR-1996: Changed to new documentation
  format.

    Updated by      Richard Gooch   15-AUG-1999: Made egcs happy.

    Last updated by Richard Gooch   11-SEP-1999: Improved documentation.


*/
#include <stdio.h>
#include <karma.h>
#include <karma_s.h>


/*  Private data  */
static unsigned int count_control_c_limit = 5;
static volatile unsigned int count_control_c = 0;

/*  Private functions  */
STATIC_FUNCTION (void print_abort, () );


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
void s_int_handler ()
/*  [SUMMARY] Recommended SIGINT handler.
    [PURPOSE] This routine will handle SIGINT signals (those generated by the
    C shell in response to a control C keypress).
    [RETURNS] Nothing.
*/
{
    if (++count_control_c > count_control_c_limit)
    {
	fprintf (stderr, "Number of control C's exceeded: %u\n\n",
		 count_control_c_limit);
	print_abort ();
    }
}   /*  End Function s_int_handler  */

/*PUBLIC_FUNCTION*/
void s_term_handler ()
/*  [SUMMARY] Recommended SIGTERM handler.
    [PURPOSE] This routine will handle SIGTERM signals.
    [RETURNS] Nothing.
*/
{
    fprintf (stderr, "Ignored sigTERM\n");
}   /*  End Function s_term_handler  */

/*PUBLIC_FUNCTION*/
flag s_check_for_int ()
/*  [SUMMARY] Check for queued SIGINTs.
    [PURPOSE] This routine will check if there is a queued interrupt (SIGINT)
    pending. A single pending interrupt is removed from the queue.
    [RETURNS] TRUE if there is a queued SIGINT, else FALSE.
*/
{
    if (count_control_c > 0)
    {
	--count_control_c;
	return (TRUE);
    }
    return (FALSE);
}   /*  End Function s_check_for_int  */


/*  Private functions follow  */

static void print_abort ()
{
    fprintf (stderr, "Aborting.\n");
    exit (RV_UNDEF_ERROR);
}   /*  End Function print_abort    */
