/*LINTLIBRARY*/
/*  main.c

    This code provides KWorldCanvas objects.

    Copyright (C) 1993-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for manipulating a World Canvas
    (window with world co-ordinates) independent of the graphics system in use.


    Written by      Richard Gooch    16-APR-1993

    Updated by      Richard Gooch    27-APR-1993

    Updated by      Richard Gooch    3-JUN-1993: Improved symmetry in
  co-ordinate conversion routines.

    Updated by      Richard Gooch    15-OCT-1993: Documented placement of
  KWorldCanvas origin.

    Updated by      Richard Gooch    21-NOV-1993: Added  canvas_draw_point_p
  ,  canvas_draw_line_p  ,  canvas_fill_ellipse_p  ,  canvas_draw_rectangle
  ,  canvas_draw_rectangle_p  ,  canvas_fill_rectangle  and
  canvas_fill_rectangle_p  routines.

    Updated by      Richard Gooch    23-NOV-1993: Added
  canvas_register_convert_func  .

    Updated by      Richard Gooch    26-NOV-1993: Added  canvas_draw_lines  ,
  canvas_draw_lines_p  and  canvas_get_pixcanvas  .

    Updated by      Richard Gooch    29-NOV-1993: Added commented test for
  magic_number  field of  win_scale_type  structure in  verify_win_scale  .Will
  be activated in version 2.0 of Karma. Added  canvas_draw_*_c  routines.

    Updated by      Richard Gooch    2-DEC-1993: Removed  canvas_draw_*_c
  routines and added  canvas_get_colour  routine.

    Updated by      Richard Gooch    6-DEC-1993: Added  canvas_specify  and
  canvas_get_specification  routines.

    Updated by      Richard Gooch    8-DEC-1993: Added  canvas_set_dressing  .

    Updated by      Richard Gooch    11-DEC-1993: Added <canvas_draw_dressing>.

    Updated by      Richard Gooch    19-MAY-1994: Modified  canvas_get_colour
  to also give RGB colour values.

    Updated by      Richard Gooch    28-JUL-1994: Added  canvas_get_cmap  .

    Updated by      Richard Gooch    2-AUG-1994: Allowed passing of NULL
  Kcolourmap  to  canvas_create  .

    Updated by      Richard Gooch    8-AUG-1994: Created  canvas_draw_rgb_image

    Updated by      Richard Gooch    2-OCT-1994: Created function
  <canvas_coord_transform>

    Updated by      Richard Gooch    22-NOV-1994: Created
  canvas_unregister_position_event_func  .

    Updated by      Richard Gooch    26-NOV-1994: Moved to
  packages/canvas/main.c

    Updated by      Richard Gooch    28-NOV-1994: Made use of  c_  package.

    Updated by      Richard Gooch    2-JAN-1995: Took account of change to
  kwin_draw_rgb_image  .

    Updated by      Richard Gooch    15-MAR-1995: Added propagation of
  PostScriptPage object to refresh callbacks.

    Updated by      Richard Gooch    30-APR-1995: Added support for intensity
  scaling function.

    Updated by      Richard Gooch    13-JUL-1995: Improved informative output
  in <negotiate_scale_change>.

    Updated by      Richard Gooch    29-AUG-1995: Created
  <canvas_get_attributes> and <canvas_set_attributes>.

    Updated by      Richard Gooch    6-SEP-1995: Made experimental use of
  <kwin_draw_pc_image>.

    Updated by      Richard Gooch    15-SEP-1995: Took account of increased
  precision of <kwin_draw_point> and <kwin_draw_line>.

    Updated by      Richard Gooch    22-SEP-1995: Fixed bug in
  <convert_from_canvas_coord>: passing wrong pointer type.

    Updated by      Richard Gooch    5-JAN-1996: Created
  <canvas_register_d_convert_func> and deprecated integer version.

    Uupdated by     Richard Gooch    19-FEB-1996: Supported list of areas to
  refresh in refresh routines.

    Updated by      Richard Gooch    14-APR-1996: Changed to new documentation
  format.

    Updated by      Richard Gooch    19-MAY-1996: Deprecated <canvas_get_size>.

    Updated by      Richard Gooch    26-MAY-1996: Cleaned code to keep
  gcc -Wall -pedantic-errors happy.

    Updated by      Richard Gooch    4-JUN-1996: Switched to left-right
  bottom-top co-ordinate specification instead of min-max x and y.

    Updated by      Richard Gooch    15-JUN-1996: Created
  <canvas_register_transforms_func> and ellipse drawing routines.

    Updated by      Richard Gooch    17-JUN-1996: Created
  <canvas_convert_to_canvas_coords> routine and deprecated
  <canvas_convert_to_canvas_coord> routine.

    Updated by      Richard Gooch    19-JUL-1996: Created
  <canvas_convert_from_canvas_coords> routine and deprecated
  <canvas_convert_from_canvas_coord> routine.

    Updated by      Richard Gooch    20-JUL-1996: Created
  <canvas_register_coords_convert_func> routine and deprecated
  <canvas_register_d_convert_func> routine.

    Updated by      Richard Gooch    10-AUG-1996: Checked for TOOBIG
  co-ordinates in <canvas_draw_lines_p>.

    Updated by      Richard Gooch    16-SEP-1996: Allow NULL win_scale for
  <canvas_create>.

    Updated by      Richard Gooch    28-SEP-1996: Fixed bug in
  <canvas_set_attributes> when setting iscale_func.

    Updated by      Richard Gooch    5-NOV-1996: Fixed <canvas_draw_lines_p> to
  divide equally in non-linear world co-ordinates rather than in linear
  co-ordinates.

    Updated by      Richard Gooch    8-DEC-1996: Added support for
  "!valueR valueI" colournames in <canvas_get_colour>. Undocumented.

    Updated by      Richard Gooch    11-DEC-1996: Fixed passing back of pixel
  value for "!valueR valueI" colournames in <canvas_get_colour>.

    Updated by      Richard Gooch    27-JAN-1997: Allow co-ordinate
  transformation function to be updated any number of times.

    Updated by      Richard Gooch    12-FEB-1997: Added CANVAS_ATT_WCS_ASTRO.

    Updated by      Richard Gooch    19-FEB-1997: Added CANVAS_ATT_AUTO_BLANK.

    Updated by      Richard Gooch    20-FEB-1997: Abort refresh if pixcanvas
  not visible in <_canvas_refresh_canvas>: deep interactions when sharing
  colourmaps (like with invisible magnifiers).

    Updated by      Richard Gooch   14-MAR-1997: Fixed bug in computation of
  radius in pixels in <canvas_fill_ellipse_p> introduced in JUL-1996.

    Updated by      Richard Gooch   16-MAR-1997: Created <canvas_get_dressing>
  and <canvas_draw_arc> functions.

    Updated by      Richard Gooch   15-APR-1997: Fixed bug in
  <negotiate_scale_change> which offset canvas area 1 pixel to the right,
  causing right dressing box line to be drawn off pixel canvas sometimes.

    Updated by      Richard Gooch   16-APR-1997: Delayed clearing of boundary
  in <canvas_resize> to <canvas_draw_dressing> if dressing is to be drawn.

    Updated by      Richard Gooch   26-MAY-1997: Fixed bugs in
  <canvas_draw_rectangle_p> and <canvas_fill_rectangle_p> functions which did
  not compute width and height properly under some co-ordinate systems.

    Updated by      Richard Gooch   21-JUL-1997: Added precompute event support

    Updated by      Richard Gooch   25-JUL-1997: Created image cache compute
  functions. Made use of new geom_clear mechanism to clear boundary area.

    Updated by      Richard Gooch   31-JUL-1997: No longer pass pointer to
  <honoured_areas> flag to precompute functions

    Updated by      Richard Gooch   4-AUG-1997: No longer add rounding offsets.

    Updated by      Richard Gooch   19-AUG-1997: Call precompute functions on
  colourmap resize.

    Updated by      Richard Gooch   20-AUG-1997: Moved code from View2Datasets
  widget to new function <canvas_convert_ellipse_coords>.

    Updated by      Richard Gooch   21-AUG-1997: Fixed long-standing bug in
  code moved into new function <canvas_convert_ellipse_coords>.

    Updated by      Richard Gooch   27-AUG-1997: Improved diagnostics in
  <canvas_draw_image>, <canvas_compute_pc_image>, <canvas_draw_rgb_image> and
  <canvas_compute_rgb_image>.

    Updated by      Richard Gooch   4-SEP-1997: Optimised <canvas_get_colour> a
  bit when colourname is "!r i" complex value.

    Updated by      Richard Gooch   25-NOV-1997: Switched to <m_copy> from
  <strncpy> in <canvas_specify> because of a bug with AIX 3.2.

    Updated by      Richard Gooch   6-FEB-1998: Took account of rotation change

    Updated by      Richard Gooch   28-FEB-1998: Made function
  <canvas_get_specification> tolerant of NULL pointers.

    Updated by      Richard Gooch   29-MAR-1998: Added pixel co-ordinates to
  world canvas position event functions.

    Updated by      Richard Gooch   7-MAY-1998: Fixed use of <m_copy> in
  <canvas_specify> to limit copying to string length.

    Updated by      Richard Gooch   10-MAY-1998: Added <<linear>> parameter to
  <canvas_convert_ellipse_coords> function.

    Updated by      Richard Gooch   17-MAY-1998: More efficient conversion and
  silent handling of illegal values.

    Updated by      Richard Gooch   20-MAY-1998: Added test for RA axis in
  <canvas_convert_ellipse_coords> function.

    Updated by      Richard Gooch   25-MAY-1998: Increased accuracy of function
  <get_ellipse_max_angle>.

    Updated by      Richard Gooch   26-MAY-1998: Bug fix in function
  <get_ellipse_max_angle>: could jump 180 degrees for hundreds of revolutions.

    Updated by      Richard Gooch   28-MAY-1998: Fixed bug in function
  <canvas_convert_ellipse_coords> where scaling was done for linear<->pixel
  rather than pixel<->world transformations. Changed test for RA axis to
  longitude axis.

    Updated by      Richard Gooch   30-MAY-1998: Changed function
  <canvas_convert_ellipse_coords> to no longer limit angle to -90,90 degrees.

    Updated by      Richard Gooch   1-JUN-1998: Fixed bug in <canvas_specify>
  where '\0' was written at end of buffer rather than end of string.

    Updated by      Richard Gooch   17-SEP-1998: Took account of some CONST
  declarations in <ex> package.

    Updated by      Richard Gooch   15-OCT-1998: Created experimental function
  <canvas_warp_pointer>.

    Updated by      Richard Gooch   10-JUN-1999: Removed some warning
  conditions from <canvas_fill_ellipse_p>.

    Updated by      Richard Gooch   2-SEP-1999: Took account of interface
  change to function <kwin_warp_pointer>.

    Updated by      Richard Gooch   22-SEP-1999: Fixed bug in function
  <negotiate_scale_change>: was not centring properly with axis labels.

    Updated by      Richard Gooch   23-SEP-1999: Changed function
  <negotiate_scale_change> to retain old offsets if possible.

    Updated by      Richard Gooch   13-OCT-1999: Removed unnecessary casts.

    Updated by      Richard Gooch   11-APR-2000: Documentation fixes.

    Updated by      Richard Gooch   27-JUN-2000: Created experimental functions
  <canvas_destroy>, <canvas_register_destroy_func> and
  <canvas_register_transforms_free_func>.

    Updated by      Richard Gooch   2-JUL-2000: Added check for NULL
  restriction name and value pointers in <canvas_destroy>.

    Updated by      Richard Gooch   6-JUL-2000: Changed function
  <negotiate_scale_change> to call size control function before getting
  dressing size.

    Updated by      Richard Gooch   7-JUL-2000: Added <<wc_only>> parameter to
  function <canvas_register_destroy_func>.

    Updated by      Richard Gooch   11-JUL-2000: Added #include <karma_kwin.h>
  and <karma_kcmap.h>

    Updated by      Richard Gooch   27-SEP-2000: Created experimental function
  <canvas_test_for_boundary>.

    Updated by      Richard Gooch   8-MAR-2001: Disabled one debugging message
  in function <canvas_test_for_boundary>.

    Updated by      Richard Gooch   10-MAR-2001: Fixed bug in function
  <_canvas_get_pixel_from_value>: was not calculating scaled min&max.

    Updated by      Richard Gooch   11-MAR-2001: Fixed bug in function
  <_canvas_get_pixel_from_value>: cope with min>max after transform.

    Updated by      Richard Gooch   18-MAR-2001: Fixed bug in function
  <canvas_test_for_boundary> introduced by recent debugging disabling.

    Updated by      Richard Gooch   21-MAR-2001: Fixed bug in function
  <canvas_destroy>: was deallocating active rather than allocated restrictions.

    Updated by      Richard Gooch   20-AUG-2001: Cosmetic change.

    Updated by      Richard Gooch   29-AUG-2001: Disabled traps for degenerate
  dimensions. Hopefully other parts of the library can cope (some changes have
  been made elsewhere).

    Updated by      Richard Gooch   4-SEP-2001: Changed <canvas_set_attributes>
  to abort if canvas is being refreshed.

    Last updated by Richard Gooch   5-NOV-2001: Added KCD_ATT_FONT_SCALE.


*/
#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#include <karma.h>
#define NEW_WIN_SCALE
#include <karma_canvas.h>
#include <karma_kcmap.h>
#include <karma_kwin.h>
#include <karma_wcs.h>
#include <karma_ds.h>
#include <karma_st.h>
#include <karma_ex.h>
#include <karma_a.h>
#include <karma_m.h>
#include <karma_c.h>


#define MAGIC_NUMBER (unsigned int) 357474556

#define DRESSING_ITERATE_MAX (unsigned int) 10
#define COORD_BUF_SIZE 2048

#define VERIFY_CANVAS(canvas) if (canvas == NULL) \
{fprintf (stderr, "NULL canvas passed\n"); \
 a_prog_bug (function_name); } \
if (canvas->magic_number != MAGIC_NUMBER) \
{fprintf (stderr, "Invalid canvas object\n"); \
 a_prog_bug (function_name); }

#define MIN(a,b) ( ((a) < (b)) ? (a) : (b) )
#define MAX(a,b) ( ((a) > (b)) ? (a) : (b) )

/*  Dressing structure: WARNING!!! Also change in dressing.c  */
typedef struct
{
    flag drawn;
    flag display;
    flag show_top_ticks;
    flag show_bottom_ticks;
    flag show_left_ticks;
    flag show_right_ticks;
    flag internal_ticks;
    flag grid;
    char *grid_colour;
    char *scale_colour;
    char *hlabel;
    char *vlabel;
    char *label_colour;
    char *title;
    char *title_colour;
    flag show_labels;
    double font_scale;
    struct dressing_info_struct *info;
} dressing_type;

/*  Private structures follow  */

struct range_struct
{
    double min;
    double max;
    flag wrapped;
};

struct world_boundary_struct
{
    flag valid;
    struct range_struct x;
    struct range_struct y;
};

struct worldcanvas_type
{
    unsigned int magic_number;
    KPixCanvas pixcanvas;
    Kcolourmap cmap;
    unsigned int cmap_num_pixels;
    unsigned long *cmap_pixels;
    struct win_scale_type win_scale;
    flag auto_min_sat;
    flag auto_max_sat;
    flag auto_blank;
    flag in_refresh;
    /*  Callback information  */
    void (*size_control_func) ();
    void *size_control_info;  /*  Leave deallocation to a destroy callback  */
    flag (*coords_convert_func) (KWorldCanvas canvas, unsigned int num_coords,
				 CONST double *xin, CONST double *yin,
				 double *xout, double *yout,
				 flag to_linear, void **info);
    flag (*deprecated_coord_d_convert_func) (KWorldCanvas canvas,
					     struct win_scale_type *win_scale,
					     double *x, double *y,
					     flag to_world, void **info);
    flag (*deprecated_coord_convert_func) (KWorldCanvas canvas,
					   struct win_scale_type *win_scale,
					   int *px, int *py,
					   double *wx, double *wy,
					   flag to_world, void **info);
    void *coord_convert_info;  /*  Leave deallocation to a destroy callback  */
    /*  Callback lists where others register an interest in this canvas   */
    KCallbackList precompute_list;
    KCallbackList refresh_list;
    KCallbackList position_list;
    KCallbackList wc_only_destroy_list;
    KCallbackList destroy_list;
    /*  Callback functions this canvas has registered with other objects  */
    KCallbackFunc precompute_func;
    KCallbackFunc refresh_func;
    KCallbackFunc position_func;
    KCallbackFunc destroy_func;
    KCallbackFunc cmap_resize_func;
    flag quash_negotiate;
    flag in_size_control_func;
    void (*coord_transforms_func) (KWorldCanvas canvas,unsigned int num_coords,
				   double *x, flag x_to_linear,
				   double *y, flag y_to_linear,
				   double left_x, double right_x,
				   double bottom_y, double top_y,
				   void **info);
    void (*coord_transforms_free_func) (void *info);
    void (*deprecated_coord_transform_func) ();
    void *coord_transform_info;
    /*  Specification information  */
    char specify_xlabel[STRING_LENGTH];
    char specify_ylabel[STRING_LENGTH];
    unsigned int num_restrictions;
    unsigned int restrictions_allocated;
    char **restriction_names;  /*  Strings of length STRING_LENGTH  */
    double *restriction_values;
    KwcsAstro ap;
    struct world_boundary_struct world_boundary;
    /*  Dressing display parameters  */
    dressing_type dressing;
};

struct refresh_struct
{
    int width;
    int height;
    flag cmap_resize;
    PostScriptPage pspage;
    unsigned int num_areas;
    KPixCanvasRefreshArea *areas;
    flag honoured_areas;
};

struct position_struct
{
    double x;
    double y;
    unsigned int event_code;
    void *e_info;
    double x_lin;
    double y_lin;
    int x_pix;
    int y_pix;
};

typedef flag (*PositionFunc) (KWorldCanvas canvas, double x, double y,
			      unsigned int event_code, void *e_info,
			      void **f_info, double x_lin, double y_lin,
			      int x_pix, int y_pix);


/*  External functions  */
EXTERN_FUNCTION (void canvas_get_dressing_size,
		 (KWorldCanvas canvas, int *p_left, int *p_right,
		  int *p_top, int *p_bottom, dressing_type *dressing) );
EXTERN_FUNCTION (void canvas_draw_dressing_internal,
		 (KWorldCanvas canvas, dressing_type *dressing) );
EXTERN_FUNCTION (void canvas_free_dressing, (dressing_type *dressing) );
EXTERN_FUNCTION (KCallbackFunc kwin_register_precompute_func,
		 (KPixCanvas canvas, void (*refresh_func) (), void *info) );


/*  Private data  */


/*  Private functions  */
STATIC_FUNCTION (void _canvas_pixcanvas_precompute_func,
		 (KPixCanvas pixcanvas, int width, int height, void **info,
		  PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas) );
STATIC_FUNCTION (void _canvas_pixcanvas_refresh_func,
		 (KPixCanvas pixcanvas, int width, int height, void **info,
		  PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (void _canvas_cmap_resize_func,
		 (Kcolourmap cmap, void **info) );
STATIC_FUNCTION (void _canvas_refresh_canvas,
		 (KWorldCanvas canvas, int width, int height,
		  flag cmap_resize, PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );
STATIC_FUNCTION (unsigned long _canvas_get_pixel_from_value,
		 (KWorldCanvas canvas, double value[2],
		  unsigned short *red, unsigned short *green,
		  unsigned short *blue) );
STATIC_FUNCTION (flag _canvas_pixcanvas_position_event,
		 (KPixCanvas pixcanvas, int x, int y, unsigned int event_code,
		  void *event_info, void **f_info) );
STATIC_FUNCTION (void _canvas_verify_win_scale,
		 (struct win_scale_type *win_scale, char *function_name) );
STATIC_FUNCTION (flag _canvas_negotiate_scale_change,
		 (KWorldCanvas canvas, int width, int height,
		  flag *boundary_clear) );
STATIC_FUNCTION (void _canvas_dressing_refresh_func,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize, void **info) );
STATIC_FUNCTION (flag _canvas_precompute_event_func,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );
STATIC_FUNCTION (flag _canvas_refresh_event_func,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );
STATIC_FUNCTION (flag _canvas_position_event_func,
		 (void *object, void *client1_data,
		  void *call_data, void *client2_data) );
STATIC_FUNCTION (void _canvas_set_sat_pixels, (KWorldCanvas canvas) );
STATIC_FUNCTION (double get_ellipse_distance,
		 (KWorldCanvas canvas, flag to_world, flag linear,
		  double angle, double xscale, double yscale,
		  double cos_rot, double sin_rot,
		  double centre_x, double centre_y,
		  double radius_x, double radius_y,
		  double cx, double cy) );
STATIC_FUNCTION (double get_ellipse_max_angle,
		 (KWorldCanvas canvas, flag to_world, flag linear,
		  double xscale, double yscale,
		  double cos_rot, double sin_rot,
		  double centre_x, double centre_y,
		  double radius_x, double radius_y,
		  double cx, double cy) );


/*  Public functions follow  */

/*PUBLIC_FUNCTION*/
KWorldCanvas canvas_create (KPixCanvas pixcanvas, Kcolourmap cmap,
			    struct win_scale_type *win_scale)
/*  [SUMMARY] Create world canvas.
    [PURPOSE] This routine will create a world canvas, ready for drawing, from
    a KPixCanvas (pixel canvas) object. The origin of a KWorldCanvas is the
    lower-left corner.
    <pixcanvas> The pixel canvas object.
    <cmap> The colourmap. If this is NULL, the canvas will only permit drawing
    using pixel values. This precludes the use of drawing PseudoColour images
    and drawing geometric primitives using data values.
    [NOTE] Resize events on the colourmap will cause the canvas to be
    refreshed.
    [NOTE] The world canvas will be automatically destroyed if the pixel canvas
    is destroyed.
    <win_scale> The window scaling information. The <<x_offset>> <<x_pixels>>
    <<y_offset>> and <<y_pixels>> fields are determined internally. If this is
    NULL a default scale is choosen.
    [RETURNS] A world canvas object on success, else NULL.
*/
{
    KWorldCanvas canvas;
    int width, height;
    struct win_scale_type win_scale_local;
    static char function_name[] = "canvas_create";

    if (pixcanvas == NULL)
    {
	fprintf (stderr, "NULL KPixCanvas passed\n");
	a_prog_bug (function_name);
    }
    if (win_scale == NULL)
    {
	canvas_init_win_scale (&win_scale_local, K_WIN_SCALE_MAGIC_NUMBER);
	win_scale = &win_scale_local;
    }
    _canvas_verify_win_scale (win_scale, function_name);
    kwin_get_size (pixcanvas, &width, &height);
    if ( ( canvas = m_calloc (sizeof *canvas) ) == NULL )
    {
	m_error_notify (function_name, "world canvas");
	return (NULL);
    }
    canvas->magic_number = MAGIC_NUMBER;
    canvas->pixcanvas = pixcanvas;
    canvas->cmap = cmap;
    if (cmap == NULL) canvas->cmap_num_pixels = 0;
    else canvas->cmap_num_pixels = kcmap_get_pixels (cmap,
						     &canvas->cmap_pixels);
    m_copy (&canvas->win_scale, win_scale, sizeof *win_scale);
    canvas->win_scale.x_offset = 0;
    canvas->win_scale.x_pixels = width;
    canvas->win_scale.y_offset = 0;
    canvas->win_scale.y_pixels = height;
    canvas->dressing.internal_ticks = TRUE;
    canvas->dressing.font_scale = 1.0;
    /*  Process precomputes from lower down  */
    canvas->precompute_func =
	kwin_register_precompute_func (pixcanvas,
				       _canvas_pixcanvas_precompute_func,
				       canvas);
    /*  Process refreshes from lower down  */
    canvas->refresh_func =
	kwin_register_refresh_func (pixcanvas, _canvas_pixcanvas_refresh_func,
				    canvas);
    /*  Process position events from lower down  */
    canvas->position_func =
	kwin_register_position_event_func (pixcanvas,
					   _canvas_pixcanvas_position_event,
					   canvas);
    /*  Process destroy events from lower down  */
    canvas->destroy_func =
	kwin_register_destroy_func (pixcanvas, canvas_destroy, canvas);
    /*  Process colourmap resizes  */
    if (cmap != NULL)
	canvas->cmap_resize_func =
	    kcmap_register_resize_func (cmap, _canvas_cmap_resize_func,canvas);
    return (canvas);
}   /*  End Function canvas_create  */

/*EXPERIMENTAL_FUNCTION*/
void canvas_destroy (KWorldCanvas canvas)
/*  [SUMMARY] Destroy a world canvas.
    <canvas> The world canvas.
    [RETURNS] Nothing.
*/
{
    flag pixcanvas_destroying;
    unsigned int count;
    static char function_name[] = "canvas_destroy";

    VERIFY_CANVAS (canvas);
    pixcanvas_destroying = c_list_dispatching (canvas->destroy_func);
    c_unregister_callback (canvas->precompute_func);
    c_unregister_callback (canvas->refresh_func);
    c_unregister_callback (canvas->position_func);
    c_unregister_callback (canvas->destroy_func);
    if (canvas->cmap_resize_func != NULL)
	c_unregister_callback (canvas->cmap_resize_func);
    if (!pixcanvas_destroying)
	c_call_callbacks (canvas->wc_only_destroy_list, NULL);
    c_call_callbacks (canvas->destroy_list, NULL);
    c_destroy_list (canvas->precompute_list);
    c_destroy_list (canvas->refresh_list);
    c_destroy_list (canvas->position_list);
    c_destroy_list (canvas->wc_only_destroy_list);
    c_destroy_list (canvas->destroy_list);
    canvas_register_transforms_func (canvas, NULL, NULL);
    canvas_free_dressing (&canvas->dressing);
    for (count = 0; count < canvas->restrictions_allocated; ++count)
	m_free (canvas->restriction_names[count]);
    if (canvas->restriction_names != NULL) m_free (canvas->restriction_names);
    if (canvas->restriction_values != NULL) m_free(canvas->restriction_values);
    canvas->magic_number = 0;
    m_free (canvas);
}   /*  End Function canvas_destroy  */

/*UNPUBLISHED_FUNCTION*/
KCallbackFunc canvas_register_precompute_func (KWorldCanvas canvas,
					       void (*precompute_func) (),
					       void *info)
/*  [SUMMARY] Register a refresh precompute callback.
    [PURPOSE] This routine will register a refresh precompute function for a
    workd canvas. The refresh precompute function will be called whenever the
    contents of the canvas need to be redrawn, prior to any drawing operations
    commencing. The callback must not perform any drawing operations: any
    drawing operations may be performed by the ordinary refresh callbacks. Many
    precompute functions may be registered per canvas. The first function
    registered is the first function called upon refresh.
    <canvas> The world canvas.
    <func> The function that is called when the canvas is to be refreshed. The
    prototype function is [<canvas_PROTO_precompute_func>].
    <info> The initial arbitrary canvas information pointer.
    [RETURNS] A KCallbackFunc object.
*/
{
    static char function_name[] = "canvas_register_precompute_func";

    VERIFY_CANVAS (canvas);
    return c_register_callback (&canvas->precompute_list,
				_canvas_precompute_event_func,
				canvas,
				info, TRUE, (void *) precompute_func, FALSE,
				FALSE);
}   /*  End Function canvas_register_precompute_func  */

/*PUBLIC_FUNCTION*/
KCallbackFunc canvas_register_refresh_func (KWorldCanvas canvas,
					    void (*refresh_func) (),void *info)
/*  [SUMMARY] Register world canvas refresh callback.
    [PURPOSE] This routine will register a refresh function for a world canvas.
    The refresh function will be called whenever the contents of the canvas
    need to be redrawn. Many refresh functions may be registered per canvas.
    The first function registered is the first function called upon refresh.
    <canvas> The world canvas.
    <refresh_func> The function that is called when the canvas is to be
    refreshed. The prototype function is [<canvas_PROTO_refresh_func>].
    <info> The initial arbitrary canvas information pointer.
    [RETURNS] A KCallbackFunc object.
*/
{
    static char function_name[] = "canvas_register_refresh_func";

    VERIFY_CANVAS (canvas);
    return c_register_callback (&canvas->refresh_list,
				_canvas_refresh_event_func,
				canvas,
				info, TRUE, (void *) refresh_func, FALSE,
				FALSE);
}   /*  End Function canvas_register_refresh_func  */

/*PUBLIC_FUNCTION*/
void canvas_register_size_control_func (KWorldCanvas canvas,
					void (*size_control_func) (),
					void *info)
/*  [SUMMARY] Register size control callback.
    [PURPOSE] This routine will register the size control function for a world
    canvas. This function will be called whenever the lower level pixel canvas
    resizes, or the associated Kcolourmap object is resized, OR a call is made
    to [<canvas_resize>]. The function may alter the window scaling information
    for the canvas. Only one size control function is permitted per canvas. The
    size control routine is called before any refresh functions are called.
    This is a means for a higher level object to take control of the world
    canvas.
    <canvas> The world canvas object.
    <size_control_func> The function that is called when the canvas is to be
    refreshed. The prototype function is [<canvas_PROTO_size_control_func>].
    <info> The initial arbitrary canvas information pointer.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "canvas_register_size_control_func";

    VERIFY_CANVAS (canvas);
    if (size_control_func == NULL) return;
    if (canvas->size_control_func != NULL)
    {
	fprintf (stderr, "size_control_func already registered\n");
	a_prog_bug (function_name);
    }
    canvas->size_control_func = size_control_func;
    canvas->size_control_info = info;
}   /*  End Function canvas_register_size_control_func  */

/*PUBLIC_FUNCTION*/
KCallbackFunc canvas_register_position_event_func (KWorldCanvas canvas,
						   flag (*func) (),
						   void *f_info)
/*  [SUMMARY] Register position event callback.
    [PURPOSE] This routine will register a position event function for a world
    canvas. The position event function will be called whenever a position
    event on the canvas has not been consumed. Many position event functions
    may be registered per canvas. The first function registered is the first
    function called upon a position event.
    <canvas> The world canvas.
    <func> The function that is called when a position event occurs. The
    prototype function is [<canvas_PROTO_position_func>].
    <f_info> The initial arbitrary function information pointer.
    [RETURNS] A KCallbackFunc object.
*/
{
    static char function_name[] = "canvas_register_position_event_func";

    VERIFY_CANVAS (canvas);
    return c_register_callback (&canvas->position_list,
				_canvas_position_event_func,
				canvas, f_info, TRUE, (void *) func, FALSE,
				TRUE);
}   /*  End Function canvas_register_position_event_func  */

/*EXPERIMENTAL_FUNCTION*/
KCallbackFunc canvas_register_destroy_func (KWorldCanvas canvas, flag wc_only,
					    void (*func) (), void *info)
/*  [SUMMARY] Register destroy callback.
    [PURPOSE] This routine will register a destroy function for a world
    canvas. The function will be called when the canvas is destroyed. Many
    destroy functions may be registered per canvas. The first function
    registered is the first function called upon destruction.
    <canvas> The world canvas.
    <wc_only> If TRUE, the function is only called for a destroy of the world
    canvas; if the pixel canvas is being destroyed, the function is not called.
    If FALSE, the function is called regardless of whether the pixel canvas is
    being destroyed.
    <func> The function that is called when the canvas is destroyed. The
    prototype function is [<canvas_PROTO_destroy_func>].
    <info> The arbitrary function information pointer.
    [NOTE] All functions registered with <<wc_only>> TRUE are called before
    those registered with <<wc_only>> FALSE.
    [RETURNS] A KCallbackFunc object. On failure, the process aborts.
*/
{
    KCallbackList *list;
    static char function_name[] = "canvas_register_destroy_func";

    VERIFY_CANVAS (canvas);
    list = wc_only ? &canvas->wc_only_destroy_list : &canvas->destroy_list;
    return c_register_callback (list, ( flag (*) () ) func, info,
				NULL, FALSE, NULL, FALSE, FALSE);
}   /*  End Function canvas_register_destroy_func  */

/*PUBLIC_FUNCTION*/
void canvas_get_attributes (KWorldCanvas canvas, ...)
/*  [SUMMARY] Get canvas attributes.
    <canvas> The world canvas.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pointer pairs. The list must be terminated with
    CANVAS_ATT_END. See [<canvas_ATTRIBUTES>] for the list of attributes.
    [RETURNS] Nothing.
*/
{
    va_list argp;
    unsigned int att_key;
    static char function_name[] = "canvas_get_attributes";

    va_start (argp, canvas);
    VERIFY_CANVAS (canvas);
    while ( ( att_key = va_arg (argp, unsigned int) ) != CANVAS_ATT_END )
    {
	switch (att_key)
	{
	  case CANVAS_ATT_X_OFFSET:
	    *( va_arg (argp, int *) ) = canvas->win_scale.x_offset;
	    break;
	  case CANVAS_ATT_Y_OFFSET:
	    *( va_arg (argp, int *) ) = canvas->win_scale.y_offset;
	    break;
	  case CANVAS_ATT_X_PIXELS:
	    *( va_arg (argp, int *) ) = canvas->win_scale.x_pixels;
	    break;
	  case CANVAS_ATT_Y_PIXELS:
	    *( va_arg (argp, int *) ) = canvas->win_scale.y_pixels;
	    break;
	  case CANVAS_ATT_BLANK_PIXEL:
	    if (canvas->auto_blank)
	    {
		kwin_get_attributes (canvas->pixcanvas,
				     KWIN_ATT_BACKGROUND_PIXEL,
				     &canvas->win_scale.blank_pixel,
				     KWIN_ATT_END);
	    }
	    *( va_arg (argp, unsigned long *) ) =canvas->win_scale.blank_pixel;
	    break;
	  case CANVAS_ATT_MIN_SAT_PIXEL:
	    *( va_arg (argp, unsigned long *) ) =
		canvas->win_scale.min_sat_pixel;
	    break;
	  case CANVAS_ATT_MAX_SAT_PIXEL:
	    *( va_arg (argp, unsigned long *) ) =
		canvas->win_scale.max_sat_pixel;
	    break;
	  case CANVAS_ATT_LEFT_X:
	    *( va_arg (argp, double *) ) = canvas->win_scale.left_x;
	    break;
	  case CANVAS_ATT_RIGHT_X:
	    *( va_arg (argp, double *) ) = canvas->win_scale.right_x;
	    break;
	  case CANVAS_ATT_BOTTOM_Y:
	    *( va_arg (argp, double *) ) = canvas->win_scale.bottom_y;
	    break;
	  case CANVAS_ATT_TOP_Y:
	    *( va_arg (argp, double *) ) = canvas->win_scale.top_y;
	    break;
	  case CANVAS_ATT_VALUE_MIN:
	    *( va_arg (argp, double *) ) = canvas->win_scale.z_min;
	    break;
	  case CANVAS_ATT_VALUE_MAX:
	    *( va_arg (argp, double *) ) = canvas->win_scale.z_max;
	    break;
	  case CANVAS_ATT_ISCALE_FUNC:
	    *( va_arg (argp, void **) ) =(void *)canvas->win_scale.iscale_func;
	    break;
	  case CANVAS_ATT_ISCALE_FREE_FUNC:
	    *( va_arg (argp, void **) ) =
		(void *) canvas->win_scale.iscale_free_info_func;
	    break;
	  case CANVAS_ATT_ISCALE_INFO:
	    *( va_arg (argp, void **) ) = canvas->win_scale.iscale_info;
	    break;
	  case CANVAS_ATT_AUTO_MIN_SAT:
	    *( va_arg (argp, flag *) ) = canvas->auto_min_sat;
	    break;
	  case CANVAS_ATT_AUTO_MAX_SAT:
	    *( va_arg (argp, flag *) ) = canvas->auto_max_sat;
	    break;
	  case CANVAS_ATT_AUTO_BLANK:
	    *( va_arg (argp, flag *) ) = canvas->auto_blank;
	    break;
	  case CANVAS_ATT_WCS_ASTRO:
	    *( va_arg (argp, KwcsAstro *) ) = canvas->ap;
	    break;
	  default:
	    fprintf (stderr, "Unknown attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	    break;
	}
    }
    va_end (argp);
}   /*  End Function canvas_get_attributes  */

/*PUBLIC_FUNCTION*/
flag canvas_set_attributes (KWorldCanvas canvas, ...)
/*  [SUMMARY] Set canvas attributes.
    <canvas> The world canvas.
    [VARARGS] The optional attributes are given as pairs of attribute-key
    attribute-value pairs. The list must be terminated with
    CANVAS_ATT_END. See [<canvas_ATTRIBUTES>] for the list of attributes.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    va_list argp;
    unsigned int att_key;
    static char function_name[] = "canvas_set_attributes";

    va_start (argp, canvas);
    VERIFY_CANVAS (canvas);
    if (canvas->in_refresh)
    {
	fprintf (stderr, "Attempt to change attributes during refresh\n");
	a_prog_bug (function_name);
    }
    while ( ( att_key = va_arg (argp, unsigned int) ) != CANVAS_ATT_END )
    {
	switch (att_key)
	{
	  case CANVAS_ATT_X_OFFSET:
	    canvas->win_scale.x_offset = va_arg (argp, int);
	    break;
	  case CANVAS_ATT_Y_OFFSET:
	    canvas->win_scale.y_offset = va_arg (argp, int);
	    break;
	  case CANVAS_ATT_X_PIXELS:
	    canvas->win_scale.x_pixels = va_arg (argp, int);
	    break;
	  case CANVAS_ATT_Y_PIXELS:
	    canvas->win_scale.y_pixels = va_arg (argp, int);
	    break;
	  case CANVAS_ATT_BLANK_PIXEL:
	    if (!canvas->auto_blank)
		canvas->win_scale.blank_pixel = va_arg (argp, unsigned long);
	    break;
	  case CANVAS_ATT_MIN_SAT_PIXEL:
	    canvas->win_scale.min_sat_pixel = va_arg (argp, unsigned long);
	    break;
	  case CANVAS_ATT_MAX_SAT_PIXEL:
	    canvas->win_scale.max_sat_pixel = va_arg (argp, unsigned long);
	    break;
	  case CANVAS_ATT_LEFT_X:
	    canvas->win_scale.left_x = va_arg (argp, double);
	    break;
	  case CANVAS_ATT_RIGHT_X:
	    canvas->win_scale.right_x = va_arg (argp, double);
	    break;
	  case CANVAS_ATT_BOTTOM_Y:
	    canvas->win_scale.bottom_y = va_arg (argp, double);
	    break;
	  case CANVAS_ATT_TOP_Y:
	    canvas->win_scale.top_y = va_arg (argp, double);
	    break;
	  case CANVAS_ATT_VALUE_MIN:
	    canvas->win_scale.z_min = va_arg (argp, double);
	    break;
	  case CANVAS_ATT_VALUE_MAX:
	    canvas->win_scale.z_max = va_arg (argp, double);
	    break;
	  case CANVAS_ATT_ISCALE_FUNC:
	    *(void **) &canvas->win_scale.iscale_func = va_arg (argp, void *);
	    break;
	  case CANVAS_ATT_ISCALE_FREE_FUNC:
	    *(void **) &canvas->win_scale.iscale_free_info_func =
		va_arg (argp, void *);
	    break;
	  case CANVAS_ATT_ISCALE_INFO:
	    canvas->win_scale.iscale_info = va_arg (argp, void *);
	    break;
	  case CANVAS_ATT_AUTO_MIN_SAT:
	    canvas->auto_min_sat = va_arg (argp, flag);
	    break;
	  case CANVAS_ATT_AUTO_MAX_SAT:
	    canvas->auto_max_sat = va_arg (argp, flag);
	    break;
	  case CANVAS_ATT_AUTO_BLANK:
	    canvas->auto_blank = va_arg (argp, flag);
	    break;
	  case CANVAS_ATT_WCS_ASTRO:
	    canvas->ap = va_arg (argp, KwcsAstro);
	    break;
	  default:
	    fprintf (stderr, "Unknown attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	    break;
	}
    }
    va_end (argp);
    _canvas_verify_win_scale (&canvas->win_scale, function_name);
    _canvas_set_sat_pixels (canvas);
    return (TRUE);
}   /*  End Function canvas_set_attributes  */

/*PUBLIC_FUNCTION*/
flag canvas_resize (KWorldCanvas canvas, struct win_scale_type *win_scale,
		    flag always_clear)
/*  [SUMMARY] Resize a world canvas.
    [PURPOSE] This routine will register a resize in the world canvas size.
    This will cause any refresh routines registered for the canvas to be
    called. The associated pixel canvas is refreshed, hence, refresh routines
    registered for the pixel canvas will also be called.
    <canvas> The world canvas object.
    <win_scale> The new window scaling information. If this is NULL no change
    is made to the window scaling information, and a simple refresh occurs.
    <always_clear> If FALSE, <<win_scale>> is NULL and there is a size
    control function registered for the world canvas, then if that function
    does not change the values of the <<x_offset>>, <<x_pixels>>, <<y_offset>>
    or <<y_pixels>> fields in the window scaling information, the associated
    pixel canvas is NOT cleared. Under all other circumstances, the pixel
    canvas is cleared prior to calling any refresh functions.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag boundary_clear = FALSE;
    int x_offset, x_pixels, y_offset, y_pixels;
    int width, height;
    KPixCanvasRefreshArea areas[5];
    static char function_name[] = "canvas_resize";

    VERIFY_CANVAS (canvas);
    FLAG_VERIFY (always_clear);
    _canvas_verify_win_scale (win_scale, function_name);
    /*  Get pixel canvas size  */
    kwin_get_size (canvas->pixcanvas, &width, &height);
    if ( (win_scale == NULL) && (!always_clear) &&
	 (canvas->size_control_func != NULL) )
    {
	/*  Call size control function first  */
	if ( !_canvas_negotiate_scale_change (canvas, width, height,
					      &boundary_clear) )
	{
	    fprintf (stderr, "%s: pixel canvas not big enough\n",
		     function_name);
	    return (FALSE);
	}
	FLAG_VERIFY (boundary_clear);
	canvas->quash_negotiate = TRUE;
    }
    if (win_scale != NULL)
    {
	m_copy (&canvas->win_scale, win_scale, sizeof *win_scale);
    }
    if (always_clear) return kwin_resize (canvas->pixcanvas, TRUE, 0, 0, 0, 0);
    /*  Clear only the boundary  */
    x_offset = canvas->win_scale.x_offset;
    x_pixels = canvas->win_scale.x_pixels;
    y_offset = canvas->win_scale.y_offset;
    y_pixels = canvas->win_scale.y_pixels;
    m_clear (areas, sizeof areas);
    /*  Define the top area  */
    areas[0].startx = 0;
    areas[0].endx = width - 1;
    areas[0].starty = 0;
    areas[0].endy = y_offset - 1;
    areas[0].geom_clear = TRUE;
    /*  Define the bottom area  */
    areas[1].startx = 0;
    areas[1].endx = width - 1;
    areas[1].starty = y_offset + y_pixels;
    areas[1].endy = height - 1;
    areas[1].geom_clear = TRUE;
    /*  Define the left area  */
    areas[2].startx = 0;
    areas[2].endx = x_offset - 1;
    areas[2].starty = y_offset;
    areas[2].endy = y_offset + y_pixels - 1;
    areas[2].geom_clear = TRUE;
    /*  Define the right area  */
    areas[3].startx = x_offset + x_pixels;
    areas[3].endx = width - 1;
    areas[3].starty = y_offset;
    areas[3].endy = y_offset + y_pixels - 1;
    areas[3].geom_clear = TRUE;
    /*  And finally define the world canvas itself  */
    areas[4].startx = x_offset;
    areas[4].endx = x_offset + x_pixels - 1;
    areas[4].starty = y_offset;
    areas[4].endy = y_offset + y_pixels - 1;
    areas[4].geom_clear = TRUE;
    return kwin_partial_refresh (canvas->pixcanvas, 5, areas, FALSE);
}   /*  End Function canvas_resize  */

/*PUBLIC_FUNCTION*/
flag canvas_specify (KWorldCanvas canvas,
		     CONST char *xlabel, CONST char *ylabel,
		     unsigned int num_restr, CONST char **restr_names,
		     CONST double *restr_values)
/*  [SUMMARY] Set specification information for a world canvas.
    [PURPOSE] This routine will change the specifications for a canvas. This
    includes the horizontal and vertical dimension labels and any restrictions.
    The entire specification is copied (i.e. the inputs may be freed).
    <canvas> The world canvas object.
    <xlabel> The horizontal dimension label.
    <ylabel> The vertical dimension label.
    <num_restr> The number of restrictions.
    <restr_names> The array of restriction names.
    <restr_values> The restriction values.
    [NOTE] This routine DOES NOT cause the canvas to be refreshed. It is highly
    recommended that the canvas is refreshed after this routine, as higher
    level packages may depend on the specification information.
    [NOTE] All string values are copied.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    int len;
    unsigned int count;
    char **names;
    double *values;
    static char function_name[] = "canvas_specify";

    VERIFY_CANVAS (canvas);
    if (xlabel == NULL) canvas->specify_xlabel[0] = '\0';
    else strcpy (canvas->specify_xlabel, xlabel);
    if (ylabel == NULL) canvas->specify_ylabel[0] = '\0';
    else strcpy (canvas->specify_ylabel, ylabel);
    if (num_restr > canvas->restrictions_allocated)
    {
	/*  Allocate restriction space  */
	if ( ( names = m_alloc (sizeof *names * num_restr) ) == NULL )
	{
	    m_error_notify (function_name, "array of restriction names");
	    return (FALSE);
	}
	if ( ( values = m_alloc (sizeof *values * num_restr) ) == NULL )
	{
	    m_error_notify (function_name, "array of restriction values");
	    m_free (names);
	    return (FALSE);
	}
	for (count = 0; count < num_restr; ++count)
	{
	    if ( ( names[count] = m_alloc (STRING_LENGTH) ) == NULL )
	    {
		m_error_notify (function_name, "restriction names");
		m_free (names);
		m_free (values);
		while (count > 0)
		{
		    m_free (names[count]);
		    --count;
		}
		return (FALSE);
	    }
	}
	/*  Free old  */
	if (canvas->restriction_names != NULL)
	{
	    for (count = 0; count < canvas->restrictions_allocated; ++count)
		m_free (canvas->restriction_names[count]);
	    m_free (canvas->restriction_names);
	}
	if (canvas->restriction_values != NULL)
	    m_free (canvas->restriction_values);
	canvas->restriction_names = names;
	canvas->restriction_values = values;
	canvas->restrictions_allocated = num_restr;
    }
    else
    {
	names = canvas->restriction_names;
	values = canvas->restriction_values;
    }
    /*  Copy the restriction data  */
    canvas->num_restrictions = num_restr;
    for (count = 0; count < num_restr; ++count)
    {
	len = strlen (restr_names[count]);
	if (len >= STRING_LENGTH) len = STRING_LENGTH - 1;
	m_copy (names[count], restr_names[count], len);
	names[count][len] = '\0';
	values[count] = restr_values[count];
    }
    return (TRUE);
}   /*  End Function canvas_specify  */

/*PUBLIC_FUNCTION*/
void canvas_get_specification (KWorldCanvas canvas, char **xlabel,
			       char **ylabel, unsigned int *num_restr,
			       char ***restr_names, double **restr_values)
/*  [SUMMARY] Get specification information for a world canvas.
    [PURPOSE] This routine will get the specifications for a canvas. This
    includes the horizontal and vertical dimension labels and any restrictions.
    Only pointers are copied (i.e. the outputs may NOT be freed).
    <canvas> The world canvas object.
    <xlabel> The pointer to the horizontal dimension label will be written
    here. This may be NULL.
    <ylabel> The pointer to the vertical dimension label will be written here.
    This may be NULL.
    <num_restr> The number of restrictions will be written here. This may be
    NULL.
    <restr_names> The pointer to the array of restriction names will be written
    here. This may be NULL.
    <restr_values> The pointer to the array of restriction values will be
    written here. This may be NULL.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "canvas_get_specification";

    VERIFY_CANVAS (canvas);
    if (xlabel != NULL)
    {
	if (canvas->specify_xlabel[0] == '\0') *xlabel = NULL;
	else *xlabel = canvas->specify_xlabel;
    }
    if (ylabel != NULL)
    {
	if (canvas->specify_ylabel[0] == '\0') *ylabel = NULL;
	else *ylabel = canvas->specify_ylabel;
    }
    if (num_restr != NULL) *num_restr = canvas->num_restrictions;
    if (restr_names != NULL) *restr_names = canvas->restriction_names;
    if (restr_values != NULL) *restr_values = canvas->restriction_values;
}   /*  End Function canvas_get_specification  */

/*PUBLIC_FUNCTION*/
void canvas_convert_to_canvas_coords (KWorldCanvas canvas, flag clip,
				      unsigned int num_coords,
				      CONST double *xin, CONST double *yin,
				      double *xout_lin, double *yout_lin,
				      double *xout, double *yout)
/*  [SUMMARY] Convert pixel co-ordinates to world co-ordinates.
    <canvas> The world canvas object.
    <clip> If TRUE, pixel co-ordinates are first clipped to the canvas
    boundaries prior to conversion.
    <num_coords> The number of co-ordinates to convert.
    <xin> The array of input horizontal pixel co-ordinates.
    <yin> The array of input vertical pixel co-ordinates.
    <xout_lin> The array of output horizontal linear world co-ordinates are
    written here. If this is NULL, nothing is written here.
    <yout_lin> The array of output vertical linear world co-ordinates are
    written here. If this is NULL, nothing is written here.
    <xout> The array of output horizontal non-linear world co-ordinates are
    written here. If this is NULL, nothing is written here.
    <yout> The array of output vertical non-linear world co-ordinates are
    written here. If this is NULL, nothing is written here.
    [RETURNS] Nothing.
*/
{
    flag converted;
    int px, py;
    unsigned int count;
    double dx, dy;
    struct win_scale_type win_scale;
    static char function_name[] = "canvas_convert_to_canvas_coords";

    VERIFY_CANVAS (canvas);
    for (count = 0; count < num_coords; ++count)
    {
	dx = xin[count];
	dy = yin[count];
	if ( clip && (dx < canvas->win_scale.x_offset) )
	{
	    dx = canvas->win_scale.x_offset;
	}
	if ( clip && (dx >= canvas->win_scale.x_offset +
		      canvas->win_scale.x_pixels) )
	{
	    dx = canvas->win_scale.x_offset + canvas->win_scale.x_pixels - 1;
	}
	if ( clip && (dy < canvas->win_scale.y_offset) )
	{
	    dy = canvas->win_scale.y_offset;
	}
	if ( clip && (dy >= canvas->win_scale.y_offset +
		      canvas->win_scale.y_pixels) )
	{
	    dy = canvas->win_scale.y_offset + canvas->win_scale.y_pixels - 1;
	}
	converted = FALSE;
	if (canvas->coords_convert_func != NULL)
	{
	    converted = ( (*canvas->coords_convert_func)
			  (canvas, 1, &dx, &dy, &dx, &dy, TRUE,
			   &canvas->coord_convert_info) );
	}
	else if (canvas->deprecated_coord_d_convert_func != NULL)
	{
	    m_copy (&win_scale, &canvas->win_scale, sizeof win_scale);
	    converted = ( (*canvas->deprecated_coord_d_convert_func)
			  (canvas, &win_scale, &dx, &dy, TRUE,
			   &canvas->coord_convert_info) );
	}
	else if (canvas->deprecated_coord_convert_func != NULL)
	{
	    m_copy (&win_scale, &canvas->win_scale, sizeof win_scale);
	    px = dx;
	    py = dy;
	    converted = ( (*canvas->deprecated_coord_convert_func)
			  (canvas, &win_scale, &px, &py, &dx, &dy, TRUE,
			   &canvas->coord_convert_info) );
	}
	if (!converted)
	{
	    /*  Convert pixel co-ordinates to linear co-ordinates  */
	    /*  Convert x  */
	    dx = dx - canvas->win_scale.x_offset;
	    dx /= (double) (canvas->win_scale.x_pixels - 1);
	    dx *= (canvas->win_scale.right_x - canvas->win_scale.left_x);
	    dx += canvas->win_scale.left_x;
	    /*  Flip vertical  */
	    dy -= canvas->win_scale.y_offset;
	    dy = canvas->win_scale.y_pixels - dy - 1;
	    /*  Convert y  */
	    dy /= (double) (canvas->win_scale.y_pixels - 1);
	    dy *= (canvas->win_scale.top_y - canvas->win_scale.bottom_y);
	    dy += canvas->win_scale.bottom_y;
	}
	/*  dx and dy should now be linear co-ordinates  */
	if (xout_lin != NULL) xout_lin[count] = dx;
	if (yout_lin != NULL) yout_lin[count] = dy;
	/*  Apply non-linear transform function  */
	if ( (xout != NULL) || (yout != NULL) )
	{
	    canvas_coords_transform (canvas, 1, &dx, FALSE, &dy, FALSE);
	    if (xout != NULL) xout[count] = dx;
	    if (yout != NULL) yout[count] = dy;
	}
    }
}   /*  End Function canvas_convert_to_canvas_coords  */

/*PUBLIC_FUNCTION*/
void canvas_convert_from_canvas_coords (KWorldCanvas canvas,
					flag clip, flag linear,
					unsigned int num_coords,
					CONST double *xin, CONST double *yin,
					double *xout, double *yout)
/*  [SUMMARY] Convert world co-ordinates to pixel co-ordinates.
    <canvas> The world canvas object.
    <clip> If TRUE, world co-ordinates are first clipped to the canvas
    boundaries prior to conversion.
    <linear> If TRUE, the input world co-ordinates are assumed to be linear,
    hence no non-linear to linear co-ordinate transformation is applied.
    <num_coords> The number of co-ordinates to convert.
    <xin> The array of input horizontal world co-ordinates.
    <yin> The array of input vertical world co-ordinates.
    <xout> The array of output horizontal pixel co-ordinates are written here.
    <yout> The array of output vertical pixel co-ordinates are written here.
    [RETURNS] Nothing.
*/
{
    flag converted = FALSE;
    int ix, iy;
    unsigned int count;
    double px, py;
    struct win_scale_type win_scale;
    static char function_name[] = "canvas_convert_from_canvas_coords";

    VERIFY_CANVAS (canvas);
    if ( (canvas->coord_transforms_func == NULL) &&
	 (canvas->deprecated_coord_transform_func == NULL) ) linear = TRUE;
    if (!linear)
    {
	/*  Input is non-linear: first convert to linear co-ordinates  */
	/*  Copy input to working array  */
	for (count = 0; count < num_coords; ++count) xout[count] = xin[count];
	for (count = 0; count < num_coords; ++count) yout[count] = yin[count];
	canvas_coords_transform (canvas, num_coords, xout, TRUE, yout, TRUE);
	if ( !clip && (canvas->coords_convert_func != NULL) )
	{
	    if ( (*canvas->coords_convert_func) (canvas, num_coords,
						 xout, yout, xout, yout, FALSE,
						 &canvas->coord_convert_info) )
		return;
	}
	/*  Co-ordinates have not been converted to pixel co-ordinates yet  */
	canvas_convert_from_canvas_coords (canvas, clip, TRUE, num_coords,
					   xout, yout, xout, yout);
	return;
    }
    /*  Linear conversion only  */
    if ( !clip && (canvas->coords_convert_func != NULL) )
    {
	if ( (*canvas->coords_convert_func) (canvas, num_coords,
					     xin, yin, xout, yout, FALSE,
					     &canvas->coord_convert_info) )
	    return;
    }
    /*  Have to do this the hard way  */
    m_copy (&win_scale, &canvas->win_scale, sizeof win_scale);
    for (count = 0; count < num_coords; ++count)
    {
	px = xin[count];
	py = yin[count];
	if (clip)
	{
	    if (win_scale.left_x < win_scale.right_x)
	    {
		if (px < win_scale.left_x) px = win_scale.left_x;
		else if (px > win_scale.right_x) px = win_scale.right_x;
	    }
	    else
	    {
		if (px > win_scale.left_x) px = win_scale.left_x;
		else if (px < win_scale.right_x) px = win_scale.right_x;
	    }
	    if (win_scale.bottom_y < win_scale.top_y)
	    {
		if (py < win_scale.bottom_y) py = win_scale.bottom_y;
		else if (py > win_scale.top_y) py = win_scale.top_y;
	    }
	    else
	    {
		if (py > win_scale.bottom_y) py = win_scale.bottom_y;
		else if (py < win_scale.top_y) py = win_scale.top_y;
	    }
	}
	if (canvas->coords_convert_func != NULL)
	{
	    converted = ( (*canvas->coords_convert_func)
			  (canvas, 1, &px, &py, &px, &py, FALSE,
			   &canvas->coord_convert_info) );
	}
	else if (canvas->deprecated_coord_d_convert_func != NULL)
	{
	    converted = ( (*canvas->deprecated_coord_d_convert_func)
			  (canvas, &win_scale, &px, &py, FALSE,
			   &canvas->coord_convert_info) );
	}
	else if (canvas->deprecated_coord_convert_func != NULL)
	{
	    converted = ( (*canvas->deprecated_coord_convert_func)
			  (canvas, &win_scale, &ix, &iy, &px, &px, FALSE,
			   &canvas->coord_convert_info) );
	    px = ix;
	    py = iy;
	}
	if (!converted)
	{
	    /*  Convert x  */
	    px = px - (double) canvas->win_scale.left_x;
	    px /= (double) (canvas->win_scale.right_x -
			    canvas->win_scale.left_x);
	    px *= (double) (canvas->win_scale.x_pixels - 1);
	    px += (double) canvas->win_scale.x_offset;
	    /*  Convert y  */
	    py = py - canvas->win_scale.bottom_y;
	    py /= canvas->win_scale.top_y - canvas->win_scale.bottom_y;
	    py *= (double) (canvas->win_scale.y_pixels - 1);
	    py = (double) (canvas->win_scale.y_offset +
			   canvas->win_scale.y_pixels - 1) - py;
	}
	xout[count] = px;
	yout[count] = py;
    }
}   /*  End Function canvas_convert_from_canvas_coords  */

/*PUBLIC_FUNCTION*/
void canvas_register_coords_convert_func (KWorldCanvas canvas,
					  flag (*coord_convert_func) (),
					  void *info)
/*  [SUMMARY] Register co-ordinate conversion function.
    [PURPOSE] This routine will register the co-ordinate conversion function
    for a world canvas. This function will be called whenever conversions
    between linear co-ordinates and pixel co-ordinates are performed. Only one
    co-ordinate conversion function is permitted per canvas. This is a means
    for a higher level object to take control of the world canvas.
    <canvas> The canvas.
    <func> The function that is called when co-ordinates are to be converted.
    The prototype function is [<canvas_PROTO_coord_convert_func>].
    <info> The initial arbitrary canvas information pointer.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "canvas_register_coords_convert_func";

    VERIFY_CANVAS (canvas);
    if ( (canvas->coords_convert_func != NULL) &&
	 (coord_convert_func != canvas->coords_convert_func) )
    {
	fprintf (stderr, "coord_convert_func already registered\n");
	a_prog_bug (function_name);
    }
    canvas->coords_convert_func = coord_convert_func;
    canvas->coord_convert_info = info;
}   /*  End Function canvas_register_coords_convert_func  */

/*PUBLIC_FUNCTION*/
void canvas_coords_transform (KWorldCanvas canvas, unsigned int num_coords,
			      double *x, flag x_to_linear,
			      double *y, flag y_to_linear)
/*  [SUMMARY] Transform world co-ordinates.
    [PURPOSE] This routine will transform co-ordinates in a world canvas using
    the registered co-ordinate transform function (see
    [<canvas_register_transforms_func>]).
    <canvas> The world canvas object.
    <num_coords> The number of co-ordinates to transform.
    <x> The array of horizontal world co-ordinates. These are modified.
    <x_to_linear> If TRUE, then the horizontal co-ordinates are converted from
    non-linear to linear, else they are converted from linear to non-linear.
    <y> The array of vertical world co-ordinates. These are modified.
    <y_to_linear> If TRUE, then the vertical co-ordinates are converted from
    non-linear to linear, else they are converted from linear to non-linear.
    [RETURNS] Nothing.
*/
{
    unsigned int count;
    static char function_name[] = "canvas_coords_transform";

    VERIFY_CANVAS (canvas);
    if (canvas->coord_transforms_func != NULL)
    {
	(*canvas->coord_transforms_func) (canvas, num_coords,
					  x, x_to_linear, y, y_to_linear,
					  canvas->win_scale.left_x,
					  canvas->win_scale.right_x,
					  canvas->win_scale.bottom_y,
					  canvas->win_scale.top_y,
					  &canvas->coord_transform_info);
	return;
    }
    if (canvas->deprecated_coord_transform_func != NULL)
    {
	if (x_to_linear != y_to_linear)
	{
	    fputs ("x_to_linear and y_to_linear do not match: deprecated ",
		   stderr);
	    fputs ("co-ordinate transform\nroutine not supported\n", stderr);
	    return;
	}
	for (count = 0; count < num_coords; ++count)
	{
	    (*canvas->deprecated_coord_transform_func)
		(x + count, y + count, x_to_linear,
		 canvas->win_scale.left_x,
		 canvas->win_scale.right_x,
		 canvas->win_scale.bottom_y,
		 canvas->win_scale.top_y,
		 &canvas->coord_transform_info);
	}
    }
}   /*  End Function canvas_coords_transform  */

/*PUBLIC_FUNCTION*/
void canvas_register_transforms_func (KWorldCanvas canvas,
				      void (*coord_transform_func) (),
				      void *info)
/*  [SUMMARY] Register non-linear co-ordinate transformation function.
    [PURPOSE] This routine will register the co-ordinate transform function
    for a world canvas. The co-ordinate transform function transforms the
    linear world co-ordinates used internally by the world canvas to be
    transformed into non-linear co-ordinates for external use. Only one
    co-ordinate transform function is permitted per canvas.
    <canvas> The world canvas object.
    <coord_transform_func> The function that is called when co-ordinates are to
    be transformed. The prototype function is
    [<canvas_PROTO_coord_transform_func>]. If an existing transformation
    function is registered, it is overwritten.
    <info> The initial arbitrary transform information pointer.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "canvas_register_transforms_func";

    VERIFY_CANVAS (canvas);
    if ( (canvas->coord_transform_info != NULL) &&
	 (canvas->coord_transforms_free_func != NULL) )
	(*canvas->coord_transforms_free_func) (canvas->coord_transform_info);
    canvas->coord_transforms_func = coord_transform_func;
    canvas->coord_transform_info = info;
    canvas->coord_transforms_free_func = NULL;
}   /*  End Function canvas_register_transforms_func  */

/*EXPERIMENTAL_FUNCTION*/
void canvas_register_transforms_free_func ( KWorldCanvas canvas,
					    void (*func) () )
/*  [SUMMARY] Register non-linear co-ordinate transformation free function.
    <canvas> The world canvas object.
    <func> The function that is called when transform data is to be freed.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "canvas_register_transforms_free_func";

    VERIFY_CANVAS (canvas);
    canvas->coord_transforms_free_func = func;
}   /*  End Function canvas_register_transforms_free_func  */

/*PUBLIC_FUNCTION*/
KPixCanvas canvas_get_pixcanvas (KWorldCanvas canvas)
/*  [SUMMARY] Get the underlying pixel canvas of a world canvas.
    <canvas> The world canvas object.
    [RETURNS] The KPixCanvas object.
*/
{
    static char function_name[] = "canvas_get_pixcanvas";

    VERIFY_CANVAS (canvas);
    return (canvas->pixcanvas);
}   /*  End Function canvas_get_pixcanvas  */

/*PUBLIC_FUNCTION*/
flag canvas_get_colour (KWorldCanvas canvas, CONST char *colourname,
			unsigned long *pixel_value, unsigned short *red,
			unsigned short *green, unsigned short *blue)
/*  [SUMMARY] Get (possibly allocating) a colourcell for a canvas.
    <canvas> The world canvas object.
    <colourname> The name of the colour to get.
    <pixel_value> The pixel value will be written here.
    [NOTE] The pixel value is valid ONLY for the underlying pixel canvas.
    ALSO: the pixel value becomes invalid EVERY time the underlying pixel
    canvas is refreshed/ resized. If the pixel canvas is refreshed/ resized,
    this routine MUST be called again.
    <red> The red intensity in the hardware colourmap for the pixel will be
    written here. If this is NULL, nothing is written here.
    <green> The green intensity in the hardware colourmap for the pixel will be
    written here. If this is NULL, nothing is written here.
    <blue> The blue intensity in the hardware colourmap for the pixel will be
    written here. If this is NULL, nothing is written here.
    [RETURNS] TRUE if the colourcell was allocated, else FALSE.
*/
{
    double value[2];
    CONST char *p;
    static char function_name[] = "canvas_get_colour";

    VERIFY_CANVAS (canvas);
    if ( (canvas->cmap != NULL) && (colourname[0] == '!') )
    {
	/*  Embedded value in colourname  */
	value[0] = ex_float (colourname + 1, &p);
	if ( (p == NULL) ||
	     ( (*p == '0') && ( (p[1] == '\0') || isspace (p[1]) ) ) )
	    value[1] = 0.0;
	else value[1] = ex_float (p, &p);
	*pixel_value = _canvas_get_pixel_from_value (canvas, value,
						     red, green, blue);
	return (TRUE);
    }
    return kwin_get_colour (canvas->pixcanvas, colourname, pixel_value,
			    red, green, blue);
}   /*  End Function canvas_get_colour  */

/*PUBLIC_FUNCTION*/
flag canvas_set_dressing (KWorldCanvas canvas, ...)
/*  [SUMMARY] Set dressing parameters for a world canvas.
    <canvas> The world canvas.
    [VARARGS] The list of parameter attribute-key attribute-value pairs. This
    list must be terminated with the value KCD_ATT_END.
    [NOTE] The dressing code is not yet complete.
    [RETURNS] TRUE if changes were made, else FALSE.
*/
{
    va_list arg_pointer;
    flag bool;
    flag were_changes = FALSE;
    unsigned int att_key;
    double dval;
    char *string, *ptr;
    static char function_name[] = "canvas_set_dressing";

    va_start (arg_pointer, canvas);
    VERIFY_CANVAS (canvas);
    while ( ( att_key = va_arg (arg_pointer, unsigned int) ) != KCD_ATT_END )
    {
	switch (att_key)
	{
	  case KCD_ATT_DISPLAY:
	    bool = va_arg (arg_pointer, flag);
	    FLAG_VERIFY (bool);
	    if (bool == canvas->dressing.display) continue;
	    were_changes = TRUE;
	    canvas->dressing.display = bool;
	    break;
	  case KCD_ATT_TOP_TICKS:
	    bool = va_arg (arg_pointer, flag);
	    FLAG_VERIFY (bool);
	    if (bool == canvas->dressing.show_top_ticks) continue;
	    were_changes = TRUE;
	    canvas->dressing.show_top_ticks = bool;
	    break;
	  case KCD_ATT_BOTTOM_TICKS:
	    bool = va_arg (arg_pointer, flag);
	    FLAG_VERIFY (bool);
	    if (bool == canvas->dressing.show_bottom_ticks) continue;
	    were_changes = TRUE;
	    canvas->dressing.show_bottom_ticks = bool;
	    break;
	  case KCD_ATT_LEFT_TICKS:
	    bool = va_arg (arg_pointer, flag);
	    FLAG_VERIFY (bool);
	    if (bool == canvas->dressing.show_left_ticks) continue;
	    were_changes = TRUE;
	    canvas->dressing.show_left_ticks = bool;
	    break;
	  case KCD_ATT_RIGHT_TICKS:
	    bool = va_arg (arg_pointer, flag);
	    FLAG_VERIFY (bool);
	    if (bool == canvas->dressing.show_right_ticks) continue;
	    were_changes = TRUE;
	    canvas->dressing.show_right_ticks = bool;
	    break;
	  case KCD_ATT_INTERNAL_TICKS:
	    bool = va_arg (arg_pointer, flag);
	    FLAG_VERIFY (bool);
	    if (bool == canvas->dressing.internal_ticks) continue;
	    were_changes = TRUE;
	    canvas->dressing.internal_ticks = bool;
	    break;
	  case KCD_ATT_GRID_COLOUR:
	    ptr = va_arg (arg_pointer, char *);
	    if ( (ptr == NULL) && (canvas->dressing.grid_colour == NULL) )
		continue;
	    if ( (ptr != NULL) && (canvas->dressing.grid_colour != NULL) &&
		(strcmp (ptr, canvas->dressing.grid_colour) == 0) ) continue;
	    if ( ( string = st_dup (ptr) ) == NULL )
	    {
		m_abort (function_name, "axes colour string");
	    }
	    if (canvas->dressing.grid_colour != NULL)
		m_free (canvas->dressing.grid_colour);
	    were_changes = TRUE;
	    canvas->dressing.grid_colour = string;
	    break;
	  case KCD_ATT_SCALE_COLOUR:
	    ptr = va_arg (arg_pointer, char *);
	    if ( (ptr == NULL) && (canvas->dressing.scale_colour == NULL) )
		continue;
	    if ( (ptr != NULL) && (canvas->dressing.scale_colour != NULL) &&
		(strcmp (ptr, canvas->dressing.scale_colour) == 0) ) continue;
	    if ( ( string = st_dup (ptr) ) == NULL )
	    {
		m_abort (function_name, "scale colour string");
	    }
	    if (canvas->dressing.scale_colour != NULL)
		m_free(canvas->dressing.scale_colour);
	    were_changes = TRUE;
	    canvas->dressing.scale_colour = string;
	    break;
	  case KCD_ATT_HLABEL:
	    ptr = va_arg (arg_pointer, char *);
	    if ( (ptr ==NULL) && (canvas->dressing.hlabel == NULL) ) continue;
	    if ( (ptr != NULL) && (canvas->dressing.hlabel != NULL) &&
		(strcmp (ptr, canvas->dressing.hlabel) == 0) ) continue;
	    if ( ( string = st_dup (ptr) ) == NULL )
	    {
		m_abort (function_name, "horizontal label string");
	    }
	    if (canvas->dressing.hlabel != NULL)
		m_free (canvas->dressing.hlabel);
	    were_changes = TRUE;
	    canvas->dressing.hlabel = string;
	    break;
	  case KCD_ATT_VLABEL:
	    ptr = va_arg (arg_pointer, char *);
	    if ( (ptr ==NULL) && (canvas->dressing.vlabel == NULL) ) continue;
	    if ( (ptr != NULL) && (canvas->dressing.vlabel != NULL) &&
		(strcmp (ptr, canvas->dressing.vlabel) == 0) ) continue;
	    if ( ( string = st_dup (ptr) ) == NULL )
	    {
		m_abort (function_name, "vertical label string");
	    }
	    if (canvas->dressing.vlabel != NULL)
		m_free (canvas->dressing.vlabel);
	    were_changes = TRUE;
	    canvas->dressing.vlabel = string;
	    break;
	  case KCD_ATT_LABEL_COLOUR:
	    ptr = va_arg (arg_pointer, char *);
	    if ( (ptr == NULL) && (canvas->dressing.label_colour == NULL) )
		continue;
	    if ( (ptr != NULL) && (canvas->dressing.label_colour != NULL) &&
		(strcmp (ptr, canvas->dressing.label_colour) == 0) ) continue;
	    if ( ( string = st_dup (ptr) ) == NULL )
	    {
		m_abort (function_name, "label colour string");
	    }
	    if (canvas->dressing.label_colour != NULL)
		m_free (canvas->dressing.label_colour);
	    were_changes = TRUE;
	    canvas->dressing.label_colour = string;
	    break;
	  case KCD_ATT_TITLE:
	    ptr = va_arg (arg_pointer, char *);
	    if ( (ptr == NULL) && (canvas->dressing.title == NULL) ) continue;
	    if ( (ptr != NULL) && (canvas->dressing.title != NULL) &&
		(strcmp (ptr, canvas->dressing.title) == 0) ) continue;
	    if ( ( string = st_dup (ptr) ) == NULL )
	    {
		m_abort (function_name, "title string");
	    }
	    if (canvas->dressing.title != NULL)m_free (canvas->dressing.title);
	    were_changes = TRUE;
	    canvas->dressing.title = string;
	    break;
	  case KCD_ATT_TITLE_COLOUR:
	    ptr = va_arg (arg_pointer, char *);
	    if ( (ptr == NULL) && (canvas->dressing.title_colour == NULL) )
		continue;
	    if ( (ptr != NULL) && (canvas->dressing.title_colour != NULL) &&
		(strcmp (ptr, canvas->dressing.title_colour) == 0) ) continue;
	    if ( ( string = st_dup (ptr) ) == NULL )
	    {
		m_abort (function_name, "title colour string");
	    }
	    if (canvas->dressing.title_colour != NULL)
		m_free (canvas->dressing.title_colour);
	    were_changes = TRUE;
	    canvas->dressing.title_colour = string;
	    break;
	  case KCD_ATT_SHOW_LABELS:
	    bool = va_arg (arg_pointer, flag);
	    FLAG_VERIFY (bool);
	    if (bool == canvas->dressing.show_labels) continue;
	    were_changes = TRUE;
	    canvas->dressing.show_labels = bool;
	    break;
	  case KCD_ATT_GRID:
	    bool = va_arg (arg_pointer, flag);
	    FLAG_VERIFY (bool);
	    if (bool == canvas->dressing.grid) continue;
	    were_changes = TRUE;
	    canvas->dressing.grid = bool;
	    break;
	  case KCD_ATT_FONT_SCALE:
	    dval = va_arg (arg_pointer, double);
	    if (dval == canvas->dressing.font_scale) continue;
	    were_changes = TRUE;
	    canvas->dressing.font_scale = dval;
	    break;
	  default:
	    fprintf (stderr, "Illegal attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	}
    }
    va_end (arg_pointer);
    return (were_changes);
}   /*  End Function canvas_set_dressing  */

/*PUBLIC_FUNCTION*/
void canvas_get_dressing (KWorldCanvas canvas, ...)
/*  [SUMMARY] Get dressing parameters for a world canvas.
    <canvas> The world canvas.
    [VARARGS] The list of parameter attribute-key attribute-value pairs. This
    list must be terminated with the value KCD_ATT_END.
    [NOTE] The dressing code is not yet complete.
    [RETURNS] Nothing.
*/
{
    va_list arg_pointer;
    unsigned int att_key;
    flag *bool_ptr;
    char **ptr;
    static char function_name[] = "canvas_get_dressing";

    va_start (arg_pointer, canvas);
    VERIFY_CANVAS (canvas);
    while ( ( att_key = va_arg (arg_pointer, unsigned int) ) != KCD_ATT_END )
    {
	switch (att_key)
	{
	  case KCD_ATT_DISPLAY:
	    bool_ptr = va_arg (arg_pointer, flag *);
	    *bool_ptr = canvas->dressing.display;
	    break;
	  case KCD_ATT_TOP_TICKS:
	    bool_ptr = va_arg (arg_pointer, flag *);
	    *bool_ptr = canvas->dressing.show_top_ticks;
	    break;
	  case KCD_ATT_BOTTOM_TICKS:
	    bool_ptr = va_arg (arg_pointer, flag *);
	    *bool_ptr = canvas->dressing.show_bottom_ticks;
	    break;
	  case KCD_ATT_LEFT_TICKS:
	    bool_ptr = va_arg (arg_pointer, flag *);
	    *bool_ptr = canvas->dressing.show_left_ticks;
	    break;
	  case KCD_ATT_RIGHT_TICKS:
	    bool_ptr = va_arg (arg_pointer, flag *);
	    *bool_ptr = canvas->dressing.show_right_ticks;
	    break;
	  case KCD_ATT_INTERNAL_TICKS:
	    bool_ptr = va_arg (arg_pointer, flag *);
	    *bool_ptr = canvas->dressing.internal_ticks;
	    break;
	  case KCD_ATT_GRID_COLOUR:
	    ptr = va_arg (arg_pointer, char **);
	    *ptr = canvas->dressing.grid_colour;
	    break;
	  case KCD_ATT_SCALE_COLOUR:
	    ptr = va_arg (arg_pointer, char **);
	    *ptr = canvas->dressing.scale_colour;
	    break;
	  case KCD_ATT_HLABEL:
	    ptr = va_arg (arg_pointer, char **);
	    *ptr = canvas->dressing.hlabel;
	    break;
	  case KCD_ATT_VLABEL:
	    ptr = va_arg (arg_pointer, char **);
	    *ptr = canvas->dressing.vlabel;
	    break;
	  case KCD_ATT_LABEL_COLOUR:
	    ptr = va_arg (arg_pointer, char **);
	    *ptr = canvas->dressing.label_colour;
	    break;
	  case KCD_ATT_TITLE:
	    ptr = va_arg (arg_pointer, char **);
	    *ptr = canvas->dressing.title;
	    break;
	  case KCD_ATT_TITLE_COLOUR:
	    ptr = va_arg (arg_pointer, char **);
	    *ptr = canvas->dressing.title_colour;
	    break;
	  case KCD_ATT_SHOW_LABELS:
	    bool_ptr = va_arg (arg_pointer, flag *);
	    *bool_ptr = canvas->dressing.show_labels;
	    break;
	  case KCD_ATT_GRID:
	    bool_ptr = va_arg (arg_pointer, flag *);
	    *bool_ptr = canvas->dressing.grid;
	    break;
	  default:
	    fprintf (stderr, "Illegal attribute key: %u\n", att_key);
	    a_prog_bug (function_name);
	}
    }
    va_end (arg_pointer);
}   /*  End Function canvas_get_dressing  */

/*PUBLIC_FUNCTION*/
void canvas_sequence_dressing_refresh (KWorldCanvas canvas)
/*  [SUMMARY] Set order in which dressing is refreshed.
    [PURPOSE] This routine will sequence the dressing (axes, etc.) refresh
    function for a world canvas. If this routine is not called, the dressing
    will be drawn after *all* registered refresh functions have been called
    (see [<canvas_register_refresh_func>]). If this routine is called, all
    preceeding refresh functions (registered with
    [<canvas_register_refresh_func>]) will be called on refresh prior to the
    dressing being drawn and all refresh functions registered after this
    routine is called will be called *after* the dressing is drawn.
    <canvas> The world canvas object.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "canvas_sequence_dressing_refresh";

    VERIFY_CANVAS (canvas);
    canvas_register_refresh_func (canvas, _canvas_dressing_refresh_func, NULL);
}   /*  End Function canvas_sequence_dressing_refresh  */

/*PUBLIC_FUNCTION*/
void canvas_draw_dressing (KWorldCanvas canvas)
/*  [SUMMARY] Draw the dressing (axes, etc.) for a world canvas.
    [NOTE] This routine is called automatically each time the canvas is
    refreshed.
    <canvas> The world canvas object.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "canvas_draw_dressing";

    VERIFY_CANVAS (canvas);
    canvas_draw_dressing_internal (canvas, &canvas->dressing);
}   /*  End Function canvas_draw_dressing  */

/*PUBLIC_FUNCTION*/
Kcolourmap canvas_get_cmap (KWorldCanvas canvas)
/*  [SUMMARY] Get the Kcolourmap object associated with a world canvas.
    <canvas> The world canvas object.
    [RETURNS] The Kcolourmap object. This may be NULL (indicating there is no
    colourmap associated with the canvas).
*/
{
    static char function_name[] = "canvas_get_cmap";

    VERIFY_CANVAS (canvas);
    return (canvas->cmap);
}   /*  End Function canvas_get_cmap  */

/*EXPERIMENTAL_FUNCTION*/
flag canvas_warp_pointer (KWorldCanvas canvas, double x, double y, flag linear,
			  flag discard)
/*  [SUMMARY] Warp the pointer to a new location.
    <x> The horizontal position to warp to.
    <y> The vertical position to warp to.
    <linear> If TRUE, input co-orindates are in linear co-ordinates.
    <discard> If TRUE, any pending events are discarded.
    [RETURNS] TRUE if warping is supported, else FALSE.
*/
{
    static char function_name[] = "canvas_warp_pointer";

    VERIFY_CANVAS (canvas);
    canvas_convert_from_canvas_coords (canvas, TRUE, linear,1, &x, &y, &x, &y);
    return kwin_warp_pointer (canvas->pixcanvas, x, y, FALSE, discard);
}   /*  End Function canvas_warp_pointer  */

/*EXPERIMENTAL_FUNCTION*/
flag canvas_test_for_boundary (KWorldCanvas canvas, unsigned int num_coords,
			       CONST double *x, CONST double *y)
/*  [SUMMARY] Test if co-ordinates lie within the canvas boundary.
    <canvas> The world canvas object.
    <x> The array of horizontal world co-ordinates.
    <y> The array of vertical world co-ordinates.
    <num_coords> The number of co-ordinates to test.
    [RETURNS] TRUE if all co-ordinates lie within the boundary, else FALSE.
*/
{
    unsigned int count;
    struct world_boundary_struct *wl;
    static char function_name[] = "canvas_test_for_boundary";

    VERIFY_CANVAS (canvas);
    wl = &canvas->world_boundary;
    if (!wl->valid)  /*  Ensure boundary cache is valid  */
    {
	double x[6], y[6];

	x[0] = canvas->win_scale.left_x;
	y[0] = canvas->win_scale.bottom_y;
	x[1] = canvas->win_scale.right_x;
	y[1] = canvas->win_scale.bottom_y;
	x[2] = canvas->win_scale.right_x;
	y[2] = canvas->win_scale.top_y;
	x[3] = canvas->win_scale.left_x;
	y[3] = canvas->win_scale.top_y;
	x[4] = (x[0] + x[1]) * 0.5;
	y[4] = y[0];
	x[5] = x[0];
	y[5] = (y[0] + y[3]) * 0.5;
	canvas_coords_transform (canvas, 6, x, FALSE, y, FALSE);
#if 0
	fprintf (stderr, "X: L: %e  centre: %e  R: %e\n",
		 x[0], x[4], x[1]);
#endif
	if ( ( x[4] < MIN (x[0], x[1]) ) || ( x[4] > MAX (x[0], x[1]) ) )
	{
	    fprintf (stderr, "X wrapped\n");
	    wl->x.wrapped = TRUE;
	}
	wl->x.min = x[0];
	wl->x.max = x[0];
	for (count = 1; count < 4; ++count)
	{
	    if (x[count] < wl->x.min) wl->x.min = x[count];
	    if (x[count] > wl->x.max) wl->x.max = x[count];
	}
	if ( ( y[5] < MIN (y[0], y[3]) ) || ( y[5] > MAX (y[0], y[3]) ) )
	{
	    fprintf (stderr, "Y wrapped\n");
	    wl->y.wrapped = TRUE;
	}
	wl->y.min = y[0];
	wl->y.max = y[0];
	for (count = 1; count < 4; ++count)
	{
	    if (y[count] < wl->y.min) wl->y.min = y[count];
	    if (y[count] > wl->y.max) wl->y.max = y[count];
	}
	wl->valid = TRUE;
    }
    for (count = 0; count < num_coords; ++count, ++x, ++y)
    {
	if (*x < wl->x.min) return (FALSE);
	if (*x > wl->x.max) return (FALSE);
	if (*y < wl->y.min) return (FALSE);
	if (*y > wl->y.max) return (FALSE);
    }
    return (TRUE);
}   /*  End Function canvas_test_for_boundary  */


/*  Drawing routines follow  */

/*PUBLIC_FUNCTION*/
flag canvas_draw_image (KWorldCanvas canvas, array_desc *arr_desc, char *slice,
			unsigned int hdim, unsigned int vdim,
			unsigned int elem_index,
			KPixCanvasImageCache *cache_ptr)
/*  [SUMMARY] Draw a PseudoColour image onto a world canvas.
    [PURPOSE] This routine will draw a 2-dimensional PseudoColour slice of a
    Karma array onto a world canvas. This slice may be tiled.
    <canvas> The world canvas object.
    <arr_desc> The array descriptor for the image.
    <slice> The start of the slice (image) data.
    <hdim> The dimension index of the horizontal dimension.
    <vdim> The dimension index of the vertical dimension.
    <elem_index> The element index of the data packets.
    <cache_ptr> The routine may produce cache data which will vastly increase
    the speed of subsequent operations on this data. The routine will write a
    pointer to this data here. Prior to process exit, a call MUST be made to
    [<kwin_free_cache_data>], otherwise shared memory segments could remain
    after the process exits.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag ok;
    unsigned int abs_start_index;
    unsigned int abs_end_index;
    unsigned int ord_start_index;
    unsigned int ord_end_index;
    unsigned int num_abs_coords;
    unsigned int num_ord_coords;
    unsigned int num_pixels;
    uaddr *hoffsets, *voffsets;
    packet_desc *pack_desc;
    dim_desc *hdim_desc, *vdim_desc;
    unsigned long *pixel_values;
    static char function_name[] = "canvas_draw_image";

    VERIFY_CANVAS (canvas);
    if (canvas->cmap == NULL)
    {
	fprintf (stderr, "Canvas has no colourmap associated with it\n");
	a_prog_bug (function_name);
    }
    num_pixels = kcmap_get_pixels (canvas->cmap, &pixel_values);
    pack_desc = arr_desc->packet;
    if (elem_index >= pack_desc->num_elements)
    {
	fprintf (stderr,"elem_index: %u greater than number of elements: %u\n",
		 elem_index, pack_desc->num_elements);
	a_prog_bug (function_name);
    }
    if (canvas->auto_blank)
    {
	kwin_get_attributes (canvas->pixcanvas,
			     KWIN_ATT_BACKGROUND_PIXEL,
			     &canvas->win_scale.blank_pixel,
			     KWIN_ATT_END);
    }
    hoffsets = arr_desc->offsets[hdim];
    voffsets = arr_desc->offsets[vdim];
    hdim_desc = arr_desc->dimensions[hdim];
    vdim_desc = arr_desc->dimensions[vdim];
    /*  Determine start and stop co-ordinates along each dimension  */
    abs_start_index = ds_get_coord_num (hdim_desc, canvas->win_scale.left_x,
					SEARCH_BIAS_CLOSEST);
    abs_end_index = ds_get_coord_num (hdim_desc, canvas->win_scale.right_x,
				      SEARCH_BIAS_CLOSEST);
    if (abs_start_index >= abs_end_index)
    {
	fprintf (stderr,"Left co-ordinate index: %u not less than right: %u\n",
		 abs_start_index, abs_end_index);
	a_prog_bug (function_name);
    }
    num_abs_coords = abs_end_index - abs_start_index + 1;
    ord_start_index = ds_get_coord_num (vdim_desc, canvas->win_scale.bottom_y,
					SEARCH_BIAS_CLOSEST);
    ord_end_index = ds_get_coord_num (vdim_desc, canvas->win_scale.top_y,
				      SEARCH_BIAS_CLOSEST);
    if (ord_start_index >= ord_end_index)
    {
	fprintf (stderr,"Bottom co-ordinate index: %u not less than top: %u\n",
		 ord_start_index, ord_end_index);
	a_prog_bug (function_name);
    }
    num_ord_coords = ord_end_index - ord_start_index + 1;
    ok = kwin_draw_pc_image (canvas->pixcanvas,
			     canvas->win_scale.x_offset,
			     canvas->win_scale.y_offset,
			     canvas->win_scale.x_pixels,
			     canvas->win_scale.y_pixels,
			     slice + ds_get_element_offset (pack_desc,
							    elem_index),
			     hoffsets + abs_start_index,
			     voffsets + ord_start_index,
			     num_abs_coords, num_ord_coords,
			     pack_desc->element_types[elem_index],
			     canvas->win_scale.conv_type,
			     num_pixels, pixel_values,
			     canvas->win_scale.blank_pixel,
			     canvas->win_scale.min_sat_pixel,
			     canvas->win_scale.max_sat_pixel,
			     canvas->win_scale.z_min,
			     canvas->win_scale.z_max,
			     canvas->win_scale.iscale_func,
			     canvas->win_scale.iscale_info,
			     cache_ptr);
    if (!ok) fprintf (stderr, "%s: error drawing PseudoColour image\n",
		      function_name);
    return (ok);
}   /*  End Function canvas_draw_image  */

/*UNPUBLISHED_FUNCTION*/
flag canvas_compute_pc_image (KWorldCanvas canvas, array_desc *arr_desc,
			      char *slice,
			      unsigned int hdim, unsigned int vdim,
			      unsigned int elem_index,
			      KPixCanvasImageCache *cache)
/*  [SUMMARY] Compute a PseudoColour image cache for a world canvas.
    [PURPOSE] This routine will compute an image cache for a 2-dimensional
    PseudoColour slice of a Karma array for a world canvas. This slice may be
    tiled.
    <canvas> The world canvas object.
    <arr_desc> The array descriptor for the image.
    <slice> The start of the slice (image) data.
    <hdim> The dimension index of the horizontal dimension.
    <vdim> The dimension index of the vertical dimension.
    <elem_index> The element index of the data packets.
    <cache> The routine will produce cache data which will vastly increase
    the speed of subsequent operations on this data. The routine will write a
    pointer to this data here. Prior to process exit, a call MUST be made to
    [<kwin_free_cache_data>], otherwise shared memory segments could remain
    after the process exits.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag ok;
    unsigned int abs_start_index;
    unsigned int abs_end_index;
    unsigned int ord_start_index;
    unsigned int ord_end_index;
    unsigned int num_abs_coords;
    unsigned int num_ord_coords;
    unsigned int num_pixels;
    uaddr *hoffsets, *voffsets;
    packet_desc *pack_desc;
    dim_desc *hdim_desc, *vdim_desc;
    unsigned long *pixel_values;
    static char function_name[] = "canvas_compute_pc_image";

    VERIFY_CANVAS (canvas);
    if (canvas->cmap == NULL)
    {
	fprintf (stderr, "Canvas has no colourmap associated with it\n");
	a_prog_bug (function_name);
    }
    num_pixels = kcmap_get_pixels (canvas->cmap, &pixel_values);
    pack_desc = arr_desc->packet;
    if (elem_index >= pack_desc->num_elements)
    {
	fprintf (stderr,"elem_index: %u greater than number of elements: %u\n",
		 elem_index, pack_desc->num_elements);
	a_prog_bug (function_name);
    }
    if (canvas->auto_blank)
    {
	kwin_get_attributes (canvas->pixcanvas,
			     KWIN_ATT_BACKGROUND_PIXEL,
			     &canvas->win_scale.blank_pixel,
			     KWIN_ATT_END);
    }
    hoffsets = arr_desc->offsets[hdim];
    voffsets = arr_desc->offsets[vdim];
    hdim_desc = arr_desc->dimensions[hdim];
    vdim_desc = arr_desc->dimensions[vdim];
    /*  Determine start and stop co-ordinates along each dimension  */
    abs_start_index = ds_get_coord_num (hdim_desc, canvas->win_scale.left_x,
					SEARCH_BIAS_CLOSEST);
    abs_end_index = ds_get_coord_num (hdim_desc, canvas->win_scale.right_x,
				      SEARCH_BIAS_CLOSEST);
#ifdef DISABLED
    if (abs_start_index >= abs_end_index)
    {
	fprintf (stderr,"Left co-ordinate index: %u not less than right: %u\n",
		 abs_start_index, abs_end_index);
	a_prog_bug (function_name);
    }
#endif
    num_abs_coords = abs_end_index - abs_start_index + 1;
    ord_start_index = ds_get_coord_num (vdim_desc, canvas->win_scale.bottom_y,
					SEARCH_BIAS_CLOSEST);
    ord_end_index = ds_get_coord_num (vdim_desc, canvas->win_scale.top_y,
				      SEARCH_BIAS_CLOSEST);
#ifdef DISABLED
    if (ord_start_index >= ord_end_index)
    {
	fprintf (stderr,"Bottom co-ordinate index: %u not less than top: %u\n",
		 ord_start_index, ord_end_index);
	a_prog_bug (function_name);
    }
#endif
    num_ord_coords = ord_end_index - ord_start_index + 1;
    ok = kwin_compute_pc_subimage (canvas->pixcanvas,
				   0, 0,
				   canvas->win_scale.x_pixels,
				   canvas->win_scale.y_pixels,
				   canvas->win_scale.x_pixels,
				   canvas->win_scale.y_pixels,
				   slice +ds_get_element_offset (pack_desc,
								 elem_index),
				   hoffsets + abs_start_index,
				   voffsets + ord_start_index,
				   num_abs_coords, num_ord_coords,
				   pack_desc->element_types[elem_index],
				   canvas->win_scale.conv_type,
				   num_pixels, pixel_values,
				   canvas->win_scale.blank_pixel,
				   canvas->win_scale.min_sat_pixel,
				   canvas->win_scale.max_sat_pixel,
				   canvas->win_scale.z_min,
				   canvas->win_scale.z_max,
				   canvas->win_scale.iscale_func,
				   canvas->win_scale.iscale_info,
				   cache);
    if (!ok) fprintf (stderr, "%s: error computing PseudoColour image\n",
		      function_name);
    return (ok);
}   /*  End Function canvas_compute_pc_image  */

/*PUBLIC_FUNCTION*/
flag canvas_draw_rgb_image (KWorldCanvas canvas, array_desc *arr_desc,
			    CONST char *slice,
			    unsigned int hdim, unsigned int vdim,
			    unsigned int red_index, unsigned int green_index,
			    unsigned int blue_index,
			    KPixCanvasImageCache *cache_ptr)
/*  [SUMMARY] Draw a TrueColour image onto a world canvas.
    [PURPOSE] This routine will draw a 2-dimensional TrueColour slice of a
    Karma array onto a world canvas. This slice may be tiled.
    <canvas> The world canvas.
    <arr_desc> The array descriptor for the image.
    <slice> The start of the slice (image) data.
    <hdim> The dimension index of the horizontal dimension.
    <vdim> The dimension index of the vertical dimension.
    <red_index> The element index of the red components.
    <green_index> The element index of the green components.
    <blue_index> The element index of the blue components.
    <cache_ptr> The routine may produce cache data which will vastly increase
    the speed of subsequent operations on this data. The routine will write a
    pointer to this data here. Prior to process exit, a call MUST be made to
    [<kwin_free_cache_data>], otherwise shared memory segments could remain
    after the process exits.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag ok;
    unsigned int abs_start_index;
    unsigned int abs_end_index;
    unsigned int ord_start_index;
    unsigned int ord_end_index;
    unsigned int num_abs_coords;
    unsigned int num_ord_coords;
    uaddr *hoffsets, *voffsets;
    CONST unsigned char *red_slice, *green_slice, *blue_slice;
    packet_desc *pack_desc;
    dim_desc *hdim_desc, *vdim_desc;
    static char function_name[] = "canvas_draw_rgb_image";

    VERIFY_CANVAS (canvas);
    pack_desc = arr_desc->packet;
    if (red_index >= pack_desc->num_elements)
    {
	fprintf (stderr, "red_index: %u greater than number of elements: %u\n",
		 red_index, pack_desc->num_elements);
	a_prog_bug (function_name);
    }
    if (pack_desc->element_types[red_index] != K_UBYTE)
    {
	fprintf (stderr, "Red component type: %u is not K_UBYTE\n",
		 pack_desc->element_types[red_index]);
	return (FALSE);
    }
    if (green_index >= pack_desc->num_elements)
    {
	fprintf(stderr,"green_index: %u greater than number of elements: %u\n",
		green_index, pack_desc->num_elements);
	a_prog_bug (function_name);
    }
    if (pack_desc->element_types[green_index] != K_UBYTE)
    {
	fprintf (stderr, "Green component type: %u is not K_UBYTE\n",
		 pack_desc->element_types[green_index]);
	return (FALSE);
    }
    if (blue_index >= pack_desc->num_elements)
    {
	fprintf (stderr,"blue_index: %u greater than number of elements: %u\n",
		 blue_index, pack_desc->num_elements);
	a_prog_bug (function_name);
    }
    if (pack_desc->element_types[blue_index] != K_UBYTE)
    {
	fprintf (stderr, "Blue component type: %u is not K_UBYTE\n",
		 pack_desc->element_types[blue_index]);
	return (FALSE);
    }
    if (arr_desc->offsets == NULL)
    {
	if ( !ds_compute_array_offsets (arr_desc) )
	{
	    a_func_abort (function_name, "error computing array offsets");
	    return (FALSE);
	}
    }
    hoffsets = arr_desc->offsets[hdim];
    voffsets = arr_desc->offsets[vdim];
    hdim_desc = arr_desc->dimensions[hdim];
    vdim_desc = arr_desc->dimensions[vdim];
    /*  Determine start and stop co-ordinates along each dimension  */
    abs_start_index = ds_get_coord_num (hdim_desc, canvas->win_scale.left_x,
					SEARCH_BIAS_CLOSEST);
    abs_end_index = ds_get_coord_num (hdim_desc, canvas->win_scale.right_x,
				      SEARCH_BIAS_CLOSEST);
    if (abs_start_index >= abs_end_index)
    {
	fprintf (stderr,"Left co-ordinate index: %u not less than right: %u\n",
		 abs_start_index, abs_end_index);
	a_prog_bug (function_name);
    }
    num_abs_coords = abs_end_index - abs_start_index + 1;
    ord_start_index = ds_get_coord_num (vdim_desc, canvas->win_scale.bottom_y,
					SEARCH_BIAS_CLOSEST);
    ord_end_index = ds_get_coord_num (vdim_desc, canvas->win_scale.top_y,
				      SEARCH_BIAS_CLOSEST);
    if (ord_start_index >= ord_end_index)
    {
	fprintf (stderr,"Bottom co-ordinate index: %u not less than top: %u\n",
		 ord_start_index, ord_end_index);
	a_prog_bug (function_name);
    }
    num_ord_coords = ord_end_index - ord_start_index + 1;
    red_slice = (CONST unsigned char *) slice;
    red_slice += ds_get_element_offset (pack_desc, red_index);
    green_slice = (CONST unsigned char *) slice;
    green_slice += ds_get_element_offset (pack_desc, green_index);
    blue_slice = (CONST unsigned char *) slice;
    blue_slice += ds_get_element_offset (pack_desc, blue_index);
    ok = kwin_draw_rgb_image (canvas->pixcanvas,
			      canvas->win_scale.x_offset,
			      canvas->win_scale.y_offset,
			      canvas->win_scale.x_pixels,
			      canvas->win_scale.y_pixels,
			      red_slice, green_slice, blue_slice,
			      hoffsets + abs_start_index,
			      voffsets + ord_start_index,
			      num_abs_coords, num_ord_coords, cache_ptr);
    if (!ok) fprintf (stderr, "%s: error drawing RGB image\n", function_name);
    return (ok);
}   /*  End Function canvas_draw_rgb_image  */

/*UNPUBLISHED_FUNCTION*/
flag canvas_compute_rgb_image (KWorldCanvas canvas, array_desc *arr_desc,
			       CONST char *slice,
			       unsigned int hdim, unsigned int vdim,
			       unsigned int red_index,
			       unsigned int green_index,
			       unsigned int blue_index,
			       KPixCanvasImageCache *cache)
/*  [SUMMARY] Compute a TrueColour image cache for a world canvas.
    [PURPOSE] This routine will compute an image cache for 2-dimensional
    TrueColour slice of a Karma array for a world canvas. This slice may be
    tiled.
    <canvas> The world canvas.
    <arr_desc> The array descriptor for the image.
    <slice> The start of the slice (image) data.
    <hdim> The dimension index of the horizontal dimension.
    <vdim> The dimension index of the vertical dimension.
    <red_index> The element index of the red components.
    <green_index> The element index of the green components.
    <blue_index> The element index of the blue components.
    <cache> The routine will produce cache data which will vastly increase
    the speed of subsequent operations on this data. The routine will write a
    pointer to this data here. Prior to process exit, a call MUST be made to
    [<kwin_free_cache_data>], otherwise shared memory segments could remain
    after the process exits.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    flag ok;
    unsigned int abs_start_index;
    unsigned int abs_end_index;
    unsigned int ord_start_index;
    unsigned int ord_end_index;
    unsigned int num_abs_coords;
    unsigned int num_ord_coords;
    uaddr *hoffsets, *voffsets;
    CONST unsigned char *red_slice, *green_slice, *blue_slice;
    packet_desc *pack_desc;
    dim_desc *hdim_desc, *vdim_desc;
    static char function_name[] = "canvas_compute_rgb_image";

    VERIFY_CANVAS (canvas);
    pack_desc = arr_desc->packet;
    if (red_index >= pack_desc->num_elements)
    {
	fprintf (stderr, "red_index: %u greater than number of elements: %u\n",
		 red_index, pack_desc->num_elements);
	a_prog_bug (function_name);
    }
    if (pack_desc->element_types[red_index] != K_UBYTE)
    {
	fprintf (stderr, "Red component type: %u is not K_UBYTE\n",
		 pack_desc->element_types[red_index]);
	return (FALSE);
    }
    if (green_index >= pack_desc->num_elements)
    {
	fprintf(stderr,"green_index: %u greater than number of elements: %u\n",
		green_index, pack_desc->num_elements);
	a_prog_bug (function_name);
    }
    if (pack_desc->element_types[green_index] != K_UBYTE)
    {
	fprintf (stderr, "Green component type: %u is not K_UBYTE\n",
		 pack_desc->element_types[green_index]);
	return (FALSE);
    }
    if (blue_index >= pack_desc->num_elements)
    {
	fprintf (stderr,"blue_index: %u greater than number of elements: %u\n",
		 blue_index, pack_desc->num_elements);
	a_prog_bug (function_name);
    }
    if (pack_desc->element_types[blue_index] != K_UBYTE)
    {
	fprintf (stderr, "Blue component type: %u is not K_UBYTE\n",
		 pack_desc->element_types[blue_index]);
	return (FALSE);
    }
    if (arr_desc->offsets == NULL)
    {
	if ( !ds_compute_array_offsets (arr_desc) )
	{
	    a_func_abort (function_name, "error computing array offsets");
	    return (FALSE);
	}
    }
    hoffsets = arr_desc->offsets[hdim];
    voffsets = arr_desc->offsets[vdim];
    hdim_desc = arr_desc->dimensions[hdim];
    vdim_desc = arr_desc->dimensions[vdim];
    /*  Determine start and stop co-ordinates along each dimension  */
    abs_start_index = ds_get_coord_num (hdim_desc, canvas->win_scale.left_x,
					SEARCH_BIAS_CLOSEST);
    abs_end_index = ds_get_coord_num (hdim_desc, canvas->win_scale.right_x,
				      SEARCH_BIAS_CLOSEST);
    if (abs_start_index >= abs_end_index)
    {
	fprintf (stderr,"Left co-ordinate index: %u not less than right: %u\n",
		 abs_start_index, abs_end_index);
	a_prog_bug (function_name);
    }
    num_abs_coords = abs_end_index - abs_start_index + 1;
    ord_start_index = ds_get_coord_num (vdim_desc, canvas->win_scale.bottom_y,
					SEARCH_BIAS_CLOSEST);
    ord_end_index = ds_get_coord_num (vdim_desc, canvas->win_scale.top_y,
				      SEARCH_BIAS_CLOSEST);
    if (ord_start_index >= ord_end_index)
    {
	fprintf (stderr,"Bottom co-ordinate index: %u not less than top: %u\n",
		 ord_start_index, ord_end_index);
	a_prog_bug (function_name);
    }
    num_ord_coords = ord_end_index - ord_start_index + 1;
    red_slice = (CONST unsigned char *) slice;
    red_slice += ds_get_element_offset (pack_desc, red_index);
    green_slice = (CONST unsigned char *) slice;
    green_slice += ds_get_element_offset (pack_desc, green_index);
    blue_slice = (CONST unsigned char *) slice;
    blue_slice += ds_get_element_offset (pack_desc, blue_index);
    ok = kwin_compute_rgb_subimage (canvas->pixcanvas,
				    0, 0,
				    canvas->win_scale.x_pixels,
				    canvas->win_scale.y_pixels,
				    canvas->win_scale.x_pixels,
				    canvas->win_scale.y_pixels,
				    red_slice, green_slice, blue_slice,
				    hoffsets + abs_start_index,
				    voffsets + ord_start_index,
				    num_abs_coords, num_ord_coords, cache);
    if (!ok) fprintf (stderr, "%s: error computing RGB image\n",function_name);
    return (ok);
}   /*  End Function canvas_compute_rgb_image  */

/*PUBLIC_FUNCTION*/
void canvas_draw_point (KWorldCanvas canvas, double x, double y,
			double value[2])
/*  [SUMMARY] Draw a single point onto a world canvas.
    <canvas> The world canvas object.
    <x> The horizontal world co-ordinate of the point.
    <y> The vertical world co-ordinate of the point.
    <value> The data value to use. This is of type K_DCOMPLEX.
    [RETURNS] Nothing.
*/
{
    unsigned long pixel_value;
    static char function_name[] = "canvas_draw_point";

    VERIFY_CANVAS (canvas);
    if (canvas->cmap == NULL)
    {
	fprintf (stderr, "Canvas has no colourmap associated with it\n");
	a_prog_bug (function_name);
    }
    pixel_value = _canvas_get_pixel_from_value (canvas, value,
						NULL, NULL, NULL);
    canvas_draw_point_p (canvas, x, y, pixel_value);
}   /*  End Function canvas_draw_point  */

/*PUBLIC_FUNCTION*/
void canvas_draw_point_p (KWorldCanvas canvas, double x, double y,
			  unsigned long pixel_value)
/*  [SUMMARY] Draw a single point onto a world canvas.
    <canvas> The world canvas object.
    <x> The horizontal world co-ordinate of the point.
    <y> The vertical world co-ordinate of the point.
    <pixel_value> The pixel value to use.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "canvas_draw_point_p";

    VERIFY_CANVAS (canvas);
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
				       &x, &y, &x, &y);
    if (x < canvas->win_scale.x_offset) return;
    if (x >= canvas->win_scale.x_offset + canvas->win_scale.x_pixels) return;
    if (y < canvas->win_scale.y_offset) return;
    if (y >= canvas->win_scale.y_offset + canvas->win_scale.y_pixels) return;
    kwin_draw_point (canvas->pixcanvas, x, y, pixel_value);
}   /*  End Function canvas_draw_point_p  */

/*PUBLIC_FUNCTION*/
void canvas_draw_line (KWorldCanvas canvas,
		       double x0, double y0, double x1, double y1,
		       double value[2])
/*  [SUMMARY] Draw a single line onto a world canvas.
    <canvas> The world canvas object.
    <x0> The horizontal world co-ordinate of the first point.
    <y0> The vertical world co-ordinate of the first point.
    <x1> The horizontal world co-ordinate of the second point.
    <y1> The vertical world co-ordinate of the second point.
    <value> The data value to use. This is of type K_DCOMPLEX.
    [RETURNS] Nothing.
*/
{
    unsigned long pixel_value;
    static char function_name[] = "canvas_draw_line";

    VERIFY_CANVAS (canvas);
    if (canvas->cmap == NULL)
    {
	fprintf (stderr, "Canvas has no colourmap associated with it\n");
	a_prog_bug (function_name);
    }
    pixel_value = _canvas_get_pixel_from_value (canvas, value,
						NULL, NULL, NULL);
    canvas_draw_line_p (canvas, x0, y0, x1, y1, pixel_value);
}   /*  End Function canvas_draw_line  */

/*PUBLIC_FUNCTION*/
void canvas_draw_line_p (KWorldCanvas canvas,
			 double x0, double y0, double x1, double y1,
			 unsigned long pixel_value)
/*  [SUMMARY] Draw a single line onto a world canvas.
    <canvas> The world canvas object.
    <x0> The horizontal world co-ordinate of the first point.
    <y0> The vertical world co-ordinate of the first point.
    <x1> The horizontal world co-ordinate of the second point.
    <y1> The vertical world co-ordinate of the second point.
    <pixel_value> The pixel value to use.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "canvas_draw_line_p";

    VERIFY_CANVAS (canvas);
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
				       &x0, &y0, &x0, &y0);
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
				       &x1, &y1, &x1, &y1);
    kwin_draw_line (canvas->pixcanvas, x0, y0, x1, y1, pixel_value);
}   /*  End Function canvas_draw_line_p  */

/*EXPERIMENTAL_FUNCTION*/
flag canvas_draw_arc (KWorldCanvas canvas,
		      double centre_x, double centre_y,
		      double radius_x, double radius_y,
		      double angle1, double angle2, double rotation,
		      double value[2], flag fill)
/*  [SUMMARY] Draw an arc onto a world canvas.
    <canvas> The world canvas object.
    <centre_x> The horizontal world co-ordinate of the centre of the ellipse.
    <centre_y> The vertical world co-ordinate of the centre of the ellipse.
    <radius_x> The horizontal radius in world co-ordinates.
    <radius_y> The vertical radius in world co-ordinates.
    <angle1> See <<XDrawArc>>.
    <angle2> See <<XDrawArc>>.
    <rotation> The rotation of the arc, in degrees.
    <value> The complex value to draw the ellipse with. This must be of type
    K_DCOMPLEX.
    <fill> If TRUE, the arc is filled, else only the outside is drawn.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned long pixel_value;
    static char function_name[] = "canvas_draw_arc";

    VERIFY_CANVAS (canvas);
    if (canvas->cmap == NULL)
    {
	fprintf (stderr, "Canvas has no colourmap associated with it\n");
	a_prog_bug (function_name);
    }
    pixel_value = _canvas_get_pixel_from_value (canvas, value,
						NULL, NULL, NULL);
    return canvas_draw_arc_p (canvas,
			      centre_x, centre_y, radius_x, radius_y,
			      angle1, angle2, rotation, pixel_value, fill);
}   /*  End Function canvas_draw_arc  */

/*EXPERIMENTAL_FUNCTION*/
flag canvas_draw_arc_p (KWorldCanvas canvas,
			double centre_x, double centre_y,
			double radius_x, double radius_y,
			int angle1, int angle2, int rotation,
			unsigned long pixel_value, flag fill)
/*  [SUMMARY] Draw an ellipse onto a world canvas.
    <canvas> The world canvas object.
    <centre_x> The horizontal world co-ordinate of the centre of the ellipse.
    <centre_y> The vertical world co-ordinate of the centre of the ellipse.
    <radius_x> The horizontal radius in world co-ordinates.
    <radius_y> The vertical radius in world co-ordinates.
    <angle1> See <<XDrawArc>>.
    <angle2> See <<XDrawArc>>.
    <rotation> The rotation of the arc, in units of 1/64 degree.
    <pixel_value> The pixel value to draw the ellipse with.
    <fill> If TRUE, the arc is filled, else only the outside is drawn.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    static char function_name[] = "canvas_draw_arc_p";

    VERIFY_CANVAS (canvas);
    /*  Offset radius value  */
    radius_x += centre_x;
    radius_y += centre_y;
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
				       &centre_x, &centre_y,
				       &centre_x, &centre_y);
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
				       &radius_x, &radius_y,
				       &radius_x, &radius_y);
    /*  Undo radius offset  */
    radius_x = fabs (radius_x - centre_x);
    radius_y = fabs (radius_y - centre_y);
    return kwin_draw_arc (canvas->pixcanvas, centre_x, centre_y,
			  radius_x, radius_y, angle1, angle1, rotation,
			  pixel_value, fill);
}   /*  End Function canvas_draw_arc_p  */

/*PUBLIC_FUNCTION*/
void canvas_draw_ellipse (KWorldCanvas canvas,
			  double centre_x, double centre_y,
			  double radius_x, double radius_y, double value[2])
/*  [SUMMARY] Draw an ellipse onto a world canvas.
    <canvas> The world canvas object.
    <centre_x> The horizontal world co-ordinate of the centre of the ellipse.
    <centre_y> The vertical world co-ordinate of the centre of the ellipse.
    <radius_x> The horizontal radius in world co-ordinates.
    <radius_y> The vertical radius in world co-ordinates.
    <value> The complex value to draw the ellipse with. This must be of type
    K_DCOMPLEX.
    [RETURNS] Nothing.
*/
{
    unsigned long pixel_value;
    static char function_name[] = "canvas_draw_ellipse";

    VERIFY_CANVAS (canvas);
    if (canvas->cmap == NULL)
    {
	fprintf (stderr, "Canvas has no colourmap associated with it\n");
	a_prog_bug (function_name);
    }
    pixel_value = _canvas_get_pixel_from_value (canvas, value,
						NULL, NULL, NULL);
    canvas_draw_ellipse_p (canvas,
			   centre_x, centre_y, radius_x, radius_y,
			   pixel_value);
}   /*  End Function canvas_draw_ellipse  */

/*PUBLIC_FUNCTION*/
void canvas_draw_ellipse_p (KWorldCanvas canvas,
			    double centre_x, double centre_y,
			    double radius_x, double radius_y,
			    unsigned long pixel_value)
/*  [SUMMARY] Draw an ellipse onto a world canvas.
    <canvas> The world canvas object.
    <centre_x> The horizontal world co-ordinate of the centre of the ellipse.
    <centre_y> The vertical world co-ordinate of the centre of the ellipse.
    <radius_x> The horizontal radius in world co-ordinates.
    <radius_y> The vertical radius in world co-ordinates.
    <pixel_value> The pixel value to draw the ellipse with.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "canvas_draw_ellipse_p";

    VERIFY_CANVAS (canvas);
    /*  Offset radius value  */
    radius_x += centre_x;
    radius_y += centre_y;
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
				       &centre_x, &centre_y,
				       &centre_x, &centre_y);
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
				       &radius_x, &radius_y,
				       &radius_x, &radius_y);
    /*  Undo radius offset  */
    radius_x = fabs (radius_x - centre_x);
    radius_y = fabs (radius_y - centre_y);
    kwin_draw_ellipse (canvas->pixcanvas, centre_x, centre_y,
		       radius_x, radius_y, pixel_value);
}   /*  End Function canvas_draw_ellipse_p  */

/*PUBLIC_FUNCTION*/
void canvas_fill_ellipse (KWorldCanvas canvas,
			  double centre_x, double centre_y,
			  double radius_x, double radius_y, double value[2])
/*  [SUMMARY] Draw a filled ellipse onto a world canvas.
    <canvas> The world canvas object.
    <centre_x> The horizontal world co-ordinate of the centre of the ellipse.
    <centre_y> The vertical world co-ordinate of the centre of the ellipse.
    <radius_x> The horizontal radius in world co-ordinates.
    <radius_y> The vertical radius in world co-ordinates.
    <value> The complex value to fill the ellipse with. This must be of type
    K_DCOMPLEX.
    [RETURNS] Nothing.
*/
{
    unsigned long pixel_value;
    static char function_name[] = "canvas_fill_ellipse";

    VERIFY_CANVAS (canvas);
    if (canvas->cmap == NULL)
    {
	fprintf (stderr, "Canvas has no colourmap associated with it\n");
	a_prog_bug (function_name);
    }
    pixel_value = _canvas_get_pixel_from_value (canvas, value,
						NULL, NULL, NULL);
    canvas_fill_ellipse_p (canvas,
			   centre_x, centre_y, radius_x, radius_y,
			   pixel_value);
}   /*  End Function canvas_fill_ellipse  */

/*PUBLIC_FUNCTION*/
void canvas_fill_ellipse_p (KWorldCanvas canvas,
			    double centre_x, double centre_y,
			    double radius_x, double radius_y,
			    unsigned long pixel_value)
/*  [SUMMARY] Draw a filled ellipse onto a world canvas.
    <canvas> The world canvas object.
    <centre_x> The horizontal world co-ordinate of the centre of the ellipse.
    <centre_y> The vertical world co-ordinate of the centre of the ellipse.
    <radius_x> The horizontal radius in world co-ordinates.
    <radius_y> The vertical radius in world co-ordinates.
    <pixel_value> The pixel value to fill the ellipse with.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "canvas_fill_ellipse_p";

    VERIFY_CANVAS (canvas);
    if ( (fabs (radius_x) > 1e9) || (fabs (radius_y) > 1e9) )
	fprintf (stderr, "%s: WARNING\nC: %e %e  R: %e %e\n",
		 function_name, centre_x, centre_y, radius_x, radius_y);
    /*  Offset radius value  */
    radius_x += centre_x;
    radius_y += centre_y;
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
				       &centre_x, &centre_y,
				       &centre_x, &centre_y);
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
				       &radius_x, &radius_y,
				       &radius_x, &radius_y);
    /*  Undo radius offset  */
    radius_x = fabs (radius_x - centre_x);
    radius_y = fabs (radius_y - centre_y);
    kwin_fill_ellipse (canvas->pixcanvas, centre_x, centre_y,
		       radius_x, radius_y, pixel_value);
}   /*  End Function canvas_fill_ellipse_p  */

/*EXPERIMENTAL_FUNCTION*/
void canvas_convert_ellipse_coords (KWorldCanvas canvas, flag to_world,
				    flag linear,
				    double *centre_x, double *centre_y,
				    double *radius_x, double *radius_y,
				    double *rotation)
/*  [SUMMARY] Convert co-ordinates for an ellipse.
    <canvas> The world canvas object.
    <to_world> If TRUE a pixel->world conversion is performed, else a
    world->pixel conversion is performed.
    <linear> If TRUE the co-ordinate conversions are pixel<->linear rather than
    pixel<->world.
    <centre_x> The horizontal co-ordinate of the centre of the ellipse.
    <centre_y> The vertical co-ordinate of the centre of the ellipse.
    <radius_x> The horizontal radius.
    <radius_y> The vertical radius.
    <rotation> The rotation of the arc, in degrees.
    [RETURNS] Nothing.
*/
{
    flag apply_ra_scaling = FALSE;
    double cos_rot, sin_rot;
    double x1, y1, xscale, yscale;
    double cx, cy, rx, ry, x, y, theta, angle;
    static char function_name[] = "canvas_convert_ellipse_coords";

    VERIFY_CANVAS (canvas);
    cos_rot = cos (*rotation * PION180);
    sin_rot = sin (*rotation * PION180);
    if ( (canvas->ap != NULL) &&
	 wcs_astro_test_if_lon (canvas->ap, canvas->specify_xlabel) )
	apply_ra_scaling = linear ? FALSE : TRUE;
    if (to_world)
    {
	if (linear)
	    canvas_convert_to_canvas_coords (canvas, FALSE, 1,
					     centre_x, centre_y, &cx, &cy,
					     NULL, NULL);
	else
	    canvas_convert_to_canvas_coords (canvas, FALSE, 1,
					     centre_x, centre_y, NULL, NULL,
					     &cx, &cy);
	xscale = apply_ra_scaling ? cos (cy * PION180) : 1.0;
    }
    else
    {
	xscale = apply_ra_scaling ? ( 1.0 / cos (*centre_y * PION180) ) : 1.0;
	canvas_convert_from_canvas_coords (canvas, FALSE, linear, 1,
					   centre_x, centre_y, &cx, &cy);
    }
    yscale = 1.0;
    angle = get_ellipse_max_angle (canvas, to_world, linear, xscale, yscale,
				   cos_rot, sin_rot, *centre_x, *centre_y,
				   *radius_x, *radius_y, cx, cy);
    /*  Now compute major and minor axes and rotation in pixel canvas  */
    x1 = *radius_x * cos (angle);
    y1 = *radius_y * sin (angle);
    x = x1 * cos_rot - y1 * sin_rot;
    y = y1 * cos_rot + x1 * sin_rot;
    x = *centre_x + x * xscale;
    y = *centre_y + y * yscale;
    if (to_world)
    {
	if (linear)
	    canvas_convert_to_canvas_coords (canvas, FALSE, 1, &x, &y, &x, &y,
					     NULL, NULL);
	else
	    canvas_convert_to_canvas_coords (canvas, FALSE, 1, &x, &y,
					     NULL, NULL, &x, &y);
    }
    else canvas_convert_from_canvas_coords (canvas, FALSE, linear, 1, &x, &y,
					    &x, &y);
    rx = sqrt ( (cx - x) * (cx - x) + (cy - y) * (cy - y) );
    x = x - cx;
    y = y - cy;
    theta = atan2 (y, x) / PION180;
    x1 = *radius_x * cos (angle + PI_ON_2);
    y1 = *radius_y * sin (angle + PI_ON_2);
    x = x1 * cos_rot - y1 * sin_rot;
    y = y1 * cos_rot + x1 * sin_rot;
    x = *centre_x + x * xscale;
    y = *centre_y + y * yscale;
    if (to_world)
    {
	if (linear)
	    canvas_convert_to_canvas_coords (canvas, FALSE, 1, &x, &y, &x, &y,
					     NULL, NULL);
	else
	    canvas_convert_to_canvas_coords (canvas, FALSE, 1, &x, &y,
					     NULL, NULL, &x, &y);
    }
    else canvas_convert_from_canvas_coords (canvas, FALSE, linear, 1, &x, &y,
					    &x, &y);
    ry = sqrt ( (cx - x) * (cx - x) + (cy - y) * (cy - y) );
    *centre_x = cx;
    *centre_y = cy;
    *radius_x = fabs (rx);
    *radius_y = fabs (ry);
    *rotation = theta;
}   /*  End Function canvas_convert_ellipse_coords  */

/*PUBLIC_FUNCTION*/
flag canvas_fill_polygon (KWorldCanvas canvas, edit_coord *coords,
			  unsigned int num_vertices, double value[2],
			  flag convex)
/*  [SUMMARY] Draw a filled polygon onto a world canvas.
    <canvas> The world canvas object.
    <coords> The array of world co-ordinates of vertices of the polygon.
    <num_vertices> The number of vertices in the polygon.
    <value> The complex value to fill the polygon with. This must be of type
    K_DCOMPLEX.
    <convex> If TRUE, then the points must form a convex polygon.
    [RETURNS] TRUE on success, else FALSE.
*/
{
    unsigned int coord_count;
    unsigned long pixel_value;
    double px, py;
    static unsigned int num_points_allocated = 0;
    static int *point_x = NULL;
    static int *point_y = NULL;
    static char function_name[] = "canvas_fill_polygon";

    VERIFY_CANVAS (canvas);
    if (canvas->cmap == NULL)
    {
	fprintf (stderr, "Canvas has no colourmap associated with it\n");
	a_prog_bug (function_name);
    }
    if (num_vertices > num_points_allocated)
    {
	if (point_x != NULL)
	{
	    m_free (point_x);
	    point_x = NULL;
	}
	if (point_y != NULL)
	{
	    m_free (point_y);
	    point_y = NULL;
	}
	num_points_allocated = 0;
	if ( ( point_x = m_alloc (sizeof *point_x * num_vertices) ) == NULL )
	{
	    m_error_notify (function_name, "array of x co-ordinates");
	    return (FALSE);
	}
	if ( ( point_y = m_alloc (sizeof *point_y * num_vertices) ) == NULL )
	{
	    m_error_notify (function_name, "array of y co-ordinates");
	    m_free (point_x);
	    point_x = NULL;
	    return (FALSE);
	}
	num_points_allocated = num_vertices;
    }
    /*  Convert world co-ordinates to pixel co-ordinates  */
    for (coord_count = 0; coord_count < num_vertices; ++coord_count)
    {
	canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
					   &coords[coord_count].abscissa,
					   &coords[coord_count].ordinate,
					   &px, &py);
	point_x[coord_count] = px;
	point_y[coord_count] = py;
    }
    pixel_value = _canvas_get_pixel_from_value (canvas, value,
						NULL, NULL, NULL);
    return kwin_fill_polygon (canvas->pixcanvas,
			      point_x, point_y, num_vertices,
			      pixel_value, convex);
}   /*  End Function canvas_fill_polygon  */

/*PUBLIC_FUNCTION*/
void canvas_draw_rectangle (KWorldCanvas canvas, double x, double y,
			    double width, double height, double value[2])
/*  [SUMMARY] Draw a single rectangle onto a world canvas.
    <canvas> The world canvas object.
    <x> The horizontal offset of the rectangle.
    <y> The vertical offset of the rectangle.
    <width> The width of the rectangle.
    <height> The height of the rectangle.
    <value> The data value to use. This is of type K_DCOMPLEX.
    [RETURNS] Nothing.
*/
{
    unsigned long pixel_value;
    static char function_name[] = "canvas_draw_rectangle";

    VERIFY_CANVAS (canvas);
    if (canvas->cmap == NULL)
    {
	fprintf (stderr, "Canvas has no colourmap associated with it\n");
	a_prog_bug (function_name);
    }
    pixel_value = _canvas_get_pixel_from_value (canvas, value,
						NULL, NULL, NULL);
    canvas_draw_rectangle_p (canvas, x, y, width, height, pixel_value);
}   /*  End Function canvas_draw_rectangle  */

/*PUBLIC_FUNCTION*/
void canvas_draw_rectangle_p (KWorldCanvas canvas, double x, double y,
			      double width, double height,
			      unsigned long pixel_value)
/*  [SUMMARY] Draw a single rectangle onto a world canvas.
    <canvas> The world canvas object.
    <x> The horizontal offset of the rectangle.
    <y> The vertical offset of the rectangle.
    <width> The width of the rectangle.
    <height> The height of the rectangle.
    <pixel_value> The pixel value to use.
    [RETURNS] Nothing.
*/
{
    double pwidth, pheight, px1, py1;
    double px, py;
    static char function_name[] = "canvas_draw_rectangle_p";

    VERIFY_CANVAS (canvas);
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
				       &x, &y, &px, &py);
    /*  Offset width and height  */
    x += width;
    y += height;
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
				       &x, &y, &px1, &py1);
    /*  Undo offset  */
    pwidth = fabs (px - px1);
    pheight = fabs (py - py1);
    kwin_draw_rectangle (canvas->pixcanvas,
			 px, py1, pwidth, pheight, pixel_value);
}   /*  End Function canvas_draw_rectangle_p  */

/*PUBLIC_FUNCTION*/
void canvas_fill_rectangle (KWorldCanvas canvas, double x, double y,
			    double width, double height, double value[2])
/*  [SUMMARY] Fill a single rectangle onto a world canvas.
    <canvas> The world canvas object.
    <x> The horizontal offset of the rectangle.
    <y> The vertical offset of the rectangle.
    <width> The width of the rectangle.
    <height> The height of the rectangle.
    <value> The data value to use. This is of type K_DCOMPLEX.
    [RETURNS] Nothing.
*/
{
    unsigned long pixel_value;
    static char function_name[] = "canvas_fill_rectangle";

    VERIFY_CANVAS (canvas);
    if (canvas->cmap == NULL)
    {
	fprintf (stderr, "Canvas has no colourmap associated with it\n");
	a_prog_bug (function_name);
    }
    pixel_value = _canvas_get_pixel_from_value (canvas, value,
						NULL, NULL, NULL);
    canvas_fill_rectangle_p (canvas, x, y, width, height, pixel_value);
}   /*  End Function canvas_fill_rectangle  */

/*PUBLIC_FUNCTION*/
void canvas_fill_rectangle_p (KWorldCanvas canvas, double x, double y,
			      double width, double height,
			      unsigned long pixel_value)
/*  [SUMMARY] Fill a single rectangle onto a world canvas.
    <canvas> The world canvas object.
    <x> The horizontal offset of the rectangle.
    <y> The vertical offset of the rectangle.
    <width> The width of the rectangle.
    <height> The height of the rectangle.
    <pixel_value> The pixel value to use.
    [RETURNS] Nothing.
*/
{
    double pwidth, pheight, px1, py1;
    double px, py;
    static char function_name[] = "canvas_fill_rectangle_p";

    VERIFY_CANVAS (canvas);
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
				       &x, &y, &px, &py);
    /*  Offset width and height  */
    x += width;
    y += height;
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1,
				       &x, &y, &px1, &py1);
    /*  Undo offset  */
    pwidth = fabs (px - px1);
    pheight = fabs (py - py1);
    kwin_fill_rectangle (canvas->pixcanvas,
			 px, py1, pwidth, pheight, pixel_value);
}   /*  End Function canvas_fill_rectangle_p  */

/*PUBLIC_FUNCTION*/
void canvas_draw_lines (KWorldCanvas canvas,
			CONST double *x_array, CONST double *y_array,
			int num_points, double value[2])
/*  [SUMMARY] Draw multiple connected lines onto a world canvas.
    <canvas> The world canvas object.
    <x_array> The horizontal world co-ordinates of the points. If this is NULL,
    the routine assumes horizontal co-ordinates equally spaced across the world
    canvas.
    <y_array> The vertical world co-ordinates of the points. If this is NULL,
    the routine assumes vertical co-ordinates equally spaced across the world
    canvas.
    <num_points> The number of points. The number of lines drawn is 1 less than
    this value.
    <value> The data value to use. This is of type K_DCOMPLEX.
    [RETURNS] Nothing.
*/
{
    unsigned long pixel_value;
    static char function_name[] = "canvas_draw_lines";

    VERIFY_CANVAS (canvas);
    if (canvas->cmap == NULL)
    {
	fprintf (stderr, "Canvas has no colourmap associated with it\n");
	a_prog_bug (function_name);
    }
    pixel_value = _canvas_get_pixel_from_value (canvas, value,
						NULL, NULL, NULL);
    canvas_draw_lines_p (canvas, x_array, y_array, num_points, pixel_value);
}   /*  End Function canvas_draw_lines  */

/*PUBLIC_FUNCTION*/
void canvas_draw_lines_p (KWorldCanvas canvas,
			  CONST double *x_array, CONST double *y_array,
			  int num_points, unsigned long pixel_value)
/*  [SUMMARY] Draw multiple connected lines onto a world canvas.
    <canvas> The world canvas object.
    <x_array> The horizontal world co-ordinates of the points. If this is NULL,
    the routine assumes horizontal co-ordinates equally spaced across the world
    canvas.
    <y_array> The vertical world co-ordinates of the points. If this is NULL,
    the routine assumes vertical co-ordinates equally spaced across the world
    canvas.
    <num_points> The number of points. The number of lines drawn is 1 less than
    this value.
    <pixel_value> The pixel value to use.
    [RETURNS] Nothing.
*/
{
    flag draw_block;
    int count, pcount;
    double xscale, yscale, wx, wy;
    double wlx, wby, wrx, wty;
    double px[COORD_BUF_SIZE], py[COORD_BUF_SIZE];
    static char function_name[] = "canvas_draw_lines_p";

    VERIFY_CANVAS (canvas);
    /*  Compute extrema in non-linear world co-ordinates  */
    wlx = canvas->win_scale.left_x;
    wby = canvas->win_scale.bottom_y;
    canvas_coords_transform (canvas, 1, &wlx, FALSE, &wby, FALSE);
    wrx = canvas->win_scale.right_x;
    wty = canvas->win_scale.top_y;
    canvas_coords_transform (canvas, 1, &wrx, FALSE, &wty, FALSE);
    xscale = (wrx - wlx) / (double) (num_points - 1);
    yscale = (wty - wby) / (double) (num_points - 1);
    /*  Divide request into managable blocks  */
    for (count = 0, pcount = 0, draw_block = FALSE; count < num_points;
	 ++count)
    {
	if (x_array == NULL) wx = wlx + xscale * (double) count;
	else if ( !( (wx = x_array[count]) < TOOBIG ) ) draw_block = TRUE;
	if (y_array == NULL) wy = wby + yscale * (double) count;
	else if ( !( (wy = y_array[count]) < TOOBIG ) ) draw_block = TRUE;
	if (!draw_block)
	{
	    px[pcount] = wx;
	    py[pcount] = wy;
	    if (++pcount >= COORD_BUF_SIZE) draw_block = TRUE;
	    if (count >= num_points - 1) draw_block = TRUE;
	}
	if (draw_block)
	{
	    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, pcount,
					       px, py, px, py);
	    kwin_draw_lines_TRANSITION (canvas->pixcanvas, px, py, pcount,
					pixel_value);
	    draw_block = FALSE;
	    pcount = 0;
	}
    }
}   /*  End Function canvas_draw_lines_p  */

/*PUBLIC_FUNCTION*/
void canvas_draw_segments (KWorldCanvas canvas,
			   CONST double *x0, CONST double *y0,
			   CONST double *x1, CONST double *y1,
			   int num_segments, double value[2])
/*  [SUMMARY] Draw multiple disjoint lines onto a world canvas.
    <canvas> The world canvas object.
    <x0> The horizontal start co-ordinates of the segments.
    <y0> The vetical start co-ordinates of the segments.
    <x1> The horizontal end co-ordinates of the segments.
    <y1> The vetical end co-ordinates of the segments.
    <num_segments> The number of segments.
    <value> The data value to use. This is of type K_DCOMPLEX.
    [RETURNS] Nothing.
*/
{
    unsigned long pixel_value;
    static char function_name[] = "canvas_draw_segments";

    VERIFY_CANVAS (canvas);
    if (canvas->cmap == NULL)
    {
	fprintf (stderr, "Canvas has no colourmap associated with it\n");
	a_prog_bug (function_name);
    }
    pixel_value = _canvas_get_pixel_from_value (canvas, value,
						NULL, NULL, NULL);
    canvas_draw_segments_p (canvas, x0, y0, x1, y1, num_segments, pixel_value);
}   /*  End Function canvas_draw_segments  */

/*PUBLIC_FUNCTION*/
void canvas_draw_segments_p (KWorldCanvas canvas,
			     CONST double *x0, CONST double *y0,
			     CONST double *x1, CONST double *y1,
			     int num_segments, unsigned long pixel_value)
/*  [SUMMARY] Draw multiple disjoint lines onto a world canvas.
    <canvas> The world canvas object.
    <x0> The horizontal start co-ordinates of the segments.
    <y0> The vetical start co-ordinates of the segments.
    <x1> The horizontal end co-ordinates of the segments.
    <y1> The vetical end co-ordinates of the segments.
    <num_segments> The number of segments.
    <pixel_value> The pixel value to use.
    [RETURNS] Nothing.
*/
{
    int num_seg;
    double px0[COORD_BUF_SIZE], py0[COORD_BUF_SIZE];
    double px1[COORD_BUF_SIZE], py1[COORD_BUF_SIZE];
    static char function_name[] = "canvas_draw_segments_p";

    VERIFY_CANVAS (canvas);
    /*  Divide request into managable blocks  */
    for (; num_segments > 0; num_segments -= num_seg,
	     x0 += num_seg, y0 += num_seg, x1 += num_seg, y1 += num_seg)
    {
	num_seg = (num_segments>COORD_BUF_SIZE) ? COORD_BUF_SIZE :num_segments;
	/*  Convert to pixel co-ordinates  */
	canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, num_seg,
					   x0, y0, px0, py0);
	canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, num_seg,
					   x1, y1, px1, py1);
	kwin_draw_segments_TRANSITION (canvas->pixcanvas, px0, py0, px1, py1,
				       num_seg, pixel_value);
    }
}   /*  End Function canvas_draw_segments_p  */


/*  Obsolete public functions follow  */

/*OBSOLETE_FUNCTION*/
void canvas_get_size (KWorldCanvas canvas, int *width, int *height,
		      struct win_scale_type *win_scale)
/*  [SUMMARY] Get the size of a world canvas.
    <canvas> The world canvas object.
    <width> The number of horizontal pixels will be written here.
    <height> The number of vertical pixels will be written here.
    <win_scale> The window scaling information will be written here.
    [RETURNS] Nothing.
*/
{
    static flag first_time = TRUE;
    static char function_name[] = "canvas_get_size";

    VERIFY_CANVAS (canvas);
    if (win_scale == NULL)
    {
	fputs ("NULL win_scale structure pointer passed\n", stderr);
	a_prog_bug (function_name);
    }
    if (first_time)
    {
	fprintf (stderr, "WARNING: the <%s> routine will be removed in Karma",
		 function_name);
	fputs (" version 2.0\nUse ", stderr);
	fputs ("the <canvas_get_attributes> routine instead\n", stderr);
	first_time = FALSE;
    }
    kwin_get_size (canvas->pixcanvas, width, height);
    m_copy (win_scale, &canvas->win_scale, sizeof *win_scale);
}   /*  End Function canvas_get_size  */

/*OBSOLETE_FUNCTION*/
flag canvas_convert_to_canvas_coord (KWorldCanvas canvas,
				     double xin, double yin,
				     double *xout, double *yout)
/*  [SUMMARY] Convert pixel co-ordinates to world co-ordinates.
    <canvas> The world canvas object.
    <xin> The lower level horizontal co-ordinate.
    <yin> The lower level vertical co-ordinate.
    <xout> The horizontal world co-ordinate will be written here.
    <yout> The vertical world co-ordinate will be written here.
    [RETURNS] TRUE if the co-ordinate lies within the canvas boundaries,
    else FALSE (although a conversion is still performed).
*/
{
    flag in_bounds = TRUE;
    static flag first_time = TRUE;
    static char function_name[] = "canvas_convert_to_canvas_coord";

    VERIFY_CANVAS (canvas);
    if (first_time)
    {
	fprintf (stderr, "WARNING: the <%s> routine will be removed in Karma",
		 function_name);
	fputs (" version 2.0\nUse the ", stderr);
	fputs ("<canvas_convert_to_canvas_coords> routine instead\n", stderr);
	first_time = FALSE;
    }
    if ( (xin < canvas->win_scale.x_offset) ||
	(xin >= canvas->win_scale.x_offset + canvas->win_scale.x_pixels) ||
	(yin < canvas->win_scale.y_offset) ||
	(yin >= canvas->win_scale.y_offset + canvas->win_scale.y_pixels) )
	in_bounds = FALSE;
    canvas_convert_to_canvas_coords (canvas, FALSE, 1, &xin, &yin,
				     NULL, NULL, xout, yout);
    return (in_bounds);
}   /*  End Function canvas_convert_to_canvas_coord  */

/*OBSOLETE_FUNCTION*/
flag canvas_convert_from_canvas_coord (KWorldCanvas canvas,
				       double xin, double yin,
				       int *xout, int *yout)
/*  [SUMMARY] Convert non-linear world co-ordinates to pixel co-ordinates
    <canvas> The world canvas object.
    <xin> The horizontal world co-ordinate.
    <yin> The vertical world co-ordinate.
    <xout> The lower level horizontal canvas co-ordinate will be written here.
    <yout> The lower level vertical canvas co-ordinate will be written here.
    [RETURNS] TRUE if the co-ordinate lies within the canvas boundaries,
    else FALSE (although a conversion is still performed).
*/
{
    static flag first_time = TRUE;
    static char function_name[] = "canvas_convert_from_canvas_coord";

    VERIFY_CANVAS (canvas);
    if (first_time)
    {
	fprintf (stderr, "WARNING: the <%s> routine will be removed in Karma",
		 function_name);
	fputs (" version 2.0\nUse the ", stderr);
	fputs ("<canvas_convert_from_canvas_coords> routine instead\n",stderr);
	first_time = FALSE;
    }
    canvas_convert_from_canvas_coords (canvas, FALSE, FALSE, 1, &xin, &yin,
				       &xin, &yin);
    *xout = xin;
    *yout = yin;
    if (xin < canvas->win_scale.x_offset) return (FALSE);
    if (xin >= canvas->win_scale.x_offset + canvas->win_scale.x_pixels)
	return (FALSE);
    if (yin < canvas->win_scale.y_offset) return (FALSE);
    if (yin >= canvas->win_scale.y_offset + canvas->win_scale.y_pixels)
	return (FALSE);
    return (TRUE);
}   /*  End Function canvas_convert_from_canvas_coord  */

/*OBSOLETE_FUNCTION*/
void canvas_register_d_convert_func (KWorldCanvas canvas,
				     flag (*coord_convert_func) (), void *info)
/*  [SUMMARY] Register co-ordinate conversion function.
    [PURPOSE] This routine will register the co-ordinate conversion function
    for a world canvas. This function will be called whenever conversions
    between world co-ordinates and pixel co-ordinates are performed. Only one
    co-ordinate conversion function is permitted per canvas. This is a means
    for a higher level object to take control of the world canvas.
    <canvas> The canvas.
    <func> The function that is called when co-ordinates are to be converted.
    The prototype function is [<canvas_PROTO_coord_convert_func>].
    <info> The initial arbitrary canvas information pointer.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "canvas_register_d_convert_func";

    VERIFY_CANVAS (canvas);
    fprintf (stderr, "WARNING: the <%s> routine will be removed in Karma",
	     function_name);
    fprintf (stderr, " version 2.0\nUse ");
    fprintf (stderr,
	     "the <canvas_register_coords_convert_func> routine instead\n");
    if (coord_convert_func == NULL) return;
    if ( (canvas->coords_convert_func != NULL) ||
	 (canvas->deprecated_coord_d_convert_func != NULL) )
    {
	fprintf (stderr, "coord_convert_func already registered\n");
	a_prog_bug (function_name);
    }
    canvas->deprecated_coord_d_convert_func = coord_convert_func;
    canvas->coord_convert_info = info;
}   /*  End Function canvas_register_d_convert_func  */

/*OBSOLETE_FUNCTION*/
void canvas_register_convert_func (KWorldCanvas canvas,
				   flag (*coord_convert_func) (), void *info)
/*  This routine will register the co-ordinate conversion function for a world
    canvas.
    This function will be called whenever conversions between world
    co-ordinates and pixel co-ordinates are performed.
    Only one co-ordinate conversion function is permitted per canvas. This is
    a means for a higher level object to take control of the world canvas.
    The canvas must be given by  canvas  .
    The function that is called when co-ordinates are to be converted must be
    pointed to by  coord_convert_func  .
    The interface to this routine is as follows:

    flag coord_convert_func (KWorldCanvas canvas,
                             struct win_scale_type *win_scale,
			     int *px, int *py, double *wx, double *wy,
			     flag to_world, void **info)
    *   This routine will modify the window scaling information for a world
        canvas.
        The canvas is given by  canvas  .
	The window scaling information is pointed to by  win_scale  .The data
	herein may be modified.
	The horizontal pixel co-ordinate storage must be pointed to by  px  .
	The vertical pixel co-ordinate storage must be pointed to by  py  .
	The horizontal world co-ordinate storage must be pointed to by  wx  .
	The vertical world co-ordinate storage must be pointed to by  wy  .
	If the value of  to_world  is TRUE, then a pixel to world co-ordinate
	transform is required, else a world to pixel co-ordinate transform is
	required.
	The arbitrary canvas information pointer is pointed to by  info  .
	The routine should return TRUE if the conversion was completed,
	else it returns FALSE indicating that the default conversions should be
	used.
    *
    ;

    The initial arbitrary canvas information pointer must be given by  info  .
    The routine returns nothing.
*/
{
    static char function_name[] = "canvas_register_convert_func";

    VERIFY_CANVAS (canvas);
    fprintf (stderr, "WARNING: the <%s> routine will be removed in Karma",
	     function_name);
    fprintf (stderr, " version 2.0\nUse ");
    fprintf (stderr,
	     "the <canvas_register_coords_convert_func> routine instead\n");
    if (coord_convert_func == NULL) return;
    if ( (canvas->coords_convert_func != NULL) ||
	 (canvas->deprecated_coord_d_convert_func != NULL) ||
	 (canvas->deprecated_coord_convert_func != NULL) )
    {
	fprintf (stderr, "coord_convert_func already registered\n");
	a_prog_bug (function_name);
    }
    canvas->deprecated_coord_convert_func = coord_convert_func;
    canvas->coord_convert_info = info;
}   /*  End Function canvas_register_convert_func  */

/*OBSOLETE_FUNCTION*/
flag canvas_coord_transform (KWorldCanvas canvas, double *x, double *y,
			     flag to_linear)
/*  [SUMMARY] Transform world co-ordinates.
    [PURPOSE] This routine will transform co-ordinates in a world canvas using
    the registered co-ordinate transform function (see
    [<canvas_register_transforms_func>]).
    <canvas> The world canvas object.
    <x> A pointer to the horizontal world co-ordinate. This is modified.
    <y> A pointer to the vertical world co-ordinate. This is modified.
    <to_linear> If TRUE then the routine will transform (possibly) non-linear
    world co-ordinates to linear world co-ordinates, else the transform is from
    linear to non-linear co-ordinates.
    [RETURNS] TRUE if the linear world co-ordinate lies within the canvas
    boundaries, else FALSE (although a conversion is still performed).
*/
{
    flag in_bounds = TRUE;
    static char function_name[] = "canvas_coord_transform";

    VERIFY_CANVAS (canvas);
    fprintf (stderr, "WARNING: the <%s> routine will be removed in Karma",
	     function_name);
    fprintf (stderr, " version 2.0\nUse ");
    fprintf (stderr, "the <canvas_coords_transform> routine instead\n");
    if (!to_linear)
    {
	/*  Input data are linear co-ordinates: test bounds  */
	if (canvas->win_scale.left_x < canvas->win_scale.right_x)
	{
	    if ( (*x < canvas->win_scale.left_x) ||
		 (*x > canvas->win_scale.right_x) ) in_bounds = FALSE;
	}
	else if ( (*x > canvas->win_scale.left_x) ||
		  (*x < canvas->win_scale.right_x) ) in_bounds = FALSE;
	if (canvas->win_scale.bottom_y < canvas->win_scale.top_y)
	{
	    if ( (*y < canvas->win_scale.bottom_y) ||
		 (*y > canvas->win_scale.top_y) ) in_bounds = FALSE;
	}
	else if ( (*y > canvas->win_scale.bottom_y) ||
		  (*y < canvas->win_scale.top_y) ) in_bounds = FALSE;
    }
    canvas_coords_transform (canvas, 1, x, to_linear, y, to_linear);
    if (to_linear)
    {
	/*  Output data are linear co-ordinates: test bounds  */
	if (canvas->win_scale.left_x < canvas->win_scale.right_x)
	{
	    if ( (*x < canvas->win_scale.left_x) ||
		 (*x > canvas->win_scale.right_x) ) in_bounds = FALSE;
	}
	else if ( (*x > canvas->win_scale.left_x) ||
		  (*x < canvas->win_scale.right_x) ) in_bounds = FALSE;
	if (canvas->win_scale.bottom_y < canvas->win_scale.top_y)
	{
	    if ( (*y < canvas->win_scale.bottom_y) ||
		 (*y > canvas->win_scale.top_y) ) in_bounds = FALSE;
	}
	else if ( (*y > canvas->win_scale.bottom_y) ||
		  (*y < canvas->win_scale.top_y) ) in_bounds = FALSE;
    }
    return (in_bounds);
}   /*  End Function canvas_coord_transform  */

/*OBSOLETE_FUNCTION*/
void canvas_register_transform_func (KWorldCanvas canvas,
				     void (*coord_transform_func) (),
				     void *info)
/*  [SUMMARY] Register non-linear co-ordinate transformation function.
    [PURPOSE] This routine will register the co-ordinate transform function
    for a world canvas. The co-ordinate transform function transforms the
    linear world co-ordinates used internally by the world canvas to be
    transformed into non-linear co-ordinates for external use. Only one
    co-ordinate transform function is permitted per canvas.
    <canvas> The world canvas object.
    <coord_transform_func> The function that is called when co-ordinates are to
    be transformed. The prototype function is
    [<canvas_PROTO_coord_transform_func>].
    <info> The initial arbitrary transform information pointer.
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "canvas_register_transform_func";

    VERIFY_CANVAS (canvas);
    fprintf (stderr, "WARNING: the <%s> routine will be removed in Karma",
	     function_name);
    fprintf (stderr, " version 2.0\nUse ");
    fprintf (stderr,"the <canvas_register_transforms_func> routine instead\n");
    if (coord_transform_func == NULL) return;
    if ( (canvas->coord_transforms_func != NULL) ||
	 (canvas->deprecated_coord_transform_func != NULL) )
    {
	fprintf (stderr, "coord_transform_func already registered\n");
	a_prog_bug (function_name);
    }
    canvas->deprecated_coord_transform_func = coord_transform_func;
    canvas->coord_transform_info = info;
}   /*  End Function canvas_register_transform_func  */


/*  Private functions follow  */

static void _canvas_pixcanvas_precompute_func
    (KPixCanvas pixcanvas, int width, int height,
     void **info, PostScriptPage pspage,
     unsigned int num_areas, KPixCanvasRefreshArea *areas)
/*  [PURPOSE] This routine is a precompute event consumer for a pixel canvas.
    <canvas> The pixel canvas.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <info> A pointer to the arbitrary information pointer.
    <pspage> The PostScriptPage object. If this is NULL, the refresh is *not*
    destined for a PostScript page.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas canvas;
    flag dummy = FALSE;
    flag visible;
    struct refresh_struct data;
    static char function_name[] = "_canvas_pixcanvas_precompute_func";

    canvas = (KWorldCanvas) *info;
    VERIFY_CANVAS (canvas);
    if (pixcanvas != canvas->pixcanvas)
    {
	fprintf (stderr, "Pixel canvases do not match\n");
	a_prog_bug (function_name);
    }
    canvas->dressing.drawn = FALSE;
    kwin_get_attributes (canvas->pixcanvas,
			 KWIN_ATT_VISIBLE, &visible,
			 KWIN_ATT_END);
    if (canvas->quash_negotiate)
    {
	canvas->quash_negotiate = FALSE;
	if (!visible) return;
    }
    else
    {
	if (!visible) return;
	if ( !_canvas_negotiate_scale_change (canvas, width, height, &dummy) )
	{
	    fprintf (stderr, "Pixel canvas not big enough\n");
	    return;
	}
    }
    /*  Call precompute functions  */
    data.width = width;
    data.height = height;
    data.cmap_resize = FALSE;
    data.pspage = pspage;
    data.num_areas = num_areas;
    data.areas = areas;
    c_call_callbacks (canvas->precompute_list, &data);
}   /*  End Function _canvas_pixcanvas_precompute_func  */

static void _canvas_pixcanvas_refresh_func (KPixCanvas pixcanvas,
					    int width,int height,
					    void **info, PostScriptPage pspage,
					    unsigned int num_areas,
					    KPixCanvasRefreshArea *areas,
					    flag *honoured_areas)
/*  [PURPOSE] This routine is a refresh event consumer for a pixel canvas.
    <canvas> The pixel canvas.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <info> A pointer to the arbitrary information pointer.
    <pspage> The PostScriptPage object. If this is NULL, the refresh is *not*
    destined for a PostScript page.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire pixel canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas canvas;
    static char function_name[] = "_canvas_pixcanvas_refresh_func";

    canvas = (KWorldCanvas) *info;
    VERIFY_CANVAS (canvas);
    if (pixcanvas != canvas->pixcanvas)
    {
	fprintf (stderr, "Pixel canvases do not match\n");
	a_prog_bug (function_name);
    }
    _canvas_refresh_canvas (canvas, width, height, FALSE, pspage,
			    num_areas, areas, honoured_areas);
}   /*  End Function _canvas_pixcanvas_refresh_func  */

static void _canvas_cmap_resize_func (Kcolourmap cmap, void **info)
/*  This routine registers a change in the size of a colourmap.
    The colourmap must be given by  cmap  .
    The arbitrary colourmap information pointer is pointed to by  info  .
    [RETURNS] Nothing.
*/
{
    KWorldCanvas canvas;
    int width, height;
    struct refresh_struct data;
    static char function_name[] = "_canvas_cmap_resize_func";

    canvas = (KWorldCanvas) *info;
    VERIFY_CANVAS (canvas);
    if (cmap != canvas->cmap)
    {
	fprintf (stderr, "Colourmaps do not match\n");
	a_prog_bug (function_name);
    }
    if (canvas->in_size_control_func) return;
    _canvas_set_sat_pixels (canvas);
    canvas->cmap_num_pixels = kcmap_get_pixels (cmap, &canvas->cmap_pixels);
    /*  Get pixel canvas size  */
    kwin_get_size (canvas->pixcanvas, &width, &height);
    /*  Call precompute functions  */
    data.width = width;
    data.height = height;
    data.cmap_resize = TRUE;
    data.pspage = NULL;
    data.num_areas = 0;
    data.areas = NULL;
    c_call_callbacks (canvas->precompute_list, &data);
    /*  Now do the refresh  */
    _canvas_refresh_canvas (canvas, width, height, TRUE, NULL, 0, NULL, NULL);
}   /*  End Function _canvas_cmap_resize_func  */

static void _canvas_refresh_canvas (KWorldCanvas canvas, int width, int height,
				    flag cmap_resize, PostScriptPage pspage,
				    unsigned int num_areas,
				    KPixCanvasRefreshArea *areas,
				    flag *honoured_areas)
/*  [PURPOSE] This routine will refresh a world canvas (by calling all
    registered refresh functions).
    <canvas> The world canvas.
    <width> The width of of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <cmap_resize> If TRUE, the refresh is due to a colourmap resize.
    <pspage> The PostScriptPage object. If this is NULL, the refresh is *not*
    destined for a PostScript page.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire pixel canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    flag visible;
    struct refresh_struct data;
    static char function_name[] = "_canvas_refresh_canvas";

    VERIFY_CANVAS (canvas);
    canvas->dressing.drawn = FALSE;
    if (canvas->auto_blank)
    {
	kwin_get_attributes (canvas->pixcanvas,
			     KWIN_ATT_VISIBLE, &visible,
			     KWIN_ATT_BACKGROUND_PIXEL,
			     &canvas->win_scale.blank_pixel,
			     KWIN_ATT_END);
    }
    else kwin_get_attributes (canvas->pixcanvas,
			      KWIN_ATT_VISIBLE, &visible,
			      KWIN_ATT_END);
    if (!visible) return;
    /*  Call refresh functions  */
    canvas->in_refresh = TRUE;
    data.width = width;
    data.height = height;
    data.cmap_resize = cmap_resize;
    data.pspage = pspage;
    data.num_areas = num_areas;
    data.areas = areas;
    data.honoured_areas = FALSE;
    c_call_callbacks (canvas->refresh_list, &data);
    if (data.honoured_areas && (honoured_areas != NULL) )
	*honoured_areas = TRUE;
    if (!canvas->dressing.drawn && canvas->dressing.display)
	canvas_draw_dressing (canvas);
    canvas->in_refresh = FALSE;
}   /*  End Function _canvas_refresh_canvas  */

static unsigned long _canvas_get_pixel_from_value (KWorldCanvas canvas,
						   double value[2],
						   unsigned short *red,
						   unsigned short *green,
						   unsigned short *blue)
/*  [SUMMARY] Get a pixel value from a data value.
    <canvas> The world canvas.
    <value> The data value.
    <red> The red intensity in the colourmap for the pixel will be
    written here. If this is NULL, nothing is written here.
    <green> The green intensity in the colourmap for the pixel will be
    written here. If this is NULL, nothing is written here.
    <blue> The blue intensity in the colourmap for the pixel will be
    written here. If this is NULL, nothing is written here.
    [RETURNS] The pixel value.
*/
{
    unsigned int index;
    unsigned long pixel_value;
    double scaled_min, scaled_max, tmp;
    struct win_scale_type *win_scale;
    flag (*iscale_func) ();
    static char function_name[] = "_canvas_get_pixel_from_value";

    VERIFY_CANVAS (canvas);
    win_scale = &canvas->win_scale;
    iscale_func = win_scale->iscale_func;
    if ( !(value[0] < TOOBIG) || !(value[1] < TOOBIG) )
    {
	if (canvas->auto_blank)
	{
	    kwin_get_attributes (canvas->pixcanvas,
				 KWIN_ATT_BACKGROUND_PIXEL,
				 &canvas->win_scale.blank_pixel,
				 KWIN_ATT_END);
	}
	return (canvas->win_scale.blank_pixel);
    }
    switch (win_scale->conv_type)
    {
      case KIMAGE_COMPLEX_CONV_REAL:
	break;
      case KIMAGE_COMPLEX_CONV_IMAG:
	value[0] = value[1];
	break;
      case KIMAGE_COMPLEX_CONV_ABS:
	value[0] = sqrt (value[0] * value[0] + value[1] * value[1]);
	break;
      case KIMAGE_COMPLEX_CONV_SQUARE_ABS:
	value[0] = value[0] * value[0] + value[1] * value[1];
	break;
      case KIMAGE_COMPLEX_CONV_PHASE:
	if ( (value[0] == 0.0) && (value[1] == 0.0) )
	{
	    value[0] = 0.0;
	}
	else
	{
	    value[0] = atan2 (value[1], value[0]);
	}
	break;
      case KIMAGE_COMPLEX_CONV_CONT_PHASE:
	fprintf (stderr, "Not finished continuous phase\n");
	return (FALSE);
	/*break;*/
      default:
	fprintf (stderr, "Illegal value of conversion: %u\n",
		 win_scale->conv_type);
	a_prog_bug (function_name);
	break;
    }
    if ( !(value[0] < TOOBIG) )
    {
	if (canvas->auto_blank)
	{
	    kwin_get_attributes (canvas->pixcanvas,
				 KWIN_ATT_BACKGROUND_PIXEL,
				 &canvas->win_scale.blank_pixel,
				 KWIN_ATT_END);
	}
	return (canvas->win_scale.blank_pixel);
    }
    /*  Take real value and convert to pixel value  */
    if (iscale_func == NULL)
    {
	scaled_min = win_scale->z_min;
	scaled_max = win_scale->z_max;
    }
    else
    {
	if ( !(*iscale_func) (&scaled_min, 0, &win_scale->z_min, 0, 1,
			      win_scale->z_min, win_scale->z_max,
			      win_scale->iscale_info) )
	{
	    fprintf (stderr, "%s: error scaling raw intensity minimum\n",
		     function_name);
	}
	if ( !(*iscale_func) (&scaled_max, 0, &win_scale->z_max, 0, 1,
			      win_scale->z_min, win_scale->z_max,
			      win_scale->iscale_info) )
	{
	    fprintf (stderr, "%s: error scaling raw intensity maximum\n",
		     function_name);
	}
	if (scaled_max < scaled_min) swap (scaled_min, scaled_max, tmp);
	if ( !(*win_scale->iscale_func) (value, 1, value, 1, 1,
					 win_scale->z_min, win_scale->z_max,
					 win_scale->iscale_info) )
	{
	    fprintf (stderr, "Error scaling data\n");
	    a_prog_bug (function_name);
	}
    }
    if (value[0] < scaled_min) pixel_value = win_scale->min_sat_pixel;
    else if (value[0] > scaled_max) pixel_value = win_scale->max_sat_pixel;
    else
    {
	index = ( (value[0] - scaled_min) * (canvas->cmap_num_pixels - 1) /
		  (scaled_max - scaled_min) + 0.5 );
	pixel_value = canvas->cmap_pixels[index];
	if ( (red != NULL) || (green != NULL) || (blue != NULL) )
	{
	    kcmap_get_rgb_value (canvas->cmap, index, red, green, blue);
	}
    }
    return (pixel_value);
}   /*  End Function _canvas_get_pixel_from_value  */

static flag _canvas_pixcanvas_position_event (KPixCanvas pixcanvas,
					      int x, int y,
					      unsigned int event_code,
					      void *event_info, void **f_info)
/*  This routine is a position event consumer for a pixel canvas.
    The pixel canvas is given by  pixcanvas  .
    The horizontal position of the event, relative to the canvas origin,
    will be given by  x  .
    The vertical position of the event, relative to the canvas origin,
    will be given by  y  .
    The arbitrary event code is given by  event_code  .
    The arbitrary event information is pointed to by  event_info  .
    The arbitrary function information pointer is pointed to by  f_info  .
    The routine returns TRUE if the event was consumed, else it returns
    FALSE indicating that the event is still to be processed.
*/
{
    KWorldCanvas canvas;
    double dx, dy;
    struct position_struct data;
    static char function_name[] = "_canvas_pixcanvas_position_event";

    canvas = (KWorldCanvas) *f_info;
    VERIFY_CANVAS (canvas);
    /*  Convert pixel co-ordinates to world co-ordinates  */
    data.x_pix = x;
    data.y_pix = y;
    dx = x;
    dy = y;
    canvas_convert_to_canvas_coords (canvas, TRUE, 1, &dx, &dy,
				     &data.x_lin, &data.y_lin,
				     &data.x, &data.y);
    /*  Call event consumer functions  */
    data.event_code = event_code;
    data.e_info = event_info;
    return c_call_callbacks (canvas->position_list, &data);
}   /*  End Function _canvas_pixcanvas_position_event  */

static flag _canvas_precompute_event_func (void *object, void *client1_data,
					   void *call_data, void *client2_data)
/*  [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    KWorldCanvas canvas;
    struct win_scale_type win_scale;
    void (*func) ();
    struct refresh_struct *data;
    static char function_name[] = "_canvas_precompute_event_func";

    canvas = (KWorldCanvas) object;
    VERIFY_CANVAS (canvas);
    data = (struct refresh_struct *) call_data;
    func = ( void (*) () ) client2_data;
    m_copy (&win_scale, &canvas->win_scale, sizeof win_scale);
    (*func) (canvas, data->width, data->height, &win_scale, canvas->cmap,
	     data->cmap_resize, client1_data, data->pspage,
	     data->num_areas, data->areas);
    return (FALSE);
}   /*  End Function _canvas_precompute_event_func  */

static flag _canvas_refresh_event_func (void *object, void *client1_data,
					void *call_data, void *client2_data)
/*  [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    KWorldCanvas canvas;
    flag honoured_areas;
    struct win_scale_type win_scale;
    void (*func) ();
    struct refresh_struct *data;
    static char function_name[] = "_canvas_refresh_event_func";

    canvas = (KWorldCanvas) object;
    VERIFY_CANVAS (canvas);
    data = (struct refresh_struct *) call_data;
    func = ( void (*) () ) client2_data;
    m_copy (&win_scale, &canvas->win_scale, sizeof win_scale);
    honoured_areas = FALSE;
    (*func) (canvas, data->width, data->height, &win_scale, canvas->cmap,
	     data->cmap_resize, client1_data, data->pspage,
	     data->num_areas, data->areas, &honoured_areas);
    if (honoured_areas) data->honoured_areas = TRUE;
    else
    {
	data->num_areas = 0;
	data->areas = NULL;
	data->honoured_areas = FALSE;
    }
    return (FALSE);
}   /*  End Function _canvas_refresh_event_func  */

static flag _canvas_position_event_func (void *object, void *client1_data,
					 void *call_data, void *client2_data)
/*  [PURPOSE] This routine is called when object callbacks are called.
    <object> The object information pointer.
    <client1_data> The first client information pointer.
    <call_data> The call information pointer.
    <client2_data> The second client information pointer.
    [RETURNS] TRUE if further callbacks should not be called, else FALSE.
*/
{
    KWorldCanvas canvas;
    PositionFunc func = (PositionFunc) client2_data;
    struct position_struct *data = call_data;
    static char function_name[] = "_canvas_position_event_func";

    canvas = (KWorldCanvas) object;
    VERIFY_CANVAS (canvas);
    return (*func) (canvas, data->x, data->y, data->event_code,
		    data->e_info, client1_data,
		    data->x_lin, data->y_lin, data->x_pix, data->y_pix);
}   /*  End Function _canvas_position_event_func  */

static void _canvas_verify_win_scale (struct win_scale_type *win_scale,
				      char *function_name)
/*  This routine will verify if window scaling information is sensible.
    The window scaling information must be pointed to by  win_scale  .
    The name of the calling function must be pointed to by  function_name  .
    The routine returns nothing (the process aborts on error).
*/
{
    if (win_scale == NULL) return;
    if (K_WIN_SCALE_MAGIC_NUMBER != win_scale->magic_number)
    {
	fprintf (stderr,"Bad magic number for  win_scale  structure\n");
	fprintf (stderr, "This will break in Karma version 2.0\n");
	fprintf (stderr,"Make sure a call is made to canvas_init_win_scale\n");
	fprintf (stderr, "and then recompile the application\n");
/*      Coming in Version 2.0 of Karma
	a_prog_bug (function_name);
*/
    }
#ifdef DISABLED
    if (win_scale->right_x == win_scale->left_x)
    {
	fprintf (stderr, "right_x: %e must not equal left_x: %e\n",
		 win_scale->right_x, win_scale->left_x);
	a_prog_bug (function_name);
    }
    if (win_scale->top_y == win_scale->bottom_y)
    {
	fprintf (stderr, "top_y: %e must not equal bottom_y: %e\n",
		 win_scale->top_y, win_scale->bottom_y);
	a_prog_bug (function_name);
    }
#endif
    if (win_scale->z_scale != K_INTENSITY_SCALE_LINEAR)
    {
	fprintf (stderr, "z_scale: %u never was supported.\n",
		 win_scale->z_scale);
	a_prog_bug (function_name);
    }
}   /*  End Function _canvas_verify_win_scale  */

static flag _canvas_negotiate_scale_change (KWorldCanvas canvas,
					    int width, int height,
					    flag *boundary_clear)
/*  [SUMMARY] Negotiate a scale change for a world canvas.
    [PURPOSE] This routine will negotiate a scale change for a world canvas.
    Thisnegotiation occurs between the size control function which may be
    registered and the dressing display control system. This negotiation
    process is, er, intricate :-)
    <canvas> The world canvas.
    <width> The width of the pixel canvas.
    <height> The height of the pixel canvas.
    <boundary_clear> If the value TRUE is written here then the
    [<canvas_resize>] routine will attempt to clear only the boundary between
    the pixel canvas and the world canvas. If the value FALSE is written here
    or nothing is written here, the [<canvas_resize>] routine will clear the
    entire pixel canvas as appropriate.
    [RETURNS] TRUE on succes, else FALSE indicating the scale change could not
    be negotiated.
*/
{
    flag dressing_toobig;
    int p_left, p_right, p_top, p_bottom;
    int tmp, old_x_pix, old_y_pix, old_x_off, old_y_off;
    unsigned int count;
    struct win_scale_type *win_scale = &canvas->win_scale;
    static char function_name[] = "_canvas_negotiate_scale_change";

    canvas->world_boundary.valid = FALSE;
    *boundary_clear = FALSE;
    if (canvas->size_control_func == NULL)
    {
	/*  Must automatically produce some window scaling information  */
	canvas_get_dressing_size (canvas, &p_left, &p_right, &p_top, &p_bottom,
				  &canvas->dressing);
	if (p_left + p_right >= width)
	{
	    fprintf (stderr,
		     "Dressing (%d pixels) too wide for pixel canvas (%d pixels)\n",
		     p_left + p_right, width);
	    return (FALSE);
	}
	if (p_top + p_bottom >= height)
	{
	    fprintf (stderr,
		     "Dressing (%d pixels) too high for pixel canvas (%d pixels)\n",
		     p_top + p_bottom, height);
	    return (FALSE);
	}
	win_scale->x_offset = p_left;
	win_scale->x_pixels = width - p_left - p_right;
	win_scale->y_offset = p_top;
	win_scale->y_pixels = height - p_top - p_bottom;
	return (TRUE);
    }
    old_x_pix = win_scale->x_pixels;
    old_y_pix = win_scale->y_pixels;
    old_x_off = win_scale->x_offset;
    old_y_off = win_scale->y_offset;
    p_left = 0;
    p_right = 0;
    p_top = 0;
    p_bottom = 0;
    for (count = 0, dressing_toobig = TRUE;
	 (count < DRESSING_ITERATE_MAX) && dressing_toobig; ++count)
    {
	/*  Call size control function first to ensure canvas boundary is
	    correct, otherwise <canvas_get_dressing_size> could fail  */
	canvas->in_size_control_func = TRUE;
	(*canvas->size_control_func) (canvas,
				      width - p_left - p_right,
				      height - p_top - p_bottom,
				      win_scale, &canvas->size_control_info,
				      boundary_clear);
	canvas->in_size_control_func = FALSE;
	win_scale->x_offset += (p_left + p_right) / 2;
	win_scale->y_offset += (p_top + p_bottom) / 2;
	_canvas_verify_win_scale (win_scale, function_name);
	canvas_get_dressing_size (canvas, &p_left, &p_right, &p_top, &p_bottom,
				  &canvas->dressing);
	if ( (p_left + p_right + win_scale->x_pixels <= width) &&
	     (p_top + p_bottom + win_scale->y_pixels <= height) )
	{
	    /*  Looks like it will fit  */
	    dressing_toobig = FALSE;
	}
    }
    if (dressing_toobig)
    {
	if (p_left + p_right + win_scale->x_pixels > width)
	{
	    fprintf (stderr,
		     "Dressing+active (%d + %d = %d) too wide for pixel canvas (%d pixels)\n",
		     p_left + p_right, win_scale->x_pixels,
		     p_left + p_right + win_scale->x_pixels,
		     width);
	    return (FALSE);
	}
	if (p_top + p_bottom + win_scale->y_pixels > height)
	{
	    fprintf (stderr,
		     "Dressing+active (%d + %d = %d) too high for pixel canvas (%d pixels)\n",
		     p_top + p_bottom, win_scale->y_pixels,
		     p_top + p_bottom + win_scale->y_pixels,
		     height);
	    return (FALSE);
	}
    }
    /*  Dressing will fit!  */
    if ( (win_scale->x_pixels == old_x_pix) &&
	 (p_left <= old_x_off) && (p_right + old_x_off + old_x_pix <= width) )
    {
	win_scale->x_offset = old_x_off;  /*  Retain old position  */
    }
    else if ( (p_left > win_scale->x_offset) ||
	      (p_right + win_scale->x_offset + win_scale->x_pixels > width) )
    {
	/*  Must move canvas so the dressing will still lie in pixel canvas  */
	tmp = width - win_scale->x_pixels - p_left - p_right;
	win_scale->x_offset = tmp / 2 + p_left;
    }
    if ( (win_scale->y_pixels == old_y_pix) &&
	 (p_top <= old_y_off) && (p_bottom + old_y_off + old_y_pix <= height) )
    {
	win_scale->y_offset = old_y_off;  /*  Retain old position  */
    }
    else if ( (p_top > win_scale->y_offset) ||
	      (p_bottom + win_scale->y_offset + win_scale->y_pixels > height) )
    {
	tmp = height - win_scale->y_pixels - p_top - p_bottom;
	win_scale->y_offset = tmp / 2 + p_top;
    }
    return (TRUE);
}   /*  End Function _canvas_negotiate_scale_change  */

static void _canvas_dressing_refresh_func (KWorldCanvas canvas,
					   int width, int height,
					   struct win_scale_type *win_scale,
					   Kcolourmap cmap, flag cmap_resize,
					   void **info)
/*  This routine is a refresh event consumer for a world canvas.
    The canvas is given by  canvas  .
    The width of the canvas in pixels is given by  width  .
    The height of the canvas in pixels is given by  height  .
    The window scaling information is pointed to by  win_scale  .
    The colourmap associated with the canvas is given by  cmap  .
    If the refresh function was called as a result of a colourmap resize
    the value of  cmap_resize  will be TRUE.
    The arbitrary canvas information pointer is pointed to by  info  .
    [RETURNS] Nothing.
*/
{
    static char function_name[] = "_canvas_dressing_refresh_func";

    VERIFY_CANVAS (canvas);
    canvas_draw_dressing (canvas);
}   /*  End Function _canvas_dressing_refresh_func  */

static void _canvas_set_sat_pixels (KWorldCanvas canvas)
/*  [PURPOSE] This routine will set the saturation pixels as appropriate.
    <canvas> The world canvas.
    [RETURNS] Nothing.
*/
{
    unsigned int num_pixels;
    unsigned long *pixel_values;

    if (canvas->cmap == NULL) return;
    if (canvas->auto_min_sat || canvas->auto_max_sat)
    {
	num_pixels = kcmap_get_pixels (canvas->cmap, &pixel_values);
	if (canvas->auto_min_sat)
	{
	    canvas->win_scale.min_sat_pixel = pixel_values[0];
	}
	if (canvas->auto_max_sat)
	{
	    if (num_pixels < 1) ++num_pixels;
	    canvas->win_scale.max_sat_pixel = pixel_values[num_pixels - 1];
	}
    }
}   /*  End Function _canvas_set_sat_pixels  */

static double get_ellipse_distance (KWorldCanvas canvas, flag to_world,
				    flag linear,
				    double angle, double xscale, double yscale,
				    double cos_rot, double sin_rot,
				    double centre_x, double centre_y,
				    double radius_x, double radius_y,
				    double cx, double cy)
/*  [SUMMARY] Get distance from ellipse centre.
    <canvas> The world canvas.
    <to_world> If TRUE a pixel->world conversion is performed, else a
    world->pixel conversion is performed.
    <linear> If TRUE the co-ordinate conversions are pixel<->linear rather than
    pixel<->world.
    <angle> The angle along the ellipse in radians.
    <xscale> The horizontal scale.
    <yscale> The vertical scale.
    <cos_rot> The cosine of the rotation angle.
    <sin_rot> The sine of the rotation angle.
    <centre_x> The input horizontal centre.
    <centre_y> The input vertical centre.
    <radius_x> The horizontal radius.
    <radius_y> The vertical radius.
    <cx> The output horizontal centre.
    <cy> The output vertical centre.
    [RETURNS] The distance.
*/
{
    double x, y, x1, y1;

    x1 = radius_x * cos (angle);
    y1 = radius_y * sin (angle);
    x = x1 * cos_rot - y1 * sin_rot;
    y = y1 * cos_rot + x1 * sin_rot;
    x = centre_x + x * xscale;
    y = centre_y + y * yscale;
    if (to_world)
    {
	if (linear)
	    canvas_convert_to_canvas_coords (canvas, FALSE, 1, &x, &y, &x, &y,
					     NULL, NULL);
	else
	    canvas_convert_to_canvas_coords (canvas, FALSE, 1, &x, &y,
					     NULL, NULL, &x, &y);
    }
    else canvas_convert_from_canvas_coords (canvas, FALSE, linear, 1,
					    &x, &y, &x, &y);
    return sqrt ( (cx - x) * (cx - x) + (cy - y) * (cy - y) );
}   /*  End Function get_ellipse_distance  */

static double get_ellipse_max_angle (KWorldCanvas canvas, flag to_world,
				     flag linear,
				     double xscale, double yscale,
				     double cos_rot, double sin_rot,
				     double centre_x, double centre_y,
				     double radius_x, double radius_y,
				     double cx, double cy)
/*  [SUMMARY] Get angle with maximum distance from ellipse centre.
    <canvas> The world canvas.
    <to_world> If TRUE a pixel->world conversion is performed, else a
    world->pixel conversion is performed.
    <linear> If TRUE the co-ordinate conversions are pixel<->linear rather than
    pixel<->world.
    <xscale> The horizontal scale.
    <yscale> The vertical scale.
    <cos_rot> The cosine of the rotation angle.
    <sin_rot> The sine of the rotation angle.
    <centre_x> The input horizontal centre.
    <centre_y> The input vertical centre.
    <radius_x> The horizontal radius.
    <radius_y> The vertical radius.
    <cx> The output horizontal centre.
    <cy> The output vertical centre.
    [RETURNS] The angle in radians.
*/
{
    double angle, angle1, angle2, jump_angle, sign;
    double distance, distance1, distance2;
    double min_jump = 1e-4;  /*  (0.367 / 64) degrees  */

    /*  First find which way is up  */
    distance = get_ellipse_distance (canvas, to_world, linear,
				     0.0, xscale, yscale,
				     cos_rot, sin_rot, centre_x, centre_y,
				     radius_x, radius_y, cx, cy);
    distance1 = get_ellipse_distance (canvas, to_world, linear,
				      -min_jump * 0.1, xscale, yscale,
				      cos_rot, sin_rot, centre_x, centre_y,
				      radius_x, radius_y, cx, cy);
    distance2 = get_ellipse_distance (canvas, to_world, linear,
				      min_jump * 0.1, xscale, yscale,
				      cos_rot, sin_rot, centre_x, centre_y,
				      radius_x, radius_y, cx, cy);
    if ( (distance >= distance1) && (distance >= distance2) ) return (0.0);
    sign = (distance2 > distance1) ? 1.0 : -1.0;
    jump_angle = PI_ON_2;
    angle = 0.0;
    while (jump_angle > min_jump)
    {
	angle1 = angle + sign * jump_angle;
	distance1 = get_ellipse_distance (canvas, to_world, linear, angle1,
					  xscale, yscale,
					  cos_rot, sin_rot, centre_x, centre_y,
					  radius_x, radius_y, cx, cy);
	angle2 = angle + sign * (jump_angle + min_jump * 0.1);
	distance2 = get_ellipse_distance (canvas, to_world, linear, angle2,
					  xscale, yscale,
					  cos_rot, sin_rot, centre_x, centre_y,
					  radius_x, radius_y, cx, cy);
	if (distance1 >= distance)
	{
	    /*  Jump point is higher  */
	    if (distance2 >= distance1)
	    {
		/*  Jump point is higher and still rising  */
		angle = angle2;
		distance = distance2;
	    }
	    else
	    {
		/*  Jump point is higher but now dropping: gone past a peak  */
		angle = angle2;
		distance = distance2;
		jump_angle *= 0.5;
		sign *= -1.0;
	    }
	}
	else
	{
	    /*  Jump point is lower: may be before or after a trough  */
	    jump_angle *= 0.5;
	}
    }
    return (angle);
}   /*  End Function get_ellipse_max_angle  */
