/*LINTLIBRARY*/
/*  ContourSimpleControl.c

    This code provides s simple contour control widget for Xt.

    Copyright (C) 1998-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for an contour simple control widget
    for Xt.


    Written by      Richard Gooch   16-SEP-1998

    Updated by      Richard Gooch   17-SEP-1998: First working version

    Last updated by Richard Gooch   7-MAY-2001: Added -ve contour colour
  control.


*/
#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Dialog.h>
#include <karma.h>
#include <k_event_codes.h>
#include <karma_contour.h>
#include <karma_iarray.h>
#include <karma_xtmisc.h>
#include <karma_ex.h>
#include <karma_m.h>
#include <karma_a.h>
#include <Xkw/ContourSimpleControlP.h>
#include <Xkw/Ktoggle.h>
#include <Xkw/SimpleSlider.h>


#define MAX_CONTOUR_LEVELS 100


struct canvas_list
{
    KWorldCanvas canvas;
    KContourImage cimage;
    struct canvas_list *next;
};

/*  Private functions  */
STATIC_FUNCTION (void ContourSimpleControl__Initialise,
		 (Widget request, Widget new) );
STATIC_FUNCTION (void unload_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void apply_levels_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void reset_levels_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void refresh_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void slider_realise_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (void ContourSimpleControl__Destroy, (Widget W) );


#define offset(field) XtOffset(ContourSimpleControlWidget, contourSimpleControl.field)

static XtResource ContourSimpleControlResources[] = 
{
    {XkwNiarray, XkwCIarray, XtRPointer, sizeof (XtPointer),
     offset (array), XtRImmediate, NULL},
};

#undef offset


ContourSimpleControlClassRec contourSimpleControlClassRec = 
{
  {
      /* CoreClassPart */
    (WidgetClass) &topLevelShellClassRec,          /* superclass */
    "ContourSimpleControl",                        /* class_name */
    sizeof (ContourSimpleControlRec),              /* widget_size */
    NULL,                                          /* class_initialise */
    NULL,                                          /* class_part_initialise */
    FALSE,                                         /* class_init */
    (XtInitProc) ContourSimpleControl__Initialise, /* initialise */
    NULL,                                          /* initialise_hook */
    XtInheritRealize,                              /* realise */
    NULL,                                          /* actions */
    0,                                             /* num_actions */
    ContourSimpleControlResources,                 /* resources */
    XtNumber (ContourSimpleControlResources),      /* num_resources */
    NULLQUARK,                                     /* xrm_class */
    TRUE,                                          /* compress_motion */
    TRUE,                                          /* compress_exposure */
    TRUE,                                          /* compress_enterleave */
    TRUE,                                          /* visible_interest */
    ContourSimpleControl__Destroy,                 /* destroy */
    XtInheritResize,                               /* resize */
    NULL,                                          /* expose */
    NULL,                                          /* set_values */
    NULL,                                          /* set_values_hook */
    XtInheritSetValuesAlmost,                      /* set_values_almost */
    NULL,                                          /* get_values_hook */
    NULL,                                          /* accept_focus */
    XtVersion,                                     /* version */
    NULL,                                          /* callback_private */
    NULL,                                          /* tm_translations */
    NULL,
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
    /* geometry_manager	  */	XtInheritGeometryManager,
    /* change_managed	  */	XtInheritChangeManaged,
    /* insert_child	  */	XtInheritInsertChild,
    /* delete_child	  */	XtInheritDeleteChild,
    /* extension	  */	NULL
  },{ /* Shell */
    /* extension	  */	NULL
  },{ /* WMShell */
    /* extension	  */	NULL
  },{ /* VendorShell */
    /* extension	  */	NULL
  },{ /* TopLevelShell */
    /* extension	  */	NULL
  },
  {  /* ContourSimpleControlClassPart */
    0 /* empty */
  }
};

WidgetClass contourSimpleControlWidgetClass = (WidgetClass)
     &contourSimpleControlClassRec;

static void ContourSimpleControl__Initialise (Widget Request, Widget New)
{
    /*ContourSimpleControlWidget request=(ContourSimpleControlWidget)Request;*/
    ContourSimpleControlWidget new = (ContourSimpleControlWidget) New;
    Widget form, w, upper_widget;
    char txt[STRING_LENGTH];
    /*static char function_name[] = "ContourSimpleControlWidget::Initialise";*/

    new->contourSimpleControl.first_canvas = NULL;
    form = XtVaCreateManagedWidget ("form", formWidgetClass, New,
				    XtNborderWidth, 0,
				    NULL);
    w = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				 XtNlabel, "Unload",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 NULL);
    XtAddCallback (w, XtNcallback, xtmisc_popdown_cbk, New);
    XtAddCallback (w, XtNcallback, unload_cbk, New);
    upper_widget = w;
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Apply Levels",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, apply_levels_cbk, New);
    w = XtVaCreateManagedWidget ("button", commandWidgetClass, form,
				 XtNlabel, "Reset Levels",
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNheight, 20,
				 XtNfromHoriz, w,
				 NULL);
    XtAddCallback (w, XtNcallback, reset_levels_cbk, New);
    w = XtVaCreateManagedWidget ("toggle", ktoggleWidgetClass, form,
				 XtNlabel, "Display",
				 XkwNcrosses, FALSE,
				 XtNstate, TRUE,
				 XtNfromHoriz, w,
				 NULL);
    new->contourSimpleControl.display_tgl = w;
    XtAddCallback (w, XtNcallback, refresh_cbk, New);
    sprintf ( txt, "Contour Levels (%s)",
	      iarray_value_name (new->contourSimpleControl.array) );
    w = XtVaCreateManagedWidget ("contourLevels", dialogWidgetClass, form,
				 XtNlabel, txt,
				 XtNborderWidth, 0,
				 XtNfromVert, upper_widget,
				 XtNvalue,
				 "%10:90+10                                  ",
				 NULL);
    upper_widget = w;
    new->contourSimpleControl.levels_dlg = w;
    w = XtVaCreateManagedWidget ("contourColour", dialogWidgetClass, form,
				 XtNlabel, "-ve Contour Colour     ",
				 XtNborderWidth, 0,
				 XtNfromVert, upper_widget,
				 XtNvalue, "grey70",
				 NULL);
    new->contourSimpleControl.negative_colour_dlg = w;
    w = XtVaCreateManagedWidget ("contourColour", dialogWidgetClass, form,
				 XtNlabel, "+ve Contour Colour    ",
				 XtNborderWidth, 0,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNvalue, "green",
				 NULL);
    upper_widget = w;
    new->contourSimpleControl.positive_colour_dlg = w;
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				 XtNfromVert, upper_widget,
				 XtNlabel, "-ve Contour Width",
				 XtNwidth, 220,
				 XkwNminimum, 0,
				 XkwNmaximum, 100,
				 XkwNmodifier, 1,
				 XtNvalue, 0,
				 XkwNscaledValuePtr,
				 &new->contourSimpleControl.contour_neg_width,
				 XkwNscaledFormat, "%.1f",
				 NULL);
    XkwSimpleSliderSetScale (w, 1e-1, 0.0, FALSE, TRUE);
    XtAddCallback (w, XkwNvalueChangeCallback, apply_levels_cbk, New);
    XtAddCallback (w, XkwNrealiseCallback, slider_realise_cbk, New);
    w = XtVaCreateManagedWidget ("slider", simpleSliderWidgetClass, form,
				 XtNfromVert, upper_widget,
				 XtNfromHoriz, w,
				 XtNlabel, "+ve Contour Width",
				 XtNwidth, 220,
				 XkwNminimum, 0,
				 XkwNmaximum, 100,
				 XkwNmodifier, 1,
				 XtNvalue, 0,
				 XkwNscaledValuePtr,
				 &new->contourSimpleControl.contour_pos_width,
				 XkwNscaledFormat, "%.1f",
				 NULL);
    XkwSimpleSliderSetScale (w, 1e-1, 0.0, FALSE, TRUE);
    XtAddCallback (w, XkwNvalueChangeCallback, apply_levels_cbk, New);
}   /*  End Function Initialise  */

static void unload_cbk (Widget W, XtPointer client_data, XtPointer call_data)
/*  This is the unload callback.
*/
{
    /*static char function_name[] = "ContourSimpleControlWidget::unload_cbk";*/

    XtDestroyWidget ( (Widget) client_data );
}   /*  End Function unload_cbk   */

static void apply_levels_cbk (Widget W, XtPointer client_data,
			      XtPointer call_data)
/*  This is the apply levels callback.
*/
{
    iarray array;
    unsigned int num_levels;
    double min, max;
    double scale, offset;
    ContourSimpleControlWidget top = (ContourSimpleControlWidget) client_data;
    struct canvas_list *curr;
    char *level_str, *negative_colour, *positive_colour;
    flag dash[MAX_CONTOUR_LEVELS];
    double contour_levels[MAX_CONTOUR_LEVELS], linewidths[MAX_CONTOUR_LEVELS];
    /*static char function_name[] =
	"ContourSimpleControlWidget::apply_levels_cbk";*/

    array = top->contourSimpleControl.array;
    iarray_get_data_scaling (array, &scale, &offset);
    iarray_get_data_range (array, &min, &max);
    XtVaGetValues (top->contourSimpleControl.negative_colour_dlg,
		   XtNvalue, &negative_colour,
		   NULL);
    XtVaGetValues (top->contourSimpleControl.positive_colour_dlg,
		   XtNvalue, &positive_colour,
		   NULL);
    XtVaGetValues (top->contourSimpleControl.levels_dlg,
		   XtNvalue, &level_str,
		   NULL);
    if ( ( num_levels =
	   contour_parse_levels (contour_levels, MAX_CONTOUR_LEVELS, level_str,
				 min, max) ) < 1 ) return;
    contour_unscale_levels (contour_levels, dash, linewidths, num_levels,
			    scale, offset,
			    top->contourSimpleControl.contour_neg_width,
			    top->contourSimpleControl.contour_pos_width);
    for (curr = top->contourSimpleControl.first_canvas; curr != NULL;
	 curr = curr->next)
    {
	unsigned int count;
	unsigned long negative_pixel_value, positive_pixel_value;
	unsigned long pixel_values[MAX_CONTOUR_LEVELS];

	canvas_get_colour (curr->canvas, negative_colour,&negative_pixel_value,
			   NULL, NULL, NULL);
	canvas_get_colour (curr->canvas, positive_colour,&positive_pixel_value,
			   NULL, NULL, NULL);
	for (count = 0; count < MAX_CONTOUR_LEVELS; ++count)
	{
	    pixel_values[count] = dash[count] ?
		negative_pixel_value : positive_pixel_value;
	}
	contour_set_levels (curr->cimage, num_levels,
			    contour_levels, pixel_values, NULL, linewidths);
    }
    refresh_cbk (NULL, client_data, NULL);
}   /*  End Function apply_levels_cbk   */

static void reset_levels_cbk (Widget w, XtPointer client_data,
			      XtPointer call_data)
/*  This is the reset levels callback.
*/
{
    ContourSimpleControlWidget top = (ContourSimpleControlWidget) client_data;
    /*static char function_name[] =
	"ContourSimpleControlWidget::reset_levels_cbk";*/

    XtVaSetValues (top->contourSimpleControl.levels_dlg,
		   XtNvalue, "%10:90+10",
		   NULL);
    apply_levels_cbk (w, top, NULL);
}   /*  End Function reset_levels_cbk   */

static void refresh_cbk (Widget w, XtPointer client_data, XtPointer call_data)
/*  This is the refresh callback.
*/
{
    Bool display;
    ContourSimpleControlWidget top = (ContourSimpleControlWidget) client_data;
    struct canvas_list *curr;
    /*static char function_name[] = "ContourSimpleControlWidget::toggle_cbk";*/

    XtVaGetValues (top->contourSimpleControl.display_tgl,
		   XtNstate, &display,
		   NULL);
    for (curr = top->contourSimpleControl.first_canvas; curr != NULL;
	 curr = curr->next)
    {
	contour_set_active (curr->cimage, display, TRUE, FALSE, FALSE, 0);
    }
}   /*  End Function refresh_cbk   */

static void slider_realise_cbk (Widget w, XtPointer client_data,
				XtPointer call_data)
/*  This is the slider realise callback.
*/
{
    ContourSimpleControlWidget top = (ContourSimpleControlWidget) client_data;

    XtVaSetValues (top->contourSimpleControl.levels_dlg,
		   XtNvalue, "%10:90+10",
		   NULL);
}   /*  End Function slider_realise_cbk   */

static void ContourSimpleControl__Destroy (Widget W)
{
    ContourSimpleControlWidget w = (ContourSimpleControlWidget) W;
    struct canvas_list *curr, *next;

    for (curr = w->contourSimpleControl.first_canvas; curr != NULL;
	 curr = next)
    {
	next = curr->next;
	contour_set_all_inactive (curr->canvas, FALSE, TRUE, (uaddr) w);
	if (curr->cimage != NULL) contour_destroy (curr->cimage);
	m_free (curr);
    }
    iarray_dealloc (w->contourSimpleControl.array);
}   /*  End Function Destroy  */


/*  Public functions follow  */

void XkwContourSimpleControlAddCanvas (Widget W, KWorldCanvas canvas)
/*  [SUMMARY] Add a KWorldCanvas to an ContourSimpleControl widget canvas list.
    <W> The widget.
    <canvas> The canvas to add to the list. The widget will then manipulate
    this canvas. If this is NULL the routine does nothing.
    [RETURNS] Nothing.
*/
{
    ContourSimpleControlWidget w;
    struct canvas_list *new;
    static char function_name[] = "XkwContourSimpleControlAddCanvas";

    if ( !XtIsContourSimpleControl (W) )
    {
	fprintf (stderr, "Invalid widget passed\n");
	a_prog_bug (function_name);
    }
    w = (ContourSimpleControlWidget) W;
    if (canvas == NULL) return;
    if ( ( new = m_alloc (sizeof *new) ) == NULL )
	m_abort (function_name, "canvas_list structure");
    new->canvas = canvas;
    if ( ( new->cimage =
	   contour_create_from_iarray (canvas, w->contourSimpleControl.array,
				       FALSE, 0, NULL, (uaddr) w) ) == NULL )
    {
	m_abort (function_name, "contour image");
    }
    new->next = w->contourSimpleControl.first_canvas;
    w->contourSimpleControl.first_canvas = new;
}   /*  End Function XkwContourSimpleControlAddCanvas  */
