/*LINTLIBRARY*/
/*  MagnifierPopup.c

    This code provides a magnifier display widget for Xt.

    Copyright (C) 1997-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for a magnifier display
    widget for Xt.


    Written by      Richard Gooch   7-DEC-1997

    Updated by      Richard Gooch   8-DEC-1997: Override baseclass
  SlaveImageDisplayPopup fastPanner resource to use slow panner.

    Updated by      Richard Gooch   9-DEC-1997: Added <all_realised> extension.

    Updated by      Richard Gooch   14-FEB-1998: Added showCrosshair resource.

    Updated by      Richard Gooch   22-FEB-1998: Created function
  <XkwMagnifierPopupAddCanvas>.

    Updated by      Richard Gooch   17-MAY-1998: Trap illegal values.

    Updated by      Richard Gooch   12-AUG-1998: Support new event
  K_CANVAS_EVENT_POINTER_WARP.

    Updated by      Richard Gooch   15-AUG-1999: Register <position_func> in
  <XkwMagnifierPopupAddCanvas> rather than <realised_cbk>.

    Updated by      Richard Gooch   11-JUL-2000: Added #include <karma_kwin.h>.

    Last updated by Richard Gooch   20-MAR-2001: Created Destroy method.


*/
#include <stdio.h>
#include <stdlib.h>
#include <X11/Xos.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#ifndef X11
#  define X11
#endif
#include <karma.h>
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_xtmisc.h>
#include <karma_kwin.h>
#include <karma_ds.h>
#include <karma_a.h>
#include <karma_m.h>
#include <Xkw/MagnifierPopupP.h>
#include <Xkw/ImageDisplay.h>


struct canvas_list
{
    KWorldCanvas canvas;
    struct canvas_list *next;
};


/*  Private functions  */
STATIC_FUNCTION (void MagnifierPopup__Initialise,
		 (Widget request, Widget new) );
STATIC_FUNCTION (void _MagnifierPopup_Destroy, (Widget W) );
STATIC_FUNCTION (void realised_cbk,
		 (Widget parent, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (flag position_func,
		 (ViewableImage vimage, double x, double y,
		  void *value, unsigned int event_code,
		  void *e_info, void **f_info,
		  double x_lin, double y_lin, unsigned int value_type) );
STATIC_FUNCTION (unsigned long get_colour,
		 (KPixCanvas canvas, CONST char *colourname) );
STATIC_FUNCTION (void magnifier_refresh,
		 (KWorldCanvas canvas, int width, int height,
		  struct win_scale_type *win_scale,
		  Kcolourmap cmap, flag cmap_resize, void **info,
		  PostScriptPage pspage,
		  unsigned int num_areas,
		  KPixCanvasRefreshArea *areas, flag *honoured_areas) );


#define offset(field) XtOffset(MagnifierPopupWidget, magnifierPopup.field)

static XtResource MagnifierPopupResources[] = 
{
    {XkwNfastPanner, XkwCFastPanner, XtRBool, sizeof (Bool),
     XtOffset (MagnifierPopupWidget, slaveImageDisplayPopup.fastPanner),
     XtRImmediate, (XtPointer) False},
    {XkwNshowCrosshair, XkwCShowCrosshair, XtRBool, sizeof (Bool),
     offset (showCrosshair), XtRImmediate, (XtPointer) False},
};

#undef offset

MagnifierPopupClassRec magnifierPopupClassRec = 
{
  {
      /* CoreClassPart */
    (WidgetClass) &slaveImageDisplayPopupClassRec, /* superclass */
    "MagnifierPopup",                            /* class_name */
    sizeof (MagnifierPopupRec),                  /* widget_size */
    NULL,                                        /* class_initialise */
    NULL,                                        /* class_part_initialise */
    FALSE,                                       /* class_init */
    (XtInitProc) MagnifierPopup__Initialise,     /* initialise */
    NULL,                                        /* initialise_hook */
    XtInheritRealize,                            /* realise */
    NULL,                                        /* actions */
    0,                                           /* num_actions */
    MagnifierPopupResources,                     /* resources */
    XtNumber (MagnifierPopupResources),          /* num_resources */
    NULLQUARK,                                   /* xrm_class */
    TRUE,                                        /* compress_motion */
    TRUE,                                        /* compress_exposure */
    TRUE,                                        /* compress_enterleave */
    TRUE,                                        /* visible_interest */
    _MagnifierPopup_Destroy,                     /* destroy */
    XtInheritResize,                             /* resize */
    NULL,                                        /* expose */
    NULL,                                        /* set_values */
    NULL,                                        /* set_values_hook */
    XtInheritSetValuesAlmost,                    /* set_values_almost */
    NULL,                                        /* get_values_hook */
    NULL,                                        /* accept_focus */
    XtVersion,                                   /* version */
    NULL,                                        /* callback_private */
    NULL,                                        /* tm_translations */
    NULL,
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
      /* geometry_manager */	XtInheritGeometryManager,
      /* change_managed	  */	XtInheritChangeManaged,
      /* insert_child	  */	XtInheritInsertChild,
      /* delete_child	  */	XtInheritDeleteChild,
      /* extension	  */	NULL
  },
  { /* Shell */
      /* extension	  */	NULL
  },
  { /* WMShell */
      /* extension	  */	NULL
  },
  { /* VendorShell */
      /* extension	  */	NULL
  },
  { /* TopLevelShell */
      /* extension	  */	NULL
  },
  {  /* SlaveImageDisplayPopupClassPart */
      0                   /* empty */
  },
  {  /* MagnifierPopupClassPart */
      0                   /* empty */
  },
};

WidgetClass magnifierPopupWidgetClass = (WidgetClass) &magnifierPopupClassRec;


static void MagnifierPopup__Initialise (Widget Request, Widget New)
{
    /*MagnifierPopupWidget request = (MagnifierPopupWidget) Request;*/
    MagnifierPopupWidget new = (MagnifierPopupWidget) New;
    /*static char function_name[] = "MagnifierPopupWidget::Initialise";*/

    new->magnifierPopup.x_lin = TOOBIG;
    new->magnifierPopup.y_lin = TOOBIG;
    new->magnifierPopup.first_canvas = NULL;
    XtAddCallback (New, XkwNrealisedCallback, realised_cbk, New);
}   /*  End Function Initialise  */

static void _MagnifierPopup_Destroy (Widget W)
{
    MagnifierPopupWidget w = (MagnifierPopupWidget) W;
    struct canvas_list *curr, *next;

    for (curr = w->magnifierPopup.first_canvas; curr != NULL; curr = next)
    {
	next = curr->next;
	m_free (curr);
    }
}   /*  End Function Destroy  */

static void realised_cbk (Widget parent, XtPointer client_data,
			  XtPointer call_data)
{
    MagnifierPopupWidget top = (MagnifierPopupWidget) client_data;
    Widget TOP = (Widget) client_data;
    SlaveImageDisplayPopupPart *slave =
	(SlaveImageDisplayPopupPart *) &top->slaveImageDisplayPopup;

    if (slave->pseudoCanvas != NULL)
    {
	canvas_register_refresh_func (slave->pseudoCanvas, magnifier_refresh,
				      top);
    }
    if (slave->directCanvas != NULL)
    {
	canvas_register_refresh_func (slave->directCanvas, magnifier_refresh,
				      top);
    }
    if (slave->trueCanvas != NULL)
    {
	canvas_register_refresh_func (slave->trueCanvas, magnifier_refresh,
				      top);
    }
    xtmisc_ext_call (TOP, "all_realised", TOP);
}   /*  End Function realised_cbk  */

static flag position_func (ViewableImage vimage, double x, double y,
			   void *value, unsigned int event_code,
			   void *e_info, void **f_info,
			   double x_lin, double y_lin, unsigned int value_type)
/*  [SUMMARY] Event handler for position events on master canvas.
    <viewimg> The active viewable image.
    <x> The horizontal world co-ordinate of the event.
    <y> The vertical world co-ordinate of the event.
    These values will have been transformed by the registered transform
    function (see  canvas_register_transform_func  ) for the associated
    world canvas.
    <value> A pointer to the data value in the viewable image corresponding
    to the event co-ordinates.
    <event_code> The arbitrary event code.
    <e_info> The arbitrary event information.
    <f_info> The arbitrary function information pointer.
    <x_lin> The linear horizontal world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <y_lin> The linear vertical world co-ordinate (the co-ordinate prior
    to the transform function being called).
    <value_type> The type of the data value. This may be K_DCOMPLEX or
    K_UB_RGB.
    [RETURNS] FALSE indicating that the event is still to be processed.
*/
{
    KWorldCanvas mag_wc;
    unsigned int hdim, vdim;
    unsigned long px, py;
    array_desc *arr_desc;
    MagnifierPopupWidget top = (MagnifierPopupWidget) *f_info;

    switch (event_code)
    {
      case K_CANVAS_EVENT_POINTER_MOVE:
      case K_CANVAS_EVENT_POINTER_WARP:
	break;
      default:
	return (FALSE);
	/*break;*/
    }
    top->magnifierPopup.x_lin = x_lin;
    top->magnifierPopup.y_lin = y_lin;
    viewimg_get_attributes (vimage,
			    VIEWIMG_VATT_ARRAY_DESC, &arr_desc,
			    VIEWIMG_VATT_HDIM, &hdim,
			    VIEWIMG_VATT_VDIM, &vdim,
			    VIEWIMG_VATT_END);
    px = ds_get_coord_num (arr_desc->dimensions[hdim], x_lin,
			   SEARCH_BIAS_CLOSEST);
    py = ds_get_coord_num (arr_desc->dimensions[vdim], y_lin,
			   SEARCH_BIAS_CLOSEST);
    mag_wc = top->slaveImageDisplayPopup.visibleCanvas;
    viewimg_set_canvas_attributes (mag_wc,
				   VIEWIMG_ATT_PAN_CENTRE_X, px,
				   VIEWIMG_ATT_PAN_CENTRE_Y, py,
				   VIEWIMG_ATT_END);
    kwin_refresh_if_visible (canvas_get_pixcanvas (mag_wc), FALSE);
    return (FALSE);
}   /*  End Function position_func  */

static unsigned long get_colour (KPixCanvas canvas, CONST char *colourname)
/*  [SUMMARY] Get colour.
    <canvas> The pixel canvas.
    <colourname> The name of the colour.
    [RETURNS] The specified colour if available, else a default colour.
*/
{
    unsigned long pixel_value;

    if ( kwin_get_colour (canvas, colourname, &pixel_value, NULL, NULL, NULL) )
    {
	return (pixel_value);
    }
    if ( kwin_get_colour (canvas, "white", &pixel_value, NULL, NULL, NULL) )
    {
	fprintf (stderr,
		 "Error allocating colour: \"%s\" defaulting to white\n",
		 colourname);
	return (pixel_value);
    }
    fprintf (stderr, "Error allocating colour: \"%s\" defaulting to 0\n",
	     colourname);
    return (0);
}   /*  End Function get_colour  */

static void magnifier_refresh (KWorldCanvas canvas, int width, int height,
			       struct win_scale_type *win_scale,
			       Kcolourmap cmap, flag cmap_resize, void **info,
			       PostScriptPage pspage,
			       unsigned int num_areas,
			       KPixCanvasRefreshArea *areas,
			       flag *honoured_areas)
/*  [SUMMARY] Refresh event callback.
    [PURPOSE] This routine is a refresh event consumer for a world canvas.
    <canvas> The world canvas being refreshed.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <win_scale> A pointer to the window scaling information.
    <cmap> The colourmap associated with the canvas.
    <cmap_resize> TRUE if the refresh function was called as a result of a
    colourmap resize, else FALSE.
    <info> A pointer to the arbitrary canvas information pointer.
    <pspage> If not NULL, the PostScriptPage object the refresh is
    redirected to.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed. Note that these
    areas are given in pixel co-ordinates.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    KPixCanvas pixcanvas;
    unsigned long pixel_value;
    double px, py;
    MagnifierPopupWidget top = (MagnifierPopupWidget) *info;

    if (!top->magnifierPopup.showCrosshair) return;
    if ( DOUBLE_IS_BLANK (top->magnifierPopup.x_lin) ) return;
    if ( DOUBLE_IS_BLANK (top->magnifierPopup.y_lin) ) return;
    pixcanvas = canvas_get_pixcanvas (canvas);
    canvas_convert_from_canvas_coords (canvas, FALSE, TRUE, 1,
				       &top->magnifierPopup.x_lin,
				       &top->magnifierPopup.y_lin,
				       &px, &py);
    pixel_value = get_colour (pixcanvas, "green");
    kwin_draw_line (pixcanvas, 0, py, width - 1, py, pixel_value);
    kwin_draw_line (pixcanvas, px, 0, px, height - 1, pixel_value);
}   /*  End Function magnifier_refresh  */


/*  Public functions follow  */

void XkwMagnifierPopupAddCanvas (Widget W, KWorldCanvas canvas)
/*  [SUMMARY] Add a KWorldCanvas to a MagnifierPopup widget canvas list.
    <W> The widget.
    <canvas> The parent canvas to add to the list. The widget will then
    manipulate this canvas. If this is NULL the routine does nothing.
    [RETURNS] Nothing.
*/
{
    MagnifierPopupWidget w;
    struct canvas_list *new;
    static char function_name[] = "XkwMagnifierPopupAddCanvas";

    if ( !XtIsMagnifierPopup (W) )
    {
	fprintf (stderr, "Invalid widget passed\n");
	a_prog_bug (function_name);
    }
    w = (MagnifierPopupWidget) W;
    if (canvas == NULL) return;
    if ( ( new = m_alloc (sizeof *new) ) == NULL )
	m_abort (function_name, "canvas_list structure");
    new->canvas = canvas;
    new->next = w->magnifierPopup.first_canvas;
    w->magnifierPopup.first_canvas = new;
    viewimg_register_position_event_func (canvas, position_func, W);
}   /*  End Function XkwMagnifierPopupAddCanvas  */
