/*LINTLIBRARY*/
/*  SimpleProgressMeter.c

    This code provides a simple progress meter widget for Xt.

    Copyright (C) 1997-2000  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains all routines needed for a simple progress meter widget
  for Xt.


    Written by      Richard Gooch   19-APR-1997

    Last updated by Richard Gooch   9-JUN-2000: Don't include multibuf.h


*/
#include <stdio.h>
#ifndef X11
#  define X11
#endif
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/XawInit.h>
#include <karma.h>
#include <karma_a.h>
#include <Xkw/SimpleProgressMeterP.h>
#include <os.h>

#define HEIGHT_AT_MIN 10


#define offset(field) XtOffsetOf(SimpleProgressMeterRec, progressMeter.field)

static XtResource resources[] =
{
    {XkwNmaximum, XkwCMaximum, XtRCardinal, sizeof (Cardinal),
     offset (maximum), XtRImmediate, 0},
    {XtNvalue, XtCValue, XtRCardinal, sizeof (Cardinal),
     offset (value), XtRImmediate, 0},
};
#undef offset

STATIC_FUNCTION (void SimpleProgressMeter__Initialise,
		 (Widget request, Widget new) );
STATIC_FUNCTION (Boolean SimpleProgressMeter__SetValues,
		 (Widget Current, Widget Request, Widget New) );
STATIC_FUNCTION (void canvas_realise_cbk, (Widget w, XtPointer client_data,
					   XtPointer call_data) );
STATIC_FUNCTION (void refresh_func,
		 (KPixCanvas canvas, int width, int height,
		  void **info, PostScriptPage pspage,
		  unsigned int num_areas, KPixCanvasRefreshArea *areas,
		  flag *honoured_areas) );


SimpleProgressMeterClassRec simpleProgressMeterClassRec =
{
    {
	/* core fields  */
	(WidgetClass) &canvasClassRec,  /*  superclass             */
	"SimpleProgressMeter",          /*  class_name             */
	sizeof (SimpleProgressMeterRec),/*  widget_size            */
	NULL,                           /*  class_initialise       */
	NULL,                           /*  class_part_initialise  */
	FALSE,                          /*  class_inited           */
	(XtInitProc) SimpleProgressMeter__Initialise,/*  initialise*/
	NULL,                           /*  initialise_hook        */
	XtInheritRealize,               /*  realise                */
	NULL,                           /*  actions                */
	0,                              /*  num_actions            */
	resources,                      /*  resources              */
	XtNumber (resources),           /*  num_resources          */
	NULLQUARK,                      /*  xrm_class              */
	TRUE,                           /*  compress_motion        */
	TRUE,                           /*  compress_exposure      */
	TRUE,                           /*  compress_enterleave    */
	TRUE,                           /*  visible_interest       */
	NULL,                           /*  destroy                */
	NULL,                           /*  resize                 */
	XtInheritExpose,                /*  expose                 */
	(XtSetValuesFunc) SimpleProgressMeter__SetValues, /*  set_values  */
	NULL,                           /*  set_values_hook        */
	XtInheritSetValuesAlmost,       /*  set_values_almost      */
	NULL,                           /*  get_values_hook        */
	NULL,                           /*  accept_focus           */
	XtVersion,                      /*  version                */
	NULL,                           /*  callback_private       */
	XtInheritTranslations,          /*  tm_table               */
	XtInheritQueryGeometry,         /*  query_geometry         */
	XtInheritDisplayAccelerator,    /*  display_accelerator    */
	NULL                            /*  extension              */
    },
    {   /* simple fields */
	XtInheritChangeSensitive        /* change_sensitive        */
    },
    {   /* canvas fields */
	0                               /* ignore                  */
    },
    {
	/*  simple progressMeter fields */
	0                               /*  ignore                 */
    }
};

WidgetClass simpleProgressMeterWidgetClass = (WidgetClass)
     &simpleProgressMeterClassRec;

static void SimpleProgressMeter__Initialise (Widget Request, Widget New)
{
    /*SimpleProgressMeterWidget request = (SimpleProgressMeterWidget)Request;*/
    SimpleProgressMeterWidget new = (SimpleProgressMeterWidget) New;
    /*static char function_name[] = "SimpleProgressMeter::Initialise";*/

    if (new->core.width < 1) new->core.width = 174;
    if (new->core.width < 50) new->core.width = 50;
    if (new->core.height < 1) new->core.height = 40;
    new->canvas.silence_unconsumed_events = TRUE;
    XtAddCallback (New, XkwNrealiseCallback, canvas_realise_cbk,
		   (XtPointer) New);
}   /*  End Function Initialise  */

static Boolean SimpleProgressMeter__SetValues (Widget Current, Widget Request,
					       Widget New)
{
    SimpleProgressMeterWidget current = (SimpleProgressMeterWidget) Current;
    SimpleProgressMeterWidget new = (SimpleProgressMeterWidget) New;
    static char function_name[] = "SimpleProgressMeter::SetValues";

    if (new->core.width < 50) new->core.width = 50;
    if (new->progressMeter.value > new->progressMeter.maximum)
    {
	fprintf (stderr, "Value: %u is greater than maximum: %u\n",
		 new->progressMeter.value, new->progressMeter.maximum);
	a_prog_bug (function_name);
    }
    if ( (new->progressMeter.value > current->progressMeter.value) &&
	 (new->progressMeter.maximum == current->progressMeter.maximum) &&
	 (new->canvas.monoPixCanvas != NULL) )
    {
	kwin_refresh_if_visible (new->canvas.monoPixCanvas, FALSE);
	return (FALSE);
    }
    return TRUE;   
}   /*  End Function SetValues  */

static void canvas_realise_cbk (Widget w, XtPointer client_data,
				XtPointer call_data)
/*  This is the canvas realise callback.
*/
{
    SimpleProgressMeterWidget top = (SimpleProgressMeterWidget) client_data;
    /*static char function_name[] = "SimpleProgressMeterWidget::canvas_realise_cbk";*/

    kwin_register_refresh_func (top->canvas.monoPixCanvas, refresh_func, top);
}   /*  End Function canvas_realise_cbk   */

static void refresh_func (KPixCanvas canvas, int width, int height,
			  void **info, PostScriptPage pspage,
			  unsigned int num_areas, KPixCanvasRefreshArea *areas,
			  flag *honoured_areas)
/*  [SUMMARY] Process a refresh event for a pixel canvas.
    <canvas> The pixel canvas.
    <width> The width of the canvas in pixels.
    <height> The height of the canvas in pixels.
    <info> A pointer to the arbitrary canvas information pointer.
    <pspage> If not NULL, the PostScriptPage object the refresh is
    redirected to.
    <num_areas> The number of areas that need to be refreshed. If this is
    0 then the entire pixel canvas needs to be refreshed.
    <areas> The list of areas that need to be refreshed.
    <honoured_areas> If the value TRUE is written here it is assumed the
    routine honoured the list of refresh areas and did not write outside
    these areas and hence the list of areas will be passed to subsequent
    registered refresh routines. If FALSE is written here (or nothing is
    written here), implying the routine refreshed the entire pixel canvas,
    subsequent refresh routines will be told to refresh the entire canvas.
    [RETURNS] Nothing.
*/
{
    int x, xmax, size, middle, y0;
    float scale, tmp;
    SimpleProgressMeterWidget top = (SimpleProgressMeterWidget) *info;

    if (width < 50) return;
    if (top->progressMeter.maximum < 1) return;
    scale = (float) (height - HEIGHT_AT_MIN) / (float) (width - 1);
    tmp = (float) top->progressMeter.value / (float)top->progressMeter.maximum;
    xmax = (int) ( tmp * (float) (width - 1) );
    middle = height / 2;
    for (x = 0; x <= xmax; ++x)
    {
	size = (float) x * scale + HEIGHT_AT_MIN;
	if (size < 1) continue;
	if (size == 1)
	{
	    kwin_draw_point (canvas, x, middle, top->canvas.foreground_pixel);
	    continue;
	}
	y0 = middle - size / 2;
	kwin_draw_line (canvas, x, y0, x, y0 + size,
			top->canvas.foreground_pixel);
    }
}   /*  End Function refresh_func  */
