/*LINTLIBRARY*/
/*  SimpleTextEntry.c

    This code provides a simple text entry widget for Xt.

    Copyright (C) 1999  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*

    This file contains all routines needed for a simple text entry widget for
    Xt.


    Written by      Richard Gooch   14-SEP-1999

    Last updated by Richard Gooch   20-SEP-1999


*/
#include <stdio.h>
#ifndef X11
#  define X11
#endif
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <karma.h>
#include <Xkw/SimpleTextEntryP.h>


#define offset(field) XtOffsetOf(SimpleTextEntryRec, simpleTextEntry.field)

static XtResource resources[] =
{
    {XkwNnewCharacterCallback, XtCCallback, XtRCallback, sizeof (XtPointer),
     offset (newCharacterCallback), XtRCallback, (XtPointer) NULL},
    {XkwNnewLineCallback, XtCCallback, XtRCallback, sizeof (XtPointer),
     offset (newLineCallback), XtRCallback, (XtPointer) NULL},
};
#undef offset

STATIC_FUNCTION (void _SimpleTextEntry_Initialise,
		 (Widget request, Widget new) );
STATIC_FUNCTION (void _SimpleTextEntry_text_cbk,
		 (Widget w, XtPointer client_data, XtPointer call_data) );


SimpleTextEntryClassRec simpleTextEntryClassRec =
{
    {
	/* core fields  */
	(WidgetClass) &asciiTextClassRec,  /*  superclass             */
	"SimpleTextEntry",                 /*  class_name             */
	sizeof (SimpleTextEntryRec),       /*  widget_size            */
	NULL,                              /*  class_initialise       */
	NULL,                              /*  class_part_initialise  */
	FALSE,                             /*  class_inited           */
	(XtInitProc) _SimpleTextEntry_Initialise,/*  initialise             */
	NULL,                              /*  initialise_hook        */
	XtInheritRealize,                  /*  realise                */
	NULL,                              /*  actions                */
	0,                                 /*  num_actions            */
	resources,                         /*  resources              */
	XtNumber (resources),              /*  num_resources          */
	NULLQUARK,                         /*  xrm_class              */
	TRUE,                              /*  compress_motion        */
	TRUE,                              /*  compress_exposure      */
	TRUE,                              /*  compress_enterleave    */
	TRUE,                              /*  visible_interest       */
	NULL,                              /*  destroy                */
	NULL,                              /*  resize                 */
	XtInheritExpose,                   /*  expose                 */
	NULL,                              /*  set_values             */
	NULL,                              /*  set_values_hook        */
	XtInheritSetValuesAlmost,          /*  set_values_almost      */
	NULL,                              /*  get_values_hook        */
	NULL,                              /*  accept_focus           */
	XtVersion,                         /*  version                */
	NULL,                              /*  callback_private       */
	XtInheritTranslations,             /*  tm_table               */
	XtInheritQueryGeometry,            /*  query_geometry         */
	XtInheritDisplayAccelerator,       /*  display_accelerator    */
	NULL                               /*  extension              */
    },
    {
	/* simple */
	XtInheritChangeSensitive,          /*  change_sensitive       */
    },
    {
	/* text */
	0,                                 /*  extension              */
    },
    {
	/* ascii */
	0,                                 /*  extension              */
    },
    {
	/*  simple text entry fields */
	0                                  /*  ignore                 */
    }
};

WidgetClass simpleTextEntryWidgetClass = (WidgetClass)&simpleTextEntryClassRec;

static void _SimpleTextEntry_Initialise (Widget Request, Widget New)
{
    /*SimpleTextEntryWidget request = (SimpleTextEntryWidget) Request;*/
    SimpleTextEntryWidget new = (SimpleTextEntryWidget) New;
    /*static char function_name[] = "SimpleTextEntry::Initialise";*/

    XtAddCallback (new->text.source, XtNcallback, _SimpleTextEntry_text_cbk,
		   New);
}   /*  End Function Initialise  */

static void _SimpleTextEntry_text_cbk (Widget w, XtPointer client_data,
				       XtPointer call_data)
{
    int offset;
    String string;
    SimpleTextEntryWidget top = client_data;
    char *ptr;
    char txt[STRING_LENGTH];
    /*static char function_name[] = "SimpleTextEntry::text_cbk";*/

    XtVaGetValues (w, XtNstring, &string, NULL);
    if ( ( ptr = strchr (string, '\n') ) == NULL )
    {
	XtCallCallbacks ( (Widget) top, XkwNnewCharacterCallback, string );
	return;
    }
    offset = ptr - string;
    strncpy (txt, string, offset);
    strcpy (txt + offset, ptr + 1);
    XtVaSetValues (w, XtNstring, txt, NULL);
    XtVaSetValues (w, XtNinsertPosition, offset, NULL);/*Doesn't seem to work*/
    XtCallCallbacks ( (Widget) top, XkwNnewLineCallback, txt );
}   /*  End Function text_cbk  */
