/*LINTLIBRARY*/
/*  SlaveImageDisplayPopup.c

    This code provides a slaveImageDisplay display widget for Xt.

    Copyright (C) 1997-2001  Richard Gooch

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  karma-request@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*  This file contains all routines needed for a slaveImageDisplay display
    widget for Xt.


    Written by      Richard Gooch   6-DEC-1997

    Updated by      Richard Gooch   7-DEC-1997

    Updated by      Richard Gooch   12-DEC-1997: Use new multiCanvas resource
  for ImageDisplay widget.

    Updated by      Richard Gooch   13-FEB-1998: Created new function
  <XkwSlaveImageDisplayPopupRefresh>.

    Updated by      Richard Gooch   22-FEB-1998: Removed master canvases array.

    Updated by      Richard Gooch   31-AUG-1998: No longer require parent
  ImageDisplay to be realised before initialisation.

    Updated by      Richard Gooch   26-AUG-1999: Fixed bug: function
  <new_visible_canvas_cbk> did not map and unmap windows.

    Updated by      Richard Gooch   12-MAY-2000: Fixed bug in function
  <im_dpy_realised_cbk>: now pass slave canvas list rather than master canvas
  list to XkwNrealisedCallback.

    Last updated by Richard Gooch   20-MAR-2001: Created Destroy method.


*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>
#include <X11/Xos.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Dialog.h>
#ifndef X11
#  define X11
#endif
#include <karma.h>
#include <k_event_codes.h>
#include <karma_viewimg.h>
#include <karma_xtmisc.h>
#include <karma_m.h>
#include <karma_a.h>
#include <Xkw/SlaveImageDisplayPopupP.h>
#include <Xkw/ImageDisplay.h>
#include <Xkw/Canvas.h>


/*  Private functions  */
STATIC_FUNCTION (void _SlaveImageDisplayPopup_Initialise,
		 (Widget request, Widget new) );
STATIC_FUNCTION (void _SlaveImageDisplayPopup_Destroy, (Widget W) );
STATIC_FUNCTION (void _SlaveImageDisplayPopup_im_dpy_realised_cbk,
		 (Widget parent, XtPointer client_data, XtPointer call_data) );
STATIC_FUNCTION (Widget clone_canvas,
		 (SlaveImageDisplayPopupWidget top, Widget orig_cnvs) );
STATIC_FUNCTION (void clone_wc,
		 (SlaveImageDisplayPopupWidget top, Widget cnvs,
		  CONST char *name, KWorldCanvas *worldcanvas,
		  KWorldCanvas **wc_ptr) );
STATIC_FUNCTION (void new_visible_canvas_cbk,
		 (Widget parent, XtPointer client_data, XtPointer call_data) );


#define offset(field) XtOffset(SlaveImageDisplayPopupWidget, slaveImageDisplayPopup.field)

static XtResource SlaveImageDisplayPopupResources[] = 
{
    {XkwNpseudoColourCanvas, XkwCWorldCanvas, XtRPointer, sizeof (XtPointer),
     offset (pseudoCanvas), XtRImmediate, NULL},
    {XkwNdirectColourCanvas, XkwCWorldCanvas, XtRPointer, sizeof (XtPointer),
     offset (directCanvas), XtRImmediate, NULL},
    {XkwNtrueColourCanvas, XkwCWorldCanvas, XtRPointer, sizeof (XtPointer),
     offset (trueCanvas), XtRImmediate, NULL},
    {XkwNpseudoColourLeftCanvas, XkwCWorldCanvas, XtRPointer,
     sizeof (XtPointer), offset (pseudoCanvasLeft), XtRImmediate, NULL},
    {XkwNpseudoColourRightCanvas, XkwCWorldCanvas, XtRPointer,
     sizeof (XtPointer), offset (pseudoCanvasRight), XtRImmediate, NULL},
    {XkwNdirectColourLeftCanvas, XkwCWorldCanvas, XtRPointer,
     sizeof (XtPointer), offset (directCanvasLeft), XtRImmediate, NULL},
    {XkwNdirectColourRightCanvas, XkwCWorldCanvas, XtRPointer,
     sizeof (XtPointer), offset (directCanvasRight), XtRImmediate, NULL},
    {XkwNtrueColourLeftCanvas, XkwCWorldCanvas, XtRPointer,
     sizeof (XtPointer), offset (trueCanvasLeft), XtRImmediate, NULL},
    {XkwNtrueColourRightCanvas, XkwCWorldCanvas, XtRPointer,
     sizeof (XtPointer), offset (trueCanvasRight), XtRImmediate, NULL},
    {XkwNvisibleCanvas, XkwCWorldCanvas, XtRPointer, sizeof (XtPointer),
     offset (visibleCanvas), XtRImmediate, NULL},
    {XkwNcanvasList, XkwCWorldCanvas, XtRPointer, sizeof (XtPointer),
     offset (canvasList), XtRImmediate, NULL},
    {XkwNfastPanner, XkwCFastPanner, XtRBool, sizeof (Bool),
     offset (fastPanner), XtRImmediate, (XtPointer) True},
    {XkwNenablePanning, XkwCEnablePanning, XtRBool, sizeof (Bool),
     offset (enablePanning), XtRImmediate, FALSE},
    {XkwNrealisedCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
     offset (realisedCallback), XtRCallback, (caddr_t) NULL},
    {XkwNverbose, XkwCVerbose, XtRBool, sizeof (Bool),
     offset (verbose), XtRImmediate, FALSE},
};

#undef offset

SlaveImageDisplayPopupClassRec slaveImageDisplayPopupClassRec = 
{
  {
      /* CoreClassPart */
    (WidgetClass) &topLevelShellClassRec,     /* superclass */
    "SlaveImageDisplayPopup",                 /* class_name */
    sizeof (SlaveImageDisplayPopupRec),       /* widget_size */
    NULL,                                     /* class_initialise */
    NULL,                                     /* class_part_initialise */
    FALSE,                                    /* class_init */
    (XtInitProc) _SlaveImageDisplayPopup_Initialise,      /* initialise */
    NULL,                                     /* initialise_hook */
    XtInheritRealize,                         /* realise */
    NULL,                                     /* actions */
    0,                                        /* num_actions */
    SlaveImageDisplayPopupResources,          /* resources */
    XtNumber (SlaveImageDisplayPopupResources),           /* num_resources */
    NULLQUARK,                                /* xrm_class */
    TRUE,                                     /* compress_motion */
    TRUE,                                     /* compress_exposure */
    TRUE,                                     /* compress_enterleave */
    TRUE,                                     /* visible_interest */
    _SlaveImageDisplayPopup_Destroy,          /* destroy */
    XtInheritResize,                          /* resize */
    NULL,                                     /* expose */
    NULL,                                     /* set_values */
    NULL,                                     /* set_values_hook */
    XtInheritSetValuesAlmost,                 /* set_values_almost */
    NULL,                                     /* get_values_hook */
    NULL,                                     /* accept_focus */
    XtVersion,                                /* version */
    NULL,                                     /* callback_private */
    NULL,                                     /* tm_translations */
    NULL,
    NULL,
    NULL,
  },
  {     /* CompositeClassPart */
      /* geometry_manager */	XtInheritGeometryManager,
      /* change_managed	  */	XtInheritChangeManaged,
      /* insert_child	  */	XtInheritInsertChild,
      /* delete_child	  */	XtInheritDeleteChild,
      /* extension	  */	NULL
  },
  { /* Shell */
      /* extension	  */	NULL
  },
  { /* WMShell */
      /* extension	  */	NULL
  },
  { /* VendorShell */
      /* extension	  */	NULL
  },
  { /* TopLevelShell */
      /* extension	  */	NULL
  },
  {  /* SlaveImageDisplayPopupClassPart */
      0                   /* empty */
  }
};

WidgetClass slaveImageDisplayPopupWidgetClass =
(WidgetClass) &slaveImageDisplayPopupClassRec;


static void _SlaveImageDisplayPopup_Initialise (Widget Request, Widget New)
{
    /*SlaveImageDisplayPopupWidget request = (SlaveImageDisplayPopupWidget) Request;*/
    SlaveImageDisplayPopupWidget new = (SlaveImageDisplayPopupWidget) New;
    Widget parent, form, w;
    static char function_name[] = "SlaveImageDisplayPopupWidget::Initialise";

    new->slaveImageDisplayPopup.canvasList = NULL;
    parent = XtParent (New);
    if ( !XtIsImageDisplay (parent) )
    {
	fprintf (stderr, "Parent widget is not class ImageDisplayWidget\n");
	a_prog_bug (function_name);
    }
    form = XtVaCreateManagedWidget ("topForm", formWidgetClass, New, NULL);
    new->slaveImageDisplayPopup.top_form = form;
    w = XtVaCreateManagedWidget ("closeButton", commandWidgetClass, form,
				 XtNlabel, "Close",
				 XtNwidth, 126,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainTop,
				 XtNleft, XtChainLeft,
				 XtNright, XtChainRight,
				 NULL);
    new->slaveImageDisplayPopup.close_btn = w;
    XtAddCallback (w, XtNcallback, xtmisc_popdown_cbk, New);
    if ( XtIsRealized (parent) )
	_SlaveImageDisplayPopup_im_dpy_realised_cbk (parent, New, NULL);
    else XtAddCallback (parent, XkwNallRealisedCallback,
			_SlaveImageDisplayPopup_im_dpy_realised_cbk, New);
    XtAddCallback (parent, XkwNnewVisibleCanvasCallback,
		   new_visible_canvas_cbk, New);
}   /*  End Function Initialise  */

static void _SlaveImageDisplayPopup_Destroy (Widget W)
{
    SlaveImageDisplayPopupWidget w = (SlaveImageDisplayPopupWidget) W;

    m_free (w->slaveImageDisplayPopup.canvasList);
}   /*  End Function Destroy  */

static void _SlaveImageDisplayPopup_im_dpy_realised_cbk (Widget parent,
							 XtPointer client_data,
							 XtPointer call_data)
{
    SlaveImageDisplayPopupWidget top =
	(SlaveImageDisplayPopupWidget) client_data;
    Widget TOP = (Widget) client_data;
    Widget multi_canvas, w;
    Widget pc_cnvs = NULL;
    Widget dc_cnvs = NULL;
    Widget tc_cnvs = NULL;
    KWorldCanvas *wc_ptr;
    static char function_name[] ="SlaveImageDisplayPopup::im_dpy_realised_cbk";

    if ( ( top->slaveImageDisplayPopup.canvasList =
	   m_calloc (sizeof *top->slaveImageDisplayPopup.canvasList * 4) )
	  == NULL )
	m_abort (function_name, "canvas list");
    wc_ptr = top->slaveImageDisplayPopup.canvasList;
    XtVaGetValues (parent, XkwNmultiCanvas, &multi_canvas, NULL);
    if (multi_canvas == NULL)
    {
	fprintf (stderr, "MultiCanvas widget not found\n");
	a_prog_bug (function_name);
    }
    /*  First create all the canvas widgets  */
    if ( ( w = XtNameToWidget (multi_canvas, "pseudoColourCanvas") ) != NULL )
	pc_cnvs = clone_canvas (top, w);
    if ( ( w = XtNameToWidget (multi_canvas, "directColourCanvas") ) != NULL )
	dc_cnvs = clone_canvas (top, w);
    if ( ( w = XtNameToWidget (multi_canvas, "trueColourCanvas") ) != NULL )
	tc_cnvs = clone_canvas (top, w);
    /*  Force all widgets to be realised  */
    XtRealizeWidget (TOP);
    /*  Now create the world canvases and so on for each canvas widget  */
    clone_wc (top, pc_cnvs, XkwNpseudoColourCanvas,
	      &top->slaveImageDisplayPopup.pseudoCanvas, &wc_ptr);
    clone_wc (top, dc_cnvs, XkwNdirectColourCanvas,
	      &top->slaveImageDisplayPopup.directCanvas, &wc_ptr);
    clone_wc (top, tc_cnvs, XkwNtrueColourCanvas,
	      &top->slaveImageDisplayPopup.trueCanvas, &wc_ptr);
    XtCallCallbacks (TOP, XkwNrealisedCallback,
		     top->slaveImageDisplayPopup.canvasList);
}   /*  End Function im_dpy_realised_cbk  */

static Widget clone_canvas (SlaveImageDisplayPopupWidget top, Widget orig_cnvs)
/*  [SUMMARY] Clone a canvas.
    <top> The SlaveImageDisplayPopupWidget.
    <orig_cnvs> The original canvas to clone.
    [RETURNS] The canvas widget.
*/
{
    unsigned long fg_pixel, bg_pixel;
    Boolean map;
    Colormap xcmap;
    Widget w;
    Widget form = top->slaveImageDisplayPopup.top_form;
    Visual *visual;

    XtVaGetValues (orig_cnvs,
		   XtNvisual, &visual,
		   XtNcolormap, &xcmap,
		   XtNbackground, &bg_pixel,
		   XtNforeground, &fg_pixel,
		   XtNmappedWhenManaged, &map,
		   NULL);
    w = XtVaCreateManagedWidget (XtName (orig_cnvs), canvasWidgetClass, form,
				 XtNfromVert,
				 top->slaveImageDisplayPopup.close_btn,
				 XtNwidth, 128,
				 XtNheight, 128,
				 XtNtop, XtChainTop,
				 XtNbottom, XtChainBottom,
				 XtNleft, XtChainLeft,
				 XtNright, XtChainRight,
				 XtNvisual, visual,
				 XtNcolormap, xcmap,
				 XtNbackground, bg_pixel,
				 XtNforeground, fg_pixel,
				 XtNmappedWhenManaged, map,
				 XtNborderWidth, 0,
				 XkwNsilenceUnconsumed, True,
				 NULL);
    return (w);
}   /*  End Function clone_canvas  */

static void clone_wc (SlaveImageDisplayPopupWidget top, Widget cnvs,
		      CONST char *name, KWorldCanvas *worldcanvas,
		      KWorldCanvas **wc_ptr)
/*  [SUMMARY] Clone a world canvas.
    <top> The SlaveImageDisplayPopupWidget.
    <cnvs> The new canvas widget.
    <name> The resource name in the ImageDisplay widget.
    <worldcanvas> The world canvas is written here.
    <wc_ptr> The world canvas is also written here and the pointer is
    incremented.
    [RETURNS] Nothing.
*/
{
    flag int_zoom, fix_aspect, fast_panner;
    Kcolourmap kcmap;
    KWorldCanvas orig_wc;
    KPixCanvas pixcanvas;
    struct win_scale_type win_scale;
    Boolean mappable;
    /*static char function_name[] = "SlaveImageDisplayPopup::clone_wc";*/

    if (cnvs == NULL) return;
    canvas_init_win_scale (&win_scale, K_WIN_SCALE_MAGIC_NUMBER);
    XtVaGetValues (XtParent ( (Widget) top ),
		   name, &orig_wc,
		   NULL);
    kcmap = canvas_get_cmap (orig_wc);
    XtVaGetValues (cnvs,
		   XkwNmonoPixCanvas, &pixcanvas,
		   XtNmappedWhenManaged, &mappable,
		   NULL);
    if ( ( *worldcanvas = canvas_create (pixcanvas, kcmap, &win_scale) )
	 == NULL )
    {
	exit (RV_UNDEF_ERROR);
    }
    **wc_ptr = *worldcanvas;
    ++*wc_ptr;
    canvas_set_attributes (*worldcanvas,
			   CANVAS_ATT_AUTO_MIN_SAT, TRUE,
			   CANVAS_ATT_AUTO_MAX_SAT, TRUE,
			   CANVAS_ATT_AUTO_BLANK, TRUE,
			   CANVAS_ATT_END);
    if (mappable) top->slaveImageDisplayPopup.visibleCanvas = *worldcanvas;
    viewimg_init (*worldcanvas);
    if (top->slaveImageDisplayPopup.enablePanning)
    {
	int_zoom = TRUE;
	fix_aspect = FALSE;
	fast_panner = top->slaveImageDisplayPopup.fastPanner;
    }
    else
    {
	int_zoom = FALSE;
	fix_aspect = TRUE;
	fast_panner = FALSE;
    }
    viewimg_set_canvas_attributes (*worldcanvas,
				   VIEWIMG_ATT_AUTO_V, FALSE,
				   VIEWIMG_ATT_INT_X, int_zoom,
				   VIEWIMG_ATT_INT_Y, int_zoom,
				   VIEWIMG_ATT_MAINTAIN_ASPECT, fix_aspect,
				   VIEWIMG_ATT_ENABLE_PANNING,
				   top->slaveImageDisplayPopup.enablePanning,
				   VIEWIMG_ATT_FAST_PANNER, fast_panner,
				   VIEWIMG_ATT_AUTO_WCS, TRUE,
				   VIEWIMG_ATT_END);
}   /*  End Function clone_wc  */

static void new_visible_canvas_cbk (Widget parent, XtPointer client_data,
				    XtPointer call_data)
{
    KWorldCanvas master_wc = (KWorldCanvas) call_data;
    KWorldCanvas master_pc, master_dc, master_tc, old_wc, new_wc;
    SlaveImageDisplayPopupWidget top =
	(SlaveImageDisplayPopupWidget) client_data;
    Widget old_cnv = NULL;      /*  Initialised to keep compiler happy  */
    Widget new_cnv = NULL;      /*  Initialised to keep compiler happy  */
    Widget form = top->slaveImageDisplayPopup.top_form;
    static char function_name[] =
	"SlaveImageDisplayPopup::new_visible_canvas_cbk";

    XtVaGetValues (parent,
		   XkwNpseudoColourCanvas, &master_pc,
		   XkwNdirectColourCanvas, &master_dc,
		   XkwNtrueColourCanvas, &master_tc,
		   NULL);
    /*  First find out the old widget that is mapped  */
    old_wc = top->slaveImageDisplayPopup.visibleCanvas;
    if (old_wc == NULL) old_cnv = NULL;
    else if (old_wc == top->slaveImageDisplayPopup.pseudoCanvas)
	old_cnv = XtNameToWidget (form, "pseudoColourCanvas");
    else if (old_wc == top->slaveImageDisplayPopup.directCanvas)
	old_cnv = XtNameToWidget (form, "directColourCanvas");
    else if (old_wc == top->slaveImageDisplayPopup.trueCanvas)
	old_cnv = XtNameToWidget (form, "trueColourCanvas");
    else
    {
	fprintf (stderr, "Visible canvas: %p unknown!\n", old_wc);
	a_prog_bug (function_name);
    }
    if (master_wc == master_pc)
    {
	new_wc = top->slaveImageDisplayPopup.pseudoCanvas;
	new_cnv = XtNameToWidget (form, "pseudoColourCanvas");
    }
    else if (master_wc == master_dc)
    {
	new_wc = top->slaveImageDisplayPopup.directCanvas;
	new_cnv = XtNameToWidget (form, "directColourCanvas");
    }
    else if (master_wc == master_tc)
    {
	new_wc = top->slaveImageDisplayPopup.trueCanvas;
	new_cnv = XtNameToWidget (form, "trueColourCanvas");
    }
    else
    {
	fprintf (stderr, "%s: unknown world canvas\n", function_name);
	return;
    }
    top->slaveImageDisplayPopup.visibleCanvas = new_wc;
    if (new_cnv == old_cnv) return;
    if (old_cnv != NULL) XtUnmapWidget (old_cnv);
    XtMapWidget (new_cnv);
}   /*  End Function new_visible_canvas_cbk  */


/*  Public functions follow  */

void XkwSlaveImageDisplayPopupRefresh (Widget W, flag clear)
/*  [PURPOSE] This routine will refresh the active display canvas for the
    SlaveImageDisplayPopup widget. If the active canvas is a stereo canvas both
    the left and right canvases are refreshed.
    <W> The widget.
    <clear> If TRUE, the canvas is cleared prior to refresh.
    [RETURNS] Nothing.
*/
{
    KWorldCanvas vc;
    flag visible;
    SlaveImageDisplayPopupWidget w;
    static char function_name[] = "XkwSlaveImageDisplayPopupRefresh";

    if ( !XtIsSlaveImageDisplayPopup (W) )
    {
	fprintf (stderr, "Invalid widget passed\n");
	a_prog_bug (function_name);
    }
    w = (SlaveImageDisplayPopupWidget) W;
    if ( (vc = w->slaveImageDisplayPopup.visibleCanvas) == NULL ) return;
    kwin_get_attributes (canvas_get_pixcanvas (vc),
			 KWIN_ATT_VISIBLE, &visible,
			 KWIN_ATT_END);
    if (!visible) return;
    if ( (vc == w->slaveImageDisplayPopup.pseudoCanvasLeft) ||
	 (vc == w->slaveImageDisplayPopup.pseudoCanvasRight) )
    {
	canvas_resize (w->slaveImageDisplayPopup.pseudoCanvasLeft, NULL,clear);
	canvas_resize (w->slaveImageDisplayPopup.pseudoCanvasRight, NULL,
		       clear);
    }
    else if ( (vc == w->slaveImageDisplayPopup.directCanvasLeft) ||
	      (vc == w->slaveImageDisplayPopup.directCanvasRight) )
    {
	canvas_resize (w->slaveImageDisplayPopup.directCanvasLeft, NULL,clear);
	canvas_resize (w->slaveImageDisplayPopup.directCanvasRight, NULL,
		       clear);
    }
    else if ( (vc == w->slaveImageDisplayPopup.trueCanvasLeft) ||
	      (vc == w->slaveImageDisplayPopup.trueCanvasRight) )
    {
	canvas_resize (w->slaveImageDisplayPopup.trueCanvasLeft, NULL, clear);
	canvas_resize (w->slaveImageDisplayPopup.trueCanvasRight, NULL, clear);
    }
    else canvas_resize (vc, NULL, clear);
}   /*  End Function XkwSlaveImageDisplayPopupRefresh  */
